// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// llSingleP.h++

#ifndef _llSingleP_H_
#define _llSingleP_H_

#include <iostream.h>
#include <assert.h>
#include "llPix.h"
#include "llSinglePNode.h"

// #define llSingleP_Debug

template < class T >
class llSingleP
{
    static int defaultMaxNodes;
    static int nodeFarmCount;
    static llSinglePNode<T>* nodeFarm;
    static llSingleP<T>* nodeFarmFreeList;
    int nodeCount;
    llSinglePNode<T>* lnode;
    llSinglePNode<T>* append( llSinglePNode<T>* );
    llSinglePNode<T>* prepend( llSinglePNode<T>* );
    llSinglePNode<T>* prepop();
    llSinglePNode<T>* getNewNode();
    static void alert( char* fn, char* s );
    static void error( char* fn, char* s );
 public:
    static void initClass( int n = 0 );
    llSingleP();
    ~llSingleP();
    void clear();
    llPix first() { return( (lnode) ? lnode->next: 0 ); }
    llPix last() { return lnode; }
    llPix next( llPix );
    T* item( llPix );
    inline T* operator()( llPix p ) { return item(p); }
    T* head() { return( (lnode) ? lnode->next->item : 0 ); }
    T* tail() { return( (lnode) ? lnode->item : 0 ); }
    T* append( T* datum );
    llSingleP<T>* append( llSingleP<T>* );
    int pop( llSingleP<T>* = nodeFarmFreeList );
    T* prepend( T* datum );
    inline int length() { return nodeCount; }
    inline int empty() { return !nodeCount; }
    inline int reallyempty() { return( (lnode) ? 0 : 1 ); }
    static ostream& printFree( ostream& os );
    static ostream& printFarm( ostream& os );
    ostream& print( ostream& os );
    int OK();
    static int freeOK() { return nodeFarmFreeList->OK(); }
    static int farmBAD();
    static inline int inFarm( llSinglePNode<T>* );
    static int freeCount() { return nodeFarmFreeList->length(); }
};

template < class T > int llSingleP<T>::defaultMaxNodes = 100;
template < class T > int llSingleP<T>::nodeFarmCount = 0;
template < class T > llSinglePNode<T>* llSingleP<T>::nodeFarm = 0;
template < class T > llSingleP<T>* llSingleP<T>::nodeFarmFreeList;

template < class T >
inline
int
llSingleP<T>::inFarm( llSinglePNode<T>* p )
{
    int index = p - nodeFarm;
    if( index < 0 || index >= nodeFarmCount )
       return 0;
    return 1;
}

template < class T >
void
llSingleP<T>::initClass( int n )
{
    if( nodeFarm && nodeFarmCount >= n )
       return;
//     cerr << "llSingleP<T>::initClass" << endl;
//     cerr << "\tn= " << n << endl;
//     cerr << "\tnodeFarm= " << nodeFarm << endl;
//     cerr << "\tnodeFarmCount= " << nodeFarmCount << endl;
//     cerr << "\tsizeof(T)= " << sizeof(T) << endl;
    if( nodeFarm )
       {
	   delete [] nodeFarm;
	   delete nodeFarmFreeList;
	   nodeFarmCount = n;
       }
    else 
       nodeFarmCount = ( n > 0 ) ? n : defaultMaxNodes;
    nodeFarm = new llSinglePNode<T>[ nodeFarmCount ];
    nodeFarmFreeList = new llSingleP<T>;
    for( int i = 0; i < nodeFarmCount; i++ )
       nodeFarmFreeList -> append( nodeFarm + i );
//     cerr << "\tnodeFarm= " << nodeFarm << endl;
//     cerr << "\tnodeFarmCount= " << nodeFarmCount << endl;
}

template < class T >
llSinglePNode<T>*
llSingleP<T>::append( llSinglePNode<T>* p )
{
    if( !p )
       // return 0;
       error( "append(node)", "null node" );

    if( !inFarm( p ) )
       error( "append(node)", "bad node" );
    nodeCount++;
    if( !lnode )
       p->next = lnode = p;
    else
       {
	   p->next = lnode->next;
	   lnode->next = p;
	   lnode = p;
       }
#ifdef llSingleP_Debug
    assert( OK() );
#endif
    return p;
}

template < class T >
llSinglePNode<T>*
llSingleP<T>::prepend( llSinglePNode<T>* p )
{
    if( !p )
       return 0;
    nodeCount++;
    if( !lnode )
       p->next = lnode = p;
    else
       {
	   p->next = lnode->next;
	   lnode->next = p;
       }
#ifdef llSingleP_Debug
    assert( OK() );
#endif
    return p;
}

template < class T >
llSinglePNode<T>*
llSingleP<T>::prepop()
{
    if( empty() )
       {
	   if( reallyempty() )
	      error( "prepop", "empty list" );
	   error( "prepop", "unsynchronized nodecount" );
	   return 0;
       }
    llSinglePNode<T>* p = lnode->next;
    if( lnode == p ) // single node in list
       lnode = 0;
    else
       lnode->next = p->next;
    nodeCount--;
#ifdef llSingleP_Debug
    assert( OK() );
#endif
    return p;
}

template < class T >
llSinglePNode<T>* 
llSingleP<T>::getNewNode()
{
    if( ! nodeFarm )
       error( "getNewNode", "class not inited (use ::initClass())" );
    if( ! nodeFarmFreeList -> empty() )
       return nodeFarmFreeList -> prepop();
    error( "getNewNode", "no more nodes" );
    return 0;
}

template < class T >
void
llSingleP<T>::alert( char* fn, char* s )
{
    cerr << "llSingleP<T>::" << fn << "(): ERROR: " << s << endl;
}

template < class T >
void
llSingleP<T>::error( char* fn, char* s )
{
    alert( fn, s );
    // exit( 1 );
    assert( 0 );
}


template < class T >
T*
llSingleP<T>::append( T* d )
{
    llSinglePNode<T>* p = getNewNode(); // error checks
    p -> item = d;
    T* v = append( p )->item;
    return v;
}

template < class T >
T*
llSingleP<T>::prepend( T* d )
{
    llSinglePNode<T>* p = getNewNode(); // error checks
    p -> item = d;
    T* v = prepend( p )->item;
    return v;
}

template < class T >
llSingleP<T>::llSingleP()
  : nodeCount( 0 ), lnode( 0 )
{
//     if( !nodeFarm )
//        error( "llSingleP<T>", "class not initialized" );
    // initClass( n );
}

template < class T >
llSingleP<T>::~llSingleP()
{
    clear();
}

// template < class T >
// llPix 
// llSingleP<T>::first()
// {
//     if( !lnode )
//        return 0;
//     return lnode->next;
// }

template < class T >
llPix
llSingleP<T>::next( llPix p )
{
    llSinglePNode<T>* pp = (llSinglePNode<T>*)p;
    if( pp == lnode )
       return 0;
    return pp->next;
}

template < class T >
T*
llSingleP<T>::item( llPix p )
{
    llSinglePNode<T>* sp = ( llSinglePNode<T>* ) p;
    if( ! inFarm( sp ) )
       error( "item", "node not in farm" );
    // return( ((llSinglePNode<T>*)p)->item );
    return( sp->item );
}

template < class T >
void
llSingleP<T>::clear()
{
    nodeFarmFreeList -> append( this );
#ifdef llSingleP_Debug
    assert( OK() );
#endif
}

template < class T >
llSingleP<T>*
llSingleP<T>::append( llSingleP<T>* l )
{
    if( l -> empty() )
       {
	   if( ! l -> reallyempty() )
	      error( "append", "nodecount synchronization in list" );
	   return this;
       }
    if( empty() )
       {
	   if( ! reallyempty() )
	      error( "append", "nodecount synchronization in this" );
	   lnode = l->lnode;
	   nodeCount = l->nodeCount;
       }
    else
       {
	   llSinglePNode<T>* p = l->lnode->next;
	   l->lnode->next = lnode->next;
	   lnode->next = p;
	   lnode = l->lnode;
	   nodeCount += l->nodeCount;
       }
    l->lnode = 0;
    l->nodeCount = 0;
#ifdef llSingleP_Debug
    assert( OK() );
#endif
    return this;
}

template < class T >
ostream&
llSingleP<T>::printFree( ostream& os )
{
    os << "free " << nodeFarmFreeList->length() << endl;
    for( llPix p = nodeFarmFreeList->first();
	 p != 0;
	 p = nodeFarmFreeList->next( p ) )
       {
	   llSinglePNode<T>* pp = (llSinglePNode<T>*)p;
	   os << "\t: (" << pp << "), "<< pp->item << ", " << pp->next << endl;
       }
    return os;
}

template < class T >
ostream&
llSingleP<T>::printFarm( ostream& os )
{
    os << "farm " << nodeFarmCount << endl;
    for( int i = 0; i < nodeFarmCount; i++ )
       {
	   llSinglePNode<T>* pp = nodeFarm + i;
	   os << "\t: (" << pp << "), "<< pp->item << ", " << pp->next << endl;
       }
    return os;
}

template < class T >
int
llSingleP<T>::pop( llSingleP<T>* l )
{
    if( empty() )
       return 1;
    l->append( prepop() );
    return 0;
}

template < class T >
ostream&
llSingleP<T>::print( ostream& os )
{
    os << "llSingleP " << length() << endl;
    if( !lnode )
       return( os << "\tempty" << endl );
    llSinglePNode<T>* pp = lnode;
    do
       {
	   pp = pp -> next;
	   os << "\t: (" << pp << "), "<< pp->item << ", " << pp->next << endl;
       }
    while( pp != lnode );
    return os;
}

template < class T >
int
llSingleP<T>::OK()
{
    if( !lnode && !nodeCount )
       return 1;
    if( !lnode && nodeCount )
       {
	   alert( "OK", "empty list, nodeCount not zero" );
	   return 0;
       }
    llSinglePNode<T>* np = lnode;
    int count = nodeCount;
    int index;
    while( count-- )
       {
	   np = np->next;
	   index = np - nodeFarm;
	   if( index < 0 || index >= nodeFarmCount )
	      {
		  alert( "OK", "node not from node farm" );
		  cerr << "\tnode= " << np << endl;
		  cerr << "\tsizeof(T)= " << sizeof(T) << endl;
		  cerr << "\ti= " << nodeCount-count-1 << endl;
		  cerr << "\tlen= " << nodeCount << endl;
		  return 0;
	      }
       }
    np = lnode;
    count = nodeCount;
    while( count-- && (np=np->next) != lnode );
    if( count || np != lnode )
       {
	   alert( "OK", "bad list" );
	   cerr << "count= " << count << endl;
	   cerr << "nodeCount= " << nodeCount << endl;
	   print( cerr );
	   return 0;
       }
    return 1;
}

template < class T >
int
llSingleP<T>::farmBAD()
{
    llSinglePNode<T>* np = nodeFarm;
    int shitCount = 0;
    for( int i = 0; i < nodeFarmCount; i++ )
       {
	   if( ! inFarm( np->next ) )
	      {
		  alert( "farmBAD", "node not from node farm" );
		  cerr << "\ti= " << i << ", ";
		  cerr << "node= " << np << ", ";
		  cerr << "next= " << np->next << ", ";
		  cerr << "farm= " << nodeFarm << ", ";
		  cerr << "cnt= " << nodeFarmCount << ", ";
		  cerr << "sizeof(T)= " << sizeof(T) << endl;
		  // cerr << "int(next)= " << int(np->next) << "; ";
		  cerr << "\tint(next)= " << *(int*)(&(np->next)) << "; ";
		  cerr << "float(next)= " << *(float*)(&(np->next)) << "; ";
		  short* shp = (short*)(&(np->next));
		  cerr << "short(next)= " << shp[0] << ", " << shp[1] << endl;
		  // char* chp = (char*)(&(np->next));
		  // cerr << "char(next)= " << chp[0] << ", " << chp[1] 
		  // << " , " << chp[2] << ", " << chp[3] << endl;
		  shitCount++;
	      }
	   np++;
       }
    return shitCount;
}    

#endif /* _llSingleP_H_ */
