// qmLUDecomposition.h++

#ifndef _qmLUDecomposition_H_
#define _qmLUDecomposition_H_

#include <iostream.h>
#include "qmMatrix.h"
#include "qmVector.h"

class qmLUDecomposition
{
    qmMatrix LUmatrix;
    int* rowOrder;

private:

    static int findMaxPivot( qmMatrixRef&, int col, int *rowOrder );
    static void forwardSubstitute( qmMatrixRef& LU, 
				   int *rowOrder,
				   real *b,
				   real *y );
    static void backwardSubstitute( qmMatrixRef& LU, 
				    int *rowOrder,
				    real *y,
				    real *x );

public:

    qmLUDecomposition() 
      : LUmatrix(), rowOrder(0) {}
    qmLUDecomposition( const qmMatrixRef& m )
      : LUmatrix(), rowOrder(0) { set( m ); }
    int set( const qmMatrixRef& );
    qmMatrixRef& getLU() { return LUmatrix; }
    int* getRowOrder() { return rowOrder; }
    static int LUDecompose( qmMatrixRef& A, int* rorder );
    static int LUSolve( qmMatrixRef& LU, 
			int* rorder,
			qmVectorRef& y,
			qmVectorRef& x );
    static int LUInvert( qmMatrixRef& LU, int* rorder, qmMatrixRef& Ai );
    int solve( qmVectorRef& y, qmVectorRef& x );
    int invert( qmMatrixRef& Ai );
    ostream& print( ostream& );
};

typedef qmLUDecomposition qmLUD;

extern ostream& operator<<( ostream&, qmLUDecomposition& );

#endif /* _qmLUDecomposition_H_ */
