// qmMatrix.c++

#include <math.h>
#include "qmMatrix.h"
#include "qmCompileOptions.h"
#include "qmMinMax.h"

qmMatrix::qmMatrix( int h, int w )
  : qmMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( h, w );
}

qmMatrix::qmMatrix( int h, int w, real *data )
  : qmMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( h, w, data );
}

qmMatrix::qmMatrix( const qmMatrix& m )
  : qmMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( m );
}

qmMatrix::qmMatrix( const qmMatrixRef& m )
  : qmMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( m );
}

qmMatrix&
qmMatrix::set( int h, int w )
{
#ifdef QmDoBoundaryChecking
    QmAssert( h > 0 && w > 0 );
#endif
    if( !( allocArray && elementCount == h*w ) )
       {
	   if( allocArray )
	      delete [] allocArray;
	   elementCount  = h*w;
	   allocArray = new real[ elementCount ];
       }
    width = w;
    array = allocArray;
    return *this;
}

qmMatrix&
qmMatrix::set( int h, int w, real *data )
{
#ifdef QmDoBoundaryChecking
    QmAssert( h > 0 && w > 0 && data );
#endif
    set( h, w );
    setData( data );
    return *this;
}

qmMatrix&
qmMatrix::set( const qmMatrixRef& qm )
{
#ifdef QmDoBoundaryChecking
    QmAssert( qm.elementCount > 0 );
#endif
    if( !( allocArray && 
	   elementCount == qm.elementCount &&
	   width == qm.width ) )
       {
	   if( allocArray && elementCount != qm.elementCount )
	      delete [] allocArray, allocArray = 0;
	   if( !allocArray )
	      allocArray = new real[ qm.elementCount ];
 	   elementCount = qm.elementCount;
	   width = qm.width;
	   array = allocArray;
       }
    copy( qm );
    return *this;
}

