// qmMatrixRef.h++

#ifndef _qmMatrixRef_H_
#define _qmMatrixRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include "qmReal.h"

class qmVectorRef;

class qmMatrixRef
{
    friend class qmVectorRef;
    friend class qmMatrix;

protected:

    int elementCount;
    int width;
    real *array;

public:

    virtual ~qmMatrixRef() {}
    qmMatrixRef( )
      : elementCount(0), width(0), array(0) {}
    qmMatrixRef( int h, int w, real* a )
      : elementCount(h*w), width(w), array(a) {}
    qmMatrixRef( const qmMatrixRef& mr )
      : elementCount( 0 ), width( 0 ), array( 0 ) { set( mr ); }

    qmMatrixRef& set( int h, int w, real* a );		// shadow
    qmMatrixRef& set( const qmMatrixRef& );		// shadow
    qmMatrixRef& copy( const qmMatrixRef& );		// copy
    qmMatrixRef& operator=( const qmMatrixRef& );	// shadow

    int getElementCount() const { return elementCount; }
    int getRowCount() const { return getHeight(); }
    int getColCount() const { return getWidth(); }
    int getHeight() const { return( (width) ? elementCount/width : 0); }
    int getWidth() const { return width; }

    real* getArray( int r=0, int c=0 ) {return( array + r*width + c ); }
    operator real*() { return array; }
    real& operator()( int r, int c ) { return array[r*width + c]; }
    real getValue( int r, int c ) const { return array[r*width + c]; }

    void setData( real * );
    void getData( real * ) const;
    qmMatrixRef& operator<<( real *p ) { setData(p); return *this; }
    void operator>>( real *p ) { getData(p); }

    qmMatrixRef& setAdd( const qmMatrixRef&, const qmMatrixRef& );
    qmMatrixRef& setSubtract(const qmMatrixRef&, const qmMatrixRef& );
    qmMatrixRef& setMultiply(const qmMatrixRef&, const qmMatrixRef& );
    qmMatrixRef& setMultiplyT( const qmMatrixRef&, const qmMatrixRef& );
    qmMatrixRef& setTMultiply( const qmMatrixRef&, const qmMatrixRef& );
    qmMatrixRef& setDivide( const qmMatrixRef&, const qmMatrixRef& );
    qmMatrixRef& setNegate( const qmMatrixRef& );
    qmMatrixRef& setTranspose( const qmMatrixRef& );
    qmMatrixRef& setShift( real, const qmMatrixRef& );
    qmMatrixRef& setScale( real, const qmMatrixRef& );
    qmMatrixRef& setScaleAddEq( real, const qmMatrixRef& );
    qmMatrixRef& setAddEq( const qmMatrixRef& op );
    qmMatrixRef& setSubtractEq( const qmMatrixRef& op );
    qmMatrixRef& setNegateEq();
    qmMatrixRef& setInvert( const qmMatrixRef& );
    qmMatrixRef& setInvertEq();
    int setInvertCheck( const qmMatrixRef& );
    int setInvertEqCheck();
    qmMatrixRef& setShiftEq( real val );
    qmMatrixRef& setScaleEq( real val );
    qmMatrixRef& setConst( real val );
    qmMatrixRef& setZero() { return setConst( 0. ); }
    qmMatrixRef& setDiagonal( real val );
    qmMatrixRef& setDiagonal( const qmVectorRef& );
    void getDiagonal( const qmVectorRef& );
    qmMatrixRef& setIdentity() { return setDiagonal(1.); }
    qmMatrixRef& setOuterProduct( const qmVectorRef&, const qmVectorRef& );
    qmMatrixRef& setAddScaledOuterProductEq( const qmVectorRef&,
					     const qmVectorRef&,
					     real=1.);
    qmMatrixRef& setVectorScale( const qmVectorRef&, const qmMatrixRef& );
    qmMatrixRef& operator+=(const qmMatrixRef& op) {return setAddEq(op); }
    qmMatrixRef& operator-=(const qmMatrixRef& op) {return setSubtractEq(op);}
    qmMatrixRef& operator+=(const real val) { return setShiftEq(val); }
    qmMatrixRef& operator-=(const real val) { return setShiftEq(-val); }
    qmMatrixRef& operator*=(const real val) { return setScaleEq(val); }
    qmMatrixRef& operator/=(const real val) { return setScaleEq(real(1)/val); }
    qmMatrixRef& operator=( const real val ) { return setConst(val); }
    int operator==( const qmMatrixRef& ) const;
    real getDeterminant3() const;
    real getDeterminant2() const;
    real getDeterminant() const;
    ostream& print( ostream&, char* = "\n" ) const;
    istream& scan( istream& );
    ostream& write( ostream& ) const;
    istream& read( istream& );
    // friend ostream& operator<<( ostream&, qmMatrixRef& );
    const qmVectorRef& getColumn( int i ) const;
    const qmVectorRef& getRow( int i ) const;
    void setColumn( int i, const qmVectorRef& );
    void setRow( int i, const qmVectorRef& );

    inline real getTrace() const;

 public:
    inline real* _getArray(int r=0,int c=0) const {return( array+r*width+c );}
    inline real& _el( int r, int c ) const { return( array[r*width+c] ); }

};

inline
real
qmMatrixRef::getTrace() const
{
    int dim = ( width <= getHeight() ) ? width : getHeight();
    real sum = 0;
    for( int i = 0; i < dim; i++ )
       sum += getValue( i, i );
    return sum;
}

extern ostream& operator<<( ostream&, const qmMatrixRef& );
extern istream& operator>>( istream&, qmMatrixRef& );

#endif /* _qmMatrixRef_H_ */
