// qmSubMatrixRef.h++

#ifndef _qmSubMatrixRef_H_
#define _qmSubMatrixRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include "qmReal.h"
#include "qmSubVectorRef.h"
#include "qmMatrixRef.h"

class qmSubMatrixRef
{
    friend class qmSubVectorRef;
    
protected:
    
    int elementCount;
    int width;
    int spacing;
    real *array;
    
public:
    
    qmSubMatrixRef( )
      : elementCount(0), width(0), spacing(0), array(0) {}
    qmSubMatrixRef( int h, int w, int line, real* a )
      : elementCount(h*w), width(w), spacing(line-w), array(a) {}
    qmSubMatrixRef( const qmMatrixRef&, int h=0, int w=0, int r=0, int c=0 );
    qmSubMatrixRef& copy( const qmSubMatrixRef& sm );
    qmSubMatrixRef& operator=( const qmSubMatrixRef& sm ) { return copy(sm); }
    int getElementCount() const { return elementCount; }
    int getRowCount() const { return getHeight(); }
    int getColCount() const { return getWidth(); }
    int getHeight() const { return( (width) ? elementCount/width : 0); }
    int getWidth() const { return width; }
    int getRowSpacing() const { return width+spacing; }
    int getColSpacing() const { return 1; }
    real* getArray( int r = 0, int c = 0 ) const { return( array + r*(width+spacing) + c ); }
    
    qmSubMatrixRef& setData( real * );
    real* getData( real * ) const;
    qmSubMatrixRef& operator<<( real *p ) { return setData(p); }
    real* operator>>( real *p ) { return getData(p); }
    
#if 0
    qmSubMatrixRef qmSubMatrixRef();
    qmSubMatrixRef qmSubMatrixRef( int ulr, int ulc, int dh, int dw );
#endif
    
    qmSubMatrixRef& setAdd( const qmSubMatrixRef&, const qmSubMatrixRef& );
    qmSubMatrixRef& setSubtract(const qmSubMatrixRef&, const qmSubMatrixRef& );
    qmSubMatrixRef& setMultiply(const qmSubMatrixRef&, const qmSubMatrixRef& );
    qmSubMatrixRef& setMultiplyT(const qmSubMatrixRef&, const qmSubMatrixRef& );
    // qmSubMatrixRef& setDivide( const qmSubMatrixRef&, const qmSubMatrixRef& );
    // qmSubMatrixRef& setNegate( const qmSubMatrixRef& );
    qmSubMatrixRef& setInvert( const qmSubMatrixRef& );
    // qmSubMatrixRef& Transpose( const qmSubMatrixRef& );
    // qmSubMatrixRef& setShift( const qmSubMatrixRef&, real val );
    // qmSubMatrixRef& setScale( const qmSubMatrixRef&, real val  );
    qmSubMatrixRef& setAddEq( const qmSubMatrixRef& op );
    // qmSubMatrixRef& setSubtractEq( const qmSubMatrixRef& op );
    // qmSubMatrixRef& setMultiplyEq( const qmSubMatrixRef& op );
    // qmSubMatrixRef& setDivideEq( const qmSubMatrixRef& op );
    // qmSubMatrixRef& setNegateEq();
    qmSubMatrixRef& setInvertEq();
    // qmSubMatrixRef& setShiftEq( real val );
    qmSubMatrixRef& setScaleEq( real val );
    qmSubMatrixRef& setConst( real val );
    qmSubMatrixRef& setDiagonal( real val );
    qmSubMatrixRef& setIdentity() { return setDiagonal( 1. ); }
    qmSubMatrixRef& setZero() { return setConst( 0. ); }
    qmSubMatrixRef& operator+=( const qmSubMatrixRef& op ) { return setAddEq(op); }
    // qmSubMatrixRef& operator-=( const qmSubMatrixRef& op ) { return setSubtractEq(op); }
    // qmSubMatrixRef& operator*=( const qmSubMatrixRef& op ) { return setMultiplyEq(op); }
    // qmSubMatrixRef& operator/=( const qmSubMatrixRef& op ) { return setDivideEq(op); }
    // qmSubMatrixRef& operator+=( const real val ) { return setShiftEq(val); }
    // qmSubMatrixRef& operator-=( const real val ) { return setShiftEq(-val); }
    // qmSubMatrixRef& operator*=( const real val ) { return setScaleEq(val); }
    qmSubMatrixRef& operator/=( const real val ) { return setScaleEq(real(1)/val); }
    // qmSubMatrixRef& operator=( const real val ) { return setConst(val); }
    
    ostream& print( ostream&, char* = "\n" );
    istream& scan( istream& );
    friend ostream& operator<<( ostream&, qmSubMatrixRef& );
    ostream& write( ostream& );
    istream& read( istream& );
    
    real& operator()( int r, int c ) { return array[r*width + c]; }
    
private:
    int FindMaxPivot( int, int * );
    int LUDecomp( int * );
    void ForwardSubstitute( real *, real *, int * );
    void BackwardSubstitute( real *, real *, int * );
    
    friend class qmMatrix;
};

#endif /* _qmSubMatrixRef_H_ */
