// qmSubVectorRef.c++

#include "qmSubVectorRef.h"

qmSubVectorRef&
qmSubVectorRef::copy( const qmSubVectorRef& sv )
{
    register real *to = array;
    register real *fm = sv.array;
    for( int i = 0; i < elementCount; i++ )
       (*to++) = (*fm++), to += spacing, fm += sv.spacing;
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setData( real *p )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) = *p++, to += spacing;
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setAdd( const qmSubVectorRef& a, const qmSubVectorRef& b )
{
    register int count = elementCount;
    register real *to = array;
    register real *fm1 = a.array;
    register real *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) + (*fm2++),
	to += spacing,
	fm1 += a.spacing,
	fm2 += b.spacing;
    return *this;
    
}

qmSubVectorRef&
qmSubVectorRef::setSubtract( const qmSubVectorRef& a, const qmSubVectorRef& b )
{
    register int count = elementCount;
    register real *to = array;
    register real *fm1 = a.array;
    register real *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) - (*fm2++),
	to += spacing,
	fm1 += a.spacing,
	fm2 += b.spacing;
    return *this;
    
}

qmSubVectorRef&
qmSubVectorRef::setMultiply( const qmSubMatrixRef& a, const qmSubVectorRef& b )
{
    register real *to = array;
    register real *fma;
    register real *fmb;
    int line = 1 + spacing;
    int bline = 1 + b.spacing;
    for( int i = 0; i < elementCount; i++ )
       {
	   real sum = 0.;
	   fma = a.getArray( i, 0 );
	   fmb = b.array;
	   for( int k = 0; k < a.width; k++ )
	      sum += (*fma++) * (*fmb), fmb += bline;
	   *to = sum;
	   to += line;
       }
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setAddEq( const qmSubVectorRef& op )
{
    register int count = elementCount;
    register real *to = array;
    register real *fm = op.array;
    while( count-- )
       (*to++) += *fm++,
	to += spacing,
	fm += op.spacing;
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setSubtractEq( const qmSubVectorRef& op )
{
    register int count = elementCount;
    register real *to = array;
    register real *fm = op.array;
    while( count-- )
       (*to++) -= *fm++,
	to += spacing,
	fm += op.spacing;
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setShiftEq( real val )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) += val,
	to += spacing;
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setScaleEq( real val )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) *= val,
	to += spacing;
    return *this;
}

qmSubVectorRef&
qmSubVectorRef::setConst( real val )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       (*to++) = val,
	to += spacing;
    return *this;
}

real
qmSubVectorRef::getInnerProduct( const qmSubVectorRef& op ) const
{
    real out = 0.;
    register int count = elementCount;
    register real *fm1 = array;
    register real *fm2 = op.array;
    while( count-- )
       out += (*fm1++) * (*fm2++),
	fm1 += spacing,
	fm2 += op.spacing;
    return out;
}

int
qmSubVectorRef::operator==( const qmSubVectorRef& op ) const
{
    register int count = elementCount;
    register real *to = array;
    register real *fm = op.array;
    while( count-- )
       if( (*to++) != (*fm++) )
	  return 0;
       else
	  to += spacing, fm += spacing;
    return 1;
}

qmSubVectorRef&
qmSubVectorRef::setAbsolute( )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       {
	   if( *to < 0 )
	      *to = (-(*to));
	   to++, to += spacing;
       }
    return *this;
}

ostream&
qmSubVectorRef::print( ostream& os )
{
    register int count = elementCount;
    register real *fm = array;
    while( count-- )
       os << " " << *fm++, fm += spacing;
    return os;
}

istream&
qmSubVectorRef::scan( istream& is )
{
    register int count = elementCount;
    register real *to = array;
    while( count-- )
       is >> *to++,
	to += spacing;
    return is;
}

ostream&
operator<<( ostream& os, qmSubVectorRef& sv )
{
    return sv.print( os );
}

istream&
operator>>( istream& is, qmSubVectorRef& sv )
{
    return sv.scan( is );
}

