// qmSubVectorRef.h++

#ifndef _qmSubVectorRef_H_
#define _qmSubVectorRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include "qmReal.h"
#include "qmSubMatrixRef.h"

class qmSubVectorRef
{
    friend class qmSubMatrixRef;
    
protected:
    
    real *array;
    int elementCount;
    int spacing; // if 0, then packed
    
public:
    
    qmSubVectorRef( )
      : array(0), elementCount(0) {}
    qmSubVectorRef( int n, int s, real* a )
      : array(a), elementCount(n), spacing(s) {}
    qmSubVectorRef& copy( const qmSubVectorRef& sv );
    qmSubVectorRef& operator=( const qmSubVectorRef& sv ) { return copy(sv); }
    int getLength() const { return elementCount; }
    int getElementCount() const { return elementCount; }
    int getSpacing() const { return spacing; }
    real* getArray( int i = 0 ) { return( array + i ); }
    real& operator()( int i ) { return array[i]; }
    qmSubVectorRef& setData( real *p );
#if 0
    real* getData( real *p );
    qmSubVectorRef& operator<<( real *p ) { setData(p); return *this; }
    real* operator>>( real *p ) { return getData(p); }
#endif
    qmSubVectorRef& setAdd( const qmSubVectorRef&, const qmSubVectorRef& );
    qmSubVectorRef& setSubtract(const qmSubVectorRef&, const qmSubVectorRef& );
    // qmSubVectorRef& setMultiply(const qmSubVectorRef&, const qmSubVectorRef& );
    qmSubVectorRef& setMultiply( const qmSubMatrixRef&, const qmSubVectorRef& );
    // qmSubVectorRef& setDivide( const qmSubVectorRef&, const qmSubVectorRef& );
    // qmSubVectorRef& setNegate( const qmSubVectorRef& );
    // qmSubVectorRef& setInvert( const qmSubVectorRef& );
    // qmSubVectorRef& Transpose( const qmSubVectorRef& );
    // qmSubVectorRef& setShift( const qmSubVectorRef&, real val );
    // qmSubVectorRef& setScale( const qmSubVectorRef&, real val  );
    qmSubVectorRef& setAddEq( const qmSubVectorRef& op );
    qmSubVectorRef& setSubtractEq( const qmSubVectorRef& op );
    // qmSubVectorRef& setMultiplyEq( const qmSubVectorRef& op );
    // qmSubVectorRef& setDivideEq( const qmSubVectorRef& op );
    // qmSubVectorRef& setNegateEq();
    // qmSubVectorRef& setInvertEq();
    qmSubVectorRef& setShiftEq( real val );
    qmSubVectorRef& setScaleEq( real val );
    qmSubVectorRef& setConst( real val );
    qmSubVectorRef& setZero() { return setConst( 0. ); }
    qmSubVectorRef& operator+=( const qmSubVectorRef& op ) { return setAddEq(op); }
    // qmSubVectorRef& operator-=( const qmSubVectorRef& op ) { return setSubtractEq(op); }
    // qmSubVectorRef& operator*=( const qmSubVectorRef& op ) { return setMultiplyEq(op); }
    // qmSubVectorRef& operator/=( const qmSubVectorRef& op ) { return setDivideEq(op); }
    // qmSubVectorRef& operator+=( const real val ) { return setShiftEq(val); }
    qmSubVectorRef& operator-=( const real val ) { return setShiftEq(-val); }
    // qmSubVectorRef& operator*=( const real val ) { return setScaleEq(val); }
    qmSubVectorRef& operator/=( const real val ) { return setScaleEq(real(1)/val); }
    // qmSubVectorRef& operator=( const real val ) { return setConst(val); }
    real getInnerProduct( const qmSubVectorRef& ) const;
    int operator==( const qmSubVectorRef& ) const;
    qmSubVectorRef& setAbsolute();
    ostream& print( ostream& );
    istream& scan( istream& );
    ostream& write( ostream& );
    istream& read( istream& );
    friend ostream& operator<<( ostream&, qmSubVectorRef& );
};

#endif /* _qmSubVectorRef_H_ */
