// qmTime.h++

#ifndef _qmTime_H_
#define _qmTime_H_

#ifdef WIN32
#  include <windows.h>
#  include <time.h>
#else
#  define _BSD_TIME // so sgi is compatible with bsd
#  include <sys/time.h>
#endif

#ifdef WIN32
#ifndef TIMEZONE_DEFINED
#define TIMEZONE_DEFINED
struct timezone {
	int	tz_minuteswest;	/* unused in win32 */
	int	tz_dsttime;	/* unused in win32 */
};
#endif
#endif

#include <iostream.h>
#include <fstream.h>

class qmTime
{
public:
  struct timeval tv;
  struct timezone tz;
  double tsec;

  void init ()
  {
#ifdef WIN32 
    timeBeginPeriod((UINT) 1);
#endif
  }

  void gettimeofdaywrapper( struct timeval *tv,struct timezone *tz )
  {
#ifdef WIN32 
    unsigned long int thetime = timeGetTime();
    tv->tv_sec = thetime/1000;
    tv->tv_usec = (thetime%1000)*1000;
#else
    gettimeofday(tv,tz);
#endif
  }

public:

     qmTime() 
       : tsec( get() ) { tz.tz_minuteswest=0; tz.tz_dsttime=0; init(); }
     qmTime( double d ) 
       : tsec( d ) { tz.tz_minuteswest=0; tz.tz_dsttime=0; init(); }
     qmTime( long sec, long usec )
       : tsec( double(sec)+double(usec)/1.e6 ) {
 	  tz.tz_minuteswest=0; tz.tz_dsttime=0; init(); }
     qmTime( const qmTime& t ) 
       : tsec( t.tsec ) { tz.tz_minuteswest=0; tz.tz_dsttime=0; init(); }
  
    operator double() { return tsec; }
    operator struct timeval() { return getTimeVal(); }
    struct timeval getTimeVal() { tv.tv_sec = int(tsec); 
				  tv.tv_usec = int((tsec - tv.tv_sec)*1.e6);
				  return tv; }
    ostream& print( ostream& os ) { getTimeVal(); 
				    return( os << tv.tv_sec << " sec, "
					    << tv.tv_usec << " usec" ); }
    double get() { gettimeofdaywrapper( &tv, &tz ); 
		   return( tv.tv_sec + tv.tv_usec/1.e6 ); }
    double set() { return( tsec = get() ); }
    void set( long sec, long usec ) { tsec=double(sec)+double(usec)/1.e6; }
    qmTime& operator/=( int N ) { tsec /= N; return *this; }
    qmTime& operator+=( qmTime& t ) { tsec += t.tsec; return *this; }
    qmTime& operator-=( qmTime& t ) { tsec -= t.tsec; return *this; }
    
};

#endif /* _qmTime_H_ */
