// qmdMatrix.c++

#include <math.h>
#include "qmdMatrix.h"
#include "qmdCompileOptions.h"
#include "qmdMinMax.h"

qmdMatrix::qmdMatrix( int h, int w )
  : qmdMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( h, w );
}

qmdMatrix::qmdMatrix( int h, int w, reald *data )
  : qmdMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( h, w, data );
}

qmdMatrix::qmdMatrix( const qmdMatrix& m )
  : qmdMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( m );
}

qmdMatrix::qmdMatrix( const qmdMatrixRef& m )
  : qmdMatrixRef( 0, 0, 0 ), allocArray( 0 )
{
    set( m );
}

qmdMatrix&
qmdMatrix::set( int h, int w )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( h > 0 && w > 0 );
#endif
    if( !( allocArray && elementCount == h*w ) )
       {
	   if( allocArray )
	      delete [] allocArray;
	   elementCount  = h*w;
	   allocArray = new reald[ elementCount ];
       }
    width = w;
    array = allocArray;
    return *this;
}

qmdMatrix&
qmdMatrix::set( int h, int w, reald *data )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( h > 0 && w > 0 && data );
#endif
    set( h, w );
    setData( data );
    return *this;
}

qmdMatrix&
qmdMatrix::set( const qmdMatrixRef& qmd )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( qmd.elementCount > 0 );
#endif
    if( !( allocArray && 
	   elementCount == qmd.elementCount &&
	   width == qmd.width ) )
       {
	   if( allocArray && elementCount != qmd.elementCount )
	      delete [] allocArray, allocArray = 0;
	   if( !allocArray )
	      allocArray = new reald[ qmd.elementCount ];
 	   elementCount = qmd.elementCount;
	   width = qmd.width;
	   array = allocArray;
       }
    copy( qmd );
    return *this;
}

