// qmdMatrixRef.h++

#ifndef _qmdMatrixRef_H_
#define _qmdMatrixRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include "qmdreald.h"

class qmdVectorRef;

class qmdMatrixRef
{
    friend class qmdVectorRef;
    friend class qmdMatrix;

protected:

    int elementCount;
    int width;
    reald *array;

public:

    virtual ~qmdMatrixRef() {}
    qmdMatrixRef( )
      : elementCount(0), width(0), array(0) {}
    qmdMatrixRef( int h, int w, reald* a )
      : elementCount(h*w), width(w), array(a) {}
    qmdMatrixRef( const qmdMatrixRef& mr )
      : elementCount( 0 ), width( 0 ), array( 0 ) { set( mr ); }

    qmdMatrixRef& set( int h, int w, reald* a );		// shadow
    qmdMatrixRef& set( const qmdMatrixRef& );		// shadow
    qmdMatrixRef& copy( const qmdMatrixRef& );		// copy
    qmdMatrixRef& operator=( const qmdMatrixRef& );	// shadow

    int getElementCount() const { return elementCount; }
    int getRowCount() const { return getHeight(); }
    int getColCount() const { return getWidth(); }
    int getHeight() const { return( (width) ? elementCount/width : 0); }
    int getWidth() const { return width; }

    reald* getArray( int r=0, int c=0 ) {return( array + r*width + c ); }
    operator reald*() { return array; }
    reald& operator()( int r, int c ) { return array[r*width + c]; }
    reald getValue( int r, int c ) const { return array[r*width + c]; }

    void setData( reald * );
    void getData( reald * ) const;
    qmdMatrixRef& operator<<( reald *p ) { setData(p); return *this; }
    void operator>>( reald *p ) { getData(p); }

    qmdMatrixRef& setAdd( const qmdMatrixRef&, const qmdMatrixRef& );
    qmdMatrixRef& setSubtract(const qmdMatrixRef&, const qmdMatrixRef& );
    qmdMatrixRef& setMultiply(const qmdMatrixRef&, const qmdMatrixRef& );
    qmdMatrixRef& setMultiplyT( const qmdMatrixRef&, const qmdMatrixRef& );
    qmdMatrixRef& setTMultiply( const qmdMatrixRef&, const qmdMatrixRef& );
    qmdMatrixRef& setDivide( const qmdMatrixRef&, const qmdMatrixRef& );
    qmdMatrixRef& setNegate( const qmdMatrixRef& );
    qmdMatrixRef& setTranspose( const qmdMatrixRef& );
    qmdMatrixRef& setShift( reald, const qmdMatrixRef& );
    qmdMatrixRef& setScale( reald, const qmdMatrixRef& );
    qmdMatrixRef& setScaleAddEq( reald, const qmdMatrixRef& );
    qmdMatrixRef& setAddEq( const qmdMatrixRef& op );
    qmdMatrixRef& setSubtractEq( const qmdMatrixRef& op );
    qmdMatrixRef& setNegateEq();
    qmdMatrixRef& setInvert( const qmdMatrixRef& );
    qmdMatrixRef& setInvertEq();
    int setInvertCheck( const qmdMatrixRef& );
    int setInvertEqCheck();
    qmdMatrixRef& setShiftEq( reald val );
    qmdMatrixRef& setScaleEq( reald val );
    qmdMatrixRef& setConst( reald val );
    qmdMatrixRef& setZero() { return setConst( 0. ); }
    qmdMatrixRef& setDiagonal( reald val );
    qmdMatrixRef& setDiagonal( const qmdVectorRef& );
    void getDiagonal( const qmdVectorRef& );
    qmdMatrixRef& setIdentity() { return setDiagonal(1.); }
    qmdMatrixRef& setOuterProduct( const qmdVectorRef&, const qmdVectorRef& );
    qmdMatrixRef& setAddScaledOuterProductEq( const qmdVectorRef&,
					     const qmdVectorRef&,
					     reald=1.);
    qmdMatrixRef& setVectorScale( const qmdVectorRef&, const qmdMatrixRef& );
    qmdMatrixRef& operator+=(const qmdMatrixRef& op) {return setAddEq(op); }
    qmdMatrixRef& operator-=(const qmdMatrixRef& op) {return setSubtractEq(op);}
    qmdMatrixRef& operator+=(const reald val) { return setShiftEq(val); }
    qmdMatrixRef& operator-=(const reald val) { return setShiftEq(-val); }
    qmdMatrixRef& operator*=(const reald val) { return setScaleEq(val); }
    qmdMatrixRef& operator/=(const reald val) { return setScaleEq(reald(1)/val); }
    qmdMatrixRef& operator=( const reald val ) { return setConst(val); }
    int operator==( const qmdMatrixRef& ) const;
    reald getDeterminant3() const;
    reald getDeterminant2() const;
    reald getDeterminant() const;
    ostream& print( ostream&, char* = "\n" ) const;
    istream& scan( istream& );
    ostream& write( ostream& ) const;
    istream& read( istream& );
    // friend ostream& operator<<( ostream&, qmdMatrixRef& );
    const qmdVectorRef& getColumn( int i ) const;
    const qmdVectorRef& getRow( int i ) const;
    void setColumn( int i, const qmdVectorRef& );
    void setRow( int i, const qmdVectorRef& );

    inline reald getTrace() const;

 public:
    inline reald* _getArray(int r=0,int c=0) const {return( array+r*width+c );}
    inline reald& _el( int r, int c ) const { return( array[r*width+c] ); }

};

inline
reald
qmdMatrixRef::getTrace() const
{
    int dim = ( width <= getHeight() ) ? width : getHeight();
    reald sum = 0;
    for( int i = 0; i < dim; i++ )
       sum += getValue( i, i );
    return sum;
}

extern ostream& operator<<( ostream&, const qmdMatrixRef& );
extern istream& operator>>( istream&, qmdMatrixRef& );

#endif /* _qmdMatrixRef_H_ */
