// qmdSubMatrixRef.h++

#ifndef _qmdSubMatrixRef_H_
#define _qmdSubMatrixRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include "qmdreald.h"
#include "qmdSubVectorRef.h"
#include "qmdMatrixRef.h"

class qmdSubMatrixRef
{
    friend class qmdSubVectorRef;
    
protected:
    
    int elementCount;
    int width;
    int spacing;
    reald *array;
    
public:
    
    qmdSubMatrixRef( )
      : elementCount(0), width(0), spacing(0), array(0) {}
    qmdSubMatrixRef( int h, int w, int line, reald* a )
      : elementCount(h*w), width(w), spacing(line-w), array(a) {}
    qmdSubMatrixRef( const qmdMatrixRef&, int h=0, int w=0, int r=0, int c=0 );
    qmdSubMatrixRef& copy( const qmdSubMatrixRef& sm );
    qmdSubMatrixRef& operator=( const qmdSubMatrixRef& sm ) { return copy(sm); }
    int getElementCount() const { return elementCount; }
    int getRowCount() const { return getHeight(); }
    int getColCount() const { return getWidth(); }
    int getHeight() const { return( (width) ? elementCount/width : 0); }
    int getWidth() const { return width; }
    int getRowSpacing() const { return width+spacing; }
    int getColSpacing() const { return 1; }
    reald* getArray( int r = 0, int c = 0 ) const { return( array + r*(width+spacing) + c ); }
    
    qmdSubMatrixRef& setData( reald * );
    reald* getData( reald * ) const;
    qmdSubMatrixRef& operator<<( reald *p ) { return setData(p); }
    reald* operator>>( reald *p ) { return getData(p); }
    
#if 0
    qmdSubMatrixRef qmdSubMatrixRef();
    qmdSubMatrixRef qmdSubMatrixRef( int ulr, int ulc, int dh, int dw );
#endif
    
    qmdSubMatrixRef& setAdd( const qmdSubMatrixRef&, const qmdSubMatrixRef& );
    qmdSubMatrixRef& setSubtract(const qmdSubMatrixRef&, const qmdSubMatrixRef& );
    qmdSubMatrixRef& setMultiply(const qmdSubMatrixRef&, const qmdSubMatrixRef& );
    qmdSubMatrixRef& setMultiplyT(const qmdSubMatrixRef&, const qmdSubMatrixRef& );
    // qmdSubMatrixRef& setDivide( const qmdSubMatrixRef&, const qmdSubMatrixRef& );
    // qmdSubMatrixRef& setNegate( const qmdSubMatrixRef& );
    qmdSubMatrixRef& setInvert( const qmdSubMatrixRef& );
    // qmdSubMatrixRef& Transpose( const qmdSubMatrixRef& );
    // qmdSubMatrixRef& setShift( const qmdSubMatrixRef&, reald val );
    // qmdSubMatrixRef& setScale( const qmdSubMatrixRef&, reald val  );
    qmdSubMatrixRef& setAddEq( const qmdSubMatrixRef& op );
    // qmdSubMatrixRef& setSubtractEq( const qmdSubMatrixRef& op );
    // qmdSubMatrixRef& setMultiplyEq( const qmdSubMatrixRef& op );
    // qmdSubMatrixRef& setDivideEq( const qmdSubMatrixRef& op );
    // qmdSubMatrixRef& setNegateEq();
    qmdSubMatrixRef& setInvertEq();
    // qmdSubMatrixRef& setShiftEq( reald val );
    qmdSubMatrixRef& setScaleEq( reald val );
    qmdSubMatrixRef& setConst( reald val );
    qmdSubMatrixRef& setDiagonal( reald val );
    qmdSubMatrixRef& setIdentity() { return setDiagonal( 1. ); }
    qmdSubMatrixRef& setZero() { return setConst( 0. ); }
    qmdSubMatrixRef& operator+=( const qmdSubMatrixRef& op ) { return setAddEq(op); }
    // qmdSubMatrixRef& operator-=( const qmdSubMatrixRef& op ) { return setSubtractEq(op); }
    // qmdSubMatrixRef& operator*=( const qmdSubMatrixRef& op ) { return setMultiplyEq(op); }
    // qmdSubMatrixRef& operator/=( const qmdSubMatrixRef& op ) { return setDivideEq(op); }
    // qmdSubMatrixRef& operator+=( const reald val ) { return setShiftEq(val); }
    // qmdSubMatrixRef& operator-=( const reald val ) { return setShiftEq(-val); }
    // qmdSubMatrixRef& operator*=( const reald val ) { return setScaleEq(val); }
    qmdSubMatrixRef& operator/=( const reald val ) { return setScaleEq(reald(1)/val); }
    // qmdSubMatrixRef& operator=( const reald val ) { return setConst(val); }
    
    ostream& print( ostream&, char* = "\n" );
    istream& scan( istream& );
    friend ostream& operator<<( ostream&, qmdSubMatrixRef& );
    ostream& write( ostream& );
    istream& read( istream& );
    
    reald& operator()( int r, int c ) { return array[r*width + c]; }
    
private:
    int FindMaxPivot( int, int * );
    int LUDecomp( int * );
    void ForwardSubstitute( reald *, reald *, int * );
    void BackwardSubstitute( reald *, reald *, int * );
    
    friend class qmdMatrix;
};

#endif /* _qmdSubMatrixRef_H_ */
