// qmdSubVectorRef.c++

#include "qmdSubVectorRef.h"

qmdSubVectorRef&
qmdSubVectorRef::copy( const qmdSubVectorRef& sv )
{
    register reald *to = array;
    register reald *fm = sv.array;
    for( int i = 0; i < elementCount; i++ )
       (*to++) = (*fm++), to += spacing, fm += sv.spacing;
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setData( reald *p )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) = *p++, to += spacing;
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setAdd( const qmdSubVectorRef& a, const qmdSubVectorRef& b )
{
    register int count = elementCount;
    register reald *to = array;
    register reald *fm1 = a.array;
    register reald *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) + (*fm2++),
	to += spacing,
	fm1 += a.spacing,
	fm2 += b.spacing;
    return *this;
    
}

qmdSubVectorRef&
qmdSubVectorRef::setSubtract( const qmdSubVectorRef& a, const qmdSubVectorRef& b )
{
    register int count = elementCount;
    register reald *to = array;
    register reald *fm1 = a.array;
    register reald *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) - (*fm2++),
	to += spacing,
	fm1 += a.spacing,
	fm2 += b.spacing;
    return *this;
    
}

qmdSubVectorRef&
qmdSubVectorRef::setMultiply( const qmdSubMatrixRef& a, const qmdSubVectorRef& b )
{
    register reald *to = array;
    register reald *fma;
    register reald *fmb;
    int line = 1 + spacing;
    int bline = 1 + b.spacing;
    for( int i = 0; i < elementCount; i++ )
       {
	   reald sum = 0.;
	   fma = a.getArray( i, 0 );
	   fmb = b.array;
	   for( int k = 0; k < a.width; k++ )
	      sum += (*fma++) * (*fmb), fmb += bline;
	   *to = sum;
	   to += line;
       }
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setAddEq( const qmdSubVectorRef& op )
{
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = op.array;
    while( count-- )
       (*to++) += *fm++,
	to += spacing,
	fm += op.spacing;
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setSubtractEq( const qmdSubVectorRef& op )
{
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = op.array;
    while( count-- )
       (*to++) -= *fm++,
	to += spacing,
	fm += op.spacing;
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setShiftEq( reald val )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) += val,
	to += spacing;
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setScaleEq( reald val )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) *= val,
	to += spacing;
    return *this;
}

qmdSubVectorRef&
qmdSubVectorRef::setConst( reald val )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) = val,
	to += spacing;
    return *this;
}

reald
qmdSubVectorRef::getInnerProduct( const qmdSubVectorRef& op ) const
{
    reald out = 0.;
    register int count = elementCount;
    register reald *fm1 = array;
    register reald *fm2 = op.array;
    while( count-- )
       out += (*fm1++) * (*fm2++),
	fm1 += spacing,
	fm2 += op.spacing;
    return out;
}

int
qmdSubVectorRef::operator==( const qmdSubVectorRef& op ) const
{
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = op.array;
    while( count-- )
       if( (*to++) != (*fm++) )
	  return 0;
       else
	  to += spacing, fm += spacing;
    return 1;
}

qmdSubVectorRef&
qmdSubVectorRef::setAbsolute( )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       {
	   if( *to < 0 )
	      *to = (-(*to));
	   to++, to += spacing;
       }
    return *this;
}

ostream&
qmdSubVectorRef::print( ostream& os )
{
    register int count = elementCount;
    register reald *fm = array;
    while( count-- )
       os << " " << *fm++, fm += spacing;
    return os;
}

istream&
qmdSubVectorRef::scan( istream& is )
{
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       is >> *to++,
	to += spacing;
    return is;
}

ostream&
operator<<( ostream& os, qmdSubVectorRef& sv )
{
    return sv.print( os );
}

istream&
operator>>( istream& is, qmdSubVectorRef& sv )
{
    return sv.scan( is );
}

