// qmdSubVectorRef.h++

#ifndef _qmdSubVectorRef_H_
#define _qmdSubVectorRef_H_

#include <stddef.h>
#include <iostream.h>
#include <assert.h>
#include "qmdreald.h"
#include "qmdSubMatrixRef.h"

class qmdSubVectorRef
{
    friend class qmdSubMatrixRef;
    
protected:
    
    reald *array;
    int elementCount;
    int spacing; // if 0, then packed
    
public:
    
    qmdSubVectorRef( )
      : array(0), elementCount(0) {}
    qmdSubVectorRef( int n, int s, reald* a )
      : array(a), elementCount(n), spacing(s) {}
    qmdSubVectorRef& copy( const qmdSubVectorRef& sv );
    qmdSubVectorRef& operator=( const qmdSubVectorRef& sv ) { return copy(sv); }
    int getLength() const { return elementCount; }
    int getElementCount() const { return elementCount; }
    int getSpacing() const { return spacing; }
    reald* getArray( int i = 0 ) { return( array + i ); }
    reald& operator()( int i ) { return array[i]; }
    qmdSubVectorRef& setData( reald *p );
#if 0
    reald* getData( reald *p );
    qmdSubVectorRef& operator<<( reald *p ) { setData(p); return *this; }
    reald* operator>>( reald *p ) { return getData(p); }
#endif
    qmdSubVectorRef& setAdd( const qmdSubVectorRef&, const qmdSubVectorRef& );
    qmdSubVectorRef& setSubtract(const qmdSubVectorRef&, const qmdSubVectorRef& );
    // qmdSubVectorRef& setMultiply(const qmdSubVectorRef&, const qmdSubVectorRef& );
    qmdSubVectorRef& setMultiply( const qmdSubMatrixRef&, const qmdSubVectorRef& );
    // qmdSubVectorRef& setDivide( const qmdSubVectorRef&, const qmdSubVectorRef& );
    // qmdSubVectorRef& setNegate( const qmdSubVectorRef& );
    // qmdSubVectorRef& setInvert( const qmdSubVectorRef& );
    // qmdSubVectorRef& Transpose( const qmdSubVectorRef& );
    // qmdSubVectorRef& setShift( const qmdSubVectorRef&, reald val );
    // qmdSubVectorRef& setScale( const qmdSubVectorRef&, reald val  );
    qmdSubVectorRef& setAddEq( const qmdSubVectorRef& op );
    qmdSubVectorRef& setSubtractEq( const qmdSubVectorRef& op );
    // qmdSubVectorRef& setMultiplyEq( const qmdSubVectorRef& op );
    // qmdSubVectorRef& setDivideEq( const qmdSubVectorRef& op );
    // qmdSubVectorRef& setNegateEq();
    // qmdSubVectorRef& setInvertEq();
    qmdSubVectorRef& setShiftEq( reald val );
    qmdSubVectorRef& setScaleEq( reald val );
    qmdSubVectorRef& setConst( reald val );
    qmdSubVectorRef& setZero() { return setConst( 0. ); }
    qmdSubVectorRef& operator+=( const qmdSubVectorRef& op ) { return setAddEq(op); }
    // qmdSubVectorRef& operator-=( const qmdSubVectorRef& op ) { return setSubtractEq(op); }
    // qmdSubVectorRef& operator*=( const qmdSubVectorRef& op ) { return setMultiplyEq(op); }
    // qmdSubVectorRef& operator/=( const qmdSubVectorRef& op ) { return setDivideEq(op); }
    // qmdSubVectorRef& operator+=( const reald val ) { return setShiftEq(val); }
    qmdSubVectorRef& operator-=( const reald val ) { return setShiftEq(-val); }
    // qmdSubVectorRef& operator*=( const reald val ) { return setScaleEq(val); }
    qmdSubVectorRef& operator/=( const reald val ) { return setScaleEq(reald(1)/val); }
    // qmdSubVectorRef& operator=( const reald val ) { return setConst(val); }
    reald getInnerProduct( const qmdSubVectorRef& ) const;
    int operator==( const qmdSubVectorRef& ) const;
    qmdSubVectorRef& setAbsolute();
    ostream& print( ostream& );
    istream& scan( istream& );
    ostream& write( ostream& );
    istream& read( istream& );
    friend ostream& operator<<( ostream&, qmdSubVectorRef& );
};

#endif /* _qmdSubVectorRef_H_ */
