// qmdVector.c++

#include "qmdVector.h"
#include "qmdVectorRef.h"
#include "qmdCompileOptions.h"

qmdVector::qmdVector( int n )
  : qmdVectorRef( ), allocArray( 0 )
{
  set( n );
}

qmdVector::qmdVector( int n, reald *data )
  : qmdVectorRef( ), allocArray( 0 )
{
  set( n, data );
}

qmdVector::qmdVector( const qmdVector& v )
  : qmdVectorRef( ), allocArray( 0 )
{
  set( v );
}

qmdVector::qmdVector( const qmdVectorRef& v )
  : qmdVectorRef( ), allocArray( 0 )
{
  set( v );
}

qmdVector&
qmdVector::set( int n )
{
#ifdef qmdDoBoundaryChecking
  qmdAssert( n > 0 );
#endif
  if( !( allocArray && elementCount == n ) )
    {
      if( allocArray )
	delete [] allocArray;
      elementCount = n;
      allocArray = new reald[ elementCount ];
    }
  array = allocArray;
  return *this;
}

qmdVector&
qmdVector::set( int n, reald *data )
{
#ifdef qmdDoBoundaryChecking
  qmdAssert( n > 0 );
  qmdAssert( data != (reald *) NULL );
#endif
  set( n );
  setData( data );
  return *this;
}

qmdVector&
qmdVector::set( const qmdVectorRef& qv )
{
#ifdef qmdDoBoundaryChecking
  qmdAssert( qv.elementCount > 0 );
  qmdAssert( qv.array != (reald *) NULL );
#endif
  if( !( allocArray && elementCount == qv.getElementCount() ) )
    {
      if( allocArray )
	delete [] allocArray;
      allocArray = new reald[ qv.getElementCount() ];
      elementCount = qv.getElementCount();
      array = allocArray;
    }
  copy( qv );
  return *this;
}
