// qmdVectorRef.c++

#include "qmdVectorRef.h"
#include "qmdCompileOptions.h"
#include <memory.h>


qmdVectorRef&
qmdVectorRef::set( int n, reald* a )
{
    elementCount = n;
    array = a;
    return *this;
}

qmdVectorRef&
qmdVectorRef::set( const qmdVectorRef& vr )
{
    memcpy( this, &vr, sizeof( qmdVectorRef ) );
    return *this;
}

qmdVectorRef&
qmdVectorRef::copy( const qmdVectorRef& vr )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount && elementCount == vr.elementCount );
#endif
    memcpy( array, vr.array, elementCount * sizeof(reald) );
    return( *this );
}

qmdVectorRef&
qmdVectorRef::operator=( const qmdVectorRef& vr )
{
#ifdef qmdWarnings
    cerr << "qmdVectorRef: WARNING: operator=() does shadow, not copy" << endl;
    cerr << "                      use set or copy instead" << endl;
#endif
    return set( vr );
}

void
qmdVectorRef::setData( const reald *p )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( array != (reald *) NULL );
#endif
    register int count = elementCount;
    register reald *to = array;
    while( count-- )
       (*to++) = *p++;
}

void
qmdVectorRef::getData( reald *p ) const
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( array != (reald *) NULL) ;
#endif
    register int count = elementCount;
    register reald *fm = array;
    while( count-- )
       *p++ = *fm++;
}

qmdVectorRef&
qmdVectorRef::setAdd( const qmdVectorRef& a, const qmdVectorRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( elementCount == a.elementCount );
    qmdAssert( a.elementCount == b.elementCount );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm1 = a.array;
    register reald *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) + (*fm2++);
    return *this;
}

qmdVectorRef&
qmdVectorRef::setSubtract( const qmdVectorRef& a, const qmdVectorRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( elementCount == a.elementCount );
    qmdAssert( a.elementCount == b.elementCount );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm1 = a.array;
    register reald *fm2 = b.array;
    while( count-- )
       (*to++) = (*fm1++) - (*fm2++);
    return *this;  
}

qmdVectorRef&
qmdVectorRef::setNegate( const qmdVectorRef& v )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *from = v.array;
    while( count-- )
       (*to++) = -(*from++);
    return *this;
}

qmdVectorRef&
qmdVectorRef::setShift( reald val, const qmdVectorRef& v )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *from = v.array;
    while( count-- )
       (*to++) = val + (*from++);
    return *this;
}

qmdVectorRef&
qmdVectorRef::setScale( reald val, const qmdVectorRef& v )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 &&
	      elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *from = v.array;
    while( count-- )
       (*to++) = val * (*from++);
    return *this;
}

qmdVectorRef&
qmdVectorRef::setScaleAddEq( reald w, const qmdVectorRef& v )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 && elementCount == v.elementCount);
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *from = v.array;
    while( count-- )
       (*to++) += w * (*from++);
    return *this;
}

qmdVectorRef&
qmdVectorRef::setAddEq( const qmdVectorRef& op )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( elementCount == op.elementCount );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = op.array;
    while( count-- )
       (*to++) += *fm++;
    return *this;
}

qmdVectorRef&
qmdVectorRef::setSubtractEq( const qmdVectorRef& op )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( elementCount == op.elementCount );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = op.array;
    while( count-- )
       (*to++) -= *fm++;
    return *this;
}

qmdVectorRef&
qmdVectorRef::setNegateEq()
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *to = array;
    register reald *fm = array;
    while( count-- )
       (*to++) = -(*fm++);
    return *this;
}

qmdVectorRef&
qmdVectorRef::setShiftEq( reald val )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       (*rp++) += val;
    return *this;
}

qmdVectorRef&
qmdVectorRef::setScaleEq( reald val )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       (*rp++) *= val;
    return *this;
}

qmdVectorRef&
qmdVectorRef::setConst( reald val )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       (*rp++) = val;
    return *this;
}

qmdVectorRef&
qmdVectorRef::setMultiply( const qmdMatrixRef& m, const qmdVectorRef& v )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( v.getLength() == m.getWidth() &&
	      v.getLength() * getLength() == m.getElementCount() );
    // (do not use getHeight(), because you can get divide by zero)
#endif
    register reald *to = getArray();
    register reald *fmM = m._getArray();
    register reald *fmV;
    register int count;
    for( int row = 0; row < getLength(); row++ )
       {
	   *to = 0.;
	   fmV = v._getArray();
	   count = m.getWidth();
	   while( count-- )
	      *to += (*fmM++) * (*fmV++);
	   to++;
       }
    return *this;
}

qmdVectorRef&
qmdVectorRef::setTMultiply( const qmdMatrixRef& m, const qmdVectorRef& v )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( getLength() == m.getWidth() &&
	      v.getLength() * getLength() == m.getElementCount() );
#endif
    register reald *to = getArray();
    register reald *fmM;
    register reald *fmV;
    register int count;
    for( int col = 0; col < getLength(); col++ )
       {
	   *to = 0.;
	   fmV = v._getArray();
	   fmM = m._getArray( 0, col );
	   count = v.getLength();
	   while( count-- )
	      *to += (*fmM) * (*fmV++), fmM+=m.getWidth();
	   to++;
       }
    return *this;
}

qmdVectorRef&
qmdVectorRef::setVectorScale( const qmdVectorRef& a, const qmdVectorRef& b )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( getLength() == a.getLength() &&
	      a.getLength() == b.getLength() );
#endif
    register reald *to = getArray();
    register reald *fmA = a._getArray();
    register reald *fmB = b._getArray();
    register int count = elementCount;
    while( count-- )
       (*to++) = (*fmA++) * (*fmB++);
    return *this;
}

reald
qmdVectorRef::getInnerProduct( const qmdVectorRef& op ) const
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
    qmdAssert( elementCount == op.elementCount );
#endif
    reald out = 0.;
    register int count = elementCount;
    register reald *fm1 = array;
    register reald *fm2 = op.array;
    while( count-- )
       out += (*fm1++) * (*fm2++);
    return out;
}

int
qmdVectorRef::operator==( const qmdVectorRef& op ) const
{
    if( elementCount != op.elementCount )
       return 0;
    for( int i = 0; i < elementCount; i++ )
       if( op.array[i] != array[i] )
	  return 0;
    return 1;
}

qmdVectorRef&
qmdVectorRef::setAbsolute( )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       {
	   if( *rp < 0 )
	      *rp = (-(*rp));
	   rp++;
       }
    return *this;
}

ostream&
qmdVectorRef::print( ostream& os ) const
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       os << " " << *rp++;
    return os;
}

istream&
qmdVectorRef::scan( istream& is )
{
#ifdef qmdDoBoundaryChecking
    qmdAssert( elementCount > 0 );
#endif
    register int count = elementCount;
    register reald *rp = array;
    while( count-- )
       is >> *rp++;
    return is;
}

ostream&
qmdVectorRef::write( ostream& os ) const
{
    return os.write( (char *)array, elementCount * sizeof(reald) );
}


istream&
qmdVectorRef::read( istream& is )
{
    return is.read( (char *)array, elementCount * sizeof(reald) );
}

qmdVectorRef&
qmdVectorRef::setCross( const qmdVectorRef& a, const qmdVectorRef& b )
{
    assert( 3==elementCount && 3==a.elementCount && 3==b.elementCount );
    array[ 0 ] = a.getValue(1)*b.getValue(2) - a.getValue(2)*b.getValue(1);
    array[ 1 ] = a.getValue(2)*b.getValue(0) - a.getValue(0)*b.getValue(2);
    array[ 2 ] = a.getValue(0)*b.getValue(1) - a.getValue(1)*b.getValue(0);
    return *this;
}


ostream&
operator<<( ostream& os, const qmdVectorRef& vr ) { return vr.print( os ); }

istream&
operator>>( istream& is, qmdVectorRef& vr ) { return vr.scan( is ); }

