#include "glos.h"
#include "render.h"
#include <stdio.h>
//#include "math.h"

HPALETTE ghpalOld, ghPalette = (HPALETTE) 0;
RECT oldrect;

#ifdef __cplusplus 
extern "C" {
#endif    

unsigned char threeto8[8] = {
    0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377
};

unsigned char twoto8[4] = {
    0, 0x55, 0xaa, 0xff
};

unsigned char oneto8[2] = {
    0, 255
};

static int defaultOverride[13] = {
    0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91
};

static PALETTEENTRY defaultPalEntry[20] = {
    { 0,   0,   0,    0 },
    { 0x80,0,   0,    0 },
    { 0,   0x80,0,    0 },
    { 0x80,0x80,0,    0 },
    { 0,   0,   0x80, 0 },
    { 0x80,0,   0x80, 0 },
    { 0,   0x80,0x80, 0 },
    { 0xC0,0xC0,0xC0, 0 },

    { 192, 220, 192,  0 },
    { 166, 202, 240,  0 },
    { 255, 251, 240,  0 },
    { 160, 160, 164,  0 },

    { 0x80,0x80,0x80, 0 },
    { 0xFF,0,   0,    0 },
    { 0,   0xFF,0,    0 },
    { 0xFF,0xFF,0,    0 },
    { 0,   0,   0xFF, 0 },
    { 0xFF,0,   0xFF, 0 },
    { 0,   0xFF,0xFF, 0 },
    { 0xFF,0xFF,0xFF, 0 }
};

unsigned char
ComponentFromIndex(int i, UINT nbits, UINT shift)
{
    unsigned char val;

    val = (unsigned char) (i >> shift);
    switch (nbits) {

    case 1:
        val &= 0x1;
        return oneto8[val];

    case 2:
        val &= 0x3;
        return twoto8[val];

    case 3:
        val &= 0x7;
        return threeto8[val];

    default:
        return 0;
    }
}

void CreateRGBPalette(HDC hDC)
{
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *pPal;
    int n, i;

    n = GetPixelFormat(hDC);
    DescribePixelFormat(hDC, n, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

    if (pfd.dwFlags & PFD_NEED_PALETTE) {
        n = 1 << pfd.cColorBits;
        pPal = (PLOGPALETTE)LocalAlloc(LMEM_FIXED, sizeof(LOGPALETTE) +
                n * sizeof(PALETTEENTRY));
        pPal->palVersion = 0x300;
        pPal->palNumEntries = n;
        for (i=0; i<n; i++) {
            pPal->palPalEntry[i].peRed =
                    ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
            pPal->palPalEntry[i].peGreen =
                    ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
            pPal->palPalEntry[i].peBlue =
                    ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
            pPal->palPalEntry[i].peFlags = 0;
        }

        /* fix up the palette to include the default GDI palette */
        if ((pfd.cColorBits == 8)                           &&
            (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
            (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
            (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)
           ) {
            for (i = 1 ; i <= 12 ; i++)
                pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
        }

        ghPalette = CreatePalette(pPal);
        LocalFree(pPal);

        ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
        n = RealizePalette(hDC);
    }
}


BOOL SetupPixelFormat(HDC hDC)
{
    static PIXELFORMATDESCRIPTOR pfd = {
	sizeof(PIXELFORMATDESCRIPTOR),	// size of this pfd
	1,				// version number
	PFD_DRAW_TO_WINDOW |		// support window
	  PFD_SUPPORT_OPENGL |		// support OpenGL
	  PFD_DOUBLEBUFFER,		// double buffered
	PFD_TYPE_RGBA,			// RGBA type
	24,				// 24-bit color depth
	0, 0, 0, 0, 0, 0,		// color bits ignored
	0,				// no alpha buffer
	0,				// shift bit ignored
	0,				// no accumulation buffer
	0, 0, 0, 0, 			// accum bits ignored
	32,				// 32-bit z-buffer	
	0,				// no stencil buffer
	0,				// no auxiliary buffer
	PFD_MAIN_PLANE,			// main layer
	0,				// reserved
	0, 0, 0				// layer masks ignored
    };
    int pixelformat;

    if ( (pixelformat = ChoosePixelFormat(hDC, &pfd)) == 0 )
    {
        MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat(hDC, pixelformat, &pfd) == FALSE)
    {
        MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    CreateRGBPalette(hDC);

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////
////   Externally Visible Methods  ////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
#define MY_ICOSAHEDRON	1
#define MY_CUBE			2
#define MY_SPHERE		3
#define X  0.525731112119133606
#define Z  0.850650808352039932
static GLfloat vdata[12][3] = {
	{-X,0.0,Z},{X,0.0,Z},{-X,0.0,-Z},{X,0.0,-Z},
	{0.0,Z,X},{0.0,Z,-X},{0.0,-Z,X},{0.0,-Z,-X},
	{Z,X,0.0},{-Z,X,0.0},{Z,-X,0.0},{-Z,-X,0.0}
};

static GLint tindices[20][3] = {
	{0,4,1},{0,9,4},{9,5,4},{4,5,8},{4,8,1},
	{8,10,1},{8,3,10},{5,3,8},{5,2,3},{2,7,3},
	{7,10,3},{7,6,10},{7,11,6},{11,0,6},{0,1,6},
	{6,1,10},{9,0,11},{9,11,2},{9,2,5},{7,2,11}
};

static GLfloat cubeVerts[8][3] = {
	{0.0,0.0,0.0},{1.0,0.0,0.0},{1.0,0.0,-1.0},{0.0,0.0,-1.0},
	{0.0,1.0,0.0},{1.0,1.0,0.0},{1.0,1.0,-1.0},{0.0,1.0,-1.0}
};

static GLint cubeInds[12][3] = {
	{0,1,5},{0,5,4}, 
	{1,2,6},{1,6,5},
	{2,3,7},{2,7,6},
	{3,0,4},{3,4,7},
	{4,5,6},{4,6,7},
	{0,1,2},{0,2,3}
};

#define I 0.5
static GLfloat cubeColors[12][3] = {
	{I,0.0,0.0},{I,0.0,0.0},
	{0.0,I,0.0},{0.0,I,0.0},
	{0.0,0.0,I},{0.0,0.0,I},
	{I,I,0.0},{I,I,0.0},
	{0.0,I,I},{0.0,I,I},
	{I,0.0,I},{I,0.0,I}
};


static float viewAngle = 20.0;
static float Rotx = 0.0;
static float Roty = 0.0;
static float Rotz = 0.0;
static float InitRotx = 0.0;
static float InitRoty = 0.0;
static float InitRotz = 0.0;
static float Distx = 0.0;
static float Disty = 0.0;
static float Distz = 0.0;
static float Angle = 0;
static float InitAngle = 0;
static float dAngle = 5;
static BOOL Cube = FALSE;
static BOOL Icos = TRUE;
static float hItems = 1;
static float vItems = 1;
static int arm_time=0, forearm_time=0;

DllExport void RenderInitialize(HWND hWnd)
{
    GLfloat	maxObjectSize, aspect;
    GLdouble	near_plane, far_plane;
	int i;

	GLfloat mat_specular[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat mat_shininess[] = {50};

	GLfloat light_position[] = {1.0, 1.0, 1.0, 0.0};
	GLfloat light_diffuse_color[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat light_ambient_color[] = {0.7, 0.7, 0.7, 1.0};
	GLfloat light_specular_color[] = {0.0, 0.0, 0.0, 1.0};

	GLfloat light_position1[] = {-10.0, 10.0, -10.0, 0.0};
	GLfloat light_diffuse_color1[] = {1.0, 1.0, 1.0};
	GLfloat light_specular_color1[] = {0.0, 0.0, 0.0};

	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse_color);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient_color);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular_color);
	glEnable(GL_LIGHT0);


	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_LIGHTING);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	GetClientRect(hWnd, &oldrect);


    glClearColor( 0.0, 0.0, 0.0, 1.0 );
    glClearDepth( 1.0 );
	
    glMatrixMode( GL_PROJECTION );
    aspect = (GLfloat) oldrect.right / oldrect.bottom;
	glLoadIdentity();
    gluPerspective( viewAngle, aspect, 1.0, 10000.0 );
    glMatrixMode( GL_MODELVIEW );

    near_plane = 1.0;
    far_plane = 10.0;
    maxObjectSize = 3.0;

	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);

///////////// Initialize the display lists


	


}

DllExport void RenderResize(HWND hWnd)
{
    RECT	rect;

    GetClientRect(hWnd, &rect);

    glViewport(0, 0, rect.right, rect.bottom);

    if((oldrect.right > rect.right) || (oldrect.bottom > rect.bottom))
	PostMessage (hWnd, WM_PAINT, 0, 0L);

    oldrect.right = rect.right;
    oldrect.bottom = rect.bottom;
}

DllExport void RenderCreate(HWND hWnd)
{
	HDC	  hDC;
	HGLRC hRC;

    hDC = GetDC(hWnd);
	SetupPixelFormat(hDC);

    hRC = wglCreateContext( hDC );
    wglMakeCurrent( hDC, hRC );
}

DllExport void RenderPaint(HWND hWnd)
{
	HDC		hDC;
	PAINTSTRUCT	ps;

	hDC = BeginPaint(hWnd, &ps);

//	RenderDrawScene(hWnd);

	EndPaint(hWnd, &ps);
}


// The WM_QUERYNEWPALETTE message informs a window that it is about to
// receive input focus. In response, the window receiving focus should
// realize its palette as a foreground palette and update its client
// area. If the window realizes its palette, it should return TRUE;
// otherwise, it should return FALSE.

DllExport BOOL RenderQueryNewPallette(HWND hWnd)
{
	HDC     hDC;

    if(ghPalette)
    {
        hDC = GetDC(hWnd);

        // Select and realize the palette

        ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
        RealizePalette(hDC);

        // Redraw the client area

        InvalidateRect(hWnd, NULL, TRUE);
        UpdateWindow(hWnd);

        if(ghpalOld)
            SelectPalette(hDC, ghpalOld, FALSE);

        ReleaseDC(hWnd, hDC);

        return TRUE;
    }

    return FALSE;
}

// The WM_PALETTECHANGED message informs all windows that the window
// with input focus has realized its logical palette, thereby changing 
// the system palette. This message allows a window without input focus
// that uses a color palette to realize its logical palettes and update
// its client area.
//
// This message is sent to all windows, including the one that changed
// the system palette and caused this message to be sent. The wParam of
// this message contains the handle of the window that caused the system
// palette to change. To avoid an infinite loop, care must be taken to
// check that the wParam of this message does not match the window's
// handle.

DllExport void RenderPalletteChanged(WPARAM wParam, HWND hWnd)
{
    HDC         hDC; 

    // Before processing this message, make sure we
    // are indeed using a palette

    if (ghPalette)
    {
        // If this application did not change the palette, select
        // and realize this application's palette

        if (wParam != (WPARAM)hWnd)
        {
            // Need the window's DC for SelectPalette/RealizePalette

            hDC = GetDC(hWnd);

            // Select and realize our palette

            ghpalOld = SelectPalette(hDC, ghPalette, FALSE);
            RealizePalette(hDC);

            // WHen updating the colors for an inactive window,
            // UpdateColors can be called because it is faster than
            // redrawing the client area (even though the results are
            // not as good)

            UpdateColors(hDC);

            // Clean up

            if (ghpalOld)
               SelectPalette(hDC, ghpalOld, FALSE);

            ReleaseDC(hWnd, hDC);
        }
    }
}


DllExport void RenderClose(HWND hWnd)
{
    HGLRC hRC;
    HDC	  hDC;

        /* release and free the device context and rendering context */
    hRC = wglGetCurrentContext();
    hDC = wglGetCurrentDC();

    wglMakeCurrent(NULL, NULL);

    if (hRC)
    	wglDeleteContext(hRC);
    if (hDC)
    	ReleaseDC(hWnd, hDC);

}


#ifdef __cplusplus 
}
#endif    
