// SkinGrabberView.cpp : implementation of the CSkinGrabberView class
//
#include "vlImage.h"

#include "stdafx.h"
#include "SkinGrabber.h"

#include "SkinGrabberDoc.h"
#include "SkinGrabberView.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/* code comes from NRC 2nd Edition */
/* modified to use doubles instead of floats */ 

#include <stdio.h>
#include <math.h>
#define SWAP(a,b) {temp=(a);(a)=(b);(b)=temp;}
#define FREE_ARG char*
#define NR_END 1

static void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
  fprintf(stderr,"Numerical Recipes run-time error...\n");
  fprintf(stderr,"%s\n",error_text);
}

static int *ivector(long nl, long nh)
  /* allocate an int vector with subscript range v[nl..nh] */
{
  int *v;

  v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
  if (!v) nrerror("allocation failure in ivector()");
  return v-nl+NR_END;
}

static void free_ivector(int *v, long nl, long nh)
  /* free an int vector allocated with ivector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}


int gaussj(double **a, int n, double **b, int m)
{
  int *indxc,*indxr,*ipiv;
  int i,icol,irow,j,k,l,ll;
  double big,dum,pivinv,temp;

  indxc=ivector(1,n);
  indxr=ivector(1,n);
  ipiv=ivector(1,n);
  for (j=1;j<=n;j++) ipiv[j]=0;
  for (i=1;i<=n;i++) {
    big=0.0;
    for (j=1;j<=n;j++)
      if (ipiv[j] != 1)
	for (k=1;k<=n;k++) {
	  if (ipiv[k] == 0) {
	    if (fabs(a[j][k]) >= big) {
	      big=fabs(a[j][k]);
	      irow=j;
	      icol=k;
	    }
	  } else if (ipiv[k] > 1) 
	    {
	      nrerror("gaussj: Singular Matrix-1");
	      return(0);
	    }
	}
    ++(ipiv[icol]);
    if (irow != icol) {
      for (l=1;l<=n;l++) SWAP(a[irow][l],a[icol][l])
			   for (l=1;l<=m;l++) SWAP(b[irow][l],b[icol][l])
			   }
    indxr[i]=irow;
    indxc[i]=icol;
    if (a[icol][icol] == 0.0) 
      {
	nrerror("gaussj: Singular Matrix-2");
	return(0);
      }
    pivinv=1.0/a[icol][icol];
    a[icol][icol]=1.0;
    for (l=1;l<=n;l++) a[icol][l] *= pivinv;
    for (l=1;l<=m;l++) b[icol][l] *= pivinv;
    for (ll=1;ll<=n;ll++)
      if (ll != icol) {
	dum=a[ll][icol];
	a[ll][icol]=0.0;
	for (l=1;l<=n;l++) a[ll][l] -= a[icol][l]*dum;
	for (l=1;l<=m;l++) b[ll][l] -= b[icol][l]*dum;
      }
  }
  for (l=n;l>=1;l--) {
    if (indxr[l] != indxc[l])
      for (k=1;k<=n;k++)
	SWAP(a[k][indxr[l]],a[k][indxc[l]]);
  }
  free_ivector(ipiv,1,n);
  free_ivector(indxr,1,n);
  free_ivector(indxc,1,n);
  return(1);
}

/////////////////////////////////////////////////////////////////////////////
// CSkinGrabberView

IMPLEMENT_DYNCREATE(CSkinGrabberView, CFormView)

BEGIN_MESSAGE_MAP(CSkinGrabberView, CFormView)
	//{{AFX_MSG_MAP(CSkinGrabberView)
	ON_BN_CLICKED(IDC_GRAB, OnGrab)
	ON_WM_LBUTTONDOWN()
	ON_BN_CLICKED(IDC_COMMIT, OnCommit)
	ON_BN_CLICKED(IDC_RESET, OnReset)
	ON_BN_CLICKED(IDC_INPUT0, OnInput0)
	ON_BN_CLICKED(IDC_INPUT1, OnInput1)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSkinGrabberView construction/destruction

CSkinGrabberView::CSkinGrabberView()
	: CFormView(CSkinGrabberView::IDD)
{
	//{{AFX_DATA_INIT(CSkinGrabberView)
	covtext = _T("");
	meantext = _T("");
	input0 = TRUE;
	input1 = FALSE;
	//}}AFX_DATA_INIT
	// TODO: add construction code here
  ZeroMemory( &m_bmi, sizeof( BITMAPINFO ) );
  m_bmp = NULL;
  ZeroMemory( &dims, sizeof(SIZE));

  Init(0);

  ZeroMemory(mean,sizeof(mean));
  ZeroMemory(cov,sizeof(cov));
  ZeroMemory(sum,sizeof(sum));
  ZeroMemory(sum2,sizeof(sum2));
  count = 0;
}

CSkinGrabberView::~CSkinGrabberView()
{
}

void CSkinGrabberView::DoDataExchange(CDataExchange* pDX)
{
	CFormView::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CSkinGrabberView)
	DDX_Control(pDX, IDC_MEANCOLOR, meancolorview);
	DDX_Control(pDX, IDC_VIEW, view);
	DDX_Text(pDX, IDC_COV, covtext);
	DDX_Text(pDX, IDC_MEAN, meantext);
	DDX_Check(pDX, IDC_INPUT0, input0);
	DDX_Check(pDX, IDC_INPUT1, input1);
	//}}AFX_DATA_MAP
}

BOOL CSkinGrabberView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CFormView::PreCreateWindow(cs);
}

void CSkinGrabberView::OnInitialUpdate()
{
	CFormView::OnInitialUpdate();
	GetParentFrame()->RecalcLayout();
	ResizeParentToFit();

  view.SetWindowPos(NULL,5,5,width+4,height+4,SWP_NOZORDER);
  OnGrab();
}

/////////////////////////////////////////////////////////////////////////////
// CSkinGrabberView diagnostics

#ifdef _DEBUG
void CSkinGrabberView::AssertValid() const
{
	CFormView::AssertValid();
}

void CSkinGrabberView::Dump(CDumpContext& dc) const
{
	CFormView::Dump(dc);
}

CSkinGrabberDoc* CSkinGrabberView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CSkinGrabberDoc)));
	return (CSkinGrabberDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CSkinGrabberView message handlers
// byte offsets
#define BO_BLUE     0
#define BO_GREEN    1
#define BO_RED      2


BOOL CSkinGrabberView::SetImage(BYTE *data, SIZE sizes)
{

    // Hide scroll bars
    //SetScrollSizes( MM_TEXT, CSize( 1, 1 ) );
    dims = sizes;

    BITMAPINFOHEADER& bih = m_bmi.bmiHeader;
    ZeroMemory( &bih, sizeof( BITMAPINFOHEADER ) );
    bih.biSize = sizeof( BITMAPINFOHEADER );
    bih.biWidth = sizes.cx;
    bih.biHeight = -sizes.cy;
    bih.biCompression = BI_RGB;
    bih.biPlanes = 1;
    bih.biBitCount = 24;

    if( m_bmp != NULL ) delete[] m_bmp;
    int imageSize = sizes.cy*(sizes.cx*3+((4-((sizes.cx*3)&3))&3));
    m_bmp = new BYTE[ imageSize ];
    if( m_bmp == NULL ) return FALSE;

    int bidx,bidy;
    int lag = 0;
    const int lagDelta = ((4-((sizes.cx*3)&3))&3);
    for( int j = 0; j < sizes.cy; j++ )
    {
        for( int i = 0; i < sizes.cx; i++ )
        {
            bidx = j*sizes.cx*3 + i*3;
            bidy = j*sizes.cx*4 + i*4;
            m_bmp[ bidx+lag+BO_RED ] = data[ bidy+BO_RED ];
            m_bmp[ bidx+lag+BO_GREEN ] = data[ bidy+BO_GREEN ];
            m_bmp[ bidx+lag+BO_BLUE ] = data[ bidy+BO_BLUE ];
        }
        lag += lagDelta;
    }

    return TRUE;
}


void CSkinGrabberView::Draw(CDC* pDC) 
{
  if( m_bmp == NULL ) return;

  CRect rectClient;
  view.GetClientRect( &rectClient );

  CRect rectTo;
  rectTo.left = rectTo.top = 1;

  rectTo.right = rectTo.left + dims.cx;
  rectTo.bottom = rectTo.top + dims.cy;


  ::StretchDIBits( pDC->m_hDC,
                   rectTo.left, rectTo.top, rectTo.Width(), rectTo.Height(),
                   0, 0, dims.cx, dims.cy,
                   m_bmp, &m_bmi, DIB_RGB_COLORS, SRCCOPY );

}

void CSkinGrabberView::OnGrab() 
{
  SIZE size;
  grabber.grab(iref);
  //iref._tobgr( rgbImage.getPixelData() );
  size.cx = width;
  size.cy = height;
  //SetImage( rgbImage.getPixelData(), size);
  SetImage( iref.getPixelData(), size);
  InvalidateRect(NULL);
}

void CSkinGrabberView::OnDraw(CDC* pDC) 
{	
  CDC * viewDC = view.GetDC();
  Draw(viewDC);
  view.ReleaseDC(viewDC);
  view.ValidateRect(NULL);
    RECT rect;
  CDC* pDC2 = meancolorview.GetDC();
  meancolorview.GetClientRect(&rect);

/*
  float toao = 30, toas = 0.9;
  float A1 = 1.164*(mean[0]-16.0);
	float B = 2.017*(mean[1]-128.0);
	float C = -0.392*(mean[1]-128.0);
	float D = -0.813*(mean[2]-128.0);
	float E = 1.596*(mean[2]-128.0);
	float BLU = toas*( A1 + B + toao );
	float GRN = toas*( A1 + C + D + toao );
	float RED = toas*( A1 + E + toao );
	BLU = vliPixel::limit( BLU, 0, 255 );
	GRN = vliPixel::limit( GRN, 0, 255 );
	RED = vliPixel::limit( RED, 0, 255 );
  pDC2->FillSolidRect(&rect,RGB(RED,GRN,BLU));
*/
//  byte r,g,b;
//  vliPixel::yuvToRgb(mean[0],mean[1],mean[2],r,g,b);
//  pDC2->FillSolidRect(&rect,RGB(r,g,b));
    pDC2->FillSolidRect(&rect,RGB(mean[0],mean[1],mean[2]));
  /*
    unsigned char b = 1.164*(mean[0]-16)+                   +2.018*(mean[1]-128);
    unsigned char g = 1.164*(mean[0]-16)-0.813*(mean[2]-128)-0.391*(mean[1]-128);
    unsigned char r = 1.164*(mean[0]-16)+1.596*(mean[2]-128);
    if(r<0) r = 0;
    if(r>255) r = 255;
    if(g<0) g = 0;
    if(g>255) g = 255;
    if(b<0) b = 0;
    if(b>255) b = 255;
    pDC2->FillSolidRect(&rect,RGB(r,g,b));
*/

  meancolorview.ReleaseDC(pDC2);
  meancolorview.ValidateRect(NULL);
}

void CSkinGrabberView::OnLButtonDown(UINT nFlags, CPoint point) 
{
  CWnd * w = ChildWindowFromPoint(point);
  if(w==&view){
    POINT p;
    p = point;
    MapWindowPoints(&view,&p,1);
    AccumulateExample(p.x,p.y);
    DrawExample(p.x,p.y);
  }
	CFormView::OnLButtonDown(nFlags, point);
}

void CSkinGrabberView::DrawExample(int x, int y)
{
  int lag = 0;
  int bidx = 0;
  const int lagDelta = ((4-((dims.cx*3)&3))&3);

  if(y<2) y = 2;
  if(x<2) x = 2;
  if(y+2>dims.cy) y = dims.cy-2;
  if(x+2>dims.cx) x = dims.cx-2;
  for( int j = y-2; j < y+2; j++ )
  {
      for( int i = x-2; i < x+2; i++ )
      {
          bidx = j*dims.cx*3 + i*3;
          m_bmp[ bidx+lag+BO_RED ] = 0;
          m_bmp[ bidx+lag+BO_GREEN ] = 255;
          m_bmp[ bidx+lag+BO_BLUE ] = 0;
      }
      lag += lagDelta;
  }
  InvalidateRect(NULL);
}

void CSkinGrabberView::AccumulateExample(int w, int h)
{
  vliRGBPixel * p;
  byte y,u,v;
  if(h<2) h = 2;
  if(w<2) w = 2;
  if(h+2>dims.cy) h = dims.cy-2;
  if(w+2>dims.cx) w = dims.cx-2;
  for( int j = h-2; j < h+2; j++ )
  {
      for( int i = w-2; i < w+2; i++ )
      {
  
          p = iref.getPixelPtr(i,j);
/*
          vliPixel::rgbToYuv(p->r,p->g,p->b,y,u,v);

          sum[0] += y;
          sum[1] += u;
          sum[2] += v;

          sum2[0][0] += y*y;
          sum2[1][1] += u*u;
          sum2[2][2] += v*v;
          sum2[0][1] += y*u;
          sum2[0][2] += y*v;
          sum2[1][2] += u*v;
*/
  
          sum[0] += p->r;
          sum[1] += p->g;
          sum[2] += p->b;

          sum2[0][0] += p->r*p->r;
          sum2[1][1] += p->g*p->g;
          sum2[2][2] += p->b*p->b;
          sum2[0][1] += p->r*p->g;
          sum2[0][2] += p->r*p->b;
          sum2[1][2] += p->g*p->b;

          count++;
      }
  }

  

  mean[0] = sum[0]/count;
  mean[1] = sum[1]/count;
  mean[2] = sum[2]/count;

  cov[0][0] = sum2[0][0]/count - mean[0]*mean[0];
  cov[1][1] = sum2[1][1]/count - mean[1]*mean[1];
  cov[2][2] = sum2[2][2]/count - mean[2]*mean[2];
  cov[0][1] = cov[1][0] = sum2[0][1]/count - mean[0]*mean[1];
  cov[0][2] = cov[2][0] = sum2[0][2]/count - mean[0]*mean[2];
  cov[1][2] = cov[2][1] = sum2[1][2]/count - mean[1]*mean[2];

  meantext.Format("%3.2f %3.2f %3.2f",mean[0],mean[1],mean[2]);
  covtext.Format("%3.2f %3.2f %3.2f\r\n%3.2f %3.2f %3.2f\r\n%3.2f %3.2f %3.2f\r\n",
    cov[0][0],cov[0][1],cov[0][2],
    cov[1][0],cov[1][1],cov[1][2],
    cov[2][0],cov[2][1],cov[2][2]);
  UpdateData(FALSE);

}

void CSkinGrabberView::OnCommit() 
{
  static double **inv_covar = NULL;
  static double **dummy = NULL;
  if (inv_covar == (double **) NULL)
    {
      inv_covar = (double **) malloc (4 * sizeof(double*));
      dummy = (double **) malloc (4 * sizeof(double*));
      inv_covar[0] = (double *) malloc (4 * sizeof(double));
      inv_covar[1] = (double *) malloc (4 * sizeof(double));
      inv_covar[2] = (double *) malloc (4 * sizeof(double));
      inv_covar[3] = (double *) malloc (4 * sizeof(double));
      dummy[0] = (double *) malloc (1 * sizeof(double));
      dummy[1] = (double *) malloc (1 * sizeof(double));
      dummy[2] = (double *) malloc (1 * sizeof(double));
      dummy[3] = (double *) malloc (1 * sizeof(double));
      dummy[0][0] = 0.0;
      dummy[1][0] = 0.0;
      dummy[2][0] = 0.0;
      dummy[3][0] = 0.0;
  }
  
  int i;
  for (i=0; i<3; i++)
  {
    for (int j=0; j<3; j++)
	  {
	    inv_covar[i+1][j+1] =  cov[i][j];
	  }
  }

  int ret = gaussj(inv_covar, 3, dummy, 1);

  for (i=0; i<3; i++)
    {
      for (int j=0; j<3; j++)
      {
	      invcov[i][j] = inv_covar[i+1][j+1];
      }
    }
 
  CFileDialog dlg(FALSE);
  dlg.m_ofn.lpstrTitle = "Select Save File";
  if(dlg.DoModal()==IDOK){
    FILE * fp = fopen(dlg.GetPathName(),"w");
    fprintf(fp,"%f %f %f\n",mean[0],mean[1],mean[2]);
    fprintf(fp,"%f %f %f\n%f %f %f\n%f %f %f\n",
    invcov[0][0],invcov[0][1],invcov[0][2],
    invcov[1][0],invcov[1][1],invcov[1][2],
    invcov[2][0],invcov[2][1],invcov[2][2]);
    fclose(fp);
  }
}

void CSkinGrabberView::OnReset() 
{
  SIZE size;
  
  size.cx = width;
  size.cy = height;
  SetImage( iref.getPixelData(), size);

  ZeroMemory(mean,sizeof(mean));
  ZeroMemory(cov,sizeof(cov));
  ZeroMemory(sum,sizeof(sum));
  ZeroMemory(sum2,sizeof(sum2));
  count = 0;	

  meantext.Format("%3.2f %3.2f %3.2f",mean[0],mean[1],mean[2]);
  covtext.Format("%3.2f %3.2f %3.2f\r\n%3.2f %3.2f %3.2f\r\n%3.2f %3.2f %3.2f\r\n",
    cov[0][0],cov[0][1],cov[0][2],
    cov[1][0],cov[1][1],cov[1][2],
    cov[2][0],cov[2][1],cov[2][2]);
  UpdateData(FALSE);

  InvalidateRect(NULL);


}

void CSkinGrabberView::OnInput0() 
{
  input0 = TRUE;
  input1 = FALSE;
  UpdateData(FALSE);
  Init(0);
  OnGrab();
}

void CSkinGrabberView::OnInput1() 
{
  input1 = TRUE;
  input0 = FALSE;
  UpdateData(FALSE);
  Init(1);
  OnGrab();
}

void CSkinGrabberView::Init(int card)
{
  grabber.close();
  grabber.init(card,2,vlImage::RGB);
  grabber.grab( iref );

  width = iref.getWidth();
  height = iref.getHeight();

  rgbImage.set(width,height);
}
