//--Putty.cpp
// Implements: CPutty Class

#include "stdafx.h"
#include <stdlib.h>
#include "glos.h"
#include <math.h>
#include "puttytypes.h"
#include "putty.h"
#include "utility.h"

Vector3f red={1,0,0};
Vector3f green={0,1,0};
Vector3f blue={0,0,1};
Vector3f black={0,0,0};

///////////////////////////////////////////////////////////////////////////
//	CPutty
//	Purpose:	Constructor
//	
CPutty::CPutty(void)
{
	//Initialize some basic variables to default values
	m_Children = NULL;
	m_NumberChildren = 0;
	m_MaxDepth = 10;
	m_CutOff = 0.0001;
	m_MaxLen = 0.15;
	m_Type = AUX_SPHERE;

    m_quadObj = gluNewQuadric ();
    gluQuadricDrawStyle (m_quadObj, GLU_FILL);
    gluQuadricNormals (m_quadObj, GLU_SMOOTH);

	//Motion Attributes
	m_ForwardIncrement = 1.0;
	m_BackwardIncrement = 1.0;
	m_RightIncrement = 1.0;
	m_LeftIncrement = 1.0;
	m_UpIncrement = 1.0;
	m_DownIncrement = 1.0;
	m_TurnRightIncrement = 10.0;
	m_TurnLeftIncrement = 10.0;
	m_FastForwardIncrement = 5.0;

	// Default DOF's
	m_DOF[0][0]=1;
	m_DOF[0][1]=0;
	m_DOF[0][2]=0;

	m_DOF[1][0]=0;
	m_DOF[1][1]=1;
	m_DOF[1][2]=0;

	m_DOF[2][0]=0;
	m_DOF[2][1]=0;
	m_DOF[2][2]=1;

	
	//Init State Vars
	m_TranslateState[0]=0.0;
	m_TranslateState[1]=0.0;
	m_TranslateState[2]=0.0;

	m_RotateState[0] = 0.0;
	m_RotateState[1] = 0.0;
	m_RotateState[2] = 0.0;
	
	m_ScaleState[0]=1.0;
	m_ScaleState[1]=1.0;
	m_ScaleState[2]=1.0;


	// Default Colors
	m_DiffuseColor[0] = 0.0;
	m_DiffuseColor[1] = 0.0;
	m_DiffuseColor[2] = 0.0;
	m_DiffuseColor[3] = 1;

	m_SpecularColor[0] = 0.0;
	m_SpecularColor[1] = 0.0;
	m_SpecularColor[2] = 0.0;
	m_SpecularColor[3] = 1;

	m_EmissionColor[0] = 0.0;
	m_EmissionColor[1] = 0.0;
	m_EmissionColor[2] = 0.0;
	m_EmissionColor[3] = 1;

	m_Shininess = 0.0;

//	m_Alpha = 0.0;

	m_Parent = NULL;
	m_ActionQueue.head = NULL;
	m_ActionQueue.tail = NULL;
	
	m_AppearanceChanged = TRUE;
	m_StateChanged = TRUE;

	m_AppearanceList = glGenLists(1);
	m_StateList = glGenLists(1);


	//Define ControlPts
	m_CPMax = 0;
	m_ControlPts = NULL;

	//Define Link Pts
	m_LinkPoint[0] = 0.0;
	m_LinkPoint[1] = 0.0;
	m_LinkPoint[2] = 0.0;

  // Transformation Matrix
  UseMatrix = FALSE;
  M[0] = 1;
  M[1] = 0;
  M[2] = 0;

  M[3] = 0;
  M[4] = 1;
  M[5] = 0;

  M[6] = 0;
  M[7] = 0;
  M[8] = 0;
}

CPutty::~CPutty() 
{
	POSITION pos;
	CString key;
	Action * pAction, *pTmp;

	gluDeleteQuadric(m_quadObj);

	for( pos = m_ActionTable.GetStartPosition(); pos != NULL; )
	{
		m_ActionTable.GetNextAssoc( pos, key, (void *&)pAction);
		while(pAction) {
			pTmp = pAction;
			pAction = pAction->next;
			delete pTmp;
		}
	}

	// Call function for all children
	for(int i=0;i<m_NumberChildren;i++){
		delete m_Children[i];
	}

	delete m_ControlPts;

}

/*
///////////////////////////////////////////////////////////////////////////
// InitActionTable
// Purpose: This procedure fills the ActionTable with this object's actions
void CPutty::InitActionTable(){
	
}
*/


///////////////////////////////////////////////////////////////////////////
// AddChild
// Purpose: To add children to this object's child list
void CPutty::AddChild(CPutty * Child){

	m_NumberChildren++;

	if (m_NumberChildren == 0){
		m_Children = new CPutty*[1];
		m_Children[0] = Child;
	}
	else{
		CPutty** temp = new CPutty* [m_NumberChildren];
		for(int i=0; i< m_NumberChildren-1;i++){
			temp[i] = m_Children[i];
		}
		temp[m_NumberChildren-1] = Child;

		delete m_Children;
		m_Children = temp;
	}
}


///////////////////////////////////////////////////////////////////////////
//	RegisterAction
//	Purpose:	This routine takes the action indexed by the given string 
//				and registers it in the Available Action list.
void CPutty::RegisterAction(CString action){
	m_AvailableActions.AddTail(action);
}

///////////////////////////////////////////////////////////////////////////
//	GetListofActions
CStringList* CPutty::GetListofActions(){
	return &m_AvailableActions;
}

///////////////////////////////////////////////////////////////////////////
//	LoadAction
//	Purpose:	This routine takes the action indexed by the given string 
//				and places it on the Action Queue.
//	Args:		String identifier for the action and the time the action was
//				loaded.
//	
void CPutty::LoadAction(CString action, float time)
{
	int range;
	float new_Threshold = 0.0f;
	int worldtranslatebool = 0;
	int previous_threshold = 0;
	Action * pNew = NULL;
	ActionNode * pNode = m_ActionQueue.head, * pPrevNode = NULL, * pTmpNode;

	//Get the head of the specified action sequence
	m_ActionTable.Lookup(action, (void *&)pNew);

	if (pNew)
		TRACE("Found Action...\n");

//	AfxDebugBreak();

	while(pNew) {

//		if (pNew->mutex == 1)
//			break;

		// Exclude this action from being used again here
		pNew->mutex = 1;
		
		if (pNew->TimeType == RANDOM_BASETIME){

//			TRACE("blah 1\n");
			range = pNew->MaxBaseTime - pNew->BaseTime;
			if (range < 0)
				range = -range;
			if (range == 0 )
				new_Threshold = pNew->BaseTime + time;
			else
				new_Threshold = rand() % range + pNew->BaseTime + time;
			
			previous_threshold = new_Threshold;
		}
		else
		if (pNew->TimeType == ADD_PREVIOUS){
//			TRACE("blah 2\n");
			new_Threshold = pNew->BaseTime + previous_threshold;
			previous_threshold = new_Threshold;
		}
		else{
//			TRACE("blah 3\n");
			new_Threshold = pNew->BaseTime + time;
			previous_threshold = new_Threshold;
		}

//AfxDebugBreak();		
		while(pNode) {
			if ((new_Threshold < pNode->ThresholdTime) &&
				!(pNew->Type == ROTATE && pNode->cur->Type == WORLD_TRANSLATE) &&
				!(pNew->Type == WORLD_TRANSLATE && pNode->cur->Type == ROTATE) &&
				!(pNew->Type == WORLD_TRANSLATE && pNode->cur->Type == WORLD_TRANSLATE)) {
	
//				worldtranslatebool = 1;
				//Insert the action state here
				if(pPrevNode) {
					pTmpNode = pPrevNode->next;
					pPrevNode->next = new ActionNode;
					pPrevNode->next->ThresholdTime = new_Threshold;
					pPrevNode->next->cur = pNew;
					pPrevNode->next->next = pTmpNode;
					pNode = pPrevNode->next;
					TRACE("Inserting...A1..time: %f\n",pNode->ThresholdTime);
				} else { //We are inserting at the head of the queue
					pTmpNode = m_ActionQueue.head;
					m_ActionQueue.head = new ActionNode;
					m_ActionQueue.head->ThresholdTime = new_Threshold;
					m_ActionQueue.head->cur = pNew;
					m_ActionQueue.head->next = pTmpNode;
					pNode = m_ActionQueue.head;
					TRACE("Inserting...A2..time: %f\n",pNode->ThresholdTime);
				}
				break;
 			}
			pPrevNode = pNode;
			pNode = pNode->next;
		}
		if (pNode == NULL) { //End of the queue so just start inserting
				//Insert the action state here
				if(pPrevNode) {
					pTmpNode = pPrevNode->next;
					pPrevNode->next = new ActionNode;
					pPrevNode->next->ThresholdTime = new_Threshold;
					pPrevNode->next->cur = pNew;
					pPrevNode->next->next = pTmpNode;//NULL
					pPrevNode = pPrevNode->next; 
					m_ActionQueue.tail = pPrevNode;
					TRACE("Inserting...B1..time: %f\n",pPrevNode->ThresholdTime);
				} else { //We are inserting at the head of the queue
					pTmpNode = m_ActionQueue.head;
					m_ActionQueue.head = new ActionNode;
					m_ActionQueue.head->ThresholdTime = new_Threshold;
					m_ActionQueue.head->cur = pNew;
					m_ActionQueue.head->next = pTmpNode;
					pPrevNode = m_ActionQueue.head;
					m_ActionQueue.tail = m_ActionQueue.head;
					TRACE("Inserting...B2..time: %f\n",m_ActionQueue.head->ThresholdTime);
				}

		}
		pNew = pNew->next;
	}

	// Call function for all children
	for(int i=0;i<m_NumberChildren;i++){
		(m_Children[i])->LoadAction(action,time);
	}

}

///////////////////////////////////////////////////////////////////////////
//	Update
//	Purpose:	This routine increments the action state machine by 
//				executing all actions on the queue with threshold time less
//				than the time passed in.  It is assumed that the Action
//				Queue is a priority queue sorted on threshold time.
//	Args:		The current absolute time
//  Returns: TRUE - if Draw() should be called
//           FALSE - if not
//	
int CPutty::Update(float time)
{
	int random_time = 0;
	ActionNode *browse, *trailer;
	CStringList ActionToSpawn;
	POSITION posit = NULL;
  int DoDraw = FALSE;

	browse = m_ActionQueue.head;
	trailer = NULL;
	
	TRACE("time %f\n", time);

	while (browse != NULL) {
		TRACE("Head Theshold time %f\n",browse->ThresholdTime);

		if (browse->cur->TimeType == RANDOM_BASETIME || browse->cur->TimeType == ADD_PREVIOUS)
			random_time = 1;
		else
			random_time = 0;

		if (browse->ThresholdTime <= time) {

			if (browse->cur->Type == SPAWN){
//				TRACE("SPAWING ACTION %s\n",browse->cur->ActionToSpawn);
				posit = ActionToSpawn.AddHead(browse->cur->ActionToSpawn);
			}
			else{
//				TRACE("Committing Action...\n");
				CommitAction(browse->cur);
        DoDraw = TRUE;
			}

			//Delete the current node if it has completed succesfully.
			
			// Head of the list
			if (browse == m_ActionQueue.head){				
				m_ActionQueue.head = browse->next;

				//Update the tail pointer
				if (browse == m_ActionQueue.tail)
					m_ActionQueue.tail = NULL;
				
				// Release mutex
				browse->cur->mutex = 0;
				
				//Remove action from the queue
				delete browse;


				//Increment pointers	
				trailer = NULL;
				browse = m_ActionQueue.head;		
			}
			// Somewhere in the middle or end
			else
			{
				trailer->next = browse->next;

				//Update the tail pointer
				if (browse == m_ActionQueue.tail)
					m_ActionQueue.tail = NULL;

				// Release mutex
				browse->cur->mutex = 0;

				//Remove action from the queue
				delete browse;
		
				//Increment pointers	
				browse=trailer->next;
			}

		} else {
			if (!random_time)
				break;
			else{
				trailer = browse;
				browse = browse->next;
			}
		}
	}

	// Spawn actions after while loop to avoid infinite loops
	while (posit){
		TRACE("SPAWNING...\n");
		LoadAction(ActionToSpawn.GetNext(posit),time);
	}
	// Call function for all children
	for(int i=0;i<m_NumberChildren;i++){
		if(m_Children[i]->Update(time))
      DoDraw = TRUE;
	}

  return DoDraw;

}

///////////////////////////////////////////////////////////////////////////
//	CancelAction
//	Purpose:	This routine takes the action indexed by the given string 
//				off the Action Queue.
//	Args:		String identifier for the action
//	
void CPutty::CancelAction(CString action)
{

	// Call function for all children
	for(int i=0;i<m_NumberChildren;i++){
		m_Children[i]->CancelAction(action);
	}

}

///////////////////////////////////////////////////////////////////////////
//	Draw
//	Purpose:	This routine causes the object to draw itself to the screen.
//				It will use the previously compiled display list unless the
//				m_Changed boolean is true.
//	
void CPutty::Draw(void)
{
	
	if (m_StateChanged) {
		glNewList(m_StateList, GL_COMPILE_AND_EXECUTE);
			//  Base Position and Orientation
			glPushMatrix();

				// Do the translation
				glTranslatef(m_TranslateState[0],m_TranslateState[1],m_TranslateState[2]);

				// Here, we rotate around a given point.
				glTranslatef(m_LinkPoint[0], m_LinkPoint[1], m_LinkPoint[2]);
				glRotatef(m_RotateState[0], m_DOF[0][0],m_DOF[0][1], m_DOF[0][2]);
				glRotatef(m_RotateState[1], m_DOF[1][0],m_DOF[1][1], m_DOF[1][2]);
				glRotatef(m_RotateState[2], m_DOF[2][0],m_DOF[2][1], m_DOF[2][2]);
				glTranslatef(-(m_LinkPoint[0]), -(m_LinkPoint[1]), -(m_LinkPoint[2]));

				// Do the scaling
				glScalef(m_ScaleState[0],m_ScaleState[1],m_ScaleState[2]);

				// Draw the axes
				glLineWidth(1.0);
				glMaterialfv(GL_FRONT, GL_EMISSION, red); 	
				glBegin(GL_LINES);
					glVertex3f(0,0,0);
					glVertex3f(3,0,0);
				glEnd();
				
				glMaterialfv(GL_FRONT, GL_EMISSION, green); 	
				glBegin(GL_LINES);
					glVertex3f(0,0,0);
					glVertex3f(0,3,0);
				glEnd();

				glMaterialfv(GL_FRONT, GL_EMISSION, blue); 	
				glBegin(GL_LINES);
					glVertex3f(0,0,0);
					glVertex3f(0,0,3);
				glEnd();


		glEndList();
		
		m_StateChanged = FALSE;
	} else {
		glCallList(m_StateList);
	}

				// Call function for all children
				for(int i=0;i<m_NumberChildren;i++){
					m_Children[i]->Draw();
				}
	
		glMaterialfv(GL_FRONT, GL_DIFFUSE, m_DiffuseColor); 	
		glMaterialfv(GL_FRONT, GL_SPECULAR, m_SpecularColor); 	
		glMaterialfv(GL_FRONT, GL_EMISSION, m_EmissionColor);
		glMaterialf(GL_FRONT, GL_SHININESS, m_Shininess);

	if(m_AppearanceChanged) {
		glNewList(m_AppearanceList, GL_COMPILE_AND_EXECUTE);

		switch(m_Type){
		case SHEET:
			DrawSheet();
			break;
		case SPHERE:
			DrawSphere();
			break;
		case AUX_SPHERE:
			gluSphere (m_quadObj, 1.0, 16, 16);
			break;
		case CYLINDER:
			glRotatef(-90,1.0,0.0,0.0);
			gluCylinder(m_quadObj, m_BottomRadius, m_TopRadius, 1.0, 16, 2);
			break;
		case COLLECTION:
			break;
		}

			
		glPopMatrix();
		glEndList();

		m_AppearanceChanged = FALSE;

	} else {
		glCallList(m_AppearanceList);
	}
	
}


///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
// Private Methods
///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////
//	CommitAction
//	Purpose:	This routine takes the action and commits its changes to
//				to the object
//	Args:		pointer to an action object
//	
void CPutty::CommitAction(Action * action)
{
	//Which type of action are we committing?
	switch(action->Type) {
	case SCALE:
		//It is more intuitive to make the below update a multiply
		//rather than an addition.  But to be able to recover ourselves
		//from a zero scale state, we need to use addition.

		//Update State vars
		m_ScaleState[0] += action->ScaleFactor[0];
		m_ScaleState[1] += action->ScaleFactor[1];
		m_ScaleState[2] += action->ScaleFactor[2];

		//Cause the state list to be recompiled
		m_StateChanged = TRUE;
		break;    
	case ROTATE:
		//CHANGE!
		
		// Update State Variables
		m_RotateState[action->RotDOF] += action->RotAngle;
		if (m_RotateState[action->RotDOF] > 360)
			m_RotateState[action->RotDOF] = m_RotateState[action->RotDOF] - 360;
		if (m_RotateState[action->RotDOF] > -360)
			m_RotateState[action->RotDOF] = m_RotateState[action->RotDOF] + 360;

		//Cause the state list to be recompiled
		m_StateChanged = TRUE;
		break;
	case TRANSLATE:
		// Update State Variables
		m_TranslateState[0] += action->Translate[0];
		m_TranslateState[1] += action->Translate[1];
		m_TranslateState[2] += action->Translate[2];

		//Cause the state list to be recompiled
		m_StateChanged = TRUE;
		break;
	case WORLD_TRANSLATE:
		// Update State Variables

		m_TranslateState[2] += action->Translate[2]*cos(PI*m_RotateState[1]/180);
		m_TranslateState[0] += action->Translate[2]*sin(PI*m_RotateState[1]/180);
		
		m_TranslateState[2] += action->Translate[0]*-cos(PI*m_RotateState[1]/180 - PI/2);
		m_TranslateState[0] += action->Translate[0]*-sin(PI*m_RotateState[1]/180 - PI/2);

		m_TranslateState[1] += action->Translate[1];

		//Cause the state list to be recompiled
		m_StateChanged = TRUE;
		break;
	case SOFT:
		//Update the control pt offsets from the action structure
		m_ControlPts[action->CtrlPtIndex].Offset[0] += action->CtrlPt.Offset[0];
		m_ControlPts[action->CtrlPtIndex].Offset[1] += action->CtrlPt.Offset[1];
		m_ControlPts[action->CtrlPtIndex].Offset[2] += action->CtrlPt.Offset[2];
		m_ControlPts[action->CtrlPtIndex].Point[0] += action->CtrlPt.Point[0];
		m_ControlPts[action->CtrlPtIndex].Point[1] += action->CtrlPt.Point[1];
		m_ControlPts[action->CtrlPtIndex].Point[2] += action->CtrlPt.Point[2];
		m_ControlPts[action->CtrlPtIndex].Rigidness += action->CtrlPt.Rigidness;
		m_ControlPts[action->CtrlPtIndex].Locality += action->CtrlPt.Locality;
		if (m_ControlPts[action->CtrlPtIndex].Rigidness <= 0)
			m_ControlPts[action->CtrlPtIndex].Rigidness = 0.01;
		if (m_ControlPts[action->CtrlPtIndex].Locality <= 0)
			m_ControlPts[action->CtrlPtIndex].Locality = 1.0;
		//Cause the appearance list to be recompiled
		m_AppearanceChanged = TRUE;

		break;
	case COLOR:
		//Update the color
		m_DiffuseColor[0] += action->DiffuseColor[0];
		m_DiffuseColor[1] += action->DiffuseColor[1];
		m_DiffuseColor[2] += action->DiffuseColor[2];

		//Color is specified outside of all display lists so don't recompile any
	}
}

///////////////////////////////////////////////////////////////////////////
//	Surf
//	Purpose:	This routine takes two points on the surface and returns
//				the point on the surface between them.
//	Args:		3D point to fill, the two points to subdivide on
//	
void CPutty::Surf(Vector3f & v, Vector3f & u)
{
	int k;
	float a, radius, sum0 = 0.0, sum1 = 0.0, sum2 = 0.0, d, b0,b1,b2;
	//First calculate the mesh of the primitive
	switch(m_Type) {
	case SPHERE:
		d = sqrt(u[0]*u[0] + u[1]*u[1] + u[2]*u[2]);
		v[0] = u[0]/d;
		v[1] = u[1]/d;
		v[2] = u[2]/d;
//		v[0] = u[0] = u[0]/d;
//		v[1] = u[1] = u[1]/d;
//		v[2] = u[2] = u[2]/d;
		break;
	case SHEET:
		v[0] = u[0];
		v[1] = u[1];
		v[2] = 0.0;
		break;
	}
	//Now deform the mesh according to the control pts	
	for(k = 0; k < m_CPMax; k++) {
		b0 = (v[0] - m_ControlPts[k].Point[0]);
		b1 = (v[1] - m_ControlPts[k].Point[1]);
		b2 = (v[2] - m_ControlPts[k].Point[2]);
		a = m_ControlPts[k].Rigidness/
			((b0*b0 + b1*b1 + b2*b2)*m_ControlPts[k].Locality + 
			m_ControlPts[k].Rigidness);
		sum0 += m_ControlPts[k].Offset[0] * a;
		sum1 += m_ControlPts[k].Offset[1] * a;
		sum2 += m_ControlPts[k].Offset[2] * a;
	}
	v[0] += sum0;
	v[1] += sum1;
	v[2] += sum2;
}

///////////////////////////////////////////////////////////////////////////
//	Deform
//	Purpose:	This routine will calculate the deformed point from
//				location u and store it in v
//	Args:		v and u
//
void CPutty::Deform(Vector3f & v, Vector3f & u)
{
	int k;
	float a, radius, sum[3] = {0.0,0.0,0.0};

	//Now deform the mesh according to the control pts	
	for(k = 0; k < m_CPMax; k++) {
		radius = ((u[0] - m_ControlPts[k].Point[0])*(u[0] - m_ControlPts[k].Point[0]) 
				+ (u[1] - m_ControlPts[k].Point[1])*(u[1] - m_ControlPts[k].Point[1]) 
				+ (u[2] - m_ControlPts[k].Point[2])*(u[2] - m_ControlPts[k].Point[2]))*m_ControlPts[k].Locality;
		a = (1 - radius/(radius + m_ControlPts[k].Rigidness));
		sum[0] += m_ControlPts[k].Offset[0] * a;
		sum[1] += m_ControlPts[k].Offset[1] * a;
		sum[2] += m_ControlPts[k].Offset[2] * a;
	}
	v[0] = u[0] + sum[0];
	v[1] = u[1] + sum[1];
	v[2] = u[2] + sum[2];
}

///////////////////////////////////////////////////////////////////////////
//	Subdivide
//	Purpose:	This routine recursively subdivides the surface primitive
//				into triangle polygons, drawing more triangles where detail
//				is needed.
//	Args:		The vertices of a triangle,
//				and the current recursion depth
//	
void CPutty::Subdivide(Vector3f & u1, Vector3f & u2, Vector3f & u3, 
					   Vector3f & v1, Vector3f & v2, Vector3f & v3, 
					   int depth)
{
	Vector3f v12, v23, v31, u12, u23, u31, n, vA, vB;
	float x[3], y[3];

	//Calculate the next subdivision
	u12[0] = 0.5*(u1[0] + u2[0]);
	u12[1] = 0.5*(u1[1] + u2[1]);
	u12[2] = 0.5*(u1[2] + u2[2]);
	u23[0] = 0.5*(u2[0] + u3[0]);
	u23[1] = 0.5*(u2[1] + u3[1]);
	u23[2] = 0.5*(u2[2] + u3[2]);
	u31[0] = 0.5*(u3[0] + u1[0]);
	u31[1] = 0.5*(u3[1] + u1[1]);
	u31[2] = 0.5*(u3[2] + u1[2]);

	Surf(v12,u12);
	Surf(v23,u23);
	Surf(v31,u31);

	if (depth == m_MaxDepth) {
//		TRACE("Got to maximum depth.\n");
		vA[0] = v12[0] - v1[0];
		vA[1] = v12[1] - v1[1];
		vA[2] = v12[2] - v1[2];
		vB[0] = v31[0] - v1[0];
		vB[1] = v31[1] - v1[1];
		vB[2] = v31[2] - v1[2];
		NormCrossProd(n, vA, vB);
		glNormal3fv(n); 
		glVertex3fv(v1);

		vA[0] = v23[0] - v2[0];
		vA[1] = v23[1] - v2[1];
		vA[2] = v23[2] - v2[2];
		vB[0] = v12[0] - v2[0];
		vB[1] = v12[1] - v2[1];
		vB[2] = v12[2] - v2[2];
		NormCrossProd(n, vA, vB);
		glNormal3fv(n); 
		glVertex3fv(v2);

		vA[0] = v31[0] - v3[0];
		vA[1] = v31[1] - v3[1];
		vA[2] = v31[2] - v3[2];
		vB[0] = v23[0] - v3[0];
		vB[1] = v23[1] - v3[1];
		vB[2] = v23[2] - v3[2];
		NormCrossProd(n, vA, vB);
		glNormal3fv(n); 
		glVertex3fv(v3);
		return;
	}
			
	//Is the Surface flat enough to stop?
	//Side 1-2
	x[0] = v12[0] - 0.5*(v1[0] + v2[0]);
	x[1] = v12[1] - 0.5*(v1[1] + v2[1]);
	x[2] = v12[2] - 0.5*(v1[2] + v2[2]);
	y[0] = v1[0] - v2[0];
	y[1] = v1[1] - v2[1];
	y[2] = v1[2] - v2[2];
	if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
		(y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
		//Side 2-3
		x[0] = v23[0] - 0.5*(v2[0] + v3[0]);
		x[1] = v23[1] - 0.5*(v2[1] + v3[1]);
		x[2] = v23[2] - 0.5*(v2[2] + v3[2]);
		y[0] = v2[0] - v3[0];
		y[1] = v2[1] - v3[1];
		y[2] = v2[2] - v3[2];
		if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
		    (y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
			//Side 3-1
			x[0] = v31[0] - 0.5*(v3[0] + v1[0]);
			x[1] = v31[1] - 0.5*(v3[1] + v1[1]);
			x[2] = v31[2] - 0.5*(v3[2] + v1[2]);
			y[0] = v3[0] - v1[0];
			y[1] = v3[1] - v1[1];
			y[2] = v3[2] - v1[2];
			if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
				(y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
					//None
//					TRACE("Got Cutoff.\n");
					vA[0] = v12[0] - v1[0];
					vA[1] = v12[1] - v1[1];
					vA[2] = v12[2] - v1[2];
					vB[0] = v31[0] - v1[0];
					vB[1] = v31[1] - v1[1];
					vB[2] = v31[2] - v1[2];
					NormCrossProd(n, vA, vB);
					glNormal3fv(n); 
					glVertex3fv(v1);

					vA[0] = v23[0] - v2[0];
					vA[1] = v23[1] - v2[1];
					vA[2] = v23[2] - v2[2];
					vB[0] = v12[0] - v2[0];
					vB[1] = v12[1] - v2[1];
					vB[2] = v12[2] - v2[2];
					NormCrossProd(n, vA, vB);
					glNormal3fv(n); 
					glVertex3fv(v2);

					vA[0] = v31[0] - v3[0];
					vA[1] = v31[1] - v3[1];
					vA[2] = v31[2] - v3[2];
					vB[0] = v23[0] - v3[0];
					vB[1] = v23[1] - v3[1];
					vB[2] = v23[2] - v3[2];
					NormCrossProd(n, vA, vB);
					glNormal3fv(n); 
					glVertex3fv(v3);
					return;
			} else {
				//Only 31
				Subdivide(u1, u2, u31, 
						  v1, v2, v31, depth + 1);
				Subdivide(u31, u2, u3, 
						  v31, v2, v3, depth + 1);
				return;
			}	
		} else {
			//Side 3-1
			x[0] = v31[0] - 0.5*(v3[0] + v1[0]);
			x[1] = v31[1] - 0.5*(v3[1] + v1[1]);
			x[2] = v31[2] - 0.5*(v3[2] + v1[2]);
			y[0] = v3[0] - v1[0];
			y[1] = v3[1] - v1[1];
			y[2] = v3[2] - v1[2];
			if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
				(y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
				//Only 23
				Subdivide(u1, u2, u23, 
						  v1, v2, v23, depth + 1);
				Subdivide(u1, u23, u3, 
						  v1, v23, v3, depth + 1);
				return;
			} else {
				//Only 23 and 31
				Subdivide(u1, u23, u31, 
						  v1, v23, v31, depth + 1);
				Subdivide(u1, u2, u23, 
						  v1, v2, v23, depth + 1);
				Subdivide(u31, u23, u3, 
						  v31, v23, v3, depth + 1);
				return;
			}
		}	
	} else {
		//Side 2-3
		x[0] = v23[0] - 0.5*(v2[0] + v3[0]);
		x[1] = v23[1] - 0.5*(v2[1] + v3[1]);
		x[2] = v23[2] - 0.5*(v2[2] + v3[2]);
		y[0] = v2[0] - v3[0];
		y[1] = v2[1] - v3[1];
		y[2] = v2[2] - v3[2];
		if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
		    (y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
			//Side 3-1
			x[0] = v31[0] - 0.5*(v3[0] + v1[0]);
			x[1] = v31[1] - 0.5*(v3[1] + v1[1]);
			x[2] = v31[2] - 0.5*(v3[2] + v1[2]);
			y[0] = v3[0] - v1[0];
			y[1] = v3[1] - v1[1];
			y[2] = v3[2] - v1[2];
			if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
				(y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
				//Only 12
				Subdivide(u12, u2, u3,
						  v12, v2, v3, depth + 1);
				Subdivide(u12, u3, u1,
						  v12, v3, v1, depth + 1);
				return;
			} else {
				//Only 12 and 31
				Subdivide(u3, u31, u2, 
						  v3, v31, v2, depth + 1);
				Subdivide(u12, u2, u31,
						  v12, v2, v31, depth + 1);
				Subdivide(u1, u12, u31, 
						  v1, v12, v31, depth + 1);
				return;
			}	
		} else {
			//Side 3-1
			x[0] = v31[0] - 0.5*(v3[0] + v1[0]);
			x[1] = v31[1] - 0.5*(v3[1] + v1[1]);
			x[2] = v31[2] - 0.5*(v3[2] + v1[2]);
			y[0] = v3[0] - v1[0];
			y[1] = v3[1] - v1[1];
			y[2] = v3[2] - v1[2];
			if ((x[0]*x[0] + x[1]*x[1] + x[2]*x[2] < m_CutOff) &&
				(y[0]*y[0] + y[1]*y[1] + y[2]*y[2] < m_MaxLen)) {
				//Only 12 and 23
				Subdivide(u1, u12, u3, 
						  v1, v12, v3, depth + 1);
				Subdivide(u2, u23, u12, 
						  v2, v23, v12, depth + 1);
				Subdivide(u12, u23, u3, 
						  v12, v23, v3, depth + 1);
				return;
			} else {
				//All Cuts			
				Subdivide(u1, u12, u31, 
						  v1, v12, v31, depth + 1);
				Subdivide(u2, u23, u12, 
						  v2, v23, v12, depth + 1);
				Subdivide(u3, u31, u23, 
						  v3, v31, v23, depth + 1);
				Subdivide(u12, u23, u31,
						  v12, v23, v31, depth + 1);
				return;
			}
		}
	}
}

///////////////////////////////////////////////////////////////////////////
//	DrawSheet
//	Purpose:	This routine initiates the drawing of a plane.  It also 
//				makes sure that the Subdivide routine tessellates enough
//				detail to detect all possible deformations.
//	
void CPutty::DrawSheet(void)
{
	float v1[3], v2[3], v3[3];
	int i,j;
	float xMin = -1.0, yMin = -1.0, xMax = 1.0, yMax = 1.0;
	float * x, * y;
	
	x = new float[m_CPMax + 2];
	y = new float[m_CPMax + 2];

	x[0] = xMin;
	x[m_CPMax + 1] = xMax;
	y[0] = yMin;
	y[m_CPMax + 1] = yMax;
	for (i = 1; i < m_CPMax + 1; i++) {
		x[i] = m_ControlPts[i - 1].Point[0];
		y[i] = m_ControlPts[i - 1].Point[1];
	}

	//Tessellate 
	glBegin(GL_TRIANGLES);
	for (i = 0; i < m_CPMax + 1; i++) {
		for (j = 0; j < m_CPMax + 1; j++) {
			v1[0] = x[i];	v1[1] = y[j];	v1[2] = 0.0;
			v2[0] = x[i+1]; v2[1] = y[j+1];	v2[2] = 0.0;
			v3[0] = x[i];	v3[1] = y[j+1];	v3[2] = 0.0;
			Deform(v1, v1);
			Deform(v2, v2);
			Deform(v3, v3);
			Subdivide(v1, v2, v3, v1,v2,v3,0);
		
			v1[0] = x[i];	v1[1] = y[j];	v1[2] = 0.0;
			v2[0] = x[i+1]; v2[1] = y[j];	v2[2] = 0.0;
			v3[0] = x[i+1];	v3[1] = y[j+1];	v3[2] = 0.0;
			Deform(v1, v1);
			Deform(v2, v2);
			Deform(v3, v3);
			Subdivide(v1, v2, v3, v1,v2,v3,0);
		}		
	}
	glEnd();
	delete x;
	delete y;
}


///////////////////////////////////////////////////////////////////////////
//	DrawSphere
//	Purpose:	This routine draws the sphere
//
#define X 0.525731112119133606
#define Z 0.850650808352039932

static Vector3f SphereData[12] = {
	{-X,0.0,Z},{X,0.0,Z},{-X,0.0,-Z},{X,0.0,-Z},
	{0.0,Z,X},{0.0,Z,-X},{0.0,-Z,X},{0.0,-Z,-X},
	{Z,X,0.0},{-Z,X,0.0},{Z,-X,0.0},{-Z,-X,0.0}
};

static Vector3i SphereIndices[20] = {
	{0,4,1},{0,9,4},{9,5,4},{4,5,8},{4,8,1},
	{8,10,1},{8,3,10},{5,3,8},{5,2,3},{2,7,3},
	{7,10,3},{7,6,10},{7,11,6},{11,0,6},{0,1,6},
	{6,1,10},{9,0,11},{9,11,2},{9,2,5},{7,2,11}
};

void CPutty::DrawSphere(void)
{
	Vector3f v1, v2, v3;
	Vector3f blue={0,0,1};
	int i;

	glBegin(GL_TRIANGLES);
	for(i = 0; i < 20; i++)
	{
		Deform(v1, SphereData[SphereIndices[i][0]]);
		Deform(v2, SphereData[SphereIndices[i][1]]);
		Deform(v3, SphereData[SphereIndices[i][2]]);
		Subdivide(v2, v1, v3, v2, v1, v3, 0);
	}
	glEnd();
/*
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, blue); 	
	for(i = 0; i < 20; i++)
	{
		Deform(v1, SphereData[SphereIndices[i][0]]);
		Deform(v2, SphereData[SphereIndices[i][1]]);
		Deform(v3, SphereData[SphereIndices[i][2]]);
		SubdivideNormals(v2, v1, v3, 0);
	}
	*/
}	



///////////////////////////////////////////////////////////////////////////
//	Translate
//	Purpose: This procedure translates objects... duh
void CPutty::Translate(int start_time, int end_time,  GLfloat x,GLfloat y,GLfloat z, CString ActionName){

	// Var Initialization
	float delta_constant = 0.0;
	Action * pAction, * pPrevAction;

	for(int i=start_time; i<=end_time; i++){

		if (i == start_time)
			delta_constant = 0;
		else
			delta_constant = 1.0 / (float)(end_time - start_time);

		// Add actions to the action queue
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = TRANSLATE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->Translate[0] = delta_constant * x;
		pAction->Translate[1] = delta_constant * y;
		pAction->Translate[2] = delta_constant * z;
			
		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}


///////////////////////////////////////////////////////////////////////////
//	TranslateSmooth
//	Purpose: This procedure translates objects smoothly
void CPutty::TranslateSmooth(int start_time, int end_time, GLfloat x,GLfloat y,GLfloat z, CString ActionName){

	// Var Initialization
	float delta_constant = 0.0;
	float old_delta_constant = 0.0;
	Action * pAction, * pPrevAction;

	for(int i=start_time; i<=end_time; i++){
		delta_constant = (float)(i - start_time) / (float)(end_time - start_time);
	
		// Smooth out the constant quadratically
		delta_constant = 1 - (1 - delta_constant) * (1 - delta_constant);

		if (i!=start_time){
			old_delta_constant = (float)(i-1 - start_time) / (float)(end_time - start_time);
		
			// Smooth out the constant quadratically
			old_delta_constant = 1 - (1 - old_delta_constant) * (1 - old_delta_constant);
		}

		delta_constant = delta_constant - old_delta_constant;

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = TRANSLATE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->Translate[0] = delta_constant * x;
		pAction->Translate[1] = delta_constant * y;
		pAction->Translate[2] = delta_constant * z;
			
		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}

///////////////////////////////////////////////////////////////////////////
//	WorldTranslate
//	Purpose: This procedure WorldTranslates objects... duh
void CPutty::WorldTranslate(int start_time, int end_time,  GLfloat x,GLfloat y,GLfloat z, CString ActionName){
			
	
	// Var Initialization
	float delta_constant = 0.0;
	Action * pAction, * pPrevAction;

	for(int i=start_time; i<=end_time; i++){

		if (i == start_time)
			delta_constant = 0;
		else
			delta_constant = 1.0 / (float)(end_time - start_time);

		// Add actions to the action queue
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = WORLD_TRANSLATE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->Translate[0] = delta_constant * x;
		pAction->Translate[1] = delta_constant * y;
		pAction->Translate[2] = delta_constant * z;

		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}


///////////////////////////////////////////////////////////////////////////
//	WorldTranslateSmooth
//	Purpose: This procedure WorldTranslates objects smoothly
void CPutty::WorldTranslateSmooth(int start_time, int end_time, GLfloat x,GLfloat y,GLfloat z, CString ActionName){

	// Var Initialization
	float delta_constant = 0.0;
	float old_delta_constant = 0.0;
	Action * pAction, * pPrevAction;

	for(int i=start_time; i<=end_time; i++){
		delta_constant = (float)(i - start_time) / (float)(end_time - start_time);
	
		// Smooth out the constant quadratically
		delta_constant = 1 - (1 - delta_constant) * (1 - delta_constant);

		if (i!=start_time){
			old_delta_constant = (float)(i-1 - start_time) / (float)(end_time - start_time);
		
			// Smooth out the constant quadratically
			old_delta_constant = 1 - (1 - old_delta_constant) * (1 - old_delta_constant);
		}

		delta_constant = delta_constant - old_delta_constant;

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = WORLD_TRANSLATE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->Translate[0] = delta_constant * x;
		pAction->Translate[1] = delta_constant * y;
		pAction->Translate[2] = delta_constant * z;
			
		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}




///////////////////////////////////////////////////////////////////////////
//	RotatePointAxis
//	Purpose: This procedure allows a rotation about an arbitrary axis
//			 originating from any point in space
void CPutty::RotatePointAxis(GLfloat angle, GLfloat x, GLfloat y, GLfloat z, GLfloat ax,GLfloat ay,GLfloat az){
	glTranslatef(x, y, z);
	glRotatef(angle, ax, ay, az);
	glTranslatef(-x, -y, -z);
}



///////////////////////////////////////////////////////////////////////////
// Rotate
// Purpose: This procedure allows us to mimic limb motion for any CPutty object
//			x,y,z is the point, and ax, ay,az is the axis originiating from that
//			point around which we rotate.
void CPutty::Rotate(int start_time, int max_time,GLfloat angle, int DOF,
					CString ActionName){
	
	float delta_angle;
	Action * pAction, * pPrevAction;

	for(int i=start_time;i<=max_time ; i++){

		
		if (i == start_time)
			delta_angle = 0;
		else
			delta_angle = 1.0/(float)(max_time - start_time);

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = ROTATE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->RotDOF = DOF;

		pAction->RotAngle = angle * delta_angle;
			
		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}

}


///////////////////////////////////////////////////////////////////////////
// RotateSmooth
// Purpose: This procedure allows us to mimic limb motion for any CPutty object
//			x,y,z is the point, and ax, ay,az is the axis originiating from that
//			point around which we rotate.
void CPutty::RotateSmooth(int start_time, int max_time, GLfloat angle, 
						  int DOF, CString ActionName){
	float delta_angle,old_delta_angle;
	Action * pAction, * pPrevAction;

	for(int i=start_time;i<=max_time ; i++){
		delta_angle = (float)(i - start_time)/(float)(max_time - start_time);
		
		// Smooth out the motion with a quadratic
		delta_angle = (1 - (1 - delta_angle) * (1 - delta_angle));


		if (i!=start_time) {
			old_delta_angle = (float)(i - 1  - start_time)/(float)(max_time - start_time);
		
			// Smooth out the motion with a quadratic
			old_delta_angle = (1 - (1 - old_delta_angle) * (1 - old_delta_angle));
		}

		delta_angle = delta_angle - old_delta_angle;

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = ROTATE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->RotDOF = DOF;

		pAction->RotAngle = angle * delta_angle;
			
		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}


///////////////////////////////////////////////////////////////////////////
//	SmoothColor
//	Purpose: This procedure smoothly alters color of the object
//  Args:	start_time		Beginning time
//			end_time		Ending time
//			r,g,b			Color values to add
//			ActionName		Identifier for action
void CPutty::SmoothColor(int start_time, int end_time, 
						 float r, float g, float b, CString ActionName){

	// Var Initialization
	float delta_constant = 0.0;
	float old_delta_constant = 0.0;
	Action * pAction, * pPrevAction;

	for(int i=start_time; i<=end_time; i++){
		delta_constant = (float)(i - start_time) / (float)(end_time - start_time);
	
		// Smooth out the constant quadratically
		delta_constant = 1 - (1 - delta_constant) * (1 - delta_constant);

		if (i!=start_time){
			old_delta_constant = (float)(i-1 - start_time) / (float)(end_time - start_time);
		
			// Smooth out the constant quadratically
			old_delta_constant = 1 - (1 - old_delta_constant) * (1 - old_delta_constant);
		}

		delta_constant = delta_constant - old_delta_constant;

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = COLOR;
		pAction->BaseTime = i;

		// Specify the control pt's current offset
		pAction->DiffuseColor[0] = r*delta_constant;
		pAction->DiffuseColor[1] = g*delta_constant;
		pAction->DiffuseColor[2] = b*delta_constant;

		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}

///////////////////////////////////////////////////////////////////////////
//	SmoothDeform
//	Purpose: This procedure deforms objects smoothly
//  Args:	start_time		Beginning time
//			end_time		Ending time
//			CtrlPtIndex		Specifies what control pt this applies to
//			xOff,yOff,zOff	The final offset vector
//			dx,dy,dz		The final position vector will be moved dx,dy,dz
//			dR				The final value to be added to base Rigidness
//			dL				The final value to be added to base Locality
//			ActionName		Identifier for action
void CPutty::SmoothDeform(int start_time, int end_time, int CtrlPtIndex, 
						  float xOff,float yOff,float zOff, 
						  float dx, float dy, float dz,
						  float dR, float dL, CString ActionName){

	// Var Initialization
	float delta_constant = 0.0;
	float old_delta_constant = 0.0;
	Action * pAction, * pPrevAction;
	Vector3f PrevOffsets = {0.0,0.0,0.0};

	for(int i=start_time; i<=end_time; i++){
		delta_constant = (float)(i - start_time) / (float)(end_time - start_time);
	
		// Smooth out the constant quadratically
		delta_constant = 1 - (1 - delta_constant) * (1 - delta_constant);

		if (i!=start_time){
			old_delta_constant = (float)(i-1 - start_time) / (float)(end_time - start_time);
		
			// Smooth out the constant quadratically
			old_delta_constant = 1 - (1 - old_delta_constant) * (1 - old_delta_constant);
		}

		delta_constant = delta_constant - old_delta_constant;

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = SOFT;
		pAction->BaseTime = i;

		// Specify the control pt's current offset
		pAction->CtrlPtIndex = CtrlPtIndex;
		pAction->CtrlPt.Offset[0] = delta_constant * xOff;
		pAction->CtrlPt.Offset[1] = delta_constant * yOff;
		pAction->CtrlPt.Offset[2] = delta_constant * zOff;
		pAction->CtrlPt.Point[0] = delta_constant * dx;
		pAction->CtrlPt.Point[1] = delta_constant * dy;
		pAction->CtrlPt.Point[2] = delta_constant * dz;
		pAction->CtrlPt.Rigidness = delta_constant * dR;
		pAction->CtrlPt.Locality = delta_constant * dL;

		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}
}


///////////////////////////////////////////////////////////////////////////
//	Scale
//	Purpose: This procedure scales objects. Note: This operation only
//		     with respect to the origin in the the OBJECT SPACE.
void CPutty::Scale(int start_time, int end_time, 
				   float scale_x, float scale_y,float scale_z, CString ActionName){
	
	float delta_scale;
	Action * pAction, * pPrevAction;

	for(int i=start_time;i<=end_time ; i++){

		
		if (i == start_time)
			delta_scale = 0;
		else
			delta_scale = 1.0/(float)(end_time - start_time);

		// Add actions to the action sequence
		pAction = new Action;
		pAction->next = NULL;
		pAction->Type = SCALE;
		pAction->ThresholdTime = i;
		pAction->BaseTime = i;

		pAction->ScaleFactor[0] = scale_x * delta_scale;
		pAction->ScaleFactor[1] = scale_y * delta_scale;
		pAction->ScaleFactor[2] = scale_z * delta_scale;
		
		if (i!=start_time) { 
			pPrevAction->next = pAction;
		}
		pPrevAction = pAction;

		//Save the head of this Action sequence to the Action Table
		if (i==start_time)
			m_ActionTable.SetAt(ActionName, pAction);
	}


}


///////////////////////////////////////////////////////////////////////////
// ConcatActions
// Purpose: This procedure will compound two actions, creating another
//			action (under the name of the first actionname) which runs both
//			one after the the other
// Modifies: The first action passed in
void CPutty::ConcatActions(CString ActionName1, CString ActionName2){
	Action * pNew = NULL;

	//Get the head of the specified action sequence
	m_ActionTable.Lookup(ActionName1, (void *&)pNew);

	while (pNew){
		// If at the end of the list
		if (!(pNew->next)){
			pNew->next = new Action;

			//Initialize a new SPAWN action
			pNew->next->ThresholdTime = pNew->ThresholdTime;
			pNew->next->BaseTime = pNew->BaseTime;

			pNew=pNew->next;
			pNew->Type = SPAWN;
			pNew->ActionToSpawn = ActionName2;
			pNew->next = NULL;
		}

		pNew = pNew->next;
	}
}



///////////////////////////////////////////////////////////////////////////
// CollectManyActions
// Purpose: This procedure will create an action which immediately spawns
//			all the passed in actions.  
// NOTE: If times is NULL then we will assume all actions start simultaneously
//		 and immediately.
void CPutty::CollectManyActions(CString ActionName, //The final action name
								int numActions,		//Number of actions
								CString * Actions,	//Array of action names to collect
								int  * times)		//Array of starting times
{
	Action * pNew, * pPrevAction;
	for(int i = 0; i < numActions; i++)
	{
		pNew = new Action;

		//Fill the action structure
		if(times)
			pNew->BaseTime = times[i];
		else
			pNew->BaseTime = 0.0;
		pNew->Type = SPAWN;
		pNew->ActionToSpawn = Actions[i];
		pNew->next = NULL;
		
		//Build the list
		if (i!=0) { 
			pPrevAction->next = pNew;
		}
		pPrevAction = pNew;

		//Save the head of this Action sequence to the Action Table
		if (i==0)
			m_ActionTable.SetAt(ActionName, pNew);
	}
}

///////////////////////////////////////////////////////////////////////////
// ConcatActions
// Purpose: This procedure will create an action which immediately spawns
//			two actions
void CPutty::CollectTwoActions(CString ActionName1, int time1, CString ActionName2, int time2, 
					CString TargetActionName){

	Action * pNew = new Action;

	pNew->BaseTime = time1;
	pNew->Type = SPAWN;
	pNew->ActionToSpawn = ActionName1;
	pNew->next = new Action;

	pNew->next->BaseTime = time2;
	pNew->next->Type = SPAWN;
	pNew->next->ActionToSpawn = ActionName2;
	pNew->next->next = NULL;

	m_ActionTable.SetAt(TargetActionName, pNew);

}


