

//--Putty.h
// Defines: CPutty Class
#include "render.h"

class CPutty
{
public:
	enum Primitives {SHEET = 1, SPHERE, AUX_SPHERE, CYLINDER, COLLECTION};
	
	//Attributes
	HTREEITEM		m_ObjectViewItem;
	GLUquadricObj * m_quadObj;

	//	Colors
	Vector4f m_DiffuseColor;
	Vector4f m_SpecularColor;
	Vector4f m_EmissionColor;

  // Transformation Matrix
  int UseMatrix;
  double M[9];
	
	//  Material Properties
	float m_Shininess;
//	float m_Alpha;
	int m_Type;
	
	//  Hierarchy
	CPutty * m_Parent;
	CPutty** m_Children;
	int m_NumberChildren;

	//  Control and Link Point Stuff
	Vector3f m_LinkPoint;
	Vector3f m_DOF[3];
	int m_MaxDepth;			// maximum level of recursion in Subdivide
	float m_CutOff;			// level of detail in Subdivide
	float m_MaxLen;			// Maximum triangle length
	int m_CPMax;			// # of control pts and offsets
	ControlPt * m_ControlPts;

	//Motion Attributes
	float m_ForwardIncrement;
	float m_BackwardIncrement;
	float m_RightIncrement;
	float m_LeftIncrement;
	float m_UpIncrement;
	float m_DownIncrement;
	float m_TurnRightIncrement;
	float m_TurnLeftIncrement;
	float m_FastForwardIncrement;

	// Hard Action State Variables
	Vector3f m_TranslateState;
	Vector3f m_RotateState;
	Vector3f m_ScaleState;

	//Cylinder State
	float m_BottomRadius;
	float m_TopRadius;

	//Bubble State
	BOOL m_BubbleOn;
	CString m_BubbleText;

	//  Action Stuff
	CMapStringToPtr m_ActionTable;
	Queue m_ActionQueue;
	CStringList		m_AvailableActions;

 	//Methods
	CPutty(void);
	~CPutty(void);
	void LoadAction(CString action, float time);
	void RegisterAction(CString action);
	CStringList* GetListofActions();
	int Update(float time);
	void CancelAction(CString action);
	void Draw(void);
	void AddChild(CPutty * Child);
	

	//Attributes
	BOOL m_AppearanceChanged;	//Has the appearance of the object changed?
	BOOL m_StateChanged;		//Has the state of the object changed?
	int m_StateList;			//State Display List, caches position and orientation transforms
	int m_AppearanceList;		//Appearance Display List, caches the triangles/color/normals

	//Methods
	void CommitAction(Action * action);
	void Surf(Vector3f & v, Vector3f & u);
	void Deform(Vector3f & v, Vector3f & u);
	void Subdivide(Vector3f & u1, Vector3f & u2, Vector3f & u3, 
						   Vector3f & v1, Vector3f & v2, Vector3f & v3, 
						   int depth);
	void DrawSheet(void);
	void DrawSphere(void);
	void DrawSphereNormals(void);
	void DrawAuxSphere(void);
	
public:
	// Action Library
	void Rotate(int start_time, int max_time, GLfloat angle, int,CString ActionName);
	void RotateSmooth(int start_time, int max_time, GLfloat angle,int, CString ActionName);
	void SmoothColor(int start_time, int end_time,  
					 float r, float g, float b, CString ActionName);	
	void Translate(int, int, GLfloat, GLfloat, GLfloat, CString);
	void TranslateSmooth(int, int, GLfloat, GLfloat, GLfloat, CString);

	void WorldTranslate(int, int, GLfloat, GLfloat, GLfloat, CString);
	void WorldTranslateSmooth(int, int, GLfloat, GLfloat, GLfloat, CString);

	void Scale(int start_time, int end_time,float scale_x, float scale_y,float scale_z, CString ActionName);

	void SmoothDeform(int start_time, int end_time, int CtrlPtIndex, 
						  float xOff,float yOff,float zOff, 
						  float dx, float dy, float dz,
						  float dR, float dL, CString ActionName);
	void ConcatActions(CString ActionName1, CString ActionName2);
	void CollectTwoActions(CString ActionName1, int time1, CString ActionName2, int time2,CString TargetActionName);
	void CollectManyActions(CString ActionName, //The final action name
							int numActions,		//Number of actions
							CString * Actions,	//Array of action names to collect
							int  * times);	//Array of starting times


	// Helper functions
	void RotatePointAxis(GLfloat angle, GLfloat x, GLfloat y, GLfloat z, GLfloat ax,GLfloat ay,GLfloat az);
};

extern "C" {
	void InitTrigArrays(void);
}
