//--types.h

#ifndef __TYPES_H
#define __TYPES_H

#define PI 3.14159


typedef float Vector3f[3];
typedef float Vector4f[4];
typedef int	Vector3i[3];


struct LinkPointTag {

	float ParentPoint[3];
	float ChildPoint[3];
	int DegreesOfFreedom;

};

typedef struct LinkPointTag LinkPoint;


struct ControlPtTag {
	Vector3f Point;
	Vector3f Offset;
	float Locality;
	float Rigidness;
};
typedef struct ControlPtTag ControlPt;

enum ActionTypes { SOFT = 1, ROTATE, TRANSLATE, SCALE, SPAWN, WORLD_TRANSLATE,COLOR};
enum TimeTypes{RANDOM_BASETIME = 1, ADD_PREVIOUS};


struct ActionTag{

	//General Action Attributes
	char Name[30];					//For finding and cancelling actions
	int Type;
	float ThresholdTime;			//HACK:if BaseTime == -69, then random time
	float BaseTime;					//     if it is random, then a random basetime will
	float MaxBaseTime;				//     be picked between Basetime and MaxBaseTime
	float EndTime;					//Used by Hard action to determine when action ends
	struct ActionTag * next;
	float TimeType;
	int mutex;					
	
	//Other StateChanges
	Vector3f DiffuseColor;
	
	//Soft Action Attributes
	int CtrlPtIndex;						//index of the control pt for which this action applies
	ControlPt CtrlPt;			

	//Hard Action Attributes (ROTATE or TRANSLATE or SCALE)
	Vector3f    LinkPt;
	int			RotDOF;  //Which degree of freedom? index: 0,1,or 2
	float		RotAngle;
	Vector3f	Translate;
	Vector3f	ScaleFactor;

	//Spawn Action Attr
	CString ActionToSpawn;

   /* Info on Hard Actions:
	* 
	* If the threshold time on a given action is -1, it is considered to be
	* 'looping action'.  A looping action has the property that it is never 
	* completed until it is 'canceled'.  It's period is
	* EndTime.  After EndTime, the action is repeated again.
	*
	* Note: IT IS UP TO YOU TO INITIALIZE ENDTIME.  IF YOU WISH TO HAVE THIS
	*	    THIS ACTION FINISH RIGHT AFTER IT IS COMMITED YOU MUST MAKE
	*		ENDTIME = STARTIME!!!!
	*/
};

typedef struct ActionTag Action;

struct ActionNodeTag {
	Action * cur;
	float ThresholdTime;
	struct ActionNodeTag * next;
};
typedef struct ActionNodeTag ActionNode;

struct QueueTag{
	ActionNode * head;			// the head of the queue, this action has 
								// the minimum threshold time of all other actions in queue
	ActionNode * tail;			// the tail of the queue
};
typedef struct QueueTag Queue;

#endif //TYPES_H