// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// StateImage.h++

#ifndef _StateImage_H_
#define _StateImage_H_

#include "multiple_ffinder.h"
#include "llSingleP.h"
#include "State.h"
template < class T >
class StateImage
{
    State<T>* array;
    int pixelCount;
    int width;
    llSingleP< State<T> > hitList;
    inline llPix hit( State<T>* );
    State<T>* nbrs[8];
    int nbrCount;
    State<T>** getNbr4( State<T>* );
    State<T>** getNbr8( State<T>* );
    inline State<T>* getState( llPix p ) { return hitList.item( p ); }
    int valid( int u, int v );
    int valid( T* );
    void error( char*, char* );
 public:
    StateImage()
      : array(0), pixelCount(0), width(0), nbrCount(0) {}
    void init( int w, int h );
    inline T* operator()( int u, int v ) { return(array[v*width+u].item()); }
    inline T* operator()( const Coords& c ) { return operator()( c.u, c.v ); }
    inline T* operator()( llPix p ) { return( getState(p)->item() ); }
    inline llPix hit( int u, int v );
    inline llPix hit( Coords& c ) { return hit( c.u, c.v ); }
    void clear();
    void cull();
    int applyfn( int (*fn)( T* ) );
    int applyfnhit( int (*fn)( T* ) );
    void hitNeighbors( llPix );
    llPix next( llPix p ) { return hitList.next( p ); }
    T* item( llPix p ) { return getState(p)->item(); }
    inline void getCoords( llPix, int& u, int& v );
    inline Coords& getCoords( llPix );
    void setLock( llPix p ) { getState(p)->setLock(); }
    ostream& print( ostream& );
    inline int getWidth() { return width; }
    inline int getHeight() { return( (width>0) ? pixelCount/width : 0 ); }
};
template < class T >
void
StateImage<T>::error( char* fn, char* msg )
{
    cerr << "StateImage<T>::" << fn << "(): ERROR: " << msg << endl;
    // exit( 1 );
    assert( 0 );
}
template < class T >
ostream&
StateImage<T>::print( ostream& os )
{
    os << "StateImage " << this << endl;
    os << "\tarray= " << array << endl;
    os << "\tpixelCount= " << pixelCount << endl;
    os << "\twidth= " << width << endl;
    os << "\tNhit= " << hitList.length() << endl;
    os << "\tNnbr= " << nbrCount << endl;
    return os;
}
template < class T >
void
StateImage<T>::init( int w, int h )
{
    if( array )
       delete array;
    width = w;
    pixelCount = w*h;
    array = new State<T>[ pixelCount ];
    hitList.initClass( pixelCount );
}
template < class T >
inline
llPix
StateImage<T>::hit( int u, int v )
{
#ifdef StateImage_Debug
    if( !array ) 
       error( "hit", "not initialized" );
    if( u<0 || u>=width || u*v >= pixelCount || u*v<0 )
       error( "hit", "out of range" );
#endif
    State<T>* sp = array + v*width + u;
    return hit( sp );
}
template < class T >
inline
llPix
StateImage<T>::hit( State<T>* sp )
{
#ifdef StateImage_Debug
    if( !array ) 
       error( "hit", "(State<T>*) not initialized" );
    if( !sp ) 
       error( "hit", "(State<T>*) null arg" );
#endif
    if( sp->isHit() )
       return 0;
    sp -> setHit();
    hitList.append( sp ); 
    return hitList.last(); 
}
template < class T >
void
StateImage<T>::clear()
{
    for( llPix p = hitList.first(); p != 0; p = hitList.next( p ) )
       getState( p ) -> clear();
    hitList.clear();
}
template < class T >
void
StateImage<T>::cull()
{
    llSingleP<T> old; // this looks broken
    old.append( hitList );
    while( ! old.empty() )
       {
	   if( old.head() -> isLocked() )
	      old.pop( hitList );
	   else
	      old.pop();
       }
}
template < class T >
int
StateImage<T>::applyfn( int (*fn)( T* ) )
{
    int count = pixelCount;
    State<T>* sp = array;
    while( count-- )
       if( fn( (sp++)->item() ) )
	  return 1;
    return 0;
}
template < class T >
int
StateImage<T>::applyfnhit( int (*fn)( T* ) )
{
    for( llPix p = hitList.first(); p != 0; p = hitList.next( p ) )
       if( fn( hitList.item( p )->item() ) )
	  return 1;
    return 0;
}
template < class T >
void
StateImage<T>::hitNeighbors( llPix p )
{
    getNbr8( getState(p) );
    for( int i = 0; i < nbrCount; i++ )
       hit( nbrs[i] );
}
template < class T >
State<T>**
StateImage<T>::getNbr4( State<T>* sp )
{
    int index = sp - array;
    if( index < 0 || index >= pixelCount )
       return 0;
    int col = index % width;
    nbrCount = 0;
    if( col > 0 )
       nbrs[ nbrCount++ ] = sp - 1;
    if( col < pixelCount-1 )
       nbrs[ nbrCount++ ] = sp + 1;
    if( index >= width )
       nbrs[ nbrCount++ ] = sp - width;
    if( index < pixelCount - width )
       nbrs[ nbrCount++ ] = sp + width;
    return nbrs;
}
template < class T >
State<T>**
StateImage<T>::getNbr8( State<T>* sp )
{
    int index = sp - array;
    if( index < 0 || index >= pixelCount )
       return 0;
    nbrs[0] = sp-1-width;
    nbrs[1] = sp-width;
    nbrs[2] = sp+1-width;
    nbrs[3] = sp+1;
    nbrs[4] = sp+1+width;
    nbrs[5] = sp+width;
    nbrs[6] = sp-1+width;
    nbrs[7] = sp-1;
    int col = index % width;
    if( col == 0 )
       nbrs[0] = nbrs[6] = nbrs[7] = 0;
    if( col == width-1 )
       nbrs[2] = nbrs[3] = nbrs[4] = 0;
    if( index < width )
       nbrs[0] = nbrs[1] = nbrs[2] = 0;
    if( index >= pixelCount-width )
       nbrs[4] = nbrs[5] = nbrs[6] = 0;
    nbrCount = 0;
    for( int i = 0; i < 8; i++ )
       if( nbrs[i] )
	  nbrs[ nbrCount++ ] = nbrs[i];
    return nbrs;
}
template < class T >
inline
void
StateImage<T>::getCoords( llPix p, int& u, int& v )
{
    getState( p )->getCoords( u, v );
}
template < class T >
inline
Coords& 
StateImage<T>::getCoords( llPix p )
{
    int index = getState(p) - array;
    return Coords( index%width, index/width );
}
#endif /* _StateImage_H_ */
