// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.	These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.	See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the	Massachusetts Institute  of Technology. 	All  
// rights reserved.
// 
// Developed	by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and	its documentation
// for	educational  and	research purposes only and without fee	is hereby
// granted, provided	that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.	If individual
// files are	separated from	this	distribution directory	structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  ffPrior permission  must be	obtained from
// MIT.  These programs shall not  be  used, rewritten, or	adapted as	the
// basis	of	a commercial	software	or	hardware product	without first
// obtaining appropriate licenses  from MIT.	MIT. makes no representations
// about the suitability of this	software for any purpose.  It is provided
// "as is" without express or implied warranty.

// ffBlobMan.c++

#include "ffBlobMan.h"
real ffBlobMan::fleshMean[3] = { 133.839, 106.329, 148.586 };
real ffBlobMan::fleshInvCovariance[9] = {
	0.00228897756499, 0.00422011948326, -0.00442830528997,
		0.00422011948326, 0.03785796579777, 0.00953987271127,
		-0.00442830528997, 0.00953987271127, 0.03405760806490 
};
real ffBlobMan::fleshCovariance[9] = {
	1155.04999999343, -179.25399999901, 200.39499999887,
		-179.25399999901, 56.23929999986, -39.06049999984,
		200.39499999887, -39.06049999984, 66.35939999981 
};
qmGaussianInv ffBlobMan::fleshClass( qmVectorRef( 3, fleshMean ),
																	qmMatrixRef( 3, 3, fleshInvCovariance ));
int ffBlobMan::maxCullffBlobs = 1000;
int ffBlobMan::cullMinffBlob = 20;
int ffBlobMan::ffBlobSpacing = 10;
int ffBlobMan::ffBlobOffset = ffBlobSpacing / 2;
int ffBlobMan::searchMargin = 5;
int ffBlobMan::verbose = 0;
int ffBlobMan::o_data[3][2] = {{ 0, -searchMargin },
{ -searchMargin, 0 },
{ searchMargin, 0 } };
int ffBlobMan::ffCellWidth = 12;
int ffBlobMan::ffCellHeight = 12;
vliRGBPixel ffBlobMan::partColors[3];
int ffBlobMan::coopMode = 3;
int ffBlobMan::crashThresh = 10;
real ffBlobMan::crashRatio = .1;
int ffBlobMan::existenceThresh = 100;
void
ffBlobMan::init( int w, int h )
{
	imageWidth = w;
	imageHeight = h;
	imageCenter(0) = w / 2;
	imageCenter(1) = h / 2;
	int bwidth = ( w / ffBlobSpacing );
	int bheight = ( h / ffBlobSpacing );
	ffBlobCount = bwidth * bheight;
	if( ffBlobList )
		delete ffBlobList;
	if( ffBlobGrid )
		delete ffBlobGrid;
	ffBlobList = new ffBlob[ ffBlobCount ];
	ffBlobGrid = new ffCoords[ ffBlobCount ];
	for( int i = 0; i < bheight; i++ )
		for( int j = 0; j < bwidth; j++ )
			ffBlobGrid[ i*bwidth + j ].set( j*ffBlobSpacing+ffBlobOffset, 
			i*ffBlobSpacing+ffBlobOffset	);
		if( !cullffBlobList )
			cullffBlobList = new int[ maxCullffBlobs ];
		for( i = 0; i < 3; i++ )
			parts[i] = new ffBlob;
		kmeans.setDimension( 2 );
		kmeans.setClusterCount( 3 );
		ffBlob::initClass( w, h, ffCellWidth, ffCellHeight );
		mode = INIT_MODE;
		partColors[ HEAD ].set( 255, 255, 0 );
		partColors[ LHAND ].set( 20, 230, 20 );
		partColors[ RHAND ].set( 230, 20, 20 );
		partVec[ HEAD ].set( 6 );
		partVec[ LHAND ].set( 6 );
		partVec[ RHAND ].set( 6 );
		crashCount[ HEAD ] = 0;
		crashCount[ LHAND ] = 0;
		crashCount[ RHAND ] = 0;
		personExists = 0;
}
int
ffBlobMan::process( vliRGBImageRef& image , ffPrior &ffPrior )
{
	switch( mode )
	{
	case INIT_MODE:
		 if( ! init( image , ffPrior ) )
				mode = TRACK_MODE;
		 break;
	case TRACK_MODE:
		 if( track( image , ffPrior ) || collide() )
				mode = INIT_MODE;
		 break;
	case INITONLY_MODE:
		 return init( image , ffPrior );
	}
	return 0;
}
void
ffBlobMan::drawBlobs( vliYUVImageRef& yuv, vliYUVPixel* color )
{
	static vliYUVPixel yellow( 210, 16, 146 );
	static vliYUVPixel cyan( 128, 146, 16 );
	if( !color )
		color = (TRACK_MODE==mode) ? &yellow : &cyan;
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawStats( yuv, color );
	
}
void
ffBlobMan::drawPoints( vliYUVImageRef& yuv, vliYUVPixel* color )
{
	static vliYUVPixel yellow( 210, 16, 146 );
	static vliYUVPixel cyan( 128, 146, 16 );
	if( !color )
		color = (TRACK_MODE==mode) ? &yellow : &cyan;
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawPoints( yuv, color );
}
void
ffBlobMan::drawBlobs( vliRGBImageRef& rgb, vliRGBPixel* color )
{
	static vliRGBPixel cyan( 0, 128, 128 );
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawStats( rgb, (color) ? color : 
	(INIT_MODE!=mode) ? &partColors[i] : &cyan );
}
void
ffBlobMan::drawPoints( vliRGBImageRef& rgb, vliRGBPixel* color )
{
	static vliRGBPixel cyan( 0, 128, 128 );
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawPoints( rgb, (color) ? color : 
	(INIT_MODE!=mode) ? &partColors[i] : &cyan );
}
istream&
ffBlobMan::readFleshStats( istream& is )
{
	qmVectorRef( 3, fleshMean ).scan( is );
	qmMatrixRef( 3, 3, fleshInvCovariance ).scan( is );
	fleshClass.set( qmVectorRef( 3, fleshMean ),
		qmMatrixRef( 3, 3, fleshInvCovariance ) );
	return is;
}
void
ffBlobMan::setFleshStats( const qmVectorRef& m, const qmMatrixRef& ic )
{
	qmVectorRef( 3, fleshMean ).copy( m );
	qmMatrixRef( 3, 3, fleshInvCovariance ).copy( ic );
	fleshClass.set( qmVectorRef( 3, fleshMean ),
		qmMatrixRef( 3, 3, fleshInvCovariance ) );
}
//#include <ieeefp.h>
static
inline
ffBlob*&
closestx( ffBlob*& b0, ffBlob*& b1, ffBlob*& b2 )
{
	real x0 = b0 -> getSpatial().getMean().getValue( 0 );
	real x1 = b1 -> getSpatial().getMean().getValue( 0 );
	real x2 =  b2 -> getSpatial().getMean().getValue( 0 );
	real d01 = x1-x0;
	d01 *= d01;
	real d02 = x2-x0;
	d02 *= d02;
	return( ( d01 < d02 ) ? b1 : b2 );
}
static
inline
void
swap( ffBlob*& b1, ffBlob*& b2 )
{
	ffBlob* tmp = b1;
	b1 = b2;
	b2 = tmp;
}
int
ffBlobMan::init( vliRGBImageRef& rgb)
{
	if( verbose )
		cerr << "INIT:" << endl;
	ffBlob::clearClass();
	int misses = 0;
	ffBlob::clearClass();
	llSingleP< ffPixel > ffPixelList;
	int totalFleshCount = 0;
	for( int k = 0; k < ffBlobCount; k++ )
	{
		 ffBlobList[k].clear();
		 misses = ffBlobList[ k ].expandWoohoo3( ffBlobGrid[k],
						 fleshClass,
						 rgb,
						 0, 0, 0, 0,
						 ffPixelList );
		 if( !misses )
		 {
			 totalFleshCount += ffPixelList.length();
			 for( llPix p = ffPixelList.first(); 
			 p != 0;
			 p = ffPixelList.next( p ) )
				 ffBlobList[k].append( ffPixelList(p)->getCoords() );
			 ffPixelList.clear();
		 }
	}
	if( verbose )
		cerr << "totalFleshCount= " << totalFleshCount << endl;
	personExists = ( totalFleshCount > existenceThresh ) ? 1 : 0;
	if( !personExists )
		return 1;
	cullffBlobCount = 0;
	for( k = 0; k < ffBlobCount; k++ )
	{
		 if( ffBlobList[k].getPointCount() > cullMinffBlob )
		 {
			 if( cullffBlobCount < maxCullffBlobs )
				 cullffBlobList[ cullffBlobCount ] = k;
			 cullffBlobCount++;
		 }
	}
	if( verbose) 
		cerr << "\t" << cullffBlobCount << " medium ffBlobs" << endl;
	if( cullffBlobCount < 3 )
		return 1;
	int itop = 0;
	int ileft = 0;
	int iright = 0;
	real ytop = ffBlobList[ cullffBlobList[0] ].getSpatial().getMean().getValue(1);
	real xleft = ffBlobList[ cullffBlobList[0]].getSpatial().getMean().getValue(0);
	real xright = xleft;
	for( k = 0; k < cullffBlobCount; k++ )
	{
		 real x =
				ffBlobList[ cullffBlobList[k] ].getSpatial().getMean().getValue(0);
		 real y = 
				ffBlobList[ cullffBlobList[k] ].getSpatial().getMean().getValue(1);
		 if( y < ytop )
				ytop = y, itop = k;
		 if( x < xleft )
				xleft = x, ileft = k;
		 if( x > xright )
				xright = x, iright = k;
	}
	if( ileft == iright )
	{
		 cerr << "init: uh-oh, left=right, abort" << endl;
		 return 1;
	}
	if( itop == ileft || itop == iright )
	{
		 int imost = 0;
		 int nmost = 0;
		 for( k = 0; k < cullffBlobCount; k++ )
		 {
			 if( k == ileft || k == iright )
				 continue;
			 int n =	ffBlobList[ cullffBlobList[k] ].getPointCount();
			 if( n > nmost )
				 nmost = n, imost = k;
		 }
		 itop = imost;
	}
	kmeans.setSampleCount( cullffBlobCount );
	kmeans.setClusterMean
		( HEAD, ffBlobList[cullffBlobList[itop]].getSpatial().getMean() );
	kmeans.setClusterMean
		( LHAND, ffBlobList[cullffBlobList[iright]].getSpatial().getMean() );
	kmeans.setClusterMean
		( RHAND, ffBlobList[cullffBlobList[ileft]].getSpatial().getMean());
	for( int i = 0; i < cullffBlobCount; i++ )
	{
		 ffBlob& ffBlob = ffBlobList[ cullffBlobList[i] ];
		 kmeans.setSample( i, ffBlob.getSpatial().getMean() );
		 kmeans.setSampleWeight( i, ffBlob.getPointCount() );
	}
	kmeans.cluster();
	ffBlob::clearClass();
	int fail;
	qmVector c( 2 );
	for( i = 0; i < 3; i++ )
	{
		 c = kmeans.getClusterMean( i );
		 parts[i] -> clear();
		 ffPixelList.clear();
		 fail = parts[i] -> expandWoohoo3( ffCoords((short)c(0), (short)c(1)),
			 fleshClass,
			 rgb,
			 0, 0, 0, 0,
			 ffPixelList );
		 if( fail )
				return 1;
		 if( ! ffPixelList.length() )
				error( "init", "whoops, no point count, no fail" );
		 for( llPix p = ffPixelList.first(); 
		 p != 0;
		 p = ffPixelList.next( p ) )
				parts[i] -> append( ffPixelList(p)->getCoords() );
	}
	qmVectorRef hdMean = parts[ HEAD ]->getSpatial().getMean();
	qmVectorRef lhMean = parts[ LHAND ]->getSpatial().getMean();
	qmVectorRef rhMean = parts[ RHAND ]->getSpatial().getMean();
	if( hdMean(1) > imageHeight/2 )
	{
		 static qmVector hl( 2 );
		 static qmVector hr( 2 );
		 hl.setSubtract( lhMean, hdMean );
		 hl.setNormalize();
		 hr.setSubtract( rhMean, hdMean );
		 hr.setNormalize();
		 if( hl(1) < -.7 || hr(1) < -.7 )
		 {	
			 alert( "init", "head is bullshit (fixing)" );
			 //swap( parts[HEAD], highest( parts[LHAND], parts[RHAND] ) );
			 swap( parts[HEAD], 
				 closestx( parts[HEAD], parts[LHAND], parts[RHAND] ) );
		 }
	}
	return 0;
}

int
ffBlobMan::init( vliRGBImageRef& rgb , ffPrior &ffPrior)
{
	if( verbose )
		cerr << "INIT:" << endl;
	ffBlob::clearClass();
	int misses = 0;
	ffBlob::clearClass();
	llSingleP< ffPixel > ffPixelList;
	int totalFleshCount = 0;
	for( int k = 0; k < ffBlobCount; k++ )
	{
		 ffBlobList[k].clear();
		 misses = ffBlobList[ k ].expandWoohoo3( ffBlobGrid[k],
						 fleshClass,
						 rgb,
						 0, 0, 0, 0,
						 ffPixelList );
		 if( !misses )
		 {
			 totalFleshCount += ffPixelList.length();
			 for( llPix p = ffPixelList.first(); 
			 p != 0;
			 p = ffPixelList.next( p ) )
				 ffBlobList[k].append( ffPixelList(p)->getCoords() );
			 ffPixelList.clear();
		 }
	}
	if( verbose )
		cerr << "totalFleshCount= " << totalFleshCount << endl;
	personExists = ( totalFleshCount > existenceThresh ) ? 1 : 0;
	if( !personExists )
		return 1;
	cullffBlobCount = 0;
	for( k = 0; k < ffBlobCount; k++ )
	{
		 if( ffBlobList[k].getPointCount() > cullMinffBlob )
		 {
			 if( cullffBlobCount < maxCullffBlobs )
				 cullffBlobList[ cullffBlobCount ] = k;
			 cullffBlobCount++;
		 }
	}
	if( verbose) 
		cerr << "\t" << cullffBlobCount << " medium ffBlobs" << endl;
	if( cullffBlobCount < 3 )
		return 1;
	int itop = 0;
	int ileft = 0;
	int iright = 0;
	real ytop = ffBlobList[ cullffBlobList[0] ].getSpatial().getMean().getValue(1);
	real xleft = ffBlobList[ cullffBlobList[0]].getSpatial().getMean().getValue(0);
	real xright = xleft;
	for( k = 0; k < cullffBlobCount; k++ )
	{
     real x =
				ffBlobList[ cullffBlobList[k] ].getSpatial().getMean().getValue(0);
		 real y = 
				ffBlobList[ cullffBlobList[k] ].getSpatial().getMean().getValue(1);
		 if( y < ytop )
				ytop = y, itop = k;
		 if( x < xleft )
				xleft = x, ileft = k;
		 if( x > xright )
				xright = x, iright = k;
	}

	if( ileft == iright )
	{
		 cerr << "init: uh-oh, left=right, abort" << endl;
		 return 1;
	}

	if( itop == ileft || itop == iright )
	{
		 int imost = 0;
		 int nmost = 0;

		 for( k = 0; k < cullffBlobCount; k++ )
		 {
			 if( k == ileft || k == iright )
				 continue;

			 int n =	ffBlobList[ cullffBlobList[k] ].getPointCount();

			 if( n > nmost )
				 nmost = n, imost = k;
		 }
		 itop = imost;
	}

	kmeans.setSampleCount( cullffBlobCount );

	kmeans.setClusterMean
		( HEAD, ffBlobList[cullffBlobList[itop]].getSpatial().getMean() );

	kmeans.setClusterMean
		( LHAND, ffBlobList[cullffBlobList[iright]].getSpatial().getMean() );

	kmeans.setClusterMean
		( RHAND, ffBlobList[cullffBlobList[ileft]].getSpatial().getMean());

	for( int i = 0; i < cullffBlobCount; i++ )
	{
		 ffBlob& ffBlob = ffBlobList[ cullffBlobList[i] ];
		 kmeans.setSample( i, ffBlob.getSpatial().getMean() );
		 kmeans.setSampleWeight( i, ffBlob.getPointCount() );
	}
	kmeans.cluster();
	ffBlob::clearClass();
	int fail;
	qmVector c( 2 );
	for( i = 0; i < 3; i++ )
	{
		 c = kmeans.getClusterMean( i );
		 parts[i] -> clear();
		 ffPixelList.clear();
		 fail = parts[i] -> expandWoohoo3( ffCoords((short)c(0), (short)c(1)),
			 fleshClass,
			 rgb,
			 0, 0, 0, 0,
			 ffPixelList );
		 if( fail )
				return 1;
		 if( ! ffPixelList.length() )
				error( "init", "whoops, no point count, no fail" );
		 for( llPix p = ffPixelList.first(); 
		 p != 0;
		 p = ffPixelList.next( p ) )
				parts[i] -> append( ffPixelList(p)->getCoords() );
	}
	qmVectorRef hdMean = parts[ HEAD ]->getSpatial().getMean();
	qmVectorRef lhMean = parts[ LHAND ]->getSpatial().getMean();
	qmVectorRef rhMean = parts[ RHAND ]->getSpatial().getMean();
	if( hdMean(1) > imageHeight/2 )
	{
		 static qmVector hl( 2 );
		 static qmVector hr( 2 );
		 hl.setSubtract( lhMean, hdMean );
		 hl.setNormalize();
		 hr.setSubtract( rhMean, hdMean );
		 hr.setNormalize();
		 if( hl(1) < -.7 || hr(1) < -.7 )
		 {	
			 alert( "init", "head is bullshit (fixing)" );
			 //swap( parts[HEAD], highest( parts[LHAND], parts[RHAND] ) );
			 swap( parts[HEAD], 
				 closestx( parts[HEAD], parts[LHAND], parts[RHAND] ) );
		 }
	}
	return 0;
}

int
ffBlobMan::track( vliRGBImageRef& image , ffPrior &ffPrior )
{
	if( coopMode )
		return( trackCoop3( image , ffPrior ) );
	alert( "track", "whoops" );
	return 1;
}
int
ffBlobMan::trackCoop3( vliRGBImageRef& image , ffPrior &ffPrior )
{
	ffBlob::clearClass();
	static qmGaussianRef space[3];
	static qmGaussianInv invSpace[3];
	static qmGaussianInvRef invSpaceRef[3];
	static real dist2[3];
	ffCoords cc;
	llSingleP< ffPixel > ffPixelList;
	for( int i = 0; i < 3; i++ )
	{
		int valid = ffPrior.valid(i);
		 if (valid)
		 {
			 space[i] = ffPrior.getSpatial(i);
			 invSpace[i].set( space[i].getDim() );
			 invSpaceRef[i] = invSpace[i];
			 if( qmStatistics::invert( space[i], invSpaceRef[i] ) )
			 {
				 invSpaceRef[i].setSingular();
			 }
		 }
		 else
		 {
			 if( ! parts[i] -> getPointCount() )
			 {
				 alert( "trackCoop3", "empty ffBlob in track!!!" );
				 return 1;
			 }
			 space[i] = parts[i] -> getSpatial();
			 invSpace[i].set( space[i].getDim() );
			 invSpaceRef[i] = invSpace[i];
			 if( qmStatistics::invert( space[i], invSpaceRef[i] ) )
			 {
				 alert( "trackCoop3", "singular class !!!!" );
				 return 1;
			 }
		 }
	}
	for( i = 0; i < 3; i++ )
		crashCount[i] = 0;
	for( i = 0; i < 3; i++ )
	{
		 cc.u = invSpaceRef[i].getMean().getValue( 0 );
		 cc.v = invSpaceRef[i].getMean().getValue( 1 );
		 ffBlob::expandWoohoo3( cc, fleshClass, image, 3, invSpaceRef, 
			 dist2, crashCount, ffPixelList);
		 parts[i] -> clear(); // !! I hope this does not affect invSpace
	}
	ffPixel* pp;
	for( llPix p = ffPixelList.first(); p != 0; p = ffPixelList.next( p ) )
	{
		 pp = ffPixelList( p );
		 if( pp->whichClass < 0 )
				error( "trackCoop3", "null class assignment" );
		 parts[ pp->whichClass ] -> append( pp->getCoords() );
	}
	for( i = 0; i < 3; i++ )
		if( ! parts[i] -> getPointCount() )
		{
			if (! ffPrior.valid(i) )
			 {
				 if (verbose) 
					 alert( "trackCoop3", "class starved." );
				 return 1;
			 }
			else
			 {
				if (verbose) 
					alert( "trackCoop3", "coasting on valid ffPrior." );
				
				//					parts[i] -> setSpatial( ffPrior.getSpatial(i) );
			 }
		}
		return 0;
}
int
ffBlobMan::collide()
{
	for( int i = 0; i < 3; i++ )
	{
		 if( real( crashCount[i] ) / real( parts[i] -> getPointCount() )
			 > crashRatio )
		 {
			 if (verbose) 
				 alert( "collide", "found a collision." );
			 return 1;
		 }
	}
	return 0;
}
ostream&
ffBlobMan::alert( char* fn, char* msg )
{
	return( cerr << "ffBlobMan: ERROR: " << fn << "(): " << msg << endl );
}

void
ffBlobMan::error( char* fn, char* msg )
{
	alert( fn, msg );
	assert( 0 );
}
const qmVectorRef&
ffBlobMan::getffBlobMean_Pics( int i )
{
	if( i < 0 || i > 2 )
		error( "getffBlobMean_Pics", "arg out of range" );
	if( ! parts[i] -> getPointCount() )
	{
		partVec[i].setZero();
		// alert( "getffBlobMean_Pics", "empty ffBlob" );
	}
	else
	{
		partVec[i].subVector( 0, 2 ).copy( parts[i] -> getSpatial().getMean() );
		partVec[i].subVector( 0, 2 ) -= imageCenter;
		partVec[i].subVector( 0, 2 ) /= imageWidth;
		partVec[i].subVector( 2, 3 ).copy( parts[i] -> getSpatialCovVect() );
		partVec[i].subVector( 2, 3 ) /= (imageWidth*imageWidth);
		partVec[i](5) = 1.0;
	}
	return partVec[i];
}
float*
ffBlobMan::getQuickPoints_Pics()
{
	getQuickPoints();
	for( int i = 0; i < 6; i++ )
		qmVectorRef( 2, quickPoints+2*i ) -= imageCenter;
	qmVectorRef( 12, quickPoints ) /= imageWidth;
	return quickPoints;
}
float*
ffBlobMan::getQuickPoints()
{
	for( int i = 0; i < 3; i++ )
		parts[i] -> getQuickPoints( quickPoints + i*4 );
	return quickPoints;
}
