// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  ffPrior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// ffCell.c++

#include "ffCell.h"
float ffCell::classifyPercentThresh = .2;
real ffCell::maha2ciel = 9.;
real ffCell::mutualThresh = 1.2;
// void
// ffCell::init( int n, ffPixel** p )
void
ffCell::init( int w, int h, ffPixel** p )
{
    ffPixelCount = w*h;
    width = w;
    array = p;
    stackCount = 0;
    stack = p+ffPixelCount;
}
int
ffCell::set( const vliRGBImageRef& vim )
{
    for( int i = 0; i < ffPixelCount; i++ )
       array[i] -> set( vim );
    return 0;
}
int 
ffCell::classify( const qmGaussianInvRef& colorClass, int sub )
{
    int tries = 0;
    int index;
    stackCount = 0;
    int height = ffPixelCount / width;
    for( int j = 0; j < height; j += sub )
       for( int i = 0; i < width; i += sub )
	  {
	      tries++;
	      index = j*width+i;
	      if( array[index] -> classify( colorClass ) )
		 stack[ stackCount++ ] = array[index];
       }
    float percent = float(stackCount) / float(tries);
    if( percent > classifyPercentThresh )
       return 1;
    return 0;
}
int
ffCell::setAndClassify( const vliRGBImageRef& vim,
		      const qmGaussianInvRef& colorClass,
		      int n,
		      qmGaussianInvRef* sp,
		      real* dist2,
		      int* crash,
		      int sub )
{
    int tries = 0;
    stackCount = 0;
    int index;
    int height = ffPixelCount / width;
    for( int j = 0; j < height; j += sub )
       for( int i = 0; i < width; i += sub )
	  {
	      tries++;
	      index = j*width+i;
	      array[index] -> set( vim );
	      if( array[index] -> classify( colorClass, n, sp, 
					    dist2, crash ) )
		 stack[ stackCount++ ] = array[index];
	  }
    float percent = float(stackCount) / float(tries);
    if( percent > classifyPercentThresh )
       return 1;
    return 0;
}
ostream&
ffCell::alert( char* fn, char* msg )
{
    return( cerr << "ffCell: ERROR: " << fn << "(): " << msg << endl );
}
void
ffCell::error( char* fn, char* msg )
{
    alert( fn, msg );
    // exit( 1 );
    assert(0);
}
