// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// Blob.c++

#include "ffinderenv.h"
#include "Blob.h"
#include "qmMinMax.h"

qmVector Blob::tmpPixelVector(3);

const qmGaussianRef&
Blob::computeSpatial()
{
    if( !spatialFlag )
       {
	   qmVector spatialSample( 2 );
	   qmSampleAcc spatialSampleAcc( 2 );
	   spatialSampleAcc.setZero();
	   Coords* cp;
	   for( llPix p = coordList.first(); p != 0; p = coordList.next( p ) )
	      {
		  cp = coordList( p );
		  spatialSample(0) = cp->u;
		  spatialSample(1) = cp->v;
		  spatialSampleAcc.addSample( spatialSample );
	      }
	   spatialSampleAcc.getGaussian( spatialStats );
	   spatialFlag = 1;
       }
    return spatialStats;
}
ostream&
Blob::alert( char* fn, char* msg )
{
    return( cerr << "Blob: ERROR: " << fn << "(): " << msg << endl );
}
void
Blob::error( char* fn, char* msg )
{
    alert( fn, msg );
    // exit( 1 );
    assert(0);
}
int
Blob::similar( const vliRGBPixel& pixel,
	       const qmGaussianInvRef& colorClass,
	       real thresh2 )
{
    real maha2 = colorClass.getMahalanobisSquare( pixelToVector( pixel ));
    if( maha2 < thresh2 )
       return 1;
    return 0;
}
int
Blob::similar( const vliRGBPixel& pixel,
	       const qmGaussianInvRef& colorClass)
{
    real maha2 = colorClass.getMahalanobisSquare( pixelToVector( pixel ));
    if( maha2 < env->simThresh2 )
       return 1;
    return 0;
}

void
Blob::drawStats( vliYUVImageRef& vim, 
		 const qmGaussianRef& g,
		 const vliYUVPixel& color )
{
    static qmMatrix invcov( 2, 2 );
    if( invcov.setInvertCheck( g.getCovariance() ) )
       // cerr << "Blob::drawStats singular inversion" << endl;
       return;
    vim.drawBlob( g.getMean().getValue(0),
		  g.getMean().getValue(1),
		  invcov.getValue(0,0),
		  invcov.getValue(1,1),
		  invcov.getValue(0,1),
		  vliYUVPixel(color) );
}

void
Blob::drawStats( vliRGBImageRef& vim, 
		 const qmGaussianRef& g,
		 const vliRGBPixel& color )
{
    qmMatrix invcov( 2, 2 );
    real det = g.getCovariance().getDeterminant2();
    if( det < 1e-14 )
       return;
    if( invcov.setInvertCheck( g.getCovariance() ) )
       cerr << "Blob::drawStats-RGB singular inversion" << endl;
    vim.drawBlob( g.getMean().getValue(0),
		  g.getMean().getValue(1),
		  invcov.getValue(0,0),
		  invcov.getValue(1,1),
		  invcov.getValue(0,1),
		  vliRGBPixel(color) );
}

void
Blob::initSearchCoords( int d )
{
    int n = 1 + d*4;
    if( n <= env->searchCoordsCount )
       return;
    if( env->searchCoords )
       delete env->searchCoords;
    env->searchCoords = new Coords[ n ];
    env->searchCoords[ env->searchCoordsCount++ ] = Coords( 0, 0 );
    for( int i = 1; i <= d; i++ )
       {
	   env->searchCoords[ env->searchCoordsCount++ ] = Coords( i, 0 );
	   env->searchCoords[ env->searchCoordsCount++ ] = Coords( -i, 0 );
	   env->searchCoords[ env->searchCoordsCount++ ] = Coords( 0, i );
	   env->searchCoords[ env->searchCoordsCount++ ] = Coords( 0, -i );
       }
}

Blob::Blob()
  : colorFlag( 0 ),
    colorStats( 3 ),
    spatialFlag( 0 ),
    spatialStats( 2 )
{}
int
Blob::operator==( Blob& b )
{
    if( getColor().getMean() == b.getColor().getMean() &&
	getColor().getInvCovariance() == b.getColor().getInvCovariance() )
       return 1;
    return 0;
}
const qmVectorRef&
Blob::getSpatialCovVect()
{
    static qmVector vec(3);
    static qmMatrix cov( 2, 2 );
    cov = getSpatial().getCovariance();
    vec( 0 ) = cov( 0, 0 );
    vec( 1 ) = cov( 1, 1 );
    vec( 2 ) = cov( 0, 1 );
    return vec;
}
void
Blob::drawPoints( vliYUVImageRef& vim, vliYUVPixel* color )
{
    static vliYUVPixel yellow( 210, 16, 146 );
    if( ! color )
       color = &yellow;
    for( llPix index = coordList.first(); index != 0; 
	 index = coordList.next(index) )
       {
	   Coords* cp = coordList( index );
	   vim.setPixel( cp->u, cp->v, *color );
       }
}
void
Blob::drawPoints( vliRGBImageRef& vim, vliRGBPixel* color )
{
    static vliRGBPixel yellow( 255, 255, 0 );
    if( ! color )
       color = &yellow;
    for( llPix index = coordList.first(); index != 0; 
	 index = coordList.next(index) )
       {
	   Coords* cp = coordList( index );
	   vim.setPixel( cp->u, cp->v, *color );
       }
}

void
Blob::drawStats( vliYUVImageRef& vim, vliYUVPixel* color )
{
    static vliYUVPixel yellow( 210, 16, 146 );
    drawStats( vim, getSpatial(), (color) ? *color : yellow );
}
void
Blob::drawStats( vliRGBImageRef& vim, vliRGBPixel* color )
{
    static vliRGBPixel yellow( 255, 255, 0 );
    // !! debug
    if( coordList.empty() )
       return;
    drawStats( vim, getSpatial(), (color) ? *color : yellow );
}
int
Blob::collide( Blob& b )
{
    // ( 5 pixels )( 1/320 )
    // const real collideThresh = ( 1./64. );
    qmVector diff( 2 );
    diff.setSubtract( getSpatial().getMean(), b.getSpatial().getMean() );
    real dist_pixels = sqrt( diff.getInnerProduct( diff ) );
    if( dist_pixels < env->collideThresh )
       return 1;
    return 0;
}
real
Blob::getEccentricity()
{
    static qmEigenStructure eig;
    if( !spatialFlag )
        {
	    // alert( "getEccentricity", "spatial stats invalid" );
	    computeSpatial();
        }
    eig.set( spatialStats.getCovariance() );
    int imin;
    qmVectorRef evals = eig.getEigValues();
    // cerr << "evals= " << evals << endl;
    evals.getMin( &imin );
    return( sqrt(evals(imin)) / sqrt(evals(1-imin)) );
}
void
Blob::getQuickPoints( float* p )  // in pixels
{
    static qmEigenStructure eig;
    static qmMatrix evec( 2, 2 );
    static qmVector eval( 2 );
    if( !spatialFlag )
       {
	   alert( "getQuickPoints", "spatial stats not valid" );
	   computeSpatial();
       }
    if( qmEigenStructure::eigenSymmetric2
	( spatialStats.getCovariance(), eval, evec ) )
      {
	alert( "getQuickPoints", "eigen2 error" );
	qmVectorRef(2,p).setZero();
	qmVectorRef(2,p+2).setZero();
      }
    else
      {
	int imax;
	int imin;
	real lmax = sqrt( eval.getMax( &imax ) );
	real lmin = sqrt( eval.getMin( &imin ) );
	real e = sqrt(eval(1-imax)) / sqrt(eval(imax));
	real alpha = ( e < .5 ) ? 1 : ( 1-2*(e-.5) );
	alpha *= 2*lmax;
	qmVectorRef(2,p).copy( spatialStats.getMean() ).setScaleAddEq
       ( -alpha, evec.getRow(imax) );
	qmVectorRef(2,p+2).copy( spatialStats.getMean() ).setScaleAddEq
	  ( alpha, evec.getRow(imax) );
      }
}
