// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.	These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.	See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the	Massachusetts Institute  of Technology. 	All  
// rights reserved.
// 
// Developed	by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and	its documentation
// for	educational  and	research purposes only and without fee	is hereby
// granted, provided	that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.	If individual
// files are	separated from	this	distribution directory	structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be	obtained from
// MIT.  These programs shall not  be  used, rewritten, or	adapted as	the
// basis	of	a commercial	software	or	hardware product	without first
// obtaining appropriate licenses  from MIT.	MIT. makes no representations
// about the suitability of this	software for any purpose.  It is provided
// "as is" without express or implied warranty.

// BlobMan.c++
#include "ffinderenv.h"
#include "BlobMan.h"

void
BlobMan::init( int w, int h )
{
	imageWidth = w;
	imageHeight = h;
	imageCenter(0) = w / 2;
	imageCenter(1) = h / 2;
	int bwidth = ( w / env->blobSpacing );
	int bheight = ( h / env->blobSpacing );
	blobCount = bwidth * bheight;
	if( blobList )
		delete blobList;
	if( blobGrid )
		delete blobGrid;
	blobList = new Blob[ blobCount ];
	blobGrid = new Coords[ blobCount ];
	for( int i = 0; i < bheight; i++ )
		for( int j = 0; j < bwidth; j++ )
			blobGrid[ i*bwidth + j ].set( j*env->blobSpacing+env->blobOffset, 
			i*env->blobSpacing+env->blobOffset	);
		if( !cullBlobList )
			cullBlobList = new int[ env->maxCullBlobs ];
		for( i = 0; i < 3; i++ )
			parts[i] = new Blob;
		kmeans.setDimension( 2 );
		kmeans.setClusterCount( 3 );
		env->initBlobClass( w, h, env->cellWidth, env->cellHeight );
		mode = INIT_MODE;
		env->partColors[ HEAD ].set( 255, 255, 0 );
		env->partColors[ LHAND ].set( 20, 230, 20 );
		env->partColors[ RHAND ].set( 230, 20, 20 );
		partVec[ HEAD ].set( 6 );
		partVec[ LHAND ].set( 6 );
		partVec[ RHAND ].set( 6 );
		crashCount[ HEAD ] = 0;
		crashCount[ LHAND ] = 0;
		crashCount[ RHAND ] = 0;
		personExists = 0;
}
int
BlobMan::process( vliRGBImageRef& image , Prior &prior )
{
	switch( mode )
	{
	case INIT_MODE:
		 if( ! init( image , prior ) )
				mode = TRACK_MODE;
		 break;
	case TRACK_MODE:
		 if( track( image , prior ) || collide() )
				mode = INIT_MODE;
		 break;
	case INITONLY_MODE:
		 return init( image , prior );
	}
	return 0;
}
void
BlobMan::drawBlobs( vliYUVImageRef& yuv, vliYUVPixel* color )
{
	static vliYUVPixel yellow( 210, 16, 146 );
	static vliYUVPixel cyan( 128, 146, 16 );
	if( !color )
		color = (TRACK_MODE==mode) ? &yellow : &cyan;
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawStats( yuv, color );
	
}
void
BlobMan::drawPoints( vliYUVImageRef& yuv, vliYUVPixel* color )
{
	static vliYUVPixel yellow( 210, 16, 146 );
	static vliYUVPixel cyan( 128, 146, 16 );
	if( !color )
		color = (TRACK_MODE==mode) ? &yellow : &cyan;
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawPoints( yuv, color );
}
void
BlobMan::drawBlobs( vliRGBImageRef& rgb, vliRGBPixel* color )
{
	static vliRGBPixel cyan( 0, 128, 128 );
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawStats( rgb, (color) ? color : 
	(INIT_MODE!=mode) ? &(env->partColors[i]) : &cyan );
}
void
BlobMan::drawPoints( vliRGBImageRef& rgb, vliRGBPixel* color )
{
	static vliRGBPixel cyan( 0, 128, 128 );
	for( int i = 0; i < 3; i++ )
		parts[i] -> drawPoints( rgb, (color) ? color : 
	(INIT_MODE!=mode) ? &(env->partColors[i]) : &cyan );
}
istream&
BlobMan::readFleshStats( istream& is )
{
	qmVectorRef( 3, env->fleshMean ).scan( is );
	qmMatrixRef( 3, 3, env->fleshInvCovariance ).scan( is );
	env->fleshClass.set( qmVectorRef( 3, env->fleshMean ),
		qmMatrixRef( 3, 3, env->fleshInvCovariance ) );
	return is;
}
void
BlobMan::setFleshStats( const qmVectorRef& m, const qmMatrixRef& ic )
{
	qmVectorRef( 3, env->fleshMean ).copy( m );
	qmMatrixRef( 3, 3, env->fleshInvCovariance ).copy( ic );
	env->fleshClass.set( qmVectorRef( 3, env->fleshMean ),
		qmMatrixRef( 3, 3, env->fleshInvCovariance ) );
}
//#include <ieeefp.h>
static
inline
Blob*&
closestx( Blob*& b0, Blob*& b1, Blob*& b2 )
{
	real x0 = b0 -> getSpatial().getMean().getValue( 0 );
	real x1 = b1 -> getSpatial().getMean().getValue( 0 );
	real x2 =  b2 -> getSpatial().getMean().getValue( 0 );
	real d01 = x1-x0;
	d01 *= d01;
	real d02 = x2-x0;
	d02 *= d02;
	return( ( d01 < d02 ) ? b1 : b2 );
}
static
inline
void
swap( Blob*& b1, Blob*& b2 )
{
	Blob* tmp = b1;
	b1 = b2;
	b2 = tmp;
}
int
BlobMan::init( vliRGBImageRef& rgb)
{
	if( env->verbose )
		cerr << "INIT:" << endl;
	env->clearBlobClass();
	int misses = 0;
	env->clearBlobClass();
	llSingleP< Pixel > pixelList;
	int totalFleshCount = 0;
	for( int k = 0; k < blobCount; k++ )
	{
		 blobList[k].clear();
		 misses = env->expandWoohoo3Blob( blobGrid[k],
						 env->fleshClass,
						 rgb,
						 0, 0, 0, 0,
						 pixelList );
		 if( !misses )
		 {
			 totalFleshCount += pixelList.length();
			 for( llPix p = pixelList.first(); 
			 p != 0;
			 p = pixelList.next( p ) )
				 blobList[k].append( pixelList(p)->getCoords() );
			 pixelList.clear();
		 }
	}
	if( env->verbose )
		cerr << "totalFleshCount= " << totalFleshCount << endl;
	personExists = ( totalFleshCount > env->existenceThresh ) ? 1 : 0;
	if( !personExists )
		return 1;
	cullBlobCount = 0;
	for( k = 0; k < blobCount; k++ )
	{
		 if( blobList[k].getPointCount() > env->cullMinBlob )
		 {
			 if( cullBlobCount < env->maxCullBlobs )
				 cullBlobList[ cullBlobCount ] = k;
			 cullBlobCount++;
		 }
	}
	if( env->verbose) 
		cerr << "\t" << cullBlobCount << " medium blobs" << endl;
	if( cullBlobCount < 3 )
		return 1;
	int itop = 0;
	int ileft = 0;
	int iright = 0;
	real ytop = blobList[ cullBlobList[0] ].getSpatial().getMean().getValue(1);
	real xleft = blobList[ cullBlobList[0]].getSpatial().getMean().getValue(0);
	real xright = xleft;
	for( k = 0; k < cullBlobCount; k++ )
	{
		 real x =
				blobList[ cullBlobList[k] ].getSpatial().getMean().getValue(0);
		 real y = 
				blobList[ cullBlobList[k] ].getSpatial().getMean().getValue(1);
		 if( y < ytop )
				ytop = y, itop = k;
		 if( x < xleft )
				xleft = x, ileft = k;
		 if( x > xright )
				xright = x, iright = k;
	}
	if( ileft == iright )
	{
		 cerr << "init: uh-oh, left=right, abort" << endl;
		 return 1;
	}
	if( itop == ileft || itop == iright )
	{
		 int imost = 0;
		 int nmost = 0;
		 for( k = 0; k < cullBlobCount; k++ )
		 {
			 if( k == ileft || k == iright )
				 continue;
			 int n =	blobList[ cullBlobList[k] ].getPointCount();
			 if( n > nmost )
				 nmost = n, imost = k;
		 }
		 itop = imost;
	}
	kmeans.setSampleCount( cullBlobCount );
	kmeans.setClusterMean
		( HEAD, blobList[cullBlobList[itop]].getSpatial().getMean() );
	kmeans.setClusterMean
		( LHAND, blobList[cullBlobList[iright]].getSpatial().getMean() );
	kmeans.setClusterMean
		( RHAND, blobList[cullBlobList[ileft]].getSpatial().getMean());
	for( int i = 0; i < cullBlobCount; i++ )
	{
		 Blob& blob = blobList[ cullBlobList[i] ];
		 kmeans.setSample( i, blob.getSpatial().getMean() );
		 kmeans.setSampleWeight( i, blob.getPointCount() );
	}
	kmeans.cluster();
	env->clearBlobClass();
	int fail;
	qmVector c( 2 );
	for( i = 0; i < 3; i++ )
	{
		 c = kmeans.getClusterMean( i );
		 parts[i] -> clear();
		 pixelList.clear();
		 fail = env->expandWoohoo3Blob( Coords((short)c(0), (short)c(1)),
			 env->fleshClass,
			 rgb,
			 0, 0, 0, 0,
			 pixelList );
		 if( fail )
				return 1;
		 if( ! pixelList.length() )
				error( "init", "whoops, no point count, no fail" );
		 for( llPix p = pixelList.first(); 
		 p != 0;
		 p = pixelList.next( p ) )
				parts[i] -> append( pixelList(p)->getCoords() );
	}
	qmVectorRef hdMean = parts[ HEAD ]->getSpatial().getMean();
	qmVectorRef lhMean = parts[ LHAND ]->getSpatial().getMean();
	qmVectorRef rhMean = parts[ RHAND ]->getSpatial().getMean();
	if( hdMean(1) > imageHeight/2 )
	{
		 static qmVector hl( 2 );
		 static qmVector hr( 2 );
		 hl.setSubtract( lhMean, hdMean );
		 hl.setNormalize();
		 hr.setSubtract( rhMean, hdMean );
		 hr.setNormalize();
		 if( hl(1) < -.7 || hr(1) < -.7 )
		 {	
			 alert( "init", "head is bullshit (fixing)" );
			 //swap( parts[HEAD], highest( parts[LHAND], parts[RHAND] ) );
			 swap( parts[HEAD], 
				 closestx( parts[HEAD], parts[LHAND], parts[RHAND] ) );
		 }
	}
	return 0;
}

int
BlobMan::init( vliRGBImageRef& rgb , Prior &prior)
{
	if( env->verbose )
		cerr << "INIT:" << endl;
	env->clearBlobClass();
	int misses = 0;
	env->clearBlobClass();
	llSingleP< Pixel > pixelList;
	int totalFleshCount = 0;
	for( int k = 0; k < blobCount; k++ )
	{
		 blobList[k].clear();
		 misses = env->expandWoohoo3Blob( blobGrid[k],
						 env->fleshClass,
						 rgb,
						 0, 0, 0, 0,
						 pixelList );
		 if( !misses )
		 {
			 totalFleshCount += pixelList.length();
			 for( llPix p = pixelList.first(); 
			 p != 0;
			 p = pixelList.next( p ) )
				 blobList[k].append( pixelList(p)->getCoords() );
			 pixelList.clear();
		 }
	}
	if( env->verbose )
		cerr << "totalFleshCount= " << totalFleshCount << endl;
	personExists = ( totalFleshCount > env->existenceThresh ) ? 1 : 0;
	if( !personExists )
		return 1;
	cullBlobCount = 0;
	for( k = 0; k < blobCount; k++ )
	{
		 if( blobList[k].getPointCount() > env->cullMinBlob )
		 {
			 if( cullBlobCount < env->maxCullBlobs )
				 cullBlobList[ cullBlobCount ] = k;
			 cullBlobCount++;
		 }
	}
	if( env->verbose) 
		cerr << "\t" << cullBlobCount << " medium blobs" << endl;
	if( cullBlobCount < 3 )
		return 1;
	int itop = 0;
	int ileft = 0;
	int iright = 0;
	real ytop = blobList[ cullBlobList[0] ].getSpatial().getMean().getValue(1);
	real xleft = blobList[ cullBlobList[0]].getSpatial().getMean().getValue(0);
	real xright = xleft;
	for( k = 0; k < cullBlobCount; k++ )
	{
     real x =
				blobList[ cullBlobList[k] ].getSpatial().getMean().getValue(0);
		 real y = 
				blobList[ cullBlobList[k] ].getSpatial().getMean().getValue(1);
		 if( y < ytop )
				ytop = y, itop = k;
		 if( x < xleft )
				xleft = x, ileft = k;
		 if( x > xright )
				xright = x, iright = k;
	}

	if( ileft == iright )
	{
		 cerr << "init: uh-oh, left=right, abort" << endl;
		 return 1;
	}

	if( itop == ileft || itop == iright )
	{
		 int imost = 0;
		 int nmost = 0;

		 for( k = 0; k < cullBlobCount; k++ )
		 {
			 if( k == ileft || k == iright )
				 continue;

			 int n =	blobList[ cullBlobList[k] ].getPointCount();

			 if( n > nmost )
				 nmost = n, imost = k;
		 }
		 itop = imost;
	}

	kmeans.setSampleCount( cullBlobCount );

	kmeans.setClusterMean
		( HEAD, blobList[cullBlobList[itop]].getSpatial().getMean() );

	kmeans.setClusterMean
		( LHAND, blobList[cullBlobList[iright]].getSpatial().getMean() );

	kmeans.setClusterMean
		( RHAND, blobList[cullBlobList[ileft]].getSpatial().getMean());

	for( int i = 0; i < cullBlobCount; i++ )
	{
		 Blob& blob = blobList[ cullBlobList[i] ];
		 kmeans.setSample( i, blob.getSpatial().getMean() );
		 kmeans.setSampleWeight( i, blob.getPointCount() );
	}
	kmeans.cluster();
	env->clearBlobClass();
	int fail;
	qmVector c( 2 );
	for( i = 0; i < 3; i++ )
	{
		 c = kmeans.getClusterMean( i );
		 parts[i] -> clear();
		 pixelList.clear();
		 fail = env -> expandWoohoo3Blob( Coords((short)c(0), (short)c(1)),
			 env->fleshClass,
			 rgb,
			 0, 0, 0, 0,
			 pixelList );
		 if( fail )
				return 1;
		 if( ! pixelList.length() )
				error( "init", "whoops, no point count, no fail" );
		 for( llPix p = pixelList.first(); 
		 p != 0;
		 p = pixelList.next( p ) )
				parts[i] -> append( pixelList(p)->getCoords() );
	}
	qmVectorRef hdMean = parts[ HEAD ]->getSpatial().getMean();
	qmVectorRef lhMean = parts[ LHAND ]->getSpatial().getMean();
	qmVectorRef rhMean = parts[ RHAND ]->getSpatial().getMean();
	if( hdMean(1) > imageHeight/2 )
	{
		 static qmVector hl( 2 );
		 static qmVector hr( 2 );
		 hl.setSubtract( lhMean, hdMean );
		 hl.setNormalize();
		 hr.setSubtract( rhMean, hdMean );
		 hr.setNormalize();
		 if( hl(1) < -.7 || hr(1) < -.7 )
		 {	
			 alert( "init", "head is bullshit (fixing)" );
			 //swap( parts[HEAD], highest( parts[LHAND], parts[RHAND] ) );
			 swap( parts[HEAD], 
				 closestx( parts[HEAD], parts[LHAND], parts[RHAND] ) );
		 }
	}
	return 0;
}

int
BlobMan::track( vliRGBImageRef& image , Prior &prior )
{
	if( env->coopMode )
		return( trackCoop3( image , prior ) );
	alert( "track", "whoops" );
	return 1;
}
int
BlobMan::trackCoop3( vliRGBImageRef& image , Prior &prior )
{
	env->clearBlobClass();
	static qmGaussianRef space[3];
	static qmGaussianInv invSpace[3];
	static qmGaussianInvRef invSpaceRef[3];
	static real dist2[3];
	Coords cc;
	llSingleP< Pixel > pixelList;
	for( int i = 0; i < 3; i++ )
	{
		int valid = prior.valid(i);
		 if (valid)
		 {
			 space[i] = prior.getSpatial(i);
			 invSpace[i].set( space[i].getDim() );
			 invSpaceRef[i] = invSpace[i];
			 if( qmStatistics::invert( space[i], invSpaceRef[i] ) )
			 {
				 invSpaceRef[i].setSingular();
			 }
		 }
		 else
		 {
			 if( ! parts[i] -> getPointCount() )
			 {
				 alert( "trackCoop3", "empty blob in track!!!" );
				 return 1;
			 }
			 space[i] = parts[i] -> getSpatial();
			 invSpace[i].set( space[i].getDim() );
			 invSpaceRef[i] = invSpace[i];
			 if( qmStatistics::invert( space[i], invSpaceRef[i] ) )
			 {
				 alert( "trackCoop3", "singular class !!!!" );
				 return 1;
			 }
		 }
	}
	for( i = 0; i < 3; i++ )
		crashCount[i] = 0;
	for( i = 0; i < 3; i++ )
	{
		 cc.u = invSpaceRef[i].getMean().getValue( 0 );
		 cc.v = invSpaceRef[i].getMean().getValue( 1 );
		 env->expandWoohoo3Blob( cc, env->fleshClass, image, 3, invSpaceRef, 
			 dist2, crashCount, pixelList);
		 parts[i] -> clear(); // !! I hope this does not affect invSpace
	}
	Pixel* pp;
	for( llPix p = pixelList.first(); p != 0; p = pixelList.next( p ) )
	{
		 pp = pixelList( p );
		 if( pp->whichClass < 0 )
				error( "trackCoop3", "null class assignment" );
		 parts[ pp->whichClass ] -> append( pp->getCoords() );
	}
	for( i = 0; i < 3; i++ )
		if( ! parts[i] -> getPointCount() )
		{
			if (! prior.valid(i) )
			 {
				 if (env->verbose) 
					 alert( "trackCoop3", "class starved." );
				 return 1;
			 }
			else
			 {
				if (env->verbose) 
					alert( "trackCoop3", "coasting on valid prior." );
				
				//					parts[i] -> setSpatial( prior.getSpatial(i) );
			 }
		}
		return 0;
}
int
BlobMan::collide()
{
	for( int i = 0; i < 3; i++ )
	{
		 if( real( crashCount[i] ) / real( parts[i] -> getPointCount() )
			 > env->crashRatio )
		 {
			 if (env->verbose) 
				 alert( "collide", "found a collision." );
			 return 1;
		 }
	}
	return 0;
}
ostream&
BlobMan::alert( char* fn, char* msg )
{
	return( cerr << "BlobMan: ERROR: " << fn << "(): " << msg << endl );
}

void
BlobMan::error( char* fn, char* msg )
{
	alert( fn, msg );
	assert( 0 );
}
const qmVectorRef&
BlobMan::getBlobMean_Pics( int i )
{
	if( i < 0 || i > 2 )
		error( "getBlobMean_Pics", "arg out of range" );
	if( ! parts[i] -> getPointCount() )
	{
		partVec[i].setZero();
		// alert( "getBlobMean_Pics", "empty blob" );
	}
	else
	{
		partVec[i].subVector( 0, 2 ).copy( parts[i] -> getSpatial().getMean() );
		partVec[i].subVector( 0, 2 ) -= imageCenter;
		partVec[i].subVector( 0, 2 ) /= imageWidth;
		partVec[i].subVector( 2, 3 ).copy( parts[i] -> getSpatialCovVect() );
		partVec[i].subVector( 2, 3 ) /= (imageWidth*imageWidth);
		partVec[i](5) = 1.0;
	}
	return partVec[i];
}
float*
BlobMan::getQuickPoints_Pics()
{
	getQuickPoints();
	for( int i = 0; i < 6; i++ )
		qmVectorRef( 2, quickPoints+2*i ) -= imageCenter;
	qmVectorRef( 12, quickPoints ) /= imageWidth;
	return quickPoints;
}
float*
BlobMan::getQuickPoints()
{
	for( int i = 0; i < 3; i++ )
		parts[i] -> getQuickPoints( quickPoints + i*4 );
	return quickPoints;
}
