// FfinderEnv.cpp: implementation of the FfinderEnv class.
//
//////////////////////////////////////////////////////////////////////

#include "FfinderEnv.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

FfinderEnv::FfinderEnv()
{
  simThresh2 = 9.;
  trackDelta = 20;
  collideThresh = 5; // pixels
  mode = Blob::WOOHOO;
  headDelta = 20;
  verbose = 0;
  expandCheck = 0;
  subWoohoo = 2;
  searchCoordsCount = 0;
  searchCoords = 0;
  deltaWoohoo = 6;
  initCount = 0;
  initWidth = 0;


  real tmp1[3] = { 133.839, 106.329, 148.586 };
  memcpy(fleshMean,tmp1,sizeof(tmp1));
  
  real tmp2[9] = {
	0.00228897756499, 0.00422011948326, -0.00442830528997,
		0.00422011948326, 0.03785796579777, 0.00953987271127,
		-0.00442830528997, 0.00953987271127, 0.03405760806490 
  };
  memcpy(fleshInvCovariance,tmp2,sizeof(tmp2));

  real tmp3[9]  = {
	1155.04999999343, -179.25399999901, 200.39499999887,
		-179.25399999901, 56.23929999986, -39.06049999984,
		200.39499999887, -39.06049999984, 66.35939999981 
  };
  memcpy(fleshCovariance,tmp3,sizeof(tmp3));

  fleshClass.set(qmVectorRef( 3, fleshMean ), qmMatrixRef( 3, 3, fleshInvCovariance ));

  maxCullBlobs = 1000;
  cullMinBlob = 20;
  blobSpacing = 10;
  blobOffset = blobSpacing / 2;
  searchMargin = 5;
  verbose = 0;
  {
    real tmp4[3][2] = {{ 0, -searchMargin },
                  { -searchMargin, 0 },
                  { searchMargin, 0 } }; 
    memcpy(&(o_data[0][0]),&(tmp4[0][0]),sizeof(tmp4));
  }

  cellWidth = 12;
  cellHeight = 12;

  coopMode = 3;
  crashThresh = 10;
  crashRatio = .1;
  existenceThresh = 100;

}

FfinderEnv::~FfinderEnv()
{

}


int
FfinderEnv::initFnBlob( Pixel* pp )
{
    pp -> coords.u = initCount % initWidth;
    pp -> coords.v = initCount / initWidth;
    initCount++;
    return 0;
}

void
FfinderEnv::initBlobClass( int w, int h, int nw, int nh )
{
    llSingleP<Coords>::initClass( w*h );
    llSingleP<Pixel>::initClass( w*h );
    stateImage.init( w, h );
    cellImage.init( &(stateImage), nw, nh );
    initCount = 0;
    initWidth = w;
    stateImage.applyfn( initFnBlob );
    rgbImage.set( w, h );
    initSearchCoords( deltaWoohoo );
}

int
FfinderEnv::expandWoohoo3Blob( Coords& seed,
		     const qmGaussianInvRef& colorClass,
		     vliRGBImageRef& vim,
		     int n,
		     qmGaussianInvRef* space,
		     real* dist2,
		     int* crash,
		     llSingleP<Pixel>& list)
{
    // assume the incoming seed is valid
    int width = cellImage.getWidth();
    int height = cellImage.getHeight();
    int success = 0;
    Cell* cp = 0;
    Coords cellCoords = cellImage.getCellCoords( seed );

    llPix p;
    Coords newCoords;
    int seedFound = 0;
    for( int i = 0; i < searchCoordsCount && !seedFound; i++ )
       {
	   newCoords = cellCoords;
	   newCoords += searchCoords[ i ];
	   if( newCoords.u < 0 || newCoords.u >= width ||
	       newCoords.v < 0 || newCoords.v >= height )
	      continue;
	   p = cellImage.hit( newCoords );
	   if( p )
	      {
		  cp = cellImage( p );
		  success = cp -> setAndClassify
		     ( vim, colorClass, n, space, 
                       dist2, crash, subWoohoo );
		  if( success )
		     seedFound = 1;
	      }
       }
    if( !seedFound )
       return 1;

    Pixel* pp = 0;
    do
       {
	   if( success )
	      {
		  while( pp = cp->pop() )
		     list.append( pp );
		  cellImage.setLock( p );
		  cellImage.hitNeighbors( p );
	      }
	   p = cellImage.next( p );
	   if( p )
	      {
		  cp = cellImage( p );
		  success = cp -> setAndClassify
		     ( vim, colorClass, n, space, 
		       dist2, crash, subWoohoo );
	      }
       }
    while( p != 0 );
    return 0; // normal exit
}