// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// Person.c++

#include "Prior.h"

vliRGBPixel Prior::partColors[3];

void 
Prior::init( long frameNumber, long timeSec, long timeUsec, int pe )
{
    person.frameNumber = frameNumber;
    person.timestampSec = timeSec;
    person.timestampUsec = timeUsec;
    person.personExists = pe;
    person.leftHand.valid = 0;
    person.rightHand.valid = 0;
    person.head.valid = 0;

    imageCenter(0) = width / 2;
    imageCenter(1) = height / 2;

    for (int i=0; i<3; i++)
      {
	spatialStats[i].setSingular();
	invSpatialStats[i].setSingular();

	partColors[i].set( 255 - BlobMan::partColors[i].red(),
			   255 - BlobMan::partColors[i].green(),
			   255 - BlobMan::partColors[i].blue());
      }
}

void
Prior::draw( vliRGBImageRef& rgb, vliRGBPixel*  )
{
  qmMatrixRef covar;
  qmVectorRef mean;

  for(int i=0; i<3; i++)
    {
      if(! spatialStats[i].isSingular())
	{
	  covar = spatialStats[i].getCovariance();
	  mean  = spatialStats[i].getMean();
	  
	  drawPrior( rgb, partColors[i], mean, covar );
	}
    }
}


void 
Prior::drawPrior( vliRGBImageRef& rgb, vliRGBPixel& c, 
		  const qmVectorRef mean, 
		  const qmMatrixRef cov)
{
  qmEigenStructure eig (cov);
  double e0 = 2.0 * sqrt(eig.getEigValues()(0));
  double e1 = 2.0 * sqrt(eig.getEigValues()(1));
  rgb.drawLine((mean.getValue(0) - e0*eig.getEigVectors()(0,0)),
	       (mean.getValue(0) + e0*eig.getEigVectors()(0,0)),
	       (mean.getValue(1) - e0*eig.getEigVectors()(1,0)),
	       (mean.getValue(1) + e0*eig.getEigVectors()(1,0)),
	       c);
  rgb.drawLine((mean.getValue(0) - e1*eig.getEigVectors()(0,1)),
	       (mean.getValue(0) + e1*eig.getEigVectors()(0,1)),
	       (mean.getValue(1) - e1*eig.getEigVectors()(1,1)),
	       (mean.getValue(1) + e1*eig.getEigVectors()(1,1)),
	       c);
}

void Prior::set(ffPerson& p)
{
  person = p;
  static qmVector m(2);
  static qmMatrix c(2,2);

  if(person.personExists)
    {
      if(person.leftHand.valid)
	{
	  person.leftHand.suu *= width*width;
	  person.leftHand.svv *= width*width;
	  person.leftHand.suv *= width*width;
	  
	  person.leftHand.u *= width;
	  person.leftHand.v *= width;
	  person.leftHand.u += imageCenter(0);
	  person.leftHand.v += imageCenter(1);
	  
	  m(0) = person.leftHand.u;
	  m(1) = person.leftHand.v;
	  
	  c(0,0) = person.leftHand.suu;
	  c(1,1) = person.leftHand.svv;
	  c(0,1) = c(1,0) = person.leftHand.suv;
	  
	  spatialStats[BlobMan::LHAND].set(m,c);
	  if (c.setInvertEqCheck())
	    {
	      spatialStats[BlobMan::LHAND].setSingular();
	      invSpatialStats[BlobMan::LHAND].setSingular();
	    }
	  else
	    {      
	      invSpatialStats[BlobMan::LHAND].set(m,c);
	    }
	}
      else
	{ 
	  spatialStats[BlobMan::LHAND].setSingular();
	  invSpatialStats[BlobMan::LHAND].setSingular();
	}
      if(person.rightHand.valid)
	{
	  person.rightHand.suu *= width*width;
	  person.rightHand.svv *= width*width;
	  person.rightHand.suv *= width*width;
	  
	  person.rightHand.u *= width;
	  person.rightHand.v *= width;
	  person.rightHand.u += imageCenter(0);
	  person.rightHand.v += imageCenter(1);
	  
	  m(0) = person.rightHand.u;
	  m(1) = person.rightHand.v;
	  
	  c(0,0) = person.rightHand.suu;
	  c(1,1) = person.rightHand.svv;
	  c(0,1) = c(1,0) = person.rightHand.suv;
	  
	  spatialStats[BlobMan::RHAND].set(m,c);
	  if (c.setInvertEqCheck())
	    {
	      spatialStats[BlobMan::RHAND].setSingular();
	      invSpatialStats[BlobMan::RHAND].setSingular();
	    }
	  else
	    {      
	      invSpatialStats[BlobMan::RHAND].set(m,c);
	    }
	}
      else
	{ 
	  spatialStats[BlobMan::RHAND].setSingular();
	  invSpatialStats[BlobMan::RHAND].setSingular();
	}
      if(person.head.valid)
	{
	  person.head.suu *= width*width;
	  person.head.svv *= width*width;
	  person.head.suv *= width*width;

	  person.head.u *= width;
	  person.head.v *= width;
	  person.head.u += imageCenter(0);
	  person.head.v += imageCenter(1);

	  m(0) = person.head.u;
	  m(1) = person.head.v;

	  c(0,0) = person.head.suu;
	  c(1,1) = person.head.svv;
	  c(0,1) = c(1,0) = person.head.suv;
      
	  spatialStats[BlobMan::HEAD].set(m,c);
	  if (c.setInvertEqCheck())
	    {
	      spatialStats[BlobMan::HEAD].setSingular();
	      invSpatialStats[BlobMan::HEAD].setSingular();
	    }
	  else
	    {      
	      invSpatialStats[BlobMan::HEAD].set(m,c);
	    }
	}
      else
	{ 
	  spatialStats[BlobMan::HEAD].setSingular();
	  invSpatialStats[BlobMan::HEAD].setSingular();
	}
    }
  else
    {
      spatialStats[BlobMan::HEAD].setSingular();
      spatialStats[BlobMan::LHAND].setSingular();
      spatialStats[BlobMan::RHAND].setSingular();
      invSpatialStats[BlobMan::HEAD].setSingular();
      invSpatialStats[BlobMan::LHAND].setSingular();
      invSpatialStats[BlobMan::RHAND].setSingular();
    }
}

int  
Prior::valid(int i)
{
  switch(i)
    {
    case BlobMan::HEAD: 
      return person.head.valid;
    case BlobMan::LHAND:
      return person.leftHand.valid;
    case BlobMan::RHAND: 
      return person.rightHand.valid;
    default:
      return 0;
    }
}

const qmGaussianRef& 
Prior::getSpatial(int i)
{
  return spatialStats[i];
}
const qmGaussianInvRef& 
Prior::getInvSpatial(int i)
{
  return invSpatialStats[i];
}
