// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// BlobMan.h++


#ifndef _BlobMan_H_
#define _BlobMan_H_
#include "ffinderenv.h"
#include "Blob.h"
#include "Coords.h"
#include "vliRGBImageRef.h"
#include "qmStatistics.h"
#include "qmKMeans.h"
class Prior;

class BlobMan : public FfinderEnv
{
public:
    static real fleshMean[3];
    static real fleshInvCovariance[9];
    static real fleshCovariance[9];
    static qmGaussianInv fleshClass;
    int imageWidth;
    int imageHeight;
    qmVector imageCenter;
    int blobCount;
    Coords* blobGrid;
    Blob* blobList;
    int cullBlobCount;
    int* cullBlobList;
    Blob* parts[3];
    qmVector partVec[3];
    qmKMeans kmeans;
    int mode;
    int init( vliRGBImageRef& image );
    int init( vliRGBImageRef& image , Prior &prior );
    int track( vliRGBImageRef& image , Prior &prior );
    int trackCoop3( vliRGBImageRef& image , Prior &prior );
    int collide();
    static ostream& alert( char* fn, char* msg );
    static void error( char* fn, char* msg );
    int crashCount[3];
    int personExists;
    float quickPoints[12];
 public:
    static int maxCullBlobs;
    static int cullMinBlob;
    static int blobSpacing;
    static int blobOffset;
    static int searchMargin;
    static int verbose;
    static int o_data[3][2];
    enum { HEAD=0, LHAND=1, RHAND=2 };
    static istream& readFleshStats( istream& );
    static void setFleshStats( const qmVectorRef& m, const qmMatrixRef& ic );
    enum { INIT_MODE, TRACK_MODE, INITONLY_MODE };
    static int cellWidth;
    static int cellHeight;
    static vliRGBPixel partColors[3];
    static int coopMode;
    static int crashThresh;
    static real crashRatio;
    static int existenceThresh;
 public:
    BlobMan()
      : imageWidth(0), imageHeight(0), imageCenter(2),
	blobCount(0), blobGrid(0), blobList(0),
	cullBlobCount(0), cullBlobList(0),
	mode(INIT_MODE) {}
    void init( int w, int h );
    int process( vliRGBImageRef& image, Prior &prior );
    void drawBlobs( vliYUVImageRef&, vliYUVPixel* = 0 );
    void drawPoints( vliYUVImageRef&, vliYUVPixel* = 0 );
    void drawBlobs( vliRGBImageRef&, vliRGBPixel* = 0 );
    void drawPoints( vliRGBImageRef&, vliRGBPixel* = 0 );
    inline int getMode() { return mode; }
    inline void setMode( int m ) { mode = m; }
    inline void initOnly() { mode = INITONLY_MODE; }
    const qmVectorRef& getBlobMean_Pics( int i );
    int getPersonExists() { return personExists; }
    float* getQuickPoints();
    float* getQuickPoints_Pics();
};

#include "Prior.h"

#endif /* _BlobMan_H_ */
