// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// Pixel.h++

#ifndef _Pixel_H_
#define _Pixel_H_
#include "ffinderenv.h"
#include "vliRGBImage.h"
#include "vliPixel.h"
#include "qmStatistics.h"
#include "Coords.h"

class Pixel : public FfinderEnv
{
  static int index_tmp;
  static real maha2_tmp;
  static real dist2min0_tmp;
  static real dist2min1_tmp;
  static int imin0_tmp;
  static int imin1_tmp;
  static real fcoord_tmp[2];
  static ostream& alert( char* fn, char* msg );
  static void error( char* fn, char* msg );
public:
  static float thresh2;
  static real mutualThresh;
  float rgb[3];
  Coords coords;
  short whichClass;
  Pixel() {}
  inline int set( const vliRGBImageRef& );
  inline int classify( const qmGaussianInvRef& );
  inline int classify( const qmGaussianInvRef& co,
		       int n, 
		       qmGaussianInvRef* sp,
		       real& dist2 );
  inline int classify( const qmGaussianInvRef& co,
		int n, 
		qmGaussianInvRef* sp,
		real* dist2,
		int* crash );
  inline Coords* getCoords() { return &coords; }
  inline float* getColor() { return rgb; }
  ostream& print( ostream& );
};
inline
int
Pixel::set( const vliRGBImageRef& vim )
{
  index_tmp = vim.ix( coords.u, coords.v );
  rgb[0] = vim.getRed( index_tmp );
  rgb[1] = vim.getGreen( index_tmp );
  rgb[2] = vim.getBlue( index_tmp );
  return 0;
}
inline
int
Pixel::classify( const qmGaussianInvRef& colorClass )
{
  maha2_tmp = colorClass.getMahalanobisSquare( qmVectorRef( 3, rgb ) );

  if( maha2_tmp < thresh2 )
    return 1;
  return 0;
}
inline
int
Pixel::classify( const qmGaussianInvRef& colorClass,
		 int n, 
		 qmGaussianInvRef* spaceClasses,
		 real& dist2 )
{
  maha2_tmp = colorClass.getMahalanobisSquare( qmVectorRef( 3, rgb ) );
  if( maha2_tmp > thresh2 )
    {
      whichClass = -1;
      return 0;
    }
  whichClass = 0;
  dist2 = 0.;
  if( n > 0 )
    {
      fcoord_tmp[0] = coords.u;
      fcoord_tmp[1] = coords.v;
      dist2 = 100000.;
      imin0_tmp = -1;
      for( int i = 0; i < n; i++ )
	{
	  qmVectorRef y(2,fcoord_tmp);
	  maha2_tmp  = spaceClasses[i].getMahalanobisSquare(y);
	  if( maha2_tmp < dist2 )
	    dist2 = maha2_tmp, imin0_tmp = i;
	}
      whichClass = imin0_tmp;
    }
  return 1;
}

inline
int
Pixel::classify( const qmGaussianInvRef& colorClass,
		 int n, 
		 qmGaussianInvRef* spaceClasses,
		 real* dist2,
		 int* crash )
{
     maha2_tmp = colorClass.getMahalanobisSquare( qmVectorRef( 3, rgb ) );

     if( maha2_tmp > thresh2 )
        {
 	   whichClass = -1;
 	   return 0;
        }
     whichClass = 0;
     if( n > 0 )
       {
 	   fcoord_tmp[0] = coords.u;
 	   fcoord_tmp[1] = coords.v;
 	   dist2min0_tmp = 100000.;
 	   dist2min1_tmp = 100000.;
 	   imin0_tmp = -1;
 	   imin1_tmp = -1;
 	   for( int i = 0; i < n; i++ )
 	      {
 		  dist2[i] = spaceClasses[i].getMahalanobisSquare
 		     ( qmVectorRef( 2, fcoord_tmp ) );
 		  if( dist2[i] < dist2min0_tmp )
 		     dist2min1_tmp = dist2min0_tmp, imin1_tmp = imin0_tmp,
 		     dist2min0_tmp = dist2[i], imin0_tmp = i;
 		  else if( dist2[i] < dist2min1_tmp )
 		     dist2min1_tmp = dist2[i], imin1_tmp = i;
 	      }
 	   if( dist2min0_tmp / dist2min1_tmp > mutualThresh )
 	      {
 		  crash[ imin0_tmp ]++;
 		  crash[ imin1_tmp ]++;
 		  whichClass = -1;
 		  return 0;
 	      }
 	   whichClass = imin0_tmp;
        }
     return 1;
 }

#endif /* _Pixel_H_ */

