// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// Blob.h++

#ifndef _Blob_H_
#define _Blob_H_
#include <iostream.h> // for print
#include "QuickMath.h"
#include "qmStatistics.h"
#include "vlImage.h"
#include "llSingleP.h"
#include "StateImage.h"
#include "Coords.h"
#include "Pixel.h"
#include "CellImage.h"
class Blob
{
 private:
    llSingleP< Coords > coordList;
    int colorFlag;			
    qmGaussianInv colorStats;
    int spatialFlag;			
    qmGaussian spatialStats;
    static ostream& alert( char* fn, char* msg );
    static void error( char* fn, char* msg );
    static int similar( const vliYUVPixel&, 
			const qmGaussianInvRef&, 
			real t = simThresh2 );
    static qmVector tmpPixelVector;
    static inline const qmVectorRef& pixelToVector( const vliYUVPixel& );
    inline void initBlob();
    static StateImage<Pixel> stateImage;
    static CellImage cellImage;
    static vliRGBImage rgbImage;
    static int searchCoordsCount;
    static Coords* searchCoords;
    static void initSearchCoords( int );
 public:
    static float simThresh2;
    static int trackDelta;
    static int headDelta;
    static real collideThresh;
    enum { REGULAR, SLOW, WOOHOO };
    static int mode;
    static int verbose;
    static int expandCheck;
    static int subWoohoo;
    static int deltaWoohoo;
    static void init() {}
    static void drawStats( vliYUVImageRef&, 
			   const qmGaussianRef&,
			   const vliYUVPixel& );
    static void drawStats( vliRGBImageRef&, 
			   const qmGaussianRef&,
			   const vliRGBPixel& );
    static void initClass( int w, int h, int nw, int nh );
    static void clearClass() { stateImage.clear(); cellImage.clear(); }
    Blob();
    int operator==( Blob& );
    inline int getPointCount() { return coordList.length(); }
    inline const qmGaussianInvRef& getColor() { return colorStats; }
    inline const qmGaussianRef& getSpatial();
    inline void setSpatial(const qmGaussianRef&);
    inline vliYUVPixel getYUVColor();
    const qmVectorRef& getSpatialCovVect();
    const qmGaussianRef& computeSpatial();
    static int expandWoohoo3( Coords& seed,
			      const qmGaussianInvRef& colorClass,
			      vliYUVImageRef& vim,
			      int n,
			      qmGaussianInvRef* space,
			      real* dist2,
			      int* crash,
			      llSingleP<Pixel>& list );
    void drawPoints( vliYUVImageRef&, vliYUVPixel* = 0 );
    void drawPoints( vliRGBImageRef&, vliRGBPixel* = 0 );
    void drawStats( vliYUVImageRef&, vliYUVPixel* = 0 );
    void drawStats( vliRGBImageRef&, vliRGBPixel* = 0 );
    int collide( Blob& );
    inline void clear() { initBlob(); }
    inline void append( Coords* c ) { coordList.append( c ); }
    real getEccentricity();
    void getQuickPoints( float* );
};

inline
const qmVectorRef&
Blob::pixelToVector( const vliYUVPixel& pixel )
{
    tmpPixelVector( 0 ) = pixel.y;
    tmpPixelVector( 1 ) = pixel.u;
    tmpPixelVector( 2 ) = pixel.v;
    return tmpPixelVector;
}

inline
void
Blob::initBlob()
{
    coordList.clear();
    colorFlag = 0;
    spatialFlag = 0;
}

inline
const qmGaussianRef&
Blob::getSpatial()
{ 
    if( coordList.empty() )
      alert( "getSpatial", "empty blob" );
    if( spatialFlag )
       return spatialStats;
    return computeSpatial();
}

inline
void 
Blob::setSpatial(const qmGaussianRef& source)
{
  initBlob();
  Coords *coord = 0;
  coordList.append(coord);
  spatialFlag = 1;
  spatialStats.copy(source);
}

inline
vliYUVPixel
Blob::getYUVColor()
{ 
    if( !colorFlag )
       cerr << "this is an ERROR which should never happen" << endl;
    return vliYUVPixel( colorStats.getMean().getValue(0),
			colorStats.getMean().getValue(1),
			colorStats.getMean().getValue(2) );
}

#endif /* _Blob_H_ */
