// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  ffPrior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// FfFile.c++

#include "FfFile.h"
#include "iotools.h"
#include <assert.h>
#include <iostream.h>
#include <fstream.h>
static char fullpath[512];
static 
inline
char*
makeFullPath( char* pname, char* fname )
{
    if( fname[0] == '/' )
       sprintf( fullpath, fname );
    else
       sprintf( fullpath, "%s/%s", pname, fname );
    return fullpath;
}
inline
istream&
FfFile::scan_color( istream& is )
{
    iotools_scanVector( is, colorMean, 3 );
    iotools_scanVector( is, colorInvCov, 9 );
    return is;
}
inline
istream&
FfFile::scan_pixel_thresh( istream& is )
{
   is >> pixel_thresh;
   return is;
}
istream& 
FfFile::scan( istream& is )
{
    char filename[512];
    ifstream file;
    if( iotools::wordpeek( is, "FILE" ) )
       {
	   is >> filename >> filename;
	   makeFullPath( pathname, filename );
	   file.open( fullpath );
	   if( !file )
	      {
		  alert( "scan", "can't open file" ) << "\t" 
						     << fullpath << endl;
		  error( "scan", "open failed" );
	      }
	   scan_color( file );
	   file.close();
       }
    else
       scan_color( is );
    if( iotools::wordpeek( is, "FILE" ) )
       {
	   is >> filename >> filename;
	   makeFullPath( pathname, filename );
	   file.open( fullpath );
	   if( !file )
	      {
		  alert( "scan", "can't open file" ) << "\t" 
						     << fullpath << endl;
		  error( "scan", "open failed" );
	      }
	   scan_pixel_thresh( file );
	   file.close();
       }
    else
        scan_pixel_thresh( is );
    return is;
}
int
FfFile::scan( char* filename )
{
    makeFullPath( pathname, filename );
    ifstream file( fullpath, ios::in|ios::nocreate );
    if( !file )
       {
	   alert( "scan", "can't open file" ) << "\t" 
					      << fullpath << endl;
	   return 1;
       }
    scan( file );
    file.close();
    return 0;
}
ostream& 
FfFile::print( ostream& os )
{
    iotools_printVector( os, colorMean, 3 );
    iotools_printVector( os, colorInvCov, 9 );
    os << pixel_thresh << endl;
    return os;
}
int
FfFile::print( char* filename )
{
    makeFullPath( pathname, filename );
    ofstream file( fullpath );
    print( file );
    file.close();
    return 0;
}
ostream&
FfFile::alert( char* fn, char* msg )
{
    return( cerr << "FfFile: ERROR: " << fn << "(): " << msg << endl );
}
void
FfFile::error( char* fn, char* msg )
{
    alert( fn, msg );
    // exit( 1 );
    assert(0);
}
