// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// lffinderServer.c++

#include "lffinderServer.h"
ffConfig lffinderServer::config;
int lffinderServer::captureFlag = 0;
int lffinderServer::doneFlag = 0;
ffPerson lffinderServer::person;
ffPerson lffinderServer::priors;
int lffinderServer::homeyLock = 0;
int lffinderServer::priorLock = 0;
int lffinderServer::currentMode = 0;
int lffinderServer::forceInit = 0;
float lffinderServer::quickPoints[12];
int lffinderServer::quitFlag = 0;

unsigned char pszProtocolSequence[100] = "ncacn_np";
unsigned char * pszSecurity         = NULL;
unsigned char pszEndpoint[100]         = "\\pipe\\lffinder";
unsigned int    cMinCalls           = 1;
unsigned int    cMaxCalls           = 20;
unsigned int    fDontWait           = FALSE;
 
void
rpc_init(){
    RPC_STATUS status;

    status = RpcServerUseProtseqEp(pszProtocolSequence,
                                   cMaxCalls,
                                   pszEndpoint,
                                   pszSecurity);  // Security descriptor
    printf("RpcServerUseProtseqEp returned 0x%x\n", status);
    if (status) {
        exit(status);
    }

    status = RpcServerRegisterIf(lffinder_v1_0_s_ifspec,  // interface to register
                                 NULL,   // MgrTypeUuid
                                 NULL);  // MgrEpv; null means use default
    printf("RpcServerRegisterIf returned 0x%x\n", status);
    if (status) {
        exit(status);
    }

}

void
service_rpc()
{
  RPC_STATUS status;
  printf("Calling RpcServerListen\n");
  status = RpcServerListen(cMinCalls,
                  cMaxCalls,
                  fDontWait);
  printf("RpcServerListen returned: 0x%x\n", status);
  printf("Calling RpcMgmtWaitServerListen\n");
  RpcMgmtWaitServerListen();  // wait operation
  printf("RpcMgmtWaitServerListen returned: 0x%x\n", status);
}

DWORD WINAPI
serverThread( void * )
{
  rpc_init();
  while( !lffinderServer::quitFlag )
     service_rpc();

  return 0;
}

void
lffinderServer::cleanup()
{
  RpcMgmtStopServerListening(NULL);
  quitFlag = 1;
}

//////////////////////////////////////////////////////////////////////////////
#ifndef RPC_MALLOC_XXX
#define RPC_MALLOC_XXX

void __RPC_FAR * __RPC_USER MIDL_user_allocate(size_t x){
  return malloc(x);
}

void __RPC_USER MIDL_user_free( void __RPC_FAR * x){
  free(x);
}

#endif

int
lffQUIT() 
{
    //RpcMgmtStopServerListening(NULL);
  lffinderServer::quitFlag = 1; // does nothing !
    //Sleep(0);
    //exit(1);
    return 1;
}
int
lffINIT( ffConfig pc)
{
    static int retval = 0;
    lffinderServer::config = pc;
    return retval;
}
int 
lffFRAME()
{
    lffinderServer::captureFlag = 1;
    lffinderServer::doneFlag = 0;
    return lffinderServer::captureFlag;
}
#include <iostream.h>
ffPerson
lffGET_FRAME_RESULT()
{
    int waitCount = 0;
    while( !lffinderServer::doneFlag && waitCount++ < 10 )
       Sleep(1);
#if 1
    if( waitCount >= 10 )
       cerr << "ffget_frame_result: timed out" << endl;
#endif
    return lffinderServer::person;
}

void 
lffSET_PRIORS(ffPerson p)
{
    while( lffinderServer::priorLock )
      {
	Sleep(1);
      }

    lffinderServer::priorLock = 1;
    lffinderServer::priors = p;
    lffinderServer::priorLock = 0;
}

ffPerson 
lffGET_CURRENT()
{
    lffinderServer::currentMode = 1;
    return lffinderServer::person;
}
void
lffforceinit()
{
    lffinderServer::forceInit = 1;
}
ffQuickPoints
lffgetquickpoints()
{
    static ffQuickPoints qp;
    memcpy( qp.pts, lffinderServer::quickPoints, 12*sizeof(float) );
    return qp;
}

lffinderServer::lffinderServer()
{
    init();
}

void
lffinderServer::init()
{
    person.leftHand.valid = 0;
    person.rightHand.valid = 0;
    person.head.valid = 0;

    priors.leftHand.valid = 0;
    priors.rightHand.valid = 0;
    priors.head.valid = 0;
}

void
lffinderServer::copy( const ffPerson& fp )
{
    if( !captureFlag && !currentMode )
       return;
    while( homeyLock )
       Sleep( 0 );
    homeyLock = 1;
    person = fp;
    homeyLock = 0;
    doneFlag = 1;
    captureFlag = 0;
}
ffPerson& 
lffinderServer::copy_priors( )
{
  static ffPerson prior_tmp;
  
    while( priorLock )
       Sleep( 1 );
    priorLock = 1;
    prior_tmp = priors;
    priorLock = 0;
    
    return (prior_tmp);
}


void lffinderServer::start()
{
  unsigned long thread_id;
  CreateThread(NULL,NULL,serverThread,NULL,NULL,&thread_id);
}

void lffinderServer::quit()
{
  lffQUIT();
}
