// rffinderClient.h++
// Christopher R. Wren Mon Mar 10 14:00:00 1997
// based on 
// SrffinderClient.h++
// Ali Azarbayejani Mon Jun  3 07:00:16 1996

#ifndef _rffinderClient_H_
#define _rffinderClient_H_

#include <iostream.h>

#include <math.h>
#include "sffinder.h"
#include "rffinder.h"
#include <QuickMath.h>

class rffinderClient
{
public:
  unsigned char * pszUuid;             //= NULL;
  unsigned char pszProtocolSequence[100]; //= "ncacn_np";
  unsigned char * pszNetworkAddress;   //= NULL;
  unsigned char pszEndpoint[100];         //= "\\pipe\\hello";
  unsigned char * pszOptions;          //= NULL;
  unsigned char * pszStringBinding;    //= NULL;
  unsigned char pszString[100];       //= "hello, world";
  unsigned long ulCode;  
  
  ffPerson person;
  int lastFrame;
  enum { GOTNONE, GOTPERSON };
  int lastGot;

  static qmMatrix world_rotation;
  static qmVector world_translation;
  static float    world_scale;
  
  static qmMatrix stereo_rotation;
  static qmVector stereo_translation;

public:
	void quit();
  enum { HEAD, LHAND, RHAND };
  enum { LEFT=0, RIGHT };

  rffinderClient ( char* serverHostname = 0 );
  ~rffinderClient();
  int connect( char* serverHostname );
  
  // nonblocking and blocking call to get entire data structure
  // over here fm server
  // 
  ffPerson* getPerson();	// no matter what - non-blocking
  ffPerson* getNextPerson();	// block till next frame
  void init();			// force an init
  void setPriors(ffPerson*);	// send an ffPerson as a priori knowledge

  inline int personExists();
  
  static inline void SetTransforms(double* translate,
				   double* rotation,
				   double scale,
				   double* stereo_t,
				   double* stereo_q);

  static inline void TransformFromWorld(int side,
					real *mean,
					real *covar);
  
  static inline void Project(ffImagePoint& prior,
			     real *mean,
			     real *covar);
  
  static  void rffinderClient::PriorFromEig(real *mean,
						 real *covar,
						 double *t,
						 double *q,
						 double *s);

  static inline void PriorFromEig(ffImagePoint& prior, 
				  double* mean,
				  double* scale,
				  double angle);
  
  static inline void setPriorFrom2D(ffImagePoint& prior, 
				    sffWorldPart& part,
				    int side);

  static inline void setPriorFrom3D(ffImagePoint& prior, 
				    sffWorldPart& part,
				    int side);
  ostream& print( ostream& );
};

inline
int
rffinderClient::personExists()
{ 
    switch( lastGot )
       {
	 case GOTPERSON: return person.personExists;
	 default: return 0;
       }
}

inline
void 
rffinderClient::SetTransforms(double* translation,
			     double* rotation,
			     double scale,
			     double* stereo_t,
			     double* stereo_q)
{
  world_scale = 1.0/scale;

  real fstereo_t[3];
  for(int i=0; i<3; i++)
    {
      world_translation(i) = (real) translation[i];

      for(int j=0; j<3; j++)
	world_rotation(i,j) = (real) rotation[3*i+j];

      fstereo_t[i] = (real) stereo_t[i];
    }
  stereo_translation.copy(qmVectorRef(3,fstereo_t));

  real fstereo_q[4];
  for(int j=0; j<4; j++)
    {
      fstereo_q[j] = (real) stereo_q[j];
    }
  qmUnitQuaternionRef squat(fstereo_q);
  squat.getRotationMatrix(stereo_rotation);
}

inline
void 
rffinderClient::TransformFromWorld(int side,
				  real *m,
				  real *c)
{
  qmVectorRef mean(3,m);
  qmVector mean_tmp(3);

  mean *= world_scale;
  mean_tmp.setTMultiply(world_rotation,mean);
  mean.setAdd(mean_tmp,world_translation);

  if (side == RIGHT)
    {
      mean_tmp.setMultiply(stereo_rotation,mean);
      mean.setAdd(mean_tmp,stereo_translation);
    }

  qmMatrixRef covar(3,3,c);
  covar *= world_scale*world_scale;
  qmMatrix covar_tmp(3,3);
  covar_tmp.setTMultiply(world_rotation, covar);
  covar.setMultiply(covar_tmp,world_rotation);

  if (side == RIGHT)
    {
      covar_tmp.setMultiply(stereo_rotation, covar);
      covar.setMultiplyT(covar_tmp,stereo_rotation);
    }

}



inline
void rffinderClient::Project(ffImagePoint& prior,
			    real *m,
			    real *c)
{
  qmVectorRef mean(3,m);
  qmMatrixRef covar(3,3,c);

  real foverz = 1.0 / (1.0+mean(2));
  prior.u = mean(0) * foverz;
  prior.v = mean(1) * foverz;

  real foverz2 = foverz * foverz;
  prior.suu = covar(0,0) * foverz2;
  prior.svv = covar(1,1) * foverz2;
  prior.suv = covar(0,1) * foverz2;
}


  
inline
void 
rffinderClient::PriorFromEig(real *m,
			    real *c,
			    double *t,
			    double *q,
			    double *s)
{
  real fq[4];
  for(int i=0; i<4; i++)
    {
      fq[i] = (real) q[i];
    }
  qmUnitQuaternionRef quat(fq);

  quat.setNormal();
  qmMatrix eigen_vectors(3,3);
  quat.getRotationMatrix(eigen_vectors);

  qmMatrix eigen_values(3,3);
  real fs[3];
  for(int j=0; j<3; j++)
    {
      fs[j] = (real) s[j];
    }
  eigen_values.setDiagonal(qmVectorRef(3,fs));
  
  qmMatrix    covar_tmp(3,3);
  qmMatrixRef     covar(3,3,c);
  covar_tmp.setMultiply(eigen_vectors,eigen_values);
  covar.setMultiplyT(covar_tmp,eigen_vectors);
  
  qmVectorRef mean(3,m);
  real ft[3];
  for(int k=0; k<3; k++)
    {
      ft[k] = (real) t[k];
    }
  mean.copy(qmVectorRef(3,ft));
}
  

inline 
void 
rffinderClient::PriorFromEig(ffImagePoint& prior, 
			    double* mean,
			    double* scale,
			    double angle)
{
   prior.valid=1;
   prior.u = mean[0];
   prior.v = mean[1];
   double a = scale[0]; a *= a;
   double b = scale[1]; b *= b;
   double c = cos(angle);
   double s = sin(angle);
   prior.suu = a*c*c+b*s*s;
   prior.svv = a*s*s+b*c*c;
   prior.suv = a*c*s-b*c*s;
}

inline
void
rffinderClient::setPriorFrom2D(ffImagePoint& prior, 
			      sffWorldPart& part,
			      int side)
{
  if (side == RIGHT)
    {    
      if (part.right.valid)
	{
	  PriorFromEig(prior, 
		       part.right.t,
		       part.right.s,
		       part.right.angle);
	  prior.valid=1;
	}
      else
	{
	  prior.valid=0;
	}  
    }
  else
    {    
      if (part.left.valid)
	{
	  PriorFromEig(prior, 
		       part.left.t,
		       part.left.s,
		       part.left.angle);
	  prior.valid=1;
	}
      else
	{
	  prior.valid=0;
	}  
    }
}

inline
void 
rffinderClient::setPriorFrom3D(ffImagePoint& prior, 
			      sffWorldPart& part,
			      int side)
{
  real mean[3];
  real covar[9];
  
  if (part.valid)
    {
      PriorFromEig(mean,
		   covar,
		   part.world_ref.t,
		   part.world_ref.q,
		   part.world_ref.s);
      
      TransformFromWorld(side,
			 mean,
			 covar);
      
      Project(prior, 
	      mean, 
	      covar);
      
      prior.valid=1;
    }
  else
    {
      prior.valid=0;
    }  
}

#endif /* _SrffinderClient_H_ */

