// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// rffinderServer.c++

#include "rffinderServer.h"
ffConfig rffinderServer::config;
int rffinderServer::captureFlag = 0;
int rffinderServer::doneFlag = 0;
ffPerson rffinderServer::person;
ffPerson rffinderServer::priors;
int rffinderServer::homeyLock = 0;
int rffinderServer::priorLock = 0;
int rffinderServer::currentMode = 0;
int rffinderServer::forceInit = 0;
float rffinderServer::quickPoints[12];
int rffinderServer::quitFlag = 0;

unsigned char pszProtocolSequence[100] = "ncacn_np";
unsigned char * pszSecurity         = NULL;
unsigned char pszEndpoint[100]         = "\\pipe\\rffinder";
unsigned int    cMinCalls           = 1;
unsigned int    cMaxCalls           = 20;
unsigned int    fDontWait           = FALSE;
 
void
rpc_init(){
    RPC_STATUS status;

    status = RpcServerUseProtseqEp(pszProtocolSequence,
                                   cMaxCalls,
                                   pszEndpoint,
                                   pszSecurity);  // Security descriptor
    printf("RpcServerUseProtseqEp returned 0x%x\n", status);
    if (status) {
        exit(status);
    }

    status = RpcServerRegisterIf(rffinder_v1_0_s_ifspec,  // interface to register
                                 NULL,   // MgrTypeUuid
                                 NULL);  // MgrEpv; null means use default
    printf("RpcServerRegisterIf returned 0x%x\n", status);
    if (status) {
        exit(status);
    }

}

void
service_rpc()
{
  RPC_STATUS status;
  printf("Calling RpcServerListen\n");
  status = RpcServerListen(cMinCalls,
                  cMaxCalls,
                  fDontWait);
  printf("RpcServerListen returned: 0x%x\n", status);
  printf("Calling RpcMgmtWaitServerListen\n");
  RpcMgmtWaitServerListen();  // wait operation
  printf("RpcMgmtWaitServerListen returned: 0x%x\n", status);
}


DWORD WINAPI
serverThread( void * )
{
  rpc_init();
  while( !rffinderServer::quitFlag )
     service_rpc();

  return 0;
}

void
rffinderServer::cleanup()
{
  RpcMgmtStopServerListening(NULL);
  quitFlag = 1;
}

//////////////////////////////////////////////////////////////////////////////


int
rffQUIT() 
{
    //RpcMgmtStopServerListening(NULL);
  rffinderServer::quitFlag = 1; // does nothing !
    Sleep(0);
 //   exit(1);
    return 1;
}
int
rffINIT( ffConfig pc)
{
    static int retval = 0;
    rffinderServer::config = pc;
    return retval;
}
int 
rffFRAME()
{
    rffinderServer::captureFlag = 1;
    rffinderServer::doneFlag = 0;
    return rffinderServer::captureFlag;
}
#include <iostream.h>
ffPerson
rffGET_FRAME_RESULT()
{
    int waitCount = 0;
    while( !rffinderServer::doneFlag && waitCount++ < 10 )
       Sleep(1);
#if 1
    if( waitCount >= 10 )
       cerr << "ffget_frame_result: timed out" << endl;
#endif
    return rffinderServer::person;
}

void 
rffSET_PRIORS(ffPerson p)
{
    while( rffinderServer::priorLock )
      {
	Sleep(1);
      }

    rffinderServer::priorLock = 1;
    rffinderServer::priors = p;
    rffinderServer::priorLock = 0;
}

ffPerson 
rffGET_CURRENT()
{
    rffinderServer::currentMode = 1;
    return rffinderServer::person;
}
void
rffforceinit()
{
    rffinderServer::forceInit = 1;
}
ffQuickPoints
rffgetquickpoints()
{
    static ffQuickPoints qp;
    memcpy( qp.pts, rffinderServer::quickPoints, 12*sizeof(float) );
    return qp;
}

rffinderServer::rffinderServer()
{
    init();
}

void
rffinderServer::init()
{
    person.leftHand.valid = 0;
    person.rightHand.valid = 0;
    person.head.valid = 0;

    priors.leftHand.valid = 0;
    priors.rightHand.valid = 0;
    priors.head.valid = 0;
}

void
rffinderServer::copy( const ffPerson& fp )
{
    if( !captureFlag && !currentMode )
       return;
    while( homeyLock )
       Sleep( 0 );
    homeyLock = 1;
    person = fp;
    homeyLock = 0;
    doneFlag = 1;
    captureFlag = 0;
}
ffPerson& 
rffinderServer::copy_priors( )
{
  static ffPerson prior_tmp;
  
    while( priorLock )
       Sleep( 1 );
    priorLock = 1;
    prior_tmp = priors;
    priorLock = 0;
    
    return (prior_tmp);
}


void rffinderServer::start()
{
  unsigned long thread_id;
  CreateThread(NULL,NULL,serverThread,NULL,NULL,&thread_id);
}

void rffinderServer::quit()
{
  rffQUIT();
}
