// $CopyrightHeader$

// Copyright (C) 1996 by Ali Azarbayejani and the Massachusetts
// Institute of Technology
// Copyright (C) 1997 by Christopher R. Wren and the Massachusetts
// Institute of Technology
//
// COMMERCIAL DISCLAIMER:
//
// This software is intended for research purposes only and contains
// unlicensed software, in original and modified forms, from various
// sources including the Free Software Foundation.  These programs
// shall not be used, rewritten, or adapted as the basis of a
// commercial software or hardware product without first obtaining
// appropriate licenses from the owners of the original software
// copyrights.
//
// In the case of copyrighted software fragments obtained from the
// Free Software Foundation: you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as
// published by the Free Software Foundation.  This software is
// distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General
// Public License for more details.  You should have received a copy
// of the GNU Library General Public License and the GNU General
// Public License along with this software; if not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


// MIT COPYRIGHT:
// 
// Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
// rights reserved.
// 
// Developed  by Ali Azarbayejani at the Media Laboratory, MIT,
// Cambridge, Massachusetts.
// 
// Permission to use, copy,  or modify this software and  its documentation
// for  educational  and  research purposes only and without fee  is hereby
// granted, provided  that this copyright notice and the original authors's
// names appear  on all copies and supporting documentation.  If individual
// files are  separated from  this  distribution directory  structure, this
// copyright notice must be included.  For any other uses of this software,
// in original or  modified form, including but not limited to distribution
// in whole or in  part, specific  prior permission  must be  obtained from
// MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
// basis  of  a commercial  software  or  hardware product  without first
// obtaining appropriate licenses  from MIT.  MIT. makes no representations
// about the suitability of this  software for any purpose.  It is provided
// "as is" without express or implied warranty.

// ffinderServer.c++

#include "ffinderServer.h"
ffConfig ffinderServer::config;
int ffinderServer::captureFlag = 0;
int ffinderServer::doneFlag = 0;
ffPerson ffinderServer::person;
ffPerson ffinderServer::priors;
int ffinderServer::homeyLock = 0;
int ffinderServer::priorLock = 0;
int ffinderServer::currentMode = 0;
int ffinderServer::forceInit = 0;
float ffinderServer::quickPoints[12];

unsigned char pszProtocolSequence[100] = "ncacn_ip_tcp";
unsigned char * pszSecurity         = NULL;
unsigned char pszEndpoint[100]         = "2046";
unsigned int    cMinCalls           = 1;
unsigned int    cMaxCalls           = 20;
unsigned int    fDontWait           = FALSE;
 
void
rpc_init(){
    RPC_STATUS status;

    status = RpcServerUseProtseqEp(pszProtocolSequence,
                                   cMaxCalls,
                                   pszEndpoint,
                                   pszSecurity);  // Security descriptor
    printf("RpcServerUseProtseqEp returned 0x%x\n", status);
    if (status) {
        exit(status);
    }

    status = RpcServerRegisterIf(ffinder_v1_0_s_ifspec,  // interface to register
                                 NULL,   // MgrTypeUuid
                                 NULL);  // MgrEpv; null means use default
    printf("RpcServerRegisterIf returned 0x%x\n", status);
    if (status) {
        exit(status);
    }

}

void
service_rpc()
{
  RPC_STATUS status;
  printf("Calling RpcServerListen\n");
  status = RpcServerListen(cMinCalls,
                  cMaxCalls,
                  fDontWait);
  printf("RpcServerListen returned: 0x%x\n", status);
  printf("Calling RpcMgmtWaitServerListen\n");
  RpcMgmtWaitServerListen();  // wait operation
  printf("RpcMgmtWaitServerListen returned: 0x%x\n", status);
}

int quitFlag = 0;
DWORD WINAPI
serverThread( void * )
{
  rpc_init();
  while( !quitFlag )
     service_rpc();

  return 0;
}

void
ffinderServer::cleanup()
{
  RpcMgmtStopServerListening(NULL);
  quitFlag = 1;
}

//////////////////////////////////////////////////////////////////////////////

void __RPC_FAR * __RPC_USER MIDL_user_allocate(size_t x){
  return malloc(x);
}

void __RPC_USER MIDL_user_free( void __RPC_FAR * x){
  free(x);
}

int
ffQUIT() 
{
    quitFlag = 1; // does nothing !
    return quitFlag;
}
int
ffINIT( ffConfig pc)
{
    static int retval = 0;
    ffinderServer::config = pc;
    return retval;
}
int 
ffFRAME()
{
    ffinderServer::captureFlag = 1;
    ffinderServer::doneFlag = 0;
    return ffinderServer::captureFlag;
}
#include <iostream.h>
ffPerson
ffGET_FRAME_RESULT()
{
    int waitCount = 0;
    while( !ffinderServer::doneFlag && waitCount++ < 10 )
       Sleep(1);
#if 1
    if( waitCount >= 10 )
       cerr << "ffget_frame_result: timed out" << endl;
#endif
    return ffinderServer::person;
}

void 
ffSET_PRIORS(ffPerson p)
{
    while( ffinderServer::priorLock )
      {
	Sleep(1);
      }

    ffinderServer::priorLock = 1;
    ffinderServer::priors = p;
    ffinderServer::priorLock = 0;
}

ffPerson 
ffGET_CURRENT()
{
    ffinderServer::currentMode = 1;
    return ffinderServer::person;
}
void
ffforceinit()
{
    ffinderServer::forceInit = 1;
}
ffQuickPoints
ffgetquickpoints()
{
    static ffQuickPoints qp;
    memcpy( qp.pts, ffinderServer::quickPoints, 12*sizeof(float) );
    return qp;
}

ffinderServer::ffinderServer()
{
    init();
}

void
ffinderServer::init()
{
    person.leftHand.valid = 0;
    person.rightHand.valid = 0;
    person.head.valid = 0;

    priors.leftHand.valid = 0;
    priors.rightHand.valid = 0;
    priors.head.valid = 0;
}

void
ffinderServer::copy( const ffPerson& fp )
{
    if( !captureFlag && !currentMode )
       return;
    while( homeyLock )
       Sleep( 0 );
    homeyLock = 1;
    person = fp;
    homeyLock = 0;
    doneFlag = 1;
    captureFlag = 0;
}
ffPerson& 
ffinderServer::copy_priors( )
{
  static ffPerson prior_tmp;
  
    while( priorLock )
       Sleep( 1 );
    priorLock = 1;
    prior_tmp = priors;
    priorLock = 0;
    
    return (prior_tmp);
}


void ffinderServer::start()
{
  unsigned long thread_id;
  CreateThread(NULL,NULL,serverThread,NULL,NULL,&thread_id);
}
