#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <windows.h>
#include <mil.h>
#include <mwinmil.h>
#include <wingdi.h>

#define BUFFERSIZEX        640
#define BUFFERSIZEY        480
#define BUFFERSIZEBAND     1
#define MAX_PATH_NAME_LEN  256

  /* MIL variables */
  MIL_ID MilApplication,  /* MIL Application identifier.  */
         MilSystem,       /* MIL System identifier.       */
         MilDisplay,      /* MIL Display identifier.      */
         MilDigitizer,    /* MIL Digitizer identifier.    */
         MilImage;        /* MIL Image buffer identifier. */

  long BufSizeX;
  long BufSizeY;
  long BufSizeBand;

void MilOpen(HWND UserWindowHandle, int x, int y)
{

         

  /* Allocate a MIL application. */
  MappAlloc(M_DEFAULT, &MilApplication);

  /* Allocate a MIL system. */
  MsysAlloc(M_DEF_SYSTEM_TYPE, M_DEV0, M_DEFAULT, &MilSystem);

  /* Allocate a MIL display. */
  MdispAlloc(MilSystem, M_DEV0, M_DEF_DISPLAY_FORMAT, M_DEFAULT, &MilDisplay);

  /* Allocate a MIL digitizer if supported and sets the target image size. */
    MdigAlloc(MilSystem, M_DEV0, M_DEF_DIGITIZER_FORMAT, M_DEFAULT, &MilDigitizer);
    MdigInquire(MilDigitizer, M_SIZE_X,    &BufSizeX);
    MdigInquire(MilDigitizer, M_SIZE_Y,    &BufSizeY);
    MdigInquire(MilDigitizer, M_SIZE_BAND, &BufSizeBand);

    //BufSizeX     = x;
    //BufSizeY     = y;
    //BufSizeBand  = BUFFERSIZEBAND;
    MdigControl(MilDigitizer,M_GRAB_SCALE,M_FILL_DISPLAY);
    //char buffer[1000];
    //MappGetError(M_CURRENT,buffer);
 // MessageBox(0,buffer,"Mil",MB_APPLMODAL|MB_ICONEXCLAMATION);
     
  /* Do not allow example to run in dual screen mode */             
  if(MdispInquire(MilDisplay, M_DISPLAY_MODE, M_NULL) != M_WINDOWED)             
    {
     MessageBox(0,"""This example does no run in dual screen mode.",
                  "MIL application example",
                  MB_APPLMODAL | MB_ICONEXCLAMATION );
    }

  /* Allocate a MIL buffer. */
  MbufAllocColor(MilSystem, BufSizeBand, BufSizeX, BufSizeY, 8+M_UNSIGNED,
             (MilDigitizer? M_IMAGE+M_DISP+M_GRAB : M_IMAGE+M_DISP), &MilImage);

  /* Clear the buffer */
  MbufClear(MilImage,0);

  /* Select the MIL buffer to be displayed in the user specified window */
  MdispSelectWindow(MilDisplay, MilImage, UserWindowHandle);
}

void MilBeginCapture(){
  /* Print a string in the image buffer using MIL.
     Note: After a MIL command writing in a MIL buffer, the display 
     will automatically update the window given to MdispSelectWindow().
  */
  MgraText(M_DEFAULT, MilImage, (BufSizeX/8)*3, BufSizeY/2,   " ------------------ ");
  MgraText(M_DEFAULT, MilImage, (BufSizeX/8)*3, BufSizeY/2+25," Welcome to MIL !!! ");
  MgraText(M_DEFAULT, MilImage, (BufSizeX/8)*3, BufSizeY/2+50," ------------------ ");

  /* Grab in the user window if supported by the system. */
  if (MilDigitizer)
     {
     /* Grab continuously. */
     MdigGrabContinuous(MilDigitizer, MilImage);
  }

}

void MilStopCapture() {

  if(MilDigitizer){
     /* Stop continuous grab. */
     MdigHalt(MilDigitizer);
     }
}

void MilClose(){
  /* Deselect the MIL buffer from the display. */
  MdispDeselect(MilDisplay, MilImage);

  /* Free allocated objects. */
  MbufFree(MilImage);

  MdispFree(MilDisplay);
  if (MilDigitizer)
     MdigFree(MilDigitizer);
  MsysFree(MilSystem);
  MappFree(MilApplication);
}
