// noErfBlob.c++

#include "noErfBlob.h"
#include "cgTriangleStereo.h"


int noErfBlob::verbose = 0;

void
noErfBlob::init( const qmdVectorRef& mL,
		 const qmdMatrixRef& c2L,
		 const qmdVectorRef& mR,
		 const qmdMatrixRef& c2R,
		 const reald b,
		 const qmdVectorRef& t,
		 const qmdUnitQuaternionRef& q )
{
    noErrorFunction::init( nMmts(), nStates(), nDStates() );
    y_obs.subVector( UL, 2 ).copy( mL );
    covL_obs.copy( c2L );
    y_obs.subVector( UR, 2 ).copy( mR );
    covR_obs.copy( c2R );
    y_obs( SuuL ) = covL_obs( 0, 0 );
    y_obs( SvvL ) = covL_obs( 1, 1 );
    y_obs( SuvL ) = covL_obs( 0, 1 );
    y_obs( SuuR ) = covR_obs( 0, 0 );
    y_obs( SvvR ) = covR_obs( 1, 1 );
    y_obs( SuvR ) = covR_obs( 0, 1 );
    beta = b;
    Trl = t;
    Qrl = q;
    Qrl.getRotationMatrix( Rrl );
    inited = 1;
}

void
noErfBlob::init( const qmdVectorRef& y,
		 const reald b,
		 const qmdVectorRef& t,
		 const qmdUnitQuaternionRef& q )
{
    noErrorFunction::init( nMmts(), nStates(), nDStates() );
    y_obs.copy( y );
    // y_obs.subVector( UL, 2 ).copy( mL );
    // covL_obs.copy( c2L );
    covL_obs( 0, 0 ) = y.getValue( SuuL );
    covL_obs( 0, 1 ) = covL_obs( 1, 0 ) = y.getValue( SuvL );
    covL_obs( 1, 1 ) = y.getValue( SvvL );
    // y_obs.subVector( UR, 2 ).copy( mR );
    // covR_obs.copy( c2R );
    covR_obs( 0, 0 ) = y.getValue( SuuR );
    covR_obs( 0, 1 ) = covR_obs( 1, 0 ) = y.getValue( SuvR );
    covR_obs( 1, 1 ) = y.getValue( SvvR );
//     y_obs( SuuL ) = covL_obs( 0, 0 );
//     y_obs( SvvL ) = covL_obs( 1, 1 );
//     y_obs( SuvL ) = covL_obs( 0, 1 );
//     y_obs( SuuR ) = covR_obs( 0, 0 );
//     y_obs( SvvR ) = covR_obs( 1, 1 );
//     y_obs( SuvR ) = covR_obs( 0, 1 );
    beta = b;
    Trl = t;
    Qrl = q;
    Qrl.getRotationMatrix( Rrl );
    inited = 1;
}

void
noErfBlob::initStereo( const reald b,
		       const qmdVectorRef& t,
		       const qmdUnitQuaternionRef& q )
{
    beta = b;
    Trl = t;
    Qrl = q;
    Qrl.getRotationMatrix( Rrl );
    inited = 1;
}

static
inline
void
makeCovVec( qmdVectorRef& v, qmdMatrixRef& m )
{
    v(0) = m(0,0); // Sxx
    v(1) = m(1,1); // Syy
    v(2) = m(2,2); // Szz
    v(3) = m(0,1); // Sxy
    v(4) = m(0,2); // Sxz
    v(5) = m(1,2); // Syz
}

static
inline
void
makeJacobianProducts( reald* p, qmdMatrixRef& Js )
{
    // p points into a (3,6) matrix; Js is (2,3)
    *p++ = Js( 0, 0 ) * Js( 0, 0 );
    *p++ = Js( 0, 1 ) * Js( 0, 1 );
    *p++ = Js( 0, 2 ) * Js( 0, 2 );
    *p++ = 2 * Js( 0, 0 ) * Js( 0, 1 );
    *p++ = 2 * Js( 0, 0 ) * Js( 0, 2 );
    *p++ = 2 * Js( 0, 1 ) * Js( 0, 2 );
    *p++ = Js( 1, 0 ) * Js( 1, 0 );
    *p++ = Js( 1, 1 ) * Js( 1, 1 );
    *p++ = Js( 1, 2 ) * Js( 1, 2 );
    *p++ = 2 * Js( 1, 0 ) * Js( 1, 1 );
    *p++ = 2 * Js( 1, 0 ) * Js( 1, 2 );
    *p++ = 2 * Js( 1, 1 ) * Js( 1, 2 );
    *p++ = Js( 0, 0 ) * Js( 1, 0 );
    *p++ = Js( 0, 1 ) * Js( 1, 1 );
    *p++ = Js( 0, 2 ) * Js( 1, 2 );
    *p++ = Js( 0, 0 ) * Js( 1, 1 ) + Js( 0, 1 ) * Js( 1, 0 );
    *p++ = Js( 0, 0 ) * Js( 1, 2 ) + Js( 0, 2 ) * Js( 1, 0 );
    *p++ = Js( 0, 1 ) * Js( 1, 2 ) + Js( 0, 2 ) * Js( 1, 1 );
}

static
inline
void
projectionJacobian( qmdMatrixRef& j, reald tx, reald ty, reald s, reald b )
{
    j( 0, 0 ) = s;
    j( 0, 1 ) = 0.;
    j( 0, 2 ) = -tx*b*s*s;
    j( 1, 0 ) = 0.;
    j( 1, 1 ) = s;;
    j( 1, 2 ) = -ty*b*s*s;
}

void
noErfBlob::forward( qmdVectorRef& dy,
		    qmdVectorRef& t,
		    qmdUnitQuaternionRef& q,
		    qmdVectorRef& l, // stddev ( = r/sqrt(5) for ellipse )
		    reald b )
{
    static qmdMatrix R( 3, 3 );
    q.getRotationMatrix( R );
    static qmdMatrix L2( 3, 3 ); // L2 is L squared, i.e. *covariance matrix*
    L2.setDiagonal( l );
    L2( 0, 0 ) *= L2( 0, 0 );
    L2( 1, 1 ) *= L2( 1, 1 );
    L2( 2, 2 ) *= L2( 2, 2 );
    _forward( dy, t, R, L2, b );
}

void
noErfBlob::forward( qmdVectorRef& dy, qmdVectorRef& x, reald b )
{
    static qmdMatrix R( 3, 3 );
    qmdUnitQuaternionRef( x.subVector(Q0,4)).getRotationMatrix( R );
    // L2 is L squared, i.e. *covariance matrix*
    static qmdMatrix L2( 3, 3 );
    L2.setDiagonal( x.subVector(LX,3) ); // L2 is L squared,*covariance matrix*
    L2( 0, 0 ) *= L2( 0, 0 );
    L2( 1, 1 ) *= L2( 1, 1 );
    L2( 2, 2 ) *= L2( 2, 2 );
    _forward( dy, x.subVector(TX,3), R, L2, b );
}

void
noErfBlob::_forward( qmdVectorRef& dy,
		     qmdVectorRef& t,
		     qmdMatrixRef& r,
		     qmdMatrixRef& l2,
		     reald b )
// _forward is for just one side
{
    // l2 is *covariance* matrix (diagonal)
    static qmdMatrix jmatP( 2, 3 );
    static qmdMatrix cmat( 3, 3 );
    static qmdVector cvec( 6 );
    static qmdMatrix tmpmat( 3, 3 );
    static qmdMatrix jmatC( 3, 6 );
    enum { U, V, SUU, SVV, SUV };
    reald s = 1. / ( 1. + t(2)*b );
    dy( U ) = t(0) * s;
    dy( V ) = t(1) * s;
    projectionJacobian( jmatP, t(0), t(1), s, b );
    makeJacobianProducts( jmatC.getArray(), jmatP );
    cmat.setMultiply( r, tmpmat.setMultiplyT( l2, r ) );
    makeCovVec( cvec, cmat );
    qmdVectorRef( 3, dy.getArray(SUU) ).setMultiply( jmatC, cvec );
}

void
noErfBlob::_forwardBoth( qmdVectorRef& y, const qmdVectorRef& x )
{
    static qmdMatrix rmatL( 3, 3 );
    static qmdVector tR( 3 );
    static qmdVector tmpvec( 3 );
    static qmdMatrix rmatR( 3, 3 );
    qmdVectorRef yL( 5, y.getArray(0) );
    qmdVectorRef yR( 5, y.getArray(5) );
    qmdVectorRef tL = x.subVector( TX, 3 );
    qmdUnitQuaternionRef( x.subVector(Q0,4) ).getRotationMatrix( rmatL );
    static qmdMatrix L2( 3, 3 );
    L2.setDiagonal( x.subVector(LX,3) ); // L2 is L squared,*covariance matrix*
    L2( 0, 0 ) *= L2( 0, 0 );
    L2( 1, 1 ) *= L2( 1, 1 );
    L2( 2, 2 ) *= L2( 2, 2 );
    _forward( yL, tL, rmatL, L2, beta );
    tR.setAdd( Trl, tmpvec.setMultiply( Rrl, tL ) );
    rmatR.setMultiply( Rrl, rmatL );
    _forward( yR, tR, rmatR, L2, beta );
}

const qmdVectorRef& 
noErfBlob::operator()( const qmdVectorRef& x )
// this is an error function = _forward(x) - _observed
{
#if 0
    static qmdMatrix rmatL( 3, 3 );
    static qmdMatrix l2mat( 3, 3 );  // *covariance* (squares)
    static qmdVector tR( 3 );
    static qmdVector tmpvec( 3 );
    static qmdMatrix rmatR( 3, 3 );
    qmdVectorRef dyL( 5, dy_return.getArray(0) );
    qmdVectorRef dyR( 5, dy_return.getArray(5) );
    qmdVectorRef tL = x.subVector( TX, 3 );
    qmdUnitQuaternionRef( x.subVector(Q0,4) ).getRotationMatrix( rmatL );
    // l2mat.setDiagonal( x.subVector( LX, 3 ) );
    l2mat.setZero();
    l2mat( 0, 0 ) = x.getValue( LX ) * x.getValue( LX );
    l2mat( 1, 1 ) = x.getValue( LY ) * x.getValue( LY );
    l2mat( 2, 2 ) = x.getValue( LZ ) + x.getValue( LZ );  // <-- !! BUG "+"
    _forward( dyL, tL, rmatL, l2mat, beta );
    tR.setAdd( Trl, tmpvec.setMultiply( Rrl, tL ) );
    rmatR.setMultiply( Rrl, rmatL );
    _forward( dyR, tR, rmatR, l2mat, beta );
#endif
    _forwardBoth( dy_return, x );
    dy_return -= y_obs;
    return dy_return;
}

#if 0
static
void
jacobian_uv_t( qmdMatrixRef& J, 
	       qmdVectorRef& t,
	       reald b,
	       qmdVectorRef& Tc, 
	       qmdMatrixRef& Rc )
{
    // tc = Tc + Rc t
    // (u,v) = ( tcx/(1+tcz b), tcy/(1+tcz b) )
    static qmdVector tc(3);
    static qmdVector tmp(3);
    static qmdMatrix Jp( 2, 3 );
    static qmdMatrix Jsub( 2, 3 );
    tc.setAdd( Tc, tmp.setMultiply( Rc, t ) );
    reald s = 1. / ( 1. + tc(2)*b );
    projectionJacobian( Jp, tc(0), tc(1), s, b );
    Jsub.setMultiply( Jp, Rc );
    reald* rp = J.getArray();
    *rp++ = Jsub(0,0);
    *rp++ = Jsub(0,1);
    *rp++ = Jsub(0,2);
    for( int i = 0; i < 6; i++ )
       *rp++ = 0.;
    *rp++ = Jsub(1,0);
    *rp++ = Jsub(2,1);
    *rp++ = Jsub(3,2);
    for( int i = 0; i < 6; i++ )
       *rp++ = 0.;
}

static
inline
qmdMatrixRef&
quadProduct( qmdMatrixRef& out, qmdMatrixRef& P, qmdMatrixRef& S )
{
    // S must be square: out = P S P^T
    qmdMatrix tmp( S.getWidth(), S.getWidth() );
    out.setMultiply( P, tmp.setMultiplyT( S, P ) );
    return out;
}

static 
void
jacobian_cov_t()
{
    static qmdMatrix A(3,3); // symmetric, Rc Rb L Rb" Rc"
    static qmdMatrix dC2_dtc( 3, 3 );
    static qmdMatrix dC2_dtb( 3, 3 );
    qmdMatrixRef& a = RcRbLRbRc;
    reald
    reald s4 = s*s*s*s;
    reald u = ( 1 + tc(2) * beta );
#if 0
    dC2_dtc( 0, 0 ) = 2*b*( -A(0,2) - b*A(0,2)*tc(2) + b*tc(0)*A(2,2) )*s4;
    dC2_dtc( 1, 0 ) = 0.;
    dC2_dtc( 2, 0 ) = -b*( A(1,2) + A(1,2)*b*tc(2) - b*A(2,2)*tc(1) )*s4;
    dC2_dtc( 0, 1 ) = 0.;
    dC2_dtc( 1, 1 ) = 2*b*( -A(1,2) - b*A(1,2)*tc(2) + b*tc(1)*A(2,2) )*s4;
    dC2_dtc( 2, 1 ) = -b*( A(0,2) + A(0,2)*b*tc(2) - b*A(2,2)*tc(0) )*s4;
    dC2_dtc( 0, 2 ) = 0.;
    dC2_dtc( 1, 2 ) = 2*b*( -A(1,2) - b*A(1,2)*tc(2) + b*tc(1)*A(2,2) )*s4;
    dC2_dtc( 2, 2 ) = -b*( A(0,2) + A(0,2)*b*tc(2) - b*A(2,2)*tc(0) )*s4;
#endif
    
    
}
#endif

void
noErfBlob::jacobian( qmdMatrixRef& J, const qmdVectorRef& x, reald b, 
		     const qmdVectorRef& Tc, const qmdMatrixRef& Rc )
{
    //       tx  ty  tz  wx  wy  wz  lx  ly  lz
    //  uM   %%  %%  %%   0   0   0   0   0   0
    //  vM   %%  %%  %%   0   0   0   0   0   0
    // suu   %%  %%  %%  %%  %%  %%  %%  %%  %%
    // svv   %%  %%  %%  %%  %%  %%  %%  %%  %%
    // suv   %%  %%  %%  %%  %%  %%  %%  %%  %%

    J.setZero();

    // dM_dt, upper left (2,3) 
    qmdVector tmp( 3 );
    qmdVector tc(3);
    tc.setAdd( Tc, tmp.setMultiply( Rc, x.subVector(TX,3) ) );
    reald u = ( 1 + tc[2] * b );
    reald u2 = u*u;
    static qmdMatrix dM_dt( 2, 3 );
    static qmdMatrix dM_dtc( 2, 3 );
    dM_dtc( 0, 0 ) = 1./u;
    dM_dtc( 0, 1 ) = 0.;
    dM_dtc( 0, 2 ) = -tc[0] * b / u2;
    dM_dtc( 1, 0 ) = 0.;
    dM_dtc( 1, 1 ) = 1./u;
    dM_dtc( 1, 2 ) = -tc[1] * b / u2;
    dM_dt.setMultiply( dM_dtc, Rc );
    qmdSubMatrixRef( J, 2, 3, 0, 0 ).setData( dM_dt.getArray() );

    // dC2_dt, lower left (3,3)
    static qmdMatrix dC2_dt( 3, 3 );
    static qmdMatrix dC2_dJ( 3, 4 );
    static qmdMatrix dJ_dtc( 4, 3 );
    static qmdMatrix L2( 3, 3 );
    L2.setDiagonal( x.subVector(LX,3) );
    L2( 0, 0 ) *= L2( 0, 0 );
    L2( 1, 1 ) *= L2( 1, 1 );
    L2( 2, 2 ) *= L2( 2, 2 );
    static qmdMatrix Rb( 3, 3 );
    qmdUnitQuaternionRef( x.subVector(Q0,4) ).getRotationMatrix( Rb );
    static qmdMatrix RbLRb( 3, 3 );
    static qmdMatrix tmpM( 3, 3 );
    static qmdMatrix A( 3, 3 );
    RbLRb.setMultiply( Rb, tmpM.setMultiplyT( L2, Rb ) );
    A.setMultiply( Rc, tmpM.setMultiplyT( RbLRb, Rc ) ); // RcRbLRb"Rc"
    dC2_dJ( 0, 0 ) = 2 * A(0,0) * dM_dtc(0,0) + 2 * A(0,2) * dM_dtc(0,2);
    dC2_dJ( 0, 1 ) = 0;
    dC2_dJ( 0, 2 ) = 2 * A(0,2) * dM_dtc(0,0) + 2 * A(2,2) * dM_dtc(0,2);
    dC2_dJ( 0, 3 ) = 0;
    dC2_dJ( 1, 0 ) = 0;
    dC2_dJ( 1, 1 ) = 2 * A(1,1) * dM_dtc(1,1) + 2 * A(1,2) * dM_dtc(1,2);
    dC2_dJ( 1, 2 ) = 0;
    dC2_dJ( 1, 3 ) = 2 * A(1,2) * dM_dtc(1,1) + 2 * A(2,2) * dM_dtc(1,2);
    dC2_dJ( 2, 0 ) = A(0,1) * dM_dtc(1,1) + A(0,2) * dM_dtc(1,2);
    dC2_dJ( 2, 1 ) = A(0,1) * dM_dtc(0,0) + A(1,2) * dM_dtc(0,2);
    dC2_dJ( 2, 2 ) = A(1,2) * dM_dtc(1,1) + A(2,2) * dM_dtc(1,2);
    dC2_dJ( 2, 3 ) = A(0,2) * dM_dtc(0,0) + A(2,2) * dM_dtc(0,2);
    dJ_dtc.setZero();
    dJ_dtc( 0, 2 ) = -b / u2;
    dJ_dtc( 1, 2 ) = dJ_dtc( 0, 2 );
    dJ_dtc( 2, 0 ) = dJ_dtc( 0, 2 );
    dJ_dtc( 2, 2 ) = 2 * tc[0] * b * b / (u*u2);
    dJ_dtc( 3, 1 ) = dJ_dtc( 0, 2 );
    dJ_dtc( 3, 2 ) = 2 * tc[1] * b * b / (u*u2);
    dC2_dt.setMultiply( tmpM.setMultiply( dC2_dJ, dJ_dtc ), Rc );
    qmdSubMatrixRef( J, 3, 3, 2, 0 ).setData( dC2_dt.getArray() );

    // dC2_dw, lower middle (3,3)
    static qmdMatrix dC2_dw( 3, 3 );
    static qmdMatrix dC2_dD( 3, 6 );
    static qmdMatrix dD_dw( 6, 3 );
    static qmdMatrix JRcRb( 2, 3 );
    qmdMatrixRef& d = JRcRb;
    // d.setMultiply( dM_dtc, Rc );
    JRcRb.setMultiply( dM_dtc, tmpM.setMultiply( Rc, Rb ) );
    if( verbose ) Rb.print( cerr << "Rb= " << endl ) << endl;
    if( verbose ) Rc.print( cerr << "Rc= " << endl ) << endl;
    if( verbose ) JRcRb.print( cerr << "JRcRb= " << endl ) << endl;
    // qmdMatrixRef& a = RbLRb;
    qmdMatrixRef& a = L2;
    dC2_dD( 0, 0 ) = 2*a(0,0)*d(0,0); //  + 2*a(0,1)*d(0,1) + 2*a(0,2)*d(0,2);
    dC2_dD( 0, 1 ) = 2*a(1,1)*d(0,1); //  + 2*a(0,1)*d(0,0) + 2*a(1,2)*d(0,2);
    dC2_dD( 0, 2 ) = 2*a(2,2)*d(0,2); //  + 2*a(0,2)*d(0,0) + 2*a(1,2)*d(0,1);
    dC2_dD( 0, 3 ) = 0;
    dC2_dD( 0, 4 ) = 0;
    dC2_dD( 0, 5 ) = 0;
    dC2_dD( 1, 0 ) = 0;
    dC2_dD( 1, 1 ) = 0;
    dC2_dD( 1, 2 ) = 0;
    dC2_dD( 1, 3 ) = 2*a(0,0)*d(1,0); //  + 2*a(0,1)*d(1,1) + 2*a(0,2)*d(1,2);
    dC2_dD( 1, 4 ) = 2*a(1,1)*d(1,1); //  + 2*a(0,1)*d(1,0) + 2*a(1,2)*d(1,2);
    dC2_dD( 1, 5 ) = 2*a(2,2)*d(1,2); //  + 2*a(0,2)*d(1,0) + 2*a(1,2)*d(1,1);
    dC2_dD( 2, 0 ) = a(0,0)*d(1,0); //  + a(0,1)*d(1,1) + a(0,2)*d(1,2);
    dC2_dD( 2, 1 ) = a(1,1)*d(1,1); //  + a(0,1)*d(1,0) + a(1,2)*d(1,2);
    dC2_dD( 2, 2 ) = a(2,2)*d(1,2); //  + a(0,2)*d(1,0) + a(1,2)*d(1,1);
    dC2_dD( 2, 3 ) = a(0,0)*d(0,0); //  + a(0,1)*d(0,1) + a(0,2)*d(0,2);
    dC2_dD( 2, 4 ) = a(1,1)*d(0,1); //  + a(0,1)*d(0,0) + a(1,2)*d(0,2);
    dC2_dD( 2, 5 ) = a(2,2)*d(0,2); //  + a(0,2)*d(0,0) + a(1,2)*d(0,1);
    if( verbose ) dC2_dD.print( cerr << "dC2_dD= " << endl ) << endl;
    dD_dw( 0, 0 ) = 0.;
    dD_dw( 0, 1 ) = d(0,2);
    dD_dw( 0, 2 ) = -d(0,1);
    dD_dw( 1, 0 ) = -d(0,2);
    dD_dw( 1, 1 ) = 0.;
    dD_dw( 1, 2 ) = d(0,0);
    dD_dw( 2, 0 ) = d(0,1);
    dD_dw( 2, 1 ) = -d(0,0);
    dD_dw( 2, 2 ) = 0.;
    dD_dw( 3, 0 ) = 0.;
    dD_dw( 3, 1 ) = d(1,2);
    dD_dw( 3, 2 ) = -d(1,1);
    dD_dw( 4, 0 ) = -d(1,2);
    dD_dw( 4, 1 ) = 0.;
    dD_dw( 4, 2 ) = d(1,0);
    dD_dw( 5, 0 ) = d(1,1);
    dD_dw( 5, 1 ) = -d(1,0);
    dD_dw( 5, 2 ) = 0.;
    if( verbose ) dD_dw.print( cerr << "dD_dw= " << endl ) << endl;
    dC2_dw.setMultiply( dC2_dD, dD_dw );
    dC2_dw.setNegateEq();     // !!
    if( verbose ) dC2_dw.print( cerr << "dC2_dw= " << endl ) << endl;
    qmdSubMatrixRef( J, 3, 3, 2, 3 ).setData( dC2_dw.getArray() );
    if( verbose ) J.print( cerr << "J (after dC2_dw)= " << endl ) << endl;

    // dC2_dL, lower right (3,3)
    // static qmdMatrix dC2_dL2( 3, 3 );
    static qmdMatrix dC2_dL( 3, 3 );	// dC2_dL = dC2_dL2 dL2_dL
    // static qmdMatrix JRcRb( 2, 3 );	// J * Rc * Rb
    static qmdMatrix Ls( 3, 3 );		// 2 * L  ( = dL2_dL )
    // JRcRb.setMultiply( d, Rb );
    reald lx = x.getValue( LX );
    reald ly = x.getValue( LY );
    reald lz = x.getValue( LZ );
    dC2_dL( 0, 0 ) = 2 * JRcRb(0,0) * JRcRb(0,0) * lx;
    dC2_dL( 0, 1 ) = 2 * JRcRb(0,1) * JRcRb(0,1) * ly;
    dC2_dL( 0, 2 ) = 2 * JRcRb(0,2) * JRcRb(0,2) * lz;
    dC2_dL( 1, 0 ) = 2 * JRcRb(1,0) * JRcRb(1,0) * lx;
    dC2_dL( 1, 1 ) = 2 * JRcRb(1,1) * JRcRb(1,1) * ly;
    dC2_dL( 1, 2 ) = 2 * JRcRb(1,2) * JRcRb(1,2) * lz;
    dC2_dL( 2, 0 ) = 2 * JRcRb(0,0) * JRcRb(1,0) * lx;
    dC2_dL( 2, 1 ) = 2 * JRcRb(0,1) * JRcRb(1,1) * ly;
    dC2_dL( 2, 2 ) = 2 * JRcRb(0,2) * JRcRb(1,2) * lz;
    // okay, now this is actually dC2/dL2  (3,3)
    // we need to multiply it by dL2/dL = diag( 2 lx, 2 ly, 2 lz )
    // Ls.setDiagonal( x.subVector( LX, 3 ) );
    // Ls *= 2.;
    // dC2_dL.setMultiply( dC2_dL2, Ls );
    qmdSubMatrixRef( J, 3, 3, 2, 6 ).setData( dC2_dL.getArray() );
}

// const qmdMatrixRef&
// noErfBlob::jacobian( const qmdVectorRef& x )
// {
//     static reald tdata[3] = { 0., 0., 0. };
//     static reald rdata[9] = { 1., 0., 0., 0., 1., 0., 0., 0., 1. };
//     // J is (10,9), with a (5,9)(5,9) split: top=Left, bottom=Right
//     qmdMatrixRef JL( 5, 9, J_return.getArray() );
//     qmdMatrixRef JR( 5, 9, J_return.getArray(5) );
//     jacobian( JL, x, beta, qmdVectorRef(3,tdata), qmdMatrixRef(3,3,rdata) );
//     jacobian( JR, x, beta, Trl, Rrl );
//     return J_return;
// }

inline
void
fillColumn( int col, 
	    qmdMatrixRef& J,
	    int icol,
	    const qmdMatrixRef& JL,
	    const qmdMatrixRef& JR )
{
    for( int i = 0; i < 5; i++ )
       J( i, col ) = JL.getValue( i, icol );
    for( i = 0; i < 5; i++ )
       J( i+5, col ) = JR.getValue( i, icol );
}

const qmdMatrixRef&
noErfBlob::jacobian( const qmdVectorRef& x )
{
    static reald tdata[3] = { 0., 0., 0. };
    static reald rdata[9] = { 1., 0., 0., 0., 1., 0., 0., 0., 1. };
    // J is (10,9), with a (5,9)(5,9) split: top=Left, bottom=Right
    static qmdMatrix JL( 5, 9 );
    static qmdMatrix JR( 5, 9 );
    jacobian( JL, x, beta, qmdVectorRef(3,tdata), qmdMatrixRef(3,3,rdata) );
    jacobian( JR, x, beta, Trl, Rrl );
    if( J_return.getWidth() != freeCount )
       cerr << "noErfBlob::jacobian: ERROR: " << J_return.getWidth()
	    << ", " << freeCount << endl;
    for( int col = 0; col < freeCount; col++ )
       fillColumn( col, J_return, freeMap[col], JL, JR );
    return J_return;
}

const qmdVectorRef&
noErfBlob::add( const qmdVectorRef& x, const qmdVectorRef& dx_part )
{
    static qmdVector dx( 9 );
    if( 9 == freeCount )
       dx.copy( dx_part );
    else
       {
	   dx.setZero();
	   for( int i = 0; i < freeCount; i++ )
	      dx( freeMap[i] ) = dx_part.getValue( i );
       }
    x_return.subVector(TX,3).setAdd( x.subVector(TX,3), dx.subVector(DTX,3) );
    if( verbose ) x.print( cerr << "x= " ) << endl;
    if( verbose ) dx.print( cerr << "dx= " ) << endl;
    if( verbose ) qmdUnitQuaternionRef( x.subVector(Q0,4) ).print( cerr << "q(x)= " ) << endl;
    if( verbose ) dx.subVector( DWX,3 ).print( cerr << "dw(x)= " ) << endl;
    qmdUnitQuaternionRef( x_return.subVector(Q0,4) ).setAddAngles
       ( qmdUnitQuaternionRef( x.subVector(Q0,4) ), dx.subVector( DWX,3 ) );
    if( verbose ) qmdUnitQuaternionRef( x_return.subVector(Q0,4) ).print( cerr << "q(xn)= " ) << endl;
    x_return.subVector(LX,3).setAdd( x.subVector(LX,3), dx.subVector(DLX,3) );
    return x_return;
}

void
noErfBlob::computeLinearInit( qmdVectorRef& x )
{
    assert( inited );
    cgTriangleStereo triangle;
    triangle.setRightCoordinateTransform
       ( Trl(0), Trl(1), Trl(2), Qrl(0), Qrl(1), Qrl(2), Qrl(3) );
    triangle.point( y_obs.subVector( UL, 2 ),
		    y_obs.subVector( UR, 2 ),
		    x.subVector( TX, 3 ) );
    // Trl.print( cerr << "noErfBlob::computeLinearInit: Trl= " ) << endl;
    // Qrl.print( cerr << "noErfBlob::computeLinearInit: Qrl= " ) << endl;
    // Rrl.print( cerr << "noErfBlob::computeLinearInit: Rrl= " ) << endl;
    
    // decompose 2D covariance
    qmdVector evalues( 2 );
    qmdMatrix evectors( 2, 2 );  // evectors in rows (dumb)
    qmdEigenStructure::eigenSymmetric2( covL_obs, evalues, evectors );
    reald& theSine = evectors(0,1);
    reald& theCosine = evectors(0,0);
    reald angle = atan2( theSine, theCosine );
    x( Q0 ) = cos( angle / 2 );
    x( QX ) = 0.;
    x( QY ) = 0.;
    x( QZ ) = sin( angle / 2 );
    // sx = a3 / sqrt(5)
    // a3 = ( 1 + tz b ) a2
    // su = a2 / sqrt(4)
    //    ==> sx = sqrt(.8) ( 1 + tz b ) su
    const reald ks = sqrt( .8 );
    reald scale = ks * ( 1 + beta * x( TZ ) );
    x( LX ) = scale * sqrt( evalues(0) );	// std devs of 3-D
    x( LY ) = scale * sqrt( evalues(1) );
    // x( LZ ) = .5 * ( x( LX ), x( LY ) );
    x( LZ ) = .5 * ( x( LX ) + x( LY ) );
}

void
noErfBlob::setFree( int ix[] )
{
    freeCount = 0;
    for( int i = 0; i < 9; i++ )
       if( ix[i] )
	  freeMap[ freeCount++ ] = i;
    noErrorFunction::init( nMmts(), nStates(), nDStates() );
}
