// noErfBlob.h++

#ifndef _noErfBlob_H_
#define _noErfBlob_H_

#include <math.h>
#include "noErrorFunction.h"

class noErfBlob : public noErrorFunction
{
 protected:
    int inited;
    qmdVector y_obs;
    qmdMatrix covL_obs;
    qmdMatrix covR_obs;
    reald beta;
    qmdVector Trl;
    qmdUnitQuaternion Qrl;
    qmdMatrix Rrl;
    int freeCount;
    int freeMap[9];
 public:
    static int verbose;
    // Mmts: (u,v)= center, (Suu,Svv,Suv)= variances & covariance
    // State: (tx,ty,tz)= center, (q)=rotation, (l)= stdev
    enum _mmts { UL, VL, SuuL, SvvL, SuvL, UR, VR, SuuR, SvvR, SuvR };
    enum _state { TX, TY, TZ, Q0, QX, QY, QZ, LX, LY, LZ };
    enum _dstate { DTX, DTY, DTZ, DWX, DWY, DWZ, DLX, DLY, DLZ };
    noErfBlob()
      : noErrorFunction( nMmts(), nStates(), 9 ),
	inited( 0 ),
	y_obs( nMmts() ),
	covL_obs( 2, 2 ),
	covR_obs( 2, 2 ),
	beta( 1. ),
	Trl( 3 ),
	Qrl(),
	Rrl( 3, 3 ),
	freeCount( 9 )
    { for( int i = 0; i < 9; i++ ) freeMap[i] = i; }
    ~noErfBlob() {}
    void init( const qmdVectorRef& mL,
	       const qmdMatrixRef& c2L, 
	       const qmdVectorRef& mR,
	       const qmdMatrixRef& c2R,
	       const reald b,
	       const qmdVectorRef& t,
	       const qmdUnitQuaternionRef& q );
    void init( const qmdVectorRef& y,
	       const reald b,
	       const qmdVectorRef& t,
	       const qmdUnitQuaternionRef& q );
    void initStereo( const reald b, 
		     const qmdVectorRef& t,
		     const qmdUnitQuaternionRef& q );
    void forward( qmdVectorRef& dy, 
		  qmdVectorRef& t, 
		  qmdUnitQuaternionRef& q,
		  qmdVectorRef& l, // stdev
		  reald b );
    void forward( qmdVectorRef& dy, qmdVectorRef& x, reald b );
    void _forward( qmdVectorRef& dy, 
		   qmdVectorRef& t, qmdMatrixRef& r, qmdMatrixRef& L2, reald b );
    void jacobian( qmdMatrixRef& J, const qmdVectorRef& x, reald b, 
		   const qmdVectorRef& Tc, const qmdMatrixRef& Rc );
    void _forwardBoth( qmdVectorRef& y, const qmdVectorRef& x );
    // void _jacobianBoth( qmdMatrixRef& J, const qmdVectorRef& x );
    virtual const qmdVectorRef& operator()( const qmdVectorRef& x );
    // **
    virtual const qmdMatrixRef& jacobian( const qmdVectorRef& x );
    // **
    virtual const qmdVectorRef& add( const qmdVectorRef& x, const qmdVectorRef& dx );
    static void setVerbose( int v ) { verbose = v; }
    inline int nMmts() { return 10; }
    inline int nStates() { return 10; }
    // **
    // static inline int nDStates() { return 9; }
    inline int nDStates() { return( freeCount ); }
    // **
    void computeLinearInit( qmdVectorRef& x );
    void setFree( int[] );
};

#endif /* _noErfBlob_H_ */
