// noErfQuirk.c++

#include "noErfQuirk.h"

qmdVector noErfQuirk::projPoint( 2 );
qmdVector noErfQuirk::camPoint( 3 );
qmdMatrix noErfQuirk::rotMat( 3, 3 );
qmdVector noErfQuirk::tmpPoint( 3 );

// void
// noErfQuirk::init( int N, int M, reald* y_data )
// {
//     noErrorFunction::init( 2*N*M, 3*N+8*M, 3*N+7*M );
//     pointCount = N;
//     frameCount = M;
//     y_obs.set( 2*N*M );
//     y_obs.copy( qmdVectorRef( 2*N*M, y_data ) );
//     int dxCount = 3*N + 7*M;
//     dxMap = new int[ dxCount ];
//     for( int i = 0; i < dxCount; i++ )
//        dxMap[i] = i;
//     // basically this function is dumb now
// }
// 
// void
// noErfQuirk::init( int N, int M, reald* y_data, int* map )
// {
//     int dxCount = 3*N + 7*M;
//     int dxFreeCount = 0;
//     dxMap = new int[ dxCount ];
//     for( int i = 0; i < dxCount; i++ )
//        dxMap[i] = ( map[i] ) ? dxFreeCount++ : -1;
//     noErrorFunction::init( 2*N*M, 3*N+8*M, dxFreeCount );
//     pointCount = N;
//     frameCount = M;
//     y_obs.set( 2*N*M );
//     y_obs.copy( qmdVectorRef( 2*N*M, y_data ) );
// }

void
noErfQuirk::init( int N, int M, int* freelist, int** ymat, 
		  qmdVectorRef& ydat, int* mlist )
{
    pointCount = N;
    frameCount = M;
    pFreeCount = 0;
    pFreeMap = new int[ N ];
    for( int k = 0; k < N; k++ )
       pFreeMap[k] = ( freelist[k] ) ? pFreeCount++ : -1;
    cerr << "noErfQuirk: Nfree= " << pFreeCount << endl;
    mFreeMap = new int[ M ];
    if( !mlist )
       {
	   mFreeCount = M-1;
	   mFreeMap[ 0 ] = -1;
	   for( int t = 1; t < M; t++ )
	      mFreeMap[ t ] = t-1;
       }
    else
       {
	   mFreeCount = 0;
	   for( int t = 0; t < M; t++ )
	      mFreeMap[t] = ( mlist[t] ) ? mFreeCount++ : -1;
       }
    yValidCount = 0;
    yValid = new int*[M];
    for( int t = 0; t < M; t++ )
       {
	   yValid[t] = new int[N];
	   for( int k = 0; k < N; k++ )
	      if( yValid[t][k] = ymat[t][k] )
		 yValidCount++;
       }
    if( 2*yValidCount != ydat.getLength() )
       cerr << "noErfQuirk::init: ERROR: dat mat don't match dat dat" << endl;
    y_obs.set( 2*yValidCount );
    y_obs.copy( ydat );
    betaFixed = 1;
    cerr << "noErfQuirk: dx [ " << getDxCount() << " ]" << endl;
    noErrorFunction::init( getYCount(), getXCount(), getDxCount() );
}

inline
const qmdVectorRef&
noErfQuirk::cameraPoint( const qmdVectorRef& t,
			 const qmdUnitQuaternionRef& q, 
			 const qmdVectorRef& p )
{
    q.getRotationMatrix( rotMat );
    camPoint.setTMultiply( rotMat, tmpPoint.setSubtract( p, t ) );
    // camPoint.print( cerr << "noErfQuirk::cameraPoint: " ) << endl;
    return camPoint;
}

inline
const qmdVectorRef&
noErfQuirk::project( reald b, const qmdVectorRef& p )
{
    reald denom = ( 1 + p.getValue(2)*b );
    projPoint( 0 ) = p.getValue(0) / denom;
    projPoint( 1 ) = p.getValue(1) / denom;
    // projPoint.print( cerr << "noErfQuirk::project: " ) << endl;
    return projPoint;
}

void
noErfQuirk::_forward( qmdVectorRef& y, const qmdVectorRef& x )
{
    qmdUnitQuaternionRef q;
    reald b;
    int iy = 0;
    int cbase = 3*pointCount;
    for( int t = 0; t < frameCount; t++ )
       for( int k =0; k < pointCount; k++ )
	  {
	      if( ! yValid[t][k] )
		 continue;
	      int ip = 3*k;
	      int ic = cbase + 8*t;
	      qmdVectorRef& yk = y.subVector( iy, 2 );
	      qmdVectorRef& p = x.subVector( ip, 3 );
	      qmdVectorRef& tr = x.subVector( ic, 3 );
	      q.set( x.subVector( ic + 3, 4 ).getArray() );
	      // c = x.subVector( cbase + 8*t, 8 );
	      b = x.getValue( ic + 7 );
	      // yk.print( cerr << "yk= " ) << endl;
	      yk.copy( project( b, cameraPoint( tr, q, p ) ) );
	      iy += 2;
	  }
}

const qmdVectorRef&
noErfQuirk::operator()( const qmdVectorRef& x )
{
    _forward( dy_return, x );
    dy_return -= y_obs;
    return dy_return;
}


#if 0
const qmdMatrixRef&
noErfQuirk::jacobian( const qmdVectorRef& x )
{
    // for fixed states, dy/ddx = 0 ??????

    static reald b;
    static qmdVectorRef tWC;
    static qmdUnitQuaternion q;
    static qmdMatrix Rt( 3, 3 );
    static qmdVectorRef pW;
    static qmdVector pC_W( 3 );
    static qmdVector pC( 3 );
    static reald distortion;
    static qmdMatrix dP_dpC( 2, 3 );
    static qmdMatrix dy_dP( 2, 3 );
    // static qmdMatrix ddRdP_dw( 3, 3 );
    static qmdMatrix dT_dw( 3, 3 );
    static qmdMatrix dy_dw( 2, 3 );
    static qmdMatrix dy_db( 2, 1 );
    
    J_return.setZero();
    dP_dpC.setZero();
    // ddRdP_dw.setZero();
    dT_dw.setZero();

    int cbase = 3*pointCount;
    int yframe = 2*pointCount;
    
    for( int t = 0; t < frameCount; t++ )
       for( int k = 0; k < pointCount; k++ )
	  {
	      int ip_x = 3*k;
	      int ic_x = cbase + 8*t;
	      int ip_dx = 3*k;
	      int ic_dx = cbase + 7*t;
	      int iy = t*yframe + 2*k;

	      // p:k

	      b = x.getValue( ic_x + 7 );
	      tWC.set( x.subVector( ic_x, 3 ) );
	      q.copy( x.subVector( ic_x + 3, 4 ) );
	      q.setInvert( q );
	      q.getRotationMatrix( Rt );
	      pW.set( x.subVector( ip_x, 3 ) );
	      pC_W.setSubtract( pW, tWC );
	      pC.setMultiply( Rt, pC_W );
	      distortion = 1. / ( 1. + pC(2)*b );
	      dP_dpC(0,0) = dP_dpC(1,1) = distortion;
	      dP_dpC(0,2) = dP_dpC(1,2) = -distortion * distortion * b;
	      dP_dpC(0,2) *= pC(0);
	      dP_dpC(1,2) *= pC(1);
	      dy_dP.setMultiply( dP_dpC, Rt );   
	      qmdSubMatrixRef
		 ( J_return, 2, 3, iy, ip_dx ).setData(dy_dP.getArray());
	      
	      // t:t

	      dy_dP.setNegateEq(); // dy_dT
	      qmdSubMatrixRef
		 ( J_return, 2, 3, iy, ic_dx ).setData(dy_dP.getArray());

	      // w:t

	      // ddRdP_dw( 0, 1 ) = pC_W(2);
	      // ddRdP_dw( 0, 2 ) = -pC_W(1);
	      // ddRdP_dw( 1, 0 ) = -pC_W(2);
	      // ddRdP_dw( 1, 2 ) = pC_W(0);
	      // ddRdP_dw( 2, 0 ) = pC_W(1);
	      // ddRdP_dw( 2, 1 ) = -pC_W(0);
	      // dT_dw.setMultiply( Rt, ddRdP_dw );
	      dT_dw( 0, 1 ) = -pC(2);
	      dT_dw( 0, 2 ) = pC(1);
	      dT_dw( 1, 0 ) = pC(2);
	      dT_dw( 1, 2 ) = -pC(0);
	      dT_dw( 2, 0 ) = -pC(1);
	      dT_dw( 2, 1 ) = pC(0);
	      dy_dw.setMultiply( dP_dpC, dT_dw );
	      qmdSubMatrixRef
		 ( J_return, 2, 3, iy, ic_dx+3 ).setData(dy_dw.getArray());

	      // b:t

	      dy_db(0,0) = dy_db(1,0) = -pC(2) * distortion * distortion;
	      dy_db(0,0) *= pC(0);
	      dy_db(1,0) *= pC(1);
	      qmdSubMatrixRef
		 ( J_return, 2, 1, iy, ic_dx+6 ).setData(dy_db.getArray());

	  }
    return J_return;
}
#endif

const qmdMatrixRef&
noErfQuirk::jacobian( const qmdVectorRef& x )
{
    // for fixed states, dy/ddx = 0 ??????

    static reald b;
    static qmdVectorRef tWC;
    static qmdUnitQuaternion q;
    static qmdMatrix Rt( 3, 3 );
    static qmdVectorRef pW;
    static qmdVector pC_W( 3 );
    static qmdVector pC( 3 );
    static reald distortion;
    static qmdMatrix dP_dpC( 2, 3 );
    static qmdMatrix dy_dP( 2, 3 );
    static qmdMatrix dT_dw( 3, 3 );
    static qmdMatrix dy_dw( 2, 3 );
    static qmdMatrix dy_db( 2, 1 );
    
    J_return.setZero();
    dP_dpC.setZero();
    dT_dw.setZero();

    int cbase = 3*pointCount;
    int iy = 0;
    
    for( int t = 0; t < frameCount; t++ )
       for( int k = 0; k < pointCount; k++ )
	  {
	      if( ! yValid[t][k] )
		 continue;
	      int ip_x = 3*k;
	      int ic_x = cbase + 8*t;
	      // int ip_dx = 3*k;
	      // int ic_dx = cbase + 7*t;
	      // int iy = t*yframe + 2*k;

	      // p:k

	      b = x.getValue( ic_x + 7 );
	      tWC.set( x.subVector( ic_x, 3 ) );
	      q.copy( x.subVector( ic_x + 3, 4 ) );
	      q.setInvert( q );
	      q.getRotationMatrix( Rt );
	      pW.set( x.subVector( ip_x, 3 ) );
	      pC_W.setSubtract( pW, tWC );
	      pC.setMultiply( Rt, pC_W );
	      distortion = 1. / ( 1. + pC(2)*b );
	      dP_dpC(0,0) = dP_dpC(1,1) = distortion;
	      dP_dpC(0,2) = dP_dpC(1,2) = -distortion * distortion * b;
	      dP_dpC(0,2) *= pC(0);
	      dP_dpC(1,2) *= pC(1);
	      dy_dP.setMultiply( dP_dpC, Rt );
	      // qmdSubMatrixRef
	      // ( J_return, 2, 3, iy, ip_dx ).setData(dy_dP.getArray());
	      if( isFreePoint( k ) )
		 setdP( J_return, iy, k, dy_dP );
	      
	      // t:t

	      dy_dP.setNegateEq(); // dy_dT
	      // qmdSubMatrixRef
	      // ( J_return, 2, 3, iy, ic_dx ).setData(dy_dP.getArray());
	      if( isFreeCamera( t ) )
		 setdT( J_return, iy, t, dy_dP );

	      // w:t

	      dT_dw( 0, 1 ) = -pC(2);
	      dT_dw( 0, 2 ) = pC(1);
	      dT_dw( 1, 0 ) = pC(2);
	      dT_dw( 1, 2 ) = -pC(0);
	      dT_dw( 2, 0 ) = -pC(1);
	      dT_dw( 2, 1 ) = pC(0);
	      dy_dw.setMultiply( dP_dpC, dT_dw );
	      // qmdSubMatrixRef
	      // ( J_return, 2, 3, iy, ic_dx+3 ).setData(dy_dw.getArray());
	      if( isFreeCamera( t ) )
		 setdW( J_return, iy, t, dy_dw );

	      // b:t

	      dy_db(0,0) = dy_db(1,0) = -pC(2) * distortion * distortion;
	      dy_db(0,0) *= pC(0);
	      dy_db(1,0) *= pC(1);
	      //qmdSubMatrixRef
	      // ( J_return, 2, 1, iy, ic_dx+6 ).setData(dy_db.getArray());
	      if( !betaFixed )
		 setdB( J_return, iy, t, dy_db );

	      iy += 2;
	  }
    return J_return;
}

// inline
// int
// noErfQuirk::isFreePointCoord( int k, int i )
// {
//     return( dxMap[ 3*k + i ] > 0 );
// }

inline
int
noErfQuirk::dxPointIndex( int k, int i )
{
    if( -1 == pFreeMap[k] )
       cerr << "noErfQuirk::dxPointIndex: ERROR " << k << endl;
    // return( dxMap[ 3*k + i ] );
    return( pFreeMap[k] * 3 + i );
}

inline
int
noErfQuirk::dxTransIndex( int t, int i )
{
    if( -1 == mFreeMap[t] )
       cerr << "noErfQuirk::dxTransIndex: ERROR " << t << endl;
    // return( dxMap[ 3*pointCount + 7*t + i ] );
    return( 3*pFreeCount + mFreeMap[t] * 6 + i );
}

inline
int
noErfQuirk::dxRotIndex( int t, int i )
{
    if( -1 == mFreeMap[t] )
       cerr << "noErfQuirk::dxRotIndex: ERROR " << t << endl;
    // return( dxMap[ 3*pointCount + 7*t + 3 + i ] );
    return( 3*pFreeCount + mFreeMap[t] * 6 + 3 + i );
}

inline
int
noErfQuirk::dxBetaIndex( int t )
{
    if( betaFixed )
       cerr << "noErfQuirk::dxBetaIndex: ERROR: no dB " << t << endl;
    // return( dxMap[ 3*pointCount + 7*t + 6 ] );
    return( 3*pFreeCount + 6*mFreeCount );
}

void
noErfQuirk::setdP( qmdMatrixRef& J_return, int iy, int k, qmdMatrixRef& dy_dP )
{
    for( int i = 0; i < 3; i++ )
       {
	   int j = dxPointIndex( k, i );
	   if( j >= 0 )
	      {
		  J_return( iy, j ) = dy_dP( 0, i ); 
		  J_return( iy+1, j ) = dy_dP( 1, i );
	      }
       }
}

void
noErfQuirk::setdT( qmdMatrixRef& J_return, int iy, int t, qmdMatrixRef& dy_dT )
{
    for( int i = 0; i < 3; i++ )
       {
	   int j = dxTransIndex( t, i );
	   if( j >= 0 )
	      {
		  J_return( iy, j ) = dy_dT( 0, i ); 
		  J_return( iy+1, j ) = dy_dT( 1, i );
	      }
       }
}

void
noErfQuirk::setdW( qmdMatrixRef& J_return, int iy, int t, qmdMatrixRef& dy_dW )
{
    for( int i = 0; i < 3; i++ )
       {
	   int j = dxRotIndex( t, i );
	   if( j >= 0 )
	      {
		  J_return( iy, j ) = dy_dW( 0, i ); 
		  J_return( iy+1, j ) = dy_dW( 1, i );
	      }
       }
}

void
noErfQuirk::setdB( qmdMatrixRef& J_return, int iy, int t, qmdMatrixRef& dy_dB )
{
    int j = dxBetaIndex( t );
    if( j >= 0 )
       {
	   J_return( iy, j ) = dy_dB( 0, 0 ); 
	   J_return( iy+1, j ) = dy_dB( 1, 0 );
       }
}

#if 0
const qmdVectorRef&
noErfQuirk::add( const qmdVectorRef& x, const qmdVectorRef& dx )
{
    // for fixed states, x + dx = x
    // should check to see if dx is in fact zero for these states

    int cbase = 3*pointCount;

    x_return.subVector( 0, 3*pointCount )
       .setAdd( x.subVector(0,3*pointCount), dx.subVector(0,3*pointCount) );

    for( int t = 0; t < frameCount; t++ )
       {
	   int ic_x = cbase + 8*t;
	   int ic_dx = cbase + 7*t;
	   x_return.subVector( ic_x, 3 ).setAdd
	      ( x.subVector(ic_x,3), dx.subVector(ic_dx,3) );

	   qmdUnitQuaternionRef( x_return.subVector(ic_x+3,4) ).setAddAngles
	      ( qmdUnitQuaternionRef( x.subVector(ic_x+3,4) ), 
		dx.subVector( ic_dx+3, 3 ) );

	   x_return( ic_x + 7 ) = 
	      x.getValue( ic_x+7 ) + dx.getValue( ic_dx+6 );
       }
    return x_return;
}
#endif

const qmdVectorRef&
noErfQuirk::dxPoints( const qmdVectorRef& dx )
{
    static qmdVector xp_delta;
    xp_delta.set( 3*pointCount );
    int ixp = 0;
    for( int k = 0; k < pointCount; k++ )
       {
	   int whichdx = pFreeMap[ k ];
	   if( -1 == whichdx )
	      {
		  xp_delta[ixp] = 0.;
		  xp_delta[ixp+1] = 0.;
		  xp_delta[ixp+2] = 0.;
	      }
	   else
	      {
		  int dxbase = 3*whichdx;
		  xp_delta[ixp] = dx.getValue( dxbase );
		  xp_delta[ixp+1] = dx.getValue( dxbase + 1 );
		  xp_delta[ixp+2] = dx.getValue( dxbase + 2 );
	      }
	   ixp += 3;
       }
//     for( int i = 0; i < 3*pointCount; i++ )
//        {
// 	   int j = dxMap[ i ];
// 	   xp_delta( i ) = ( j >= 0 ) ? dx.getValue( j ) : 0.;
//        }
    return xp_delta;
}

static qmdVector dxtmp_vector( 3 );

inline
const qmdVectorRef&
noErfQuirk::dxTrans( const qmdVectorRef& dx, int t )
{
    int whichdt = mFreeMap[ t ];
    if( -1 == whichdt )
       {
	   dxtmp_vector( 0 ) = 0.;
	   dxtmp_vector( 1 ) = 0.;
	   dxtmp_vector( 2 ) = 0.;
       }
    else
       {
	   int dtbase = 3*pFreeCount + 6*whichdt;
	   dxtmp_vector( 0 ) = dx.getValue( dtbase );
	   dxtmp_vector( 1 ) = dx.getValue( dtbase + 1 );
	   dxtmp_vector( 2 ) = dx.getValue( dtbase + 2 );
       }
    return dxtmp_vector;
}

inline
const qmdVectorRef&
noErfQuirk::dxRot( const qmdVectorRef& dx, int t )
{
    int whichdt = mFreeMap[ t ];
    if( -1 == whichdt )
       {
	   dxtmp_vector( 0 ) = 0.;
	   dxtmp_vector( 1 ) = 0.;
	   dxtmp_vector( 2 ) = 0.;
       }
    else
       {
	   int dtbase = 3*pFreeCount + 6*whichdt + 3;
	   dxtmp_vector( 0 ) = dx.getValue( dtbase );
	   dxtmp_vector( 1 ) = dx.getValue( dtbase + 1 );
	   dxtmp_vector( 2 ) = dx.getValue( dtbase + 2 );
       }
    return dxtmp_vector;
//     for( int i = 0; i < 3; i++ )
//        {
// 	   int j = dxMap[ 3*pointCount + 7*t + 3 + i ];
// 	   dxtmp_vector( i ) = ( j >= 0 ) ? dx.getValue( j ) : 0.;
//        }
//     return dxtmp_vector;
}

inline
reald
noErfQuirk::dxBeta( const qmdVectorRef& dx )
{
//     int j = dxMap[ 3*pointCount + 7*t + 6 ];
//     return( ( j >= 0 ) ? dx.getValue( j ) : 0. );
    return( (betaFixed) ? 0. : dx.getValue( 3*pFreeCount + 6*mFreeCount ) );
}

const qmdVectorRef&
noErfQuirk::add( const qmdVectorRef& x, const qmdVectorRef& dx )
{
    int cbase = 3*pointCount;

    x_return.subVector( 0, 3*pointCount ).setAdd
       ( x.subVector( 0, 3*pointCount ), dxPoints( dx ) );

    for( int t = 0; t < frameCount; t++ )
       {
	   int ic_x = cbase + 8*t;

	   x_return.subVector( ic_x, 3 ).setAdd
	      ( x.subVector( ic_x, 3 ), dxTrans( dx, t ) );

	   qmdUnitQuaternionRef( x_return.subVector(ic_x+3,4) ).setAddAngles
	      ( qmdUnitQuaternionRef( x.subVector(ic_x+3,4) ), 
		dxRot( dx, t ) );

	   x_return( ic_x + 7 ) = x.getValue( ic_x + 7 ) + dxBeta( dx );
       }
    return x_return;
}

ostream&
noErfQuirk::printState( ostream& os, qmdVectorRef& x, int N, int M )
{
    int index = 0;
    for( int k = 0; k < N; k++ )
       {
	   x.subVector( index, 3 ).print( os << "\t" ) << endl;
	   index += 3;
       }
    for( int t = 0; t < M; t++ )
       {
	   x.subVector( index, 8 ).print( os << "\t" ) << endl;
	   index += 8;
       }
    return os;
}

istream&
noErfQuirk::scanState( istream& is, qmdVectorRef& x, int N, int M )
{
    int index = 0;
    for( int k = 0; k < N; k++ )
       {
	   x.subVector( index, 3 ).scan( is );
	   index += 3;
       }
    for( int t = 0; t < M; t++ )
       {
	   x.subVector( index, 8 ).scan( is );
	   index += 8;
       }
    return is;
}

// ostream&
// noErfQuirk::printObs( ostream& os, qmdVectorRef& y, int N, int M )
// {
//     int index = 0;
//     for( int t = 0; t < M; t++ )
//        {
// 	   for( int k = 0; k < N; k++ )
// 	      {
// 		  y.subVector( index, 2 ).print( os << "\t" ) << endl;
// 		  index += 2;
// 	      }
// 	   os << endl;
//        }
//     return os;
// }
// 
// istream&
// noErfQuirk::scanObs( istream& is, qmdVectorRef& y, int N, int M )
// {
//     int index = 0;
//     for( int t = 0; t < M; t++ )
//        for( int k = 0; k < N; k++ )
// 	  {
// 	      y.subVector( index, 2 ).scan( is );
// 	      index += 2;
// 	  }
//     return is;
// }
