// noErfQuirk.h++

#ifndef _noErfQuirk_H_
#define _noErfQuirk_H_

    // x = ( P:0, ... , P:N-1, C:0, ... , C:M-1 )
    //    P:k = ( X:k, Y:k, Z:k )
    //    C:t = ( TX:t, TY:t, TZ:t, Q0:t, QX:t, QY:t, QZ:t, B:t )
    //
    // dx = ( P:0, ... , P:N-1, C:0, ... , C:M-1 )
    //    dP:k = ( dX:k, dY:k, dZ:k )
    //    dC:t = ( dTX:t, dTY:t, dTZ:t, dWX:t, dWY:t, dWZ:t, dB:t )
    //
    // y = ( (Y:0, ..., Y:Nt-1):t=0, ... , (Y:0, ..., Y:Nt-1):t=M-1 )
    // 

#include "noErrorFunction.h"

class noErfQuirk : public noErrorFunction
{
    int pointCount;
    int frameCount;
    int pFreeCount;
    int* pFreeMap;
    int mFreeCount;
    int* mFreeMap;
    int yValidCount;
    int** yValid;
    int betaFixed;
    qmdVector y_obs;

    inline int getYCount() { return 2*yValidCount; }
    inline int getXCount() { return( 3*pointCount + 8*frameCount ); }
    inline int getDxCount() { return( 3*pFreeCount + 6*mFreeCount +
				      ((betaFixed) ? 0 : 1 )); }

    inline int isFreePoint( int k ) { return( -1 != pFreeMap[k] ); }
    inline int isFreeCamera( int t ) { return( -1 != mFreeMap[t] ); }

    inline const qmdVectorRef& cameraPoint( const qmdVectorRef& t,
					   const qmdUnitQuaternionRef& q,
					   const qmdVectorRef& p );
    inline const qmdVectorRef& project( reald b, const qmdVectorRef& p );
    static qmdVector projPoint;
    static qmdVector camPoint;
    static qmdMatrix rotMat;
    static qmdVector tmpPoint;
    inline int dxPointIndex( int k, int i );
    inline int dxTransIndex( int t, int i );
    inline int dxRotIndex( int t, int i );
    inline int dxBetaIndex( int t );
    void setdP( qmdMatrixRef&, int iy, int k, qmdMatrixRef& dyddx );
    void setdT( qmdMatrixRef&, int iy, int t, qmdMatrixRef& dyddx );
    void setdW( qmdMatrixRef&, int iy, int t, qmdMatrixRef& dyddx );
    void setdB( qmdMatrixRef&, int iy, int t, qmdMatrixRef& dyddx );
    const qmdVectorRef& dxPoints( const qmdVectorRef& dx );
    inline const qmdVectorRef& dxTrans( const qmdVectorRef& dx, int t );
    inline const qmdVectorRef& dxRot( const qmdVectorRef& dx, int t );
    inline reald dxBeta( const qmdVectorRef& dx );
public:
    noErfQuirk()
      : noErrorFunction(),
	pointCount(0), frameCount(0) {}
    enum points { X, Y, Z };
    enum cameras { TX, TY, TZ, Q0, QX, QY, QZ, B };
    // void init( int N, int M, reald* y_data );
    // void init( int N, int M, reald* y_data, int* map );
    // void init( int N, int M, int* plist, int** ylist, qmdVectorRef& y );
    void init( int N, int M, int* plist, int** ylist, qmdVectorRef& y,
	       int* mlist = 0 );
    void _forward( qmdVectorRef& y, const qmdVectorRef& x );
    const qmdVectorRef& operator()( const qmdVectorRef& x );
    const qmdMatrixRef& jacobian( const qmdVectorRef& x );
    const qmdVectorRef& add( const qmdVectorRef& x,
			    const qmdVectorRef& dx );
    static ostream& printState( ostream&, qmdVectorRef&, int N, int M );
    // static ostream& printObs( ostream&, qmdVectorRef&, int N, int M );
    static istream& scanState( istream&, qmdVectorRef&, int N, int M );
    // static istream& scanObs( istream&, qmdVectorRef&, int N, int M );
    static inline qmdVectorRef get3DPoint( const qmdVectorRef&, int k );
    static inline qmdVectorRef getCamera( const qmdVectorRef&, int t, int N );
    static inline qmdVectorRef getTrans( const qmdVectorRef&, int t, int N );
    static inline qmdUnitQuaternionRef getRot ( const qmdVectorRef&, int, int );
    static inline reald getBeta( const qmdVectorRef&, int t, int N );
};

inline
qmdVectorRef
noErfQuirk::get3DPoint( const qmdVectorRef& x, int k )
{
    return x.subVector( 3*k, 3 );
}

inline
qmdVectorRef
noErfQuirk::getCamera( const qmdVectorRef& x, int t, int N )
{
    return x.subVector( 3*N + 8*t, 8 );
}

inline
qmdVectorRef
noErfQuirk::getTrans( const qmdVectorRef& x, int t, int N )
{
    return x.subVector( 3*N + 8*t, 3 );
}

inline
qmdUnitQuaternionRef
noErfQuirk::getRot( const qmdVectorRef& x, int t, int N )
{
    return qmdUnitQuaternionRef( x.subVector( 3*N + 8*t + 3, 4 ) );
}

inline
reald
noErfQuirk::getBeta( const qmdVectorRef& x, int t, int N )
{
    return x.getValue( 3*N + 8*t + 7 );
}

#endif /* _noErfQuirk_H_ */

