// cgTriangle.c++

#include "cgTriangle.h"

reald cgTriangle::defaultSV = 0.;
int cgTriangle::svinvert = 1;

cgTriangle::cgTriangle()
  : T(3), R(3,3), bL(1.), bR(1.), state(1), q()
{
  T(0) = -1.;
  T(1) = 0.;
  T(2) = 0.;
  R.setIdentity();
  q.setIdentity();
}

cgTriangle::cgTriangle( char *filename )
  : T(3), R(3,3), bL(1.), bR(1.), state(1), q()
{
  if( FScan( filename ) )
     {
       cerr << "cgTriangle::cgTriangle: ERROR: Read of " 
	    << filename << " failed!" << endl;
       state = 0;
     }
}

istream&
cgTriangle::scan( istream& is )
{
  // qmdUnitQuaternion q;
  T.scan( is );
  q.scan( is );
  q.getRotationMatrix( R );
  is >> bL;
  is >> bR;

  if( bR > .01 )
     T(2) /= bR;
  else
     cerr << "oops in cgTriangle::scan" << endl;
  return is;
}

int
cgTriangle::FScan( char *filename )
{
  fstream file( filename, ios::in );
  if( !file )
     return 1;
  scan( file );
  file.close();
  return 0;
}

reald
cgTriangle::Point( reald uL, reald vL, reald uR, reald vR, qmdVectorRef& pL,
		   reald* residual )
{
  static qmdVector Aty(2);
  static qmdMatrix AtAinv(2,2);
  static qmdVector negY(3);
  static qmdVector rayL(3);
  static qmdVector rayR(3);
  static qmdVector rayL_R(3);
  static qmdVector pointL(3);
  static qmdVector pointR(3);
  static qmdVector a(2);

  reald discrim;

  pointL(0) = uL;
  pointL(1) = vL;
  pointL(2) = 0.;

  rayL(0) = uL * bL;
  rayL(1) = vL * bL;
  rayL(2) = 1.;

  pointR(0) = uR;
  pointR(1) = vR;
  pointR(2) = 0.;

  rayR(0) = uR * bR;
  rayR(1) = vR * bR;
  rayR(2) = 1.;

  // Construct AtA inverse

  rayL_R.setMultiply( R, rayL );
  // AtA( 0, 0 ) = rayL.getInnerProduct( rayL );
  // AtA( 0, 1 ) = -( rayR.getInnerProduct( qmd::setMultiply( temp, R, rayL ) ) );
  // AtA( 1, 0 ) = A(0,1);
  // AtA( 1, 1 ) = rayR.getInnerProduct( rayR );
  AtAinv( 0, 0 ) = rayR.getInnerProduct( rayR );
  AtAinv( 0, 1 ) = ( rayR.getInnerProduct( rayL_R ) );
  AtAinv( 1, 0 ) = AtAinv(0,1);
  AtAinv( 1, 1 ) = rayL.getInnerProduct( rayL );
  discrim = AtAinv(0,0)*AtAinv(1,1) - AtAinv(0,1)*AtAinv(1,0);
  AtAinv /= discrim;
  
  // Construct Aty

  // this is actually negative y!
  negY.setMultiply( R, pointL );
  negY.setAddEq( T );
  negY.setSubtractEq( pointR );  // negY currently has ( R qL + T - qR )
  Aty( 0 ) = - rayL_R.getInnerProduct( negY );
  Aty( 1 ) = rayR.getInnerProduct( negY );

  a.setMultiply( AtAinv, Aty );

  pL.copy( rayL ).setScaleEq( a(0) ).setAddEq( pointL );

  if( residual )
     {
	 qmdMatrix A( 3, 2 );
	 qmdVector negRayR( 3 );
	 qmdVector error( 3 );
	 negRayR.setNegate( rayR );
	 A.setColumn( 0, rayL_R );
	 A.setColumn( 1, negRayR );
	 error.setMultiply( A, a );
	 error.setAddEq( negY );
	 *residual = sqrt( error.getInnerProduct( error ) );
     }
#if 0
  // Testing
  cerr << "AtAinv= " << AtAinv << endl;
  cerr << "Aty= " << Aty << endl;
  qmdVector e(3);
  qmdVector pR_R(3);
  qmdVector pL_R(3);
  qmdMatrix A(3,2);
  A( 0, 0 ) = rayL_R(0);
  A( 1, 0 ) = rayL_R(1);
  A( 2, 0 ) = rayL_R(2);
  A( 0, 1 ) = -rayR(0);
  A( 1, 1 ) = -rayR(1);
  A( 2, 1 ) = -rayR(2);
  e.setMultiply( A, a ); // e = Ax
  e.setAddEq( negY );	   // e = Ax - y
  cerr << "3-D error (pRL_R)= " << e << endl;
  cerr << "  (aL, aR)= " << a << endl;
  pR_R.copy( rayR ).setScaleEq( a(1) ).setAddEq( pointR );
  cerr << "  pR_R = " << pR_R << endl;
  cerr << "  rayL = " << rayL << endl;
  cerr << "  pointL = " << pointL << endl;
  pL_R.setMultiply( R, pL ).setAddEq( T );
  cerr << "  pL = " << pL << endl;
  cerr << "  pL_R = " << pL_R << endl;
  cerr << "  R= " << R << endl;
  cerr << "  T= " << T << endl;
  
  cerr << endl;
#endif
  return a(1); // right alpha
}

#if 0
reald
cgTriangle::FakePoint( reald uL, reald vL, reald uR, reald vR, qmdVectorRef& pL )
{
  pL(0) = ( 1. + bL ) * uL;
  pL(1) = ( 1. + bL ) * vL;
  pL(2) = 1.5;
  return 1.;
}
#endif

static void
my_triangle( qmdVectorRef& T, qmdMatrixRef& R, reald bL, reald bR, 
	  reald uRight, reald vRight,  	// actually left
	  reald uLeft, reald vLeft,	// actually right
	  qmdVectorRef& P, reald& aR)
{
  qmdVector y(2);
  qmdMatrix A(2,2);
  qmdVector vL(3);
  qmdVector vR(3);
  qmdVector rL(3);
  qmdVector rR(3);
  qmdVector temp(3);
  qmdVector b(3);
  qmdVector a(2);

  vR(0) = uRight * bR;
  vR(1) = vRight * bR;
  vR(2) = 1.;

  vL(0) = uLeft * bL;
  vL(1) = vLeft * bL;
  vL(2) = 1.;

  A(0,0) = vR.getInnerProduct( vR );
  A(0,1) = -( vL.getInnerProduct( temp.setMultiply( R, vR ) ));
  A(1,0) = A(0,1);
  A(1,1) = vL.getInnerProduct( vL );

  cerr << "my AtA (before invert)= " << A << endl;

  rR(0) = uRight;
  rR(1) = vRight;
  rR(2) = 0.;

  rL(0) = uLeft;
  rL(1) = vLeft;
  rL(2) = 0.;

  b.copy( rL );
  b.setSubtractEq( T );
  b.setSubtractEq( temp.setMultiply( R, rR ) );

  y(0) = b.getInnerProduct( temp.setMultiply( R, vR ) );
  y(1) = -( b.getInnerProduct( vL ) );

  a.setMultiply( A.setInvertEq(), y );

  aR = a(0);
  reald& aL = a(1);
  qmdVector pL(3);
  qmdVector pR(3);
  qmdVector pLR(3);

  pL.setAdd( rL, vL.setScaleEq( aL ) );
  pR.setAdd( rR, vR.setScaleEq( aR ) );
  pLR.setAdd( T, temp.setMultiply( R, pR ) );
  P.setAdd( pL, pLR ).setScaleEq( .5 );
#if 1
  cerr << "my AtA (after invert)= " << A << endl;
  cerr << "my Aty= " << y << endl;
  cerr << "my pL= " << pL << endl;
  cerr << "my pR= " << pR << endl;
#endif
}

reald
cgTriangle::PointAlt( reald uL, reald vL, reald uR, reald vR, qmdVectorRef& pL )
{
  reald ar;
  my_triangle( T, R, bR, bL, uL, vL, uR, vR, pL, ar );
  return ar;
}

istream&
cgTriangle::scanEstRecord( istream& is )
{
  T.scan( is );
  q.scan( is );
  q.getRotationMatrix( R );
  is >> bL;
  bR = bL;
  if( bR > .01 )
     T(2) /= bR;
  else
     cerr << "oops in cgTriangle::scan" << endl;

  // Throw away the structure estimate
  reald fjunk;
  int ijunk;
  int count;
  is >> fjunk >> fjunk;
  is >> count;
  while( count-- )
     is >> ijunk >> fjunk;

  return is;
}

void
cgTriangle::ellipse( qmdVectorRef& mL, qmdMatrixRef& CL,
		     qmdVectorRef& mR, qmdMatrixRef& CR,
		     qmdVectorRef& poL, qmdMatrixRef& C3L )
{
    // Use 2-D means (qoL, qoR) and relative orientation (R, t)
    // to approximately estimate the 3-D mean (po)
    // 
    qmdVector poR(3);
    Point( mL(0), mL(1), mR(0), mR(1), poL );
    poR.setMultiply( Rotation(), poL ).setAddEq( Translation() );

    // Generate the 2x3 Jacobians (JL, JR) of the projection,
    // linearized about the 3-D mean (po).
    // 
    qmdMatrix JL( 2, 3 );	// dqL = JL dpL
    qmdMatrix JR( 2, 3 );	// dqR = JR dpR
    qmdMatrix JRL( 2, 3 );	// dqR = JRL dpL = JR R dpL
    projectionJacobian( JL, poL, BL() );
    projectionJacobian( JR, poR, BR() );
    JRL.setMultiply( JR, Rotation() );
    // JL.print( cerr << "JL= " ) << endl;
    // JR.print( cerr << "JR= " ) << endl;
    // JRL.print( cerr << "JRL= " ) << endl;

    // Fourth, construct combined jacobian model that maps
    // 3-D parameters (vX, vY, vZ, cXY, cXZ, cYZ ) to the
    // 2-D parameters (vLx, vLy, cLxy, vRx, vRy, cRxy )
    //
    qmdMatrix J( 6, 6 );
    qmdVector params2( 6 );
    combineJacobians( J, JL, JRL );
    // J.print( cerr << "J= " ) << endl;
    params2( 0 ) = CL( 0, 0 );
    params2( 1 ) = CL( 1, 1 );
    params2( 2 ) = CL( 0, 1 );
    params2( 3 ) = CR( 0, 0 );
    params2( 4 ) = CR( 1, 1 );
    params2( 5 ) = CR( 0, 1 );
    // params2.print( cerr << "p2= " ) << endl;
    compute3DCovariance( C3L, J, params2 );
    // C3L.print( cerr << "C3L= " ) << endl;
}

void
cgTriangle::ellipse( qmdVectorRef& mL, qmdMatrixRef& CL,
		     qmdVectorRef& mR, qmdMatrixRef& CR,
		     qmdVectorRef& poL, qmdMatrixRef& C3L,
		     qmdMatrixRef& Ut, qmdVectorRef& s, qmdMatrixRef& Vt )
{
    // Use 2-D means (qoL, qoR) and relative orientation (R, t)
    // to approximately estimate the 3-D mean (po)
    // 
    qmdVector poR(3);
    Point( mL(0), mL(1), mR(0), mR(1), poL );
    poR.setMultiply( Rotation(), poL ).setAddEq( Translation() );

    // Generate the 2x3 Jacobians (JL, JR) of the projection,
    // linearized about the 3-D mean (po).
    // 
    qmdMatrix JL( 2, 3 );	// dqL = JL dpL
    qmdMatrix JR( 2, 3 );	// dqR = JR dpR
    qmdMatrix JRL( 2, 3 );	// dqR = JRL dpL = JR R dpL
    projectionJacobian( JL, poL, BL() );
    projectionJacobian( JR, poR, BR() );
    JRL.setMultiply( JR, Rotation() );
    // JL.print( cerr << "JL= " ) << endl;
    // JR.print( cerr << "JR= " ) << endl;
    // JRL.print( cerr << "JRL= " ) << endl;

    // Fourth, construct combined jacobian model that maps
    // 3-D parameters (vX, vY, vZ, cXY, cXZ, cYZ ) to the
    // 2-D parameters (vLx, vLy, cLxy, vRx, vRy, cRxy )
    //
    qmdMatrix J( 6, 6 );
    qmdVector params2( 6 );
    combineJacobians( J, JL, JRL );
    // J.print( cerr << "J= " ) << endl;
    params2( 0 ) = CL( 0, 0 );
    params2( 1 ) = CL( 1, 1 );
    params2( 2 ) = CL( 0, 1 );
    params2( 3 ) = CR( 0, 0 );
    params2( 4 ) = CR( 1, 1 );
    params2( 5 ) = CR( 0, 1 );
    // params2.print( cerr << "p2= " ) << endl;
    compute3DCovariance( C3L, J, params2, Ut, s, Vt );
    // C3L.print( cerr << "C3L= " ) << endl;
}

void
cgTriangle::projectionJacobian( qmdMatrixRef& J, qmdVectorRef& m, reald b )
{
    double distortion = 1. + m(2)*b;
    double distortion2 = distortion*distortion;
    J( 0, 0 ) = 1. / distortion;
    J( 0, 1 ) = 0.;
    J( 0, 2 ) = -m(0)*b / distortion2;
    J( 1, 0 ) = 0.;
    J( 1, 1 ) = J( 0, 0 );
    J( 1, 2 ) = -m(1)*b / distortion2;
}

static
inline
void
makeJacobianProducts( reald* p, qmdMatrixRef& Js )
{
    // a 3x6 matrix
    *p++ = Js( 0, 0 ) * Js( 0, 0 );
    *p++ = Js( 0, 1 ) * Js( 0, 1 );
    *p++ = Js( 0, 2 ) * Js( 0, 2 );
    *p++ = 2 * Js( 0, 0 ) * Js( 0, 1 );
    *p++ = 2 * Js( 0, 0 ) * Js( 0, 2 );
    *p++ = 2 * Js( 0, 1 ) * Js( 0, 2 );

    *p++ = Js( 1, 0 ) * Js( 1, 0 );
    *p++ = Js( 1, 1 ) * Js( 1, 1 );
    *p++ = Js( 1, 2 ) * Js( 1, 2 );
    *p++ = 2 * Js( 1, 0 ) * Js( 1, 1 );
    *p++ = 2 * Js( 1, 0 ) * Js( 1, 2 );
    *p++ = 2 * Js( 1, 1 ) * Js( 1, 2 );

    *p++ = Js( 0, 0 ) * Js( 1, 0 );
    *p++ = Js( 0, 1 ) * Js( 1, 1 );
    *p++ = Js( 0, 2 ) * Js( 1, 2 );
    *p++ = Js( 0, 0 ) * Js( 1, 1 ) + Js( 0, 1 ) * Js( 1, 0 );
    *p++ = Js( 0, 0 ) * Js( 1, 2 ) + Js( 0, 2 ) * Js( 1, 0 );
    *p++ = Js( 0, 1 ) * Js( 1, 2 ) + Js( 0, 2 ) * Js( 1, 1 );
}

void
cgTriangle::combineJacobians( qmdMatrixRef& J, qmdMatrixRef& JL, qmdMatrixRef& JRL )
{
    makeJacobianProducts( J.getArray( 0, 0 ), JL );
    makeJacobianProducts( J.getArray( 3, 0 ), JRL );
}

void
cgTriangle::compute3DCovariance( qmdMatrixRef& C,
				 qmdMatrixRef& J,
				 qmdVectorRef& p )
{
    const reald tooSmall = .001;
    qmdVector p3( 6 );
    reald& sX = p3( 0 );
    reald& sY = p3( 1 );
    reald& sZ = p3( 2 );
    reald& sXY = p3( 3 );
    reald& sXZ = p3( 4 );
    reald& sYZ = p3( 5 );
    // -- Compute SVD
    qmdSVD svdJ( J );
    cerr << "SVD= " << endl << svdJ << endl;
    // -- Nuke bad directions
    qmdMatrix S( 6, 6 );
    S.setDiagonal( svdJ.getSingularValues() );
    for( int i = 0; i < 6; i++ )
       // S(i,i) = ( S(i,i) > tooSmall ) ?  (1. / S(i,i)) : defaultSV;
       if( S(i,i) > tooSmall || !svinvert )
	  S(i,i) = (1. / S(i,i));
       else
	  {
	      S(i,i) = defaultSV;
	      cerr << "singular direction " << i << endl;
	  }
    qmdVector Utp( 6 );
    qmdVector SUtp( 6 );
    Utp.setTMultiply( svdJ.getU(), p );
    SUtp.setMultiply( S, Utp );
    p3.setMultiply( svdJ.getV(), SUtp );
    // -- Fill in 3-D covariance matrix
    C( 0, 0 ) = sX;
    C( 0, 1 ) = sXY;
    C( 0, 2 ) = sXZ;
    C( 1, 0 ) = sXY;
    C( 1, 1 ) = sY;
    C( 1, 2 ) = sYZ;
    C( 2, 0 ) = sXZ;
    C( 2, 1 ) = sYZ;
    C( 2, 2 ) = sZ;
}

void
cgTriangle::compute3DCovariance( qmdMatrixRef& C,
				 qmdMatrixRef& J,
				 qmdVectorRef& p,
				 qmdMatrixRef& Ut,
				 qmdVectorRef& s,
				 qmdMatrixRef& Vt )
{
    const reald tooSmall = .001;

    qmdVector p3( 6 );
    reald& sX = p3( 0 );
    reald& sY = p3( 1 );
    reald& sZ = p3( 2 );
    reald& sXY = p3( 3 );
    reald& sXZ = p3( 4 );
    reald& sYZ = p3( 5 );

    // -- Compute SVD
    qmdSVD svdJ( J );
    cerr << "SVD= " << endl << svdJ << endl;

    // -- Nuke bad directions
    qmdMatrix S( 6, 6 );
    S.setDiagonal( svdJ.getSingularValues() );
    for( int i = 0; i < 6; i++ )
       S(i,i) = ( S(i,i) > tooSmall ) ?  (1. / S(i,i)) : defaultSV;
    qmdVector Utp( 6 );
    qmdVector SUtp( 6 );
    Utp.setTMultiply( svdJ.getU(), p );
    SUtp.setMultiply( S, Utp );
    p3.setMultiply( svdJ.getV(), SUtp );

    // -- Fill in 3-D covariance matrix
    C( 0, 0 ) = sX;
    C( 0, 1 ) = sXY;
    C( 0, 2 ) = sXZ;
    C( 1, 0 ) = sXY;
    C( 1, 1 ) = sY;
    C( 1, 2 ) = sYZ;
    C( 2, 0 ) = sXZ;
    C( 2, 1 ) = sYZ;
    C( 2, 2 ) = sZ;

    // -- Output SVD
    Ut.setTranspose( svdJ.getU() );
    Vt.setTranspose( svdJ.getV() );
    s.copy( svdJ.getSingularValues() );
}

ostream&
cgTriangle::print( ostream& os )
{
    os << "T= " << endl << T << endl;
    os << "R= " << endl << R << endl;
    os << "bL= " << endl << bL << endl;
    os << "bR= " << endl << bR << endl;
    os << "q= " << endl << q << endl;
    return os;
}

