// cgTriangleStereo.c++

#include "cgTriangleStereo.h"

cgTriangleStereo::cgTriangleStereo()
  : cgTriangle()
{
    tL[0] = 0.;
    tL[1] = 0.;
    tL[2] = 0.;
    qL[0] = 1.;
    qL[1] = 0.;
    qL[2] = 0.;
    qL[3] = 0.;
    tR[0] = 1.;
    tR[1] = 0.;
    tR[2] = 0.;
    qR[0] = 1.;
    qR[1] = 0.;
    qR[2] = 0.;
    qR[3] = 0.;
}

void
cgTriangleStereo::computeTriangle( reald* TL, reald* QL, reald* TR, reald* QR,
				   reald* t, reald* q )
{
    qmdUnitQuaternion qRi;
    qmdMatrix RRi( 3, 3 );
    qmdVector tL_minus_tR( 3 );
    qRi.setInvert( QR );
    qRi.getRotationMatrix( RRi );
    qmdUnitQuaternionRef(q).setMultiply( qRi, QL );
    tL_minus_tR.setSubtract( qmdVectorRef( 3, TL), qmdVectorRef( 3, TR ) );
    qmdVectorRef( 3, t ).setMultiply( RRi, tL_minus_tR );
}

void
cgTriangleStereo::computeTriangle()
{
#if 0
    qmdMatrix Rleft( 3, 3 );
    qmdMatrix Rright( 3, 3 );
    qmdUnitQuaternionRef( qL ).getRotationMatrix( Rleft );
    qmdUnitQuaternionRef( qR ).getRotationMatrix( Rright );
    qmdVectorRef tLeft( 3, tL );
    qmdVectorRef tRight( 3, tR );
    qmdMatrix Rrl( 3, 3 );
    // Rright.setInvertEq();
    // Rrl.setMultiply( Rright, Rleft );
    Rrl.setTMultiply( Rright, Rleft );
    qmdVector Trl( 3 );
    qmdVector Trlw( 3 );
    Trl.setTMultiply( Rright, Trlw.setSubtract( tLeft, tRight ) );
    Translation().copy( Trl );
    Rotation().copy( Rrl );
#endif
    computeTriangle( tL, qL, tR, qR, cgTriangle::T, cgTriangle::q );
    cgTriangle::q.getRotationMatrix( cgTriangle::R );
}

void
cgTriangleStereo::setLeftCoordinateTransform( reald tx, reald ty, reald tz,
					      reald q0, reald q1,
					      reald q2, reald q3 )
{
    static qmdMatrix R( 3, 3 );
    static qmdVector t( 3 );
    t(0) = tx;
    t(1) = ty;
    t(2) = tz;
    qL[0] = q0;
    qL[1] = -q1;
    qL[2] = -q2;
    qL[3] = -q3;
    qmdUnitQuaternionRef( qL ).getRotationMatrix( R );
    qmdVectorRef( 3, tL ).setMultiply( R, t ).setNegateEq();
    computeTriangle();
}

void
cgTriangleStereo::setRightCoordinateTransform( reald tx, reald ty, reald tz,
					       reald q0, reald q1,
					       reald q2, reald q3 )
{
    static qmdMatrix R( 3, 3 );
    static qmdVector t( 3 );
    t(0) = tx;
    t(1) = ty;
    t(2) = tz;
    qR[0] = q0;
    qR[1] = -q1;
    qR[2] = -q2;
    qR[3] = -q3;
    qmdUnitQuaternionRef( qR ).getRotationMatrix( R );
    qmdVectorRef( 3, tR ).setMultiply( R, t ).setNegateEq();
    computeTriangle();
}

