// noIterativeReweightedLeastSquares.c++

#include "noIterativeReweightedLeastSquares.h"

reald noIterativeReweightedLeastSquares::defaultAbsTolerance = .001;

noIterativeReweightedLeastSquares::noIterativeReweightedLeastSquares()
  : erf(0),
    x(),
    dy(),
    J(),
    Ri(),
    Pi(),
    JtRi(),
    tmp(),
    paren(),
    pseudoinverse(),
    dx(),
    absTolerance( defaultAbsTolerance ),
    verbose( 0 ),
    iterationLimit( 1000 )
{}

void
noIterativeReweightedLeastSquares::init( noErrorFunction& a )
{
    erf = &a;
    x.set( a.getNx() );
    dy.set( a.getNdy() );
    J.set( a.getNdy(), a.getNdx() );
    Ri.set( a.getNdy(), a.getNdy() );
    Ri.setIdentity();
    Pi.set( a.getNdx(), a.getNdx() );
    Pi.setIdentity();
    JtRi.set( a.getNdx(), a.getNdy() );
    tmp.set( a.getNdx(), a.getNdx() );
    paren.set( a.getNdx(), a.getNdx() );
    pseudoinverse.set( a.getNdx(), a.getNdy() );
    dx.set( a.getNdx() );
    absTolerance = defaultAbsTolerance;
    iterationLimit = 1000;
    verbose = 0;
}

#if 0
const qmdVectorRef&
noIterativeReweightedLeastSquares::estimate()
{
    // dx = ( Pi + J` Ri J )^-1 J` Ri dy
    noErrorFunction& a = *erf;
    dy = a( x );
    J = a.jacobian( x );
    JtRi.setTMultiply( J, Ri );
    paren.setAdd( Pi, tmp.setMultiply( JtRi, J ) );
    paren.setInvertEq();
    pseudoinverse.setMultiply( paren, JtRi );
    dx.setMultiply( pseudoinverse, dy );
    x = a.add( x, dx );
    return x;
}
#endif

const qmdVectorRef&
noIterativeReweightedLeastSquares::estimate()
{
    // dx = ( Pi + J` Ri J )^-1 J` Ri dy
    noErrorFunction& a = *erf;
    dy = a( x );
    reald loss = sqrt( dy.getInnerProduct( dy ) );
    int iterationCount = 0;
    while( loss > absTolerance && iterationCount < iterationLimit )
       {
	   J = a.jacobian( x );
	   JtRi.setTMultiply( J, Ri );
	   paren.setAdd( Pi, tmp.setMultiply( JtRi, J ) );
	   paren.setInvertEq();
	   pseudoinverse.setMultiply( paren, JtRi );
	   dx.setMultiply( pseudoinverse, dy );
	   x = a.add( x, dx );
	   dy = a( x );
	   loss = sqrt( dy.getInnerProduct( dy ) );
	   iterationCount++;
       }
    return x;
}
