// noLevenbergMarquardt.h++

#ifndef _noLevenbergMarquardt_H_
#define _noLevenbergMarquardt_H_

#include <math.h>
#include "noOptimizer.h"
#include "noErrorFunction.h"
#include "QuickMathDouble.h"

class noLevenbergMarquardt : public noOptimizer
{
    static reald defaultU;
    static reald defaultV;
    static reald defaultTolerance;
    static reald defaultAbsTolerance;
protected:
    noErrorFunction* erf;
    reald u;
    reald v;
    qmdVector dy;
    qmdVector Rivec;
    qmdVector x;
    qmdMatrix J;
    qmdVector dx;
    qmdVector new_x;
    qmdVector new_dy;
    reald new_loss;
    int reduced;
    reald loss;
    qmdVector JtDy;
    qmdVector RiDy;
    qmdMatrix RiJ;
    reald magJtDy;
    reald gradientTolerance;
    reald absoluteTolerance;
    qmdVector solveDx;
    qmdMatrix solveH;
    qmdMatrix solveJtJ;
    qmdMatrix solveuI;
    qmdVector solveJtDy;
    int verbose;
    int iterationLimit;
    int echoExitStats;
    int estimateStatus;
    int solveStatus;
protected:
    reald computeLoss( qmdVectorRef& dy );
    const qmdVectorRef& solve( qmdMatrixRef& J, qmdVectorRef& dy, reald u );
public:
    enum { SolveSuccess, SolveFail };
    enum { None, Success, Overflow, Singular };
    noLevenbergMarquardt();
    void init( noErrorFunction& a );
    const qmdVectorRef& estimate();
    inline void setGradientTolerance( reald );
    inline void setAbsoluteTolerance( reald );
    inline void setRelativeTolerance( const qmdVectorRef& );
    inline void setState( const qmdVectorRef& );
    inline void setVerbose( int v = 1 ) { verbose = v; }
    inline void setIterationLimit( int l ) { iterationLimit = l; }
    inline void setEchoExitStats( int e = 1 ) { echoExitStats = e; }
    inline int getEstimateStatus() { return estimateStatus; }
    inline int estimateOK() { return( (Success==estimateStatus) ? 1 : 0 ); }
};

inline
void
noLevenbergMarquardt::setGradientTolerance( reald tol )
{
    gradientTolerance = tol;
}

inline
void
noLevenbergMarquardt::setAbsoluteTolerance( reald tol )
{
    absoluteTolerance = tol;
}

inline
void
noLevenbergMarquardt::setRelativeTolerance( const qmdVectorRef& ri )
{
    Rivec.copy( ri );
}

inline
void
noLevenbergMarquardt::setState( const qmdVectorRef& x_set )
{
    x = x_set;
}

#endif /* _noLevenbergMarquardt_H_ */
