// qmGaussianRef.h++

#ifndef _qmGaussianRef_H_
#define _qmGaussianRef_H_

#include "QuickMath.h"

class qmGaussianRef
{
    static ostream& alert( ostream&, char* msg );
    inline real _getMahalanobisSquare( const qmVectorRef& ) const;
    inline real _getLogLikelihood( const qmVectorRef& ) const;
protected:
    inline int _testSingular() const;
    inline void testSetSingular();

    unsigned int state;
    qmVectorRef mean;
    qmMatrixRef covariance;

public:

    static real MinDet;  // distribution considered singular if det < MinDet
    static int getStorageSize( int n ) { return( sizeof(real)*n*(n+1) ); }
    enum { NONSINGULAR=0, SINGULAR=1, NOT_INITED=2, NAN=3 }; // NAN not impl

    qmGaussianRef()
      : state( NOT_INITED ), mean( ), covariance( ) {}
    qmGaussianRef( const int dim, real* a )
      : state( NOT_INITED ), mean( dim, a ), covariance( dim, dim, a+dim ) {}
    qmGaussianRef( const qmGaussianRef& g )
      : state( g.state ), mean( g.mean ), covariance( g.covariance ) {}
    qmGaussianRef( const qmVectorRef& v, const qmMatrixRef& m )
      : state( NOT_INITED ), mean( v ), covariance( m ) { testSetSingular(); }

    void set( const int dim, real* a );
    void set( const qmGaussianRef& );
    void copy( const qmGaussianRef& );
			  
    inline int getDim() const { return mean.getLength(); }
    const qmVectorRef& getMean() const { return mean; } 
    void setMean( const qmVectorRef & v ) { mean.copy( v ); }
    const qmMatrixRef& getCovariance() const { return covariance; }
    void setCovariance( const qmMatrixRef& m );
    inline void set( const qmVectorRef&, const qmMatrixRef& );
    int getState() const { return state; }
    int isSingular() const { return( SINGULAR == state ); }

    real getMahalanobisSquare( const qmVectorRef& ) const;
    real getMahalanobis( const qmVectorRef& ) const;
    real getLikelihood( const qmVectorRef& ) const;
    real getLogLikelihood( const qmVectorRef& ) const;

    ostream& write( ostream& );
    istream& read( istream& );
    ostream& printPretty( ostream&, char* = "\n" );
    // istream& scan( istream& );

    inline void setSingular();
};

inline
void
qmGaussianRef::set( const qmVectorRef& v, const qmMatrixRef& m )
{
    setMean( v );
    setCovariance( m );
}

inline
int
qmGaussianRef::_testSingular() const
{
    if( covariance.getDeterminant() < MinDet )
       return 1;
    return 0;
}

inline
void
qmGaussianRef::testSetSingular()
{
    if( _testSingular() )
       state = SINGULAR;
    else
       state = NONSINGULAR;
}

inline
void
qmGaussianRef::setSingular()
{
  state = SINGULAR;
}

#endif /* _qmGaussianRef_H_ */
