// qmSampleAccRef.c++

#include "qmSampleAccRef.h"

ostream&
qmSampleAccRef::alert( ostream& os, char* msg )
{
    return( os << "qmSampleAccRef: ALERT: " << msg << endl );
}

void
qmSampleAccRef::set( const int dim, real* memchunk )
{
    sampleCount = 0;
    sampleSum.set( dim, memchunk );
    sampleSumSquares.set( dim, dim, memchunk+dim );
}

void
qmSampleAccRef::set( const qmSampleAccRef& s )
{
    sampleCount = s.sampleCount;
    sampleSum.set( s.sampleSum );
    sampleSumSquares.set( s.sampleSumSquares );
}

void
qmSampleAccRef::setZero()
{
    sampleCount=0;
    sampleSum.setZero();
    sampleSumSquares.setZero();
}

int
qmSampleAccRef::addSample( const qmVectorRef& sample )
{
    sampleSum.setAddEq( sample );
    sampleSumSquares.setAddScaledOuterProductEq( sample, sample );
    return( ++sampleCount );
}

void
qmSampleAccRef::getGaussian( qmGaussianRef& g ) const
{
    if( sampleCount <= 0 )
       {
	   alert( cerr, "getGaussian: no Gaussian to get (aborted)" );
	   return;
       }
    qmVector auxv;
    qmMatrix auxm;
    auxv.set( sampleSum );
    auxm.set( sampleSumSquares );
    auxv /= real( sampleCount );
    auxm /= real( sampleCount );
    auxm.setAddScaledOuterProductEq( auxv, auxv, (-1) );
    g.set( auxv, auxm );
}

int
qmSampleAccRef::getGaussianInv( qmGaussianInvRef& g ) const
{
    if( sampleCount <= 0 )
       {
	   alert( cerr, "getGaussianInv: no Gaussian to get (aborted)" );
	   return 1;
       }
    qmVector auxv;
    qmMatrix auxm;
    auxv.set( sampleSum );
    auxm.set( sampleSumSquares );
    auxv /= real( sampleCount );
    auxm /= real( sampleCount );
    auxm.setAddScaledOuterProductEq( auxv, auxv, (-1) );
    if( auxm.setInvertEqCheck() )
       {
	   auxm.setDiagonal( QmMaxReal );
	   g.set( auxv, auxm );
	   return 1;
       }
    g.set( auxv, auxm );
    return 0;
}

qmSampleAccRef& 
qmSampleAccRef::setMerge( const qmSampleAccRef& g )
{
    real totalSample = sampleCount + g.getSampleCount();
    real w = g.getSampleCount() / totalSample;
    sampleSum.setScaleEq( 1.-w );
    sampleSum.setScaleAddEq( w, g.sampleSum );
    sampleSumSquares.setScaleEq( 1-w );
    sampleSumSquares.setScaleAddEq( w, g.sampleSumSquares );
    return *this;
}

real
qmSampleAccRef::getMeanElement() const
{
    assert( sampleCount > 0 );
    int totalSamples = sampleCount * sampleSum.getLength();
    real totalSum = sampleSum.getSum();
    return totalSum / totalSamples;
}

real
qmSampleAccRef::getRMSElement() const
{
    assert( sampleCount > 0 );
    int totalSamples = sampleCount * sampleSum.getLength();
    real totalSumSq = sampleSumSquares.getTrace();
    return sqrt( totalSumSq / totalSamples );
}

real
qmSampleAccRef::getSTDElement() const
{
    assert( sampleCount > 0 );
    int totalSamples = sampleCount * sampleSum.getLength();
    real meanSum = sampleSum.getSum() / totalSamples;
    real meanSumSq = sampleSumSquares.getTrace() / totalSamples;
    return sqrt( meanSumSq - meanSum*meanSum );
}
