// qmSampleAccRef.h++

#ifndef _qmSampleAccRef_H_
#define _qmSampleAccRef_H_

#include "QuickMath.h"
#include "qmGaussianRef.h"
#include "qmGaussianInvRef.h"

class qmSampleAccRef 
{
    static ostream& alert( ostream&, char* msg );
protected:

    unsigned int sampleCount;
    qmVectorRef sampleSum;
    qmMatrixRef sampleSumSquares;

public:
    static int getStorageSize( const int n ) { return(sizeof(real)*n*(n+1) ); }
    qmSampleAccRef()
      : sampleCount(0), sampleSum(), sampleSumSquares() {}
    qmSampleAccRef( const int dim, real* memchunk )
      : sampleCount( 0 ),
	sampleSum( dim, memchunk ),
	sampleSumSquares( dim, dim, memchunk+dim) { setZero(); }
    qmSampleAccRef( const qmSampleAccRef& s )
      : sampleCount( s.sampleCount ),
	sampleSum( s.sampleSum ),
	sampleSumSquares( s.sampleSumSquares ) {}
    void set( const int dim, real* memchunk );
    void set( const qmSampleAccRef& );
    void setZero();
    int addSample( const qmVectorRef& );
    void getGaussian( qmGaussianRef& g ) const;
    int getGaussianInv( qmGaussianInvRef& g ) const;
    qmSampleAccRef& setMerge( const qmSampleAccRef& );
    int getDim() const { return sampleSum.getLength(); }
    int getSampleCount() const { return sampleCount; }
    const qmVectorRef& getSampleSum() { return sampleSum; }
    const qmMatrixRef& getSampleSumSquares() { return sampleSumSquares; }
    real getMeanElement() const;
    real getRMSElement() const;
    real getSTDElement() const;
    // ostream& print( ostream&, char* = "\n" );
};

#endif /* _qmSampleAccRef_H_ */
