// qmdGaussianRef.h++

#ifndef _qmdGaussianRef_H_
#define _qmdGaussianRef_H_

#include "QuickMathDouble.h"

class qmdGaussianRef
{
    static ostream& alert( ostream&, char* msg );
    inline reald _getMahalanobisSquare( const qmdVectorRef& ) const;
    inline reald _getLogLikelihood( const qmdVectorRef& ) const;
protected:
    inline int _testSingular() const;
    inline void testSetSingular();

    unsigned int state;
    qmdVectorRef mean;
    qmdMatrixRef covariance;

public:

    static reald MinDet;  // distribution considered singular if det < MinDet
    static int getStorageSize( int n ) { return( sizeof(reald)*n*(n+1) ); }
    enum { NONSINGULAR=0, SINGULAR=1, NOT_INITED=2, NAN=3 }; // NAN not impl

    qmdGaussianRef()
      : state( NOT_INITED ), mean( ), covariance( ) {}
    qmdGaussianRef( const int dim, reald* a )
      : state( NOT_INITED ), mean( dim, a ), covariance( dim, dim, a+dim ) {}
    qmdGaussianRef( const qmdGaussianRef& g )
      : state( g.state ), mean( g.mean ), covariance( g.covariance ) {}
    qmdGaussianRef( const qmdVectorRef& v, const qmdMatrixRef& m )
      : state( NOT_INITED ), mean( v ), covariance( m ) { testSetSingular(); }

    void set( const int dim, reald* a );
    void set( const qmdGaussianRef& );
    void copy( const qmdGaussianRef& );
			  
    inline int getDim() const { return mean.getLength(); }
    const qmdVectorRef& getMean() const { return mean; } 
    void setMean( const qmdVectorRef & v ) { mean.copy( v ); }
    const qmdMatrixRef& getCovariance() const { return covariance; }
    void setCovariance( const qmdMatrixRef& m );
    inline void set( const qmdVectorRef&, const qmdMatrixRef& );
    int getState() const { return state; }
    int isSingular() const { return( SINGULAR == state ); }

    reald getMahalanobisSquare( const qmdVectorRef& ) const;
    reald getMahalanobis( const qmdVectorRef& ) const;
    reald getLikelihood( const qmdVectorRef& ) const;
    reald getLogLikelihood( const qmdVectorRef& ) const;

    ostream& write( ostream& );
    istream& read( istream& );
    ostream& printPretty( ostream&, char* = "\n" );
    // istream& scan( istream& );

    inline void setSingular();
};

inline
void
qmdGaussianRef::set( const qmdVectorRef& v, const qmdMatrixRef& m )
{
    setMean( v );
    setCovariance( m );
}

inline
int
qmdGaussianRef::_testSingular() const
{
    if( covariance.getDeterminant() < MinDet )
       return 1;
    return 0;
}

inline
void
qmdGaussianRef::testSetSingular()
{
    if( _testSingular() )
       state = SINGULAR;
    else
       state = NONSINGULAR;
}

inline
void
qmdGaussianRef::setSingular()
{
  state = SINGULAR;
}

#endif /* _qmdGaussianRef_H_ */
