// qmdSampleAccRef.c++

#include "qmdSampleAccRef.h"

ostream&
qmdSampleAccRef::alert( ostream& os, char* msg )
{
    return( os << "qmdSampleAccRef: ALERT: " << msg << endl );
}

void
qmdSampleAccRef::set( const int dim, reald* memchunk )
{
    sampleCount = 0;
    sampleSum.set( dim, memchunk );
    sampleSumSquares.set( dim, dim, memchunk+dim );
}

void
qmdSampleAccRef::set( const qmdSampleAccRef& s )
{
    sampleCount = s.sampleCount;
    sampleSum.set( s.sampleSum );
    sampleSumSquares.set( s.sampleSumSquares );
}

void
qmdSampleAccRef::setZero()
{
    sampleCount=0;
    sampleSum.setZero();
    sampleSumSquares.setZero();
}

int
qmdSampleAccRef::addSample( const qmdVectorRef& sample )
{
    sampleSum.setAddEq( sample );
    sampleSumSquares.setAddScaledOuterProductEq( sample, sample );
    return( ++sampleCount );
}

void
qmdSampleAccRef::getGaussian( qmdGaussianRef& g ) const
{
    if( sampleCount <= 0 )
       {
	   alert( cerr, "getGaussian: no Gaussian to get (aborted)" );
	   return;
       }
    qmdVector auxv;
    qmdMatrix auxm;
    auxv.set( sampleSum );
    auxm.set( sampleSumSquares );
    auxv /= reald( sampleCount );
    auxm /= reald( sampleCount );
    auxm.setAddScaledOuterProductEq( auxv, auxv, (-1) );
    g.set( auxv, auxm );
}

int
qmdSampleAccRef::getGaussianInv( qmdGaussianInvRef& g ) const
{
    if( sampleCount <= 0 )
       {
	   alert( cerr, "getGaussianInv: no Gaussian to get (aborted)" );
	   return 1;
       }
    qmdVector auxv;
    qmdMatrix auxm;
    auxv.set( sampleSum );
    auxm.set( sampleSumSquares );
    auxv /= reald( sampleCount );
    auxm /= reald( sampleCount );
    auxm.setAddScaledOuterProductEq( auxv, auxv, (-1) );
    if( auxm.setInvertEqCheck() )
       {
	   auxm.setDiagonal( qmdMaxReal );
	   g.set( auxv, auxm );
	   return 1;
       }
    g.set( auxv, auxm );
    return 0;
}

qmdSampleAccRef& 
qmdSampleAccRef::setMerge( const qmdSampleAccRef& g )
{
    reald totalSample = sampleCount + g.getSampleCount();
    reald w = g.getSampleCount() / totalSample;
    sampleSum.setScaleEq( 1.-w );
    sampleSum.setScaleAddEq( w, g.sampleSum );
    sampleSumSquares.setScaleEq( 1-w );
    sampleSumSquares.setScaleAddEq( w, g.sampleSumSquares );
    return *this;
}

reald
qmdSampleAccRef::getMeanElement() const
{
    assert( sampleCount > 0 );
    int totalSamples = sampleCount * sampleSum.getLength();
    reald totalSum = sampleSum.getSum();
    return totalSum / totalSamples;
}

reald
qmdSampleAccRef::getRMSElement() const
{
    assert( sampleCount > 0 );
    int totalSamples = sampleCount * sampleSum.getLength();
    reald totalSumSq = sampleSumSquares.getTrace();
    return sqrt( totalSumSq / totalSamples );
}

reald
qmdSampleAccRef::getSTDElement() const
{
    assert( sampleCount > 0 );
    int totalSamples = sampleCount * sampleSum.getLength();
    reald meanSum = sampleSum.getSum() / totalSamples;
    reald meanSumSq = sampleSumSquares.getTrace() / totalSamples;
    return sqrt( meanSumSq - meanSum*meanSum );
}
