// qmdSampleAccRef.h++

#ifndef _qmdSampleAccRef_H_
#define _qmdSampleAccRef_H_

#include "QuickMathDouble.h"
#include "qmdGaussianRef.h"
#include "qmdGaussianInvRef.h"

class qmdSampleAccRef 
{
    static ostream& alert( ostream&, char* msg );
protected:

    unsigned int sampleCount;
    qmdVectorRef sampleSum;
    qmdMatrixRef sampleSumSquares;

public:
    static int getStorageSize( const int n ) { return(sizeof(reald)*n*(n+1) ); }
    qmdSampleAccRef()
      : sampleCount(0), sampleSum(), sampleSumSquares() {}
    qmdSampleAccRef( const int dim, reald* memchunk )
      : sampleCount( 0 ),
	sampleSum( dim, memchunk ),
	sampleSumSquares( dim, dim, memchunk+dim) { setZero(); }
    qmdSampleAccRef( const qmdSampleAccRef& s )
      : sampleCount( s.sampleCount ),
	sampleSum( s.sampleSum ),
	sampleSumSquares( s.sampleSumSquares ) {}
    void set( const int dim, reald* memchunk );
    void set( const qmdSampleAccRef& );
    void setZero();
    int addSample( const qmdVectorRef& );
    void getGaussian( qmdGaussianRef& g ) const;
    int getGaussianInv( qmdGaussianInvRef& g ) const;
    qmdSampleAccRef& setMerge( const qmdSampleAccRef& );
    int getDim() const { return sampleSum.getLength(); }
    int getSampleCount() const { return sampleCount; }
    const qmdVectorRef& getSampleSum() { return sampleSum; }
    const qmdMatrixRef& getSampleSumSquares() { return sampleSumSquares; }
    reald getMeanElement() const;
    reald getRMSElement() const;
    reald getSTDElement() const;
    // ostream& print( ostream&, char* = "\n" );
};

#endif /* _qmdSampleAccRef_H_ */
