// ComputeGeometry.c++

#include "ComputeGeometry.h"
reald ComputeGeometry::ballSize[3] = { .1, .05, .05 };
reald ComputeGeometry::lmAbsTol = .1;
int ComputeGeometry::lmIterLimit = 100;
reald ComputeGeometry::yakkaT = 6;	// inches
reald ComputeGeometry::yakkaW = 10;	// degrees
reald ComputeGeometry::yakkaL = 0;	// inches
reald ComputeGeometry::yakkaDT = 2;	// inches
reald ComputeGeometry::yakkaDW = 2;	// degrees
reald ComputeGeometry::yakkaDL = 0;	// inches
reald ComputeGeometry::yakkaDLs = 1.0;	// ratio
reald ComputeGeometry::yakkaUV = .01;	// units
reald ComputeGeometry::yakkaS = .0001;	// units
int ComputeGeometry::wakkaFooMode = 1;
reald ComputeGeometry::wakkaFoo[3] = { .03, .03, .06 };
reald ComputeGeometry::wakkaHoo[3] = { .03, .03, .06 };
int ComputeGeometry::kalmanMode = 0;
reald ComputeGeometry::mmtNoiseVar = 1;
reald ComputeGeometry::dynNoiseVar = 1;
int ComputeGeometry::vtiIter = 3;
float ComputeGeometry::calibTol = .001;
float ComputeGeometry::calibAbstol = .001;
ComputeGeometry::ComputeGeometry()
  : personMode( 1 ),
    Rwc( 3, 3 ),
    Rwc_inv( 3, 3 ),
    Twc( 3 ),
    scaleToWorld( reald(1.) ),
    pct( 3 )
{
    for( int i = 0; i < 3; i++ )
       {
	 mode[i] = NOMODE;
	 prevMode[i] = NOMODE;
	 yakkaX[i] = 0;
	 blobX[i] = 0;
       }
    Rwc.setIdentity();
    Rwc_inv.setIdentity();
    Twc.setZero();
}
void
ComputeGeometry::initNoMode(int i)
{
    mode[i] = NOMODE;
}
void
ComputeGeometry::initTriangle(int i)
{
    mode[i] = TRIANGLE;
}
void
ComputeGeometry::initBlobs(int i)
{
    mode[i] = BLOB;
    if( !blobX[i] )
      blobX[i] = new qmdVector( 10 );
}
void
ComputeGeometry::initYakka(int i)
{
    mode[i] = YAKKA;
    if( !yakkaX[i] )
      yakkaX[i] = new qmdVector( 10 );
}
void
ComputeGeometry::computeTriangle( int i,
				  qmdVectorRef* y, 
				  int* val,
				  qmdVectorRef* t, 
				  qmdUnitQuaternionRef* q,
				  qmdVectorRef* l )
{
  if( val[i] )
    triangle.point( y[i].subVector( 0, 2 ),
		    y[i].subVector( 5, 2 ),
		    t[i] );
  else
    t[i].setZero();
  q[i].setIdentity();
  l[i].setConst( ballSize[i] );
}
void
ComputeGeometry::initCalibration( reald* t, reald* q, reald b )
{
    float* p = calibData;
    for( int i = 0; i < 3; i++ )
       *p++ = t[i];
    for( i = 0; i < 4; i++ )
       *p++ = q[i];
    *p = b;
    triangle.setRightCoordinateTransform
       ( t[0], t[1], t[2], q[0], q[1], q[2], q[3] );
    triangle.setRightBeta( b );
    triangle.setLeftBeta( b );
}
void
ComputeGeometry::computeBlob( int i,
			      qmdVectorRef* y, 
			      int* val,
			      qmdVectorRef* t, 
			      qmdUnitQuaternionRef* q,
			      qmdVectorRef* l )
{
    static noErfBlob erfBlob;
    static noLevenbergMarquardt lmBlob;
    static int inited = 0;
    const reald root5 = sqrt( 5. );
    if( !inited )
      {
	lmBlob.init( erfBlob );
	inited = 1;
      }
    if( !val[i] )
      {
	t[i].setZero();
	q[i].setIdentity();
	l[i].setZero();
	return;
      }
    erfBlob.init( y[i],
		  triangle.BL(), 
		  triangle.Translation(),
		  triangle.Q() );
    erfBlob.computeLinearInit( *blobX[i] );
    lmBlob.setState( *blobX[i] );
    lmBlob.setAbsoluteTolerance( lmAbsTol );
    lmBlob.setIterationLimit( lmIterLimit );

    *blobX[i] = lmBlob.estimate();

    t[i].copy( (*blobX[i]).subVector( noErfBlob::TX, 3 ) );
    q[i].copy( (*blobX[i]).subVector( noErfBlob::Q0, 4 ) );
    l[i].copy( (*blobX[i]).subVector( noErfBlob::LX, 3 ) );
}
void
ComputeGeometry::yakkaNormalizeBlobs( int which, qmdVector** b )
{
    static qmdVectorRef dim;
    static int maxdim;
    static reald maxval;
    static int mindim;
    static reald minval;
    dim.set( b[HEAD] -> subVector( noErfBlob::LX, 3 ) );
    maxval = dim.getMax( &maxdim );
    minval = dim.getMin( &mindim );
    if (which == HEAD)
      for( int i = 0; i < 3; i++ )
	if( i != maxdim && i != mindim )
	  dim.setConst( dim(i) );

    qmdVectorRef dimL = b[LEFTHAND] -> subVector( noErfBlob::LX, 3 );
    qmdVectorRef dimR = b[RIGHTHAND] -> subVector( noErfBlob::LX, 3 );
    int maxL;
    int maxR;
    reald maxleft = dimL.getMax( &maxL );
    reald maxrite = dimR.getMax( &maxR );
    maxval = .5 * ( maxleft + maxrite );
    reald minLeft = 0.;
    reald minRite = 0.;
    for( int i = 0; i < 3; i++ )
       {
	   if( i != maxL )
	      minLeft += dimL(i);
	   if( i != maxR )
	      minRite += dimR(i);
       }
    minval = .25 * ( minLeft + minRite );
    for( i = 0; i < 3; i++ )
       {
	 if (which == LEFTHAND)
	   dimL(i) = ( i == maxL ) ? maxval : minval;
	 if (which == RIGHTHAND)
	   dimR(i) = ( i == maxR ) ? maxval : minval;
       }
}
void
ComputeGeometry::computeYakkaYakka( int which,
				    qmdVectorRef* y, 
				    int* val,
				    qmdVectorRef* t, 
				    qmdUnitQuaternionRef* q,
				    qmdVectorRef* l )
{
    static noErfBlob erfBlob;
    static noYakkaYakka yakkas[ 3 ];
    static int yakkaInit[ 3 ];
    const reald root5 = sqrt( 5. );
    if( BLOB == prevMode[which] )
       yakkaNormalizeBlobs( which , blobX );

    if( !val[which] )
      {
	t[which].setZero();
	q[which].setIdentity();
	l[which].setZero();
	return;
      }
    qmdVectorRef& xblob = (*yakkaX[which]);
    erfBlob.init( y[ which ], 
		  triangle.BL(), 
		  triangle.Translation(),
		  triangle.Q() );
    if( ! yakkaInit[ which ] || NOMODE == prevMode[which] )
      {
	static qmdMatrix P( 9, 9 );
	yakkas[ which ].init( erfBlob );
	erfBlob.computeLinearInit( *yakkaX[which] );
	if( wakkaFooMode )
	  {
	    if( HEAD == which )
	      xblob.subVector( noErfBlob::LX, 3 ).setData
		( wakkaHoo );
	    else
	      xblob.subVector( noErfBlob::LX, 3 ).setData
		( wakkaFoo );
	    
	  }
	yakkas[ which ].setState( xblob );
	P.setZero();
	reald s2 = scaleToWorld*scaleToWorld;
	reald dtor = M_PI / 180.;
	for( int i = 0; i < 3; i++ )
	  P( i, i ) = yakkaT*yakkaT / s2;
	for( ; i < 6; i++ )
	  P( i, i ) = yakkaW*yakkaW*dtor*dtor;
	for( ; i < 9; i++ )
	  P( i, i ) = yakkaL*yakkaL / s2;
	yakkas[ which ].setCovariance( P );
	yakkas[ which ].Q.setZero();
	for( i = 0; i < 3; i++ )
	  yakkas[ which ].Q( i, i ) = yakkaDT*yakkaDT/s2;
	for( ; i < 6; i++ )
	  yakkas[ which ].Q( i, i ) = yakkaDW*yakkaDW/s2;

	yakkas[ which ].Q( i, i ) =          yakkaDL*yakkaDL/s2;  i++;
	yakkas[ which ].Q( i, i ) = yakkaDLs*yakkaDL*yakkaDL/s2;  i++;
	yakkas[ which ].Q( i, i ) = yakkaDLs*yakkaDL*yakkaDL/s2;  i++;

	yakkas[ which ].R.setZero();
	for( i = 0; i < 2; i++ )
	  yakkas[ which ].R( i, i ) = yakkaUV*yakkaUV;
	for( ; i < 5; i++ )
	  yakkas[ which ].R( i, i ) = yakkaS*yakkaS;
	for( i = 0; i < 7; i++ )
	  yakkas[ which ].R( i, i ) = yakkaUV*yakkaUV;
	for( ; i < 10; i++ )
	  yakkas[ which ].R( i, i ) = yakkaS*yakkaS;
	yakkaInit[ which ] = 1;
      }
    else if( BLOB == prevMode[which] )
      {
	yakkas[ which ].setState( *blobX[which] );
      }
    xblob = yakkas[ which ].estimate();
    t[ which ].copy( xblob.subVector( noErfBlob::TX, 3 ) );
    q[ which ].copy( xblob.subVector( noErfBlob::Q0, 4 ) );
    l[ which ].copy( xblob.subVector( noErfBlob::LX, 3 ) );
}

void
ComputeGeometry::computeNotrealdly( int i,
				   qmdVectorRef* , 
				   int* ,
				   qmdVectorRef* t, 
				   qmdUnitQuaternionRef* q,
				   qmdVectorRef* l )
{
    const reald dfltPos[3] = { 0, .6, 0 };
    const reald dfltSiz[3] = { .1, .05, .05 };

    t[i].setData( dfltPos );
    q[i].setIdentity();
    l[i].setData( dfltSiz );
}

void
ComputeGeometry::setMode( int i, int m )
{
    switch( m )
       {
	case TRIANGLE:
	   initTriangle(i);
	   break;
	case BLOB:
	   initBlobs(i);
	   break;
	case YAKKA:
	   initYakka(i);
	   break;
	default:
	   initNoMode(i);
       }
}
inline
void
ComputeGeometry::cameraToWorld( const qmdVectorRef& pcv, qmdVectorRef& pw )
{
    pct.setSubtract( pcv, Twc );
    pw.setMultiply( Rwc, pct );
    pw.setScaleEq( scaleToWorld );
}
inline
void
ComputeGeometry::worldToCamera( const qmdVectorRef& pw, qmdVectorRef& pc )
{
    pct.setScale( 1./scaleToWorld, pw );
    pc.setMultiply( Rwc_inv, pct );
    pc += Twc;
}
void
ComputeGeometry::convertToWorld( int* val,
				 qmdVectorRef* t,
				 qmdUnitQuaternionRef* q,
				 qmdVectorRef* l,
				 qmdVectorRef* tw,
				 qmdUnitQuaternionRef* qw,
				 qmdVectorRef* lw )
{
    static qmdMatrix Rc( 3, 3 );
    static qmdMatrix Rw( 3, 3 );
    for( int i = 0; i < 3; i++ )
       {
	   if( !val[i] )
	      continue;
	   cameraToWorld( t[i], tw[i] );
	   lw[i].setScale( scaleToWorld, l[i] );
	   q[i].getRotationMatrix( Rc );
	   Rw.setMultiply( Rwc, Rc );
	   qw[i].set( Rw );
       }
}
void
ComputeGeometry::setWorldTransform( qmdVectorRef& t, qmdMatrixRef& r, reald s )
{
    Twc.copy( t );
    Rwc.copy( r );
    Rwc_inv.setInvert( r );
    scaleToWorld = s; 
}
ostream&
ComputeGeometry::alert( char* fn, char* msg )
{
    return( cerr << "ComputeGeometry: ALERT: " << fn << "(): " << msg << endl );
}

void
ComputeGeometry::error( char* fn, char* msg )
{
    cerr << "ComputeGeometry: ERROR: " << fn << "(): " << msg << endl;
    assert(0);
}
void
ComputeGeometry::calibrate( int ncorr, float* data )
{
    noErfRelativeOrientation erf;
    qmdVector uL( ncorr );
    qmdVector vL( ncorr );
    qmdVector uR( ncorr );
    qmdVector vR( ncorr );
    for( int i = 0; i < ncorr; i++ )
       {
	   uL(i) = data[0];
	   vL(i) = data[1];
	   uR(i) = data[2];
	   vR(i) = data[3];
	   data += 4;
       }
    qmdVector actualState( 7 + ncorr );
    qmdVectorRef trel( 3, actualState.getArray() );
    qmdUnitQuaternionRef qrel
       ( actualState.getArray( noErfRelativeOrientation::Q0 ) );
    qmdVectorRef alpha
       ( ncorr, actualState.getArray( noErfRelativeOrientation::ALPHA ) );
    qmdVector state;
    state.set( 7 + ncorr );
    xinit[0] = 0;
    xinit[1] = 0;
    xinit[2] = 0;
    xinit[3] = 1;
    xinit[4] = 0;
    xinit[5] = 0;
    xinit[6] = 0;
    xinit[7] = 0;
    float fixAlpha = 2;
    state( noErfRelativeOrientation::TX ) = xinit[0];
    state( noErfRelativeOrientation::TY ) = xinit[1];
    state( noErfRelativeOrientation::TZB ) = xinit[2];
    state( noErfRelativeOrientation::Q0 ) = xinit[3];
    state( noErfRelativeOrientation::QX ) = xinit[4];
    state( noErfRelativeOrientation::QY ) = xinit[5];
    state( noErfRelativeOrientation::QZ ) = xinit[6];
    for( i = 0; i < ncorr; i++ )
      state( noErfRelativeOrientation::ALPHA + i ) = fixAlpha;
    noLevenbergMarquardt estimator;
    erf.init( uL, vL, uR, vR );
    reald beta = 1;
    erf.setBeta( beta );
    estimator.init( erf );
    estimator.setGradientTolerance( calibTol );
    estimator.setAbsoluteTolerance( calibAbstol );
    estimator.setIterationLimit( 30 );
    estimator.setState( state );
    state = estimator.estimate();
    cerr << state << endl;
    initCalibration( state.getArray(), state.getArray( 3 ), beta );
}

void
ComputeGeometry::calibrate( int ncorr, float* data , char * filename)
{
    noErfRelativeOrientation erf;
    qmdVector uL( ncorr );
    qmdVector vL( ncorr );
    qmdVector uR( ncorr );
    qmdVector vR( ncorr );
    for( int i = 0; i < ncorr; i++ )
       {
	   uL(i) = data[0];
	   vL(i) = data[1];
	   uR(i) = data[2];
	   vR(i) = data[3];
	   data += 4;
       }
    qmdVector actualState( 7 + ncorr );
    qmdVectorRef trel( 3, actualState.getArray() );
    qmdUnitQuaternionRef qrel
       ( actualState.getArray( noErfRelativeOrientation::Q0 ) );
    qmdVectorRef alpha
       ( ncorr, actualState.getArray( noErfRelativeOrientation::ALPHA ) );
    qmdVector state;
    state.set( 7 + ncorr );
    xinit[0] = 0;
    xinit[1] = 0;
    xinit[2] = 0;
    xinit[3] = 1;
    xinit[4] = 0;
    xinit[5] = 0;
    xinit[6] = 0;
    xinit[7] = 0;
    float fixAlpha = 2;
    state( noErfRelativeOrientation::TX ) = xinit[0];
    state( noErfRelativeOrientation::TY ) = xinit[1];
    state( noErfRelativeOrientation::TZB ) = xinit[2];
    state( noErfRelativeOrientation::Q0 ) = xinit[3];
    state( noErfRelativeOrientation::QX ) = xinit[4];
    state( noErfRelativeOrientation::QY ) = xinit[5];
    state( noErfRelativeOrientation::QZ ) = xinit[6];
    for( i = 0; i < ncorr; i++ )
      state( noErfRelativeOrientation::ALPHA + i ) = fixAlpha;
    noLevenbergMarquardt estimator;
    erf.init( uL, vL, uR, vR );
    reald beta = 1;
    erf.setBeta( beta );
    estimator.init( erf );
    estimator.setGradientTolerance( calibTol );
    estimator.setAbsoluteTolerance( calibAbstol );
    estimator.setIterationLimit( 30 );
    estimator.setState( state );
    state = estimator.estimate();
    cerr << state << endl;
    fstream out(filename,ios::out);
    out << state << endl;
    out.close();
    initCalibration( state.getArray(), state.getArray( 3 ), beta );
}

void
ComputeGeometry::compute2DQuickPoints( qmdVectorRef& mean, qmdMatrixRef& cov,
				       qmdVectorRef& qpNeg, qmdVectorRef& qpPos )
{
    static qmdEigenStructure eig;
    static qmdMatrix evec( 2, 2 );
    static qmdVector eval( 2 );
    if( qmdEigenStructure::eigenSymmetric2( cov, eval, evec ) )
       alert( "compute2DQuickPoints", "eigen2" );
    int imax;
    int imin;
    reald lmax = sqrt( eval.getMax( &imax ) );
    reald lmin = sqrt( eval.getMin( &imin ) );
    reald e = sqrt(eval(1-imax)) / sqrt(eval(imax));
    reald alpha = ( e < .5 ) ? 1 : ( 1-2*(e-.5) );
    alpha *= 2*lmax;
    qpNeg.copy( mean ).setScaleAddEq( -alpha, evec.getRow(imax) );
    qpPos.copy( mean ).setScaleAddEq( alpha, evec.getRow(imax) );
}

void
ComputeGeometry::computeQuickPoint( reald* yL, reald* yR, 
				    reald* qpNeg, reald* qpPos )
{
    static qmdVector qLNeg( 2 );
    static qmdVector qLPos( 2 );
    static qmdVector qRNeg( 2 );
    static qmdVector qRPos( 2 );
    static qmdMatrix cov( 2, 2 );
    cov(0,0) = yL[2];
    cov(0,1) = cov(1,0) = yL[4];
    cov(1,1) = yL[3];
    compute2DQuickPoints( qmdVectorRef(2,yL), cov, qLNeg, qLPos );
    cov(0,0) = yR[2];
    cov(0,1) = cov(1,0) = yR[4];
    cov(1,1) = yR[3];
    compute2DQuickPoints( qmdVectorRef(2,yR), cov, qRNeg, qRPos );
    reald res_same_0;
    reald res_same_1;
    reald res_diff_0;
    reald res_diff_1;
    static qmdVector qpNP( 3 );
    static qmdVector qpPN( 3 );
    triangle.point( qLNeg, qRNeg, qmdVectorRef( 3, qpNeg ), &res_same_0 );
    triangle.point( qLPos, qRPos, qmdVectorRef( 3, qpPos ), &res_same_1 );
    triangle.point( qLNeg, qRPos, qpNP, &res_diff_0 );
    triangle.point( qLPos, qRNeg, qpPN, &res_diff_1 );
    if( res_diff_0 + res_diff_1 < res_diff_0 + res_diff_1 )
       {
	   qmdVectorRef( 3, qpNeg ).copy( qpNP );
	   qmdVectorRef( 3, qpPos ).copy( qpPN );
       }
}

int
ComputeGeometry::computeQuickPoints( qmdVectorRef* y, int* val, reald* qp )
{
    for( int i = 0; i < 3; i++ )
       { 
	   if( !val[i] )
	      continue;
	   computeQuickPoint( y[i].getArray(0), y[i].getArray(5), 
			      qp+6*i, qp+6*i+3 );
       }
    return 0;
}
