// ComputeGeometry.h++

#ifndef _ComputeGeometry_H_
#define _ComputeGeometry_H_

#include "QuickMathDouble.h"
#include "noLevenbergMarquardt.h"
#include "noYakkaYakka.h"
#include "noErfBlob.h"
#include "noErfRelativeOrientation.h"
//#include <sys/time.h>
#include "cgTriangleStereo.h"

class ComputeGeometry
{
    int mode[3];
    int prevMode[3];
    int personMode;
    cgTriangleStereo triangle;
    static reald ballSize[3];
    void computeTriangle( int which,
			  qmdVectorRef* y, 
			  int* val,
			  qmdVectorRef* t, 
			  qmdUnitQuaternionRef* q,
			  qmdVectorRef* l );
    void computeBlob( int which,
		      qmdVectorRef* y, 
		      int* val,
		      qmdVectorRef* t, 
		      qmdUnitQuaternionRef* q,
		      qmdVectorRef* l );
    void computeNotrealdly( int which,
			   qmdVectorRef* y, 
			   int* val,
			   qmdVectorRef* t, 
			   qmdUnitQuaternionRef* q,
			   qmdVectorRef* l );
    qmdVector* yakkaX[3];
    qmdVector* blobX[3];
    void computeYakkaYakka( int which,
			    qmdVectorRef* y, 
			    int* val,
			    qmdVectorRef* t, 
			    qmdUnitQuaternionRef* q,
			    qmdVectorRef* l );
    qmdMatrix Rwc;
    qmdMatrix Rwc_inv;
    qmdVector Twc;
    reald scaleToWorld;
    qmdVector pct;
    inline void cameraToWorld( const qmdVectorRef& pcv, qmdVectorRef& pw );
    inline void worldToCamera( const qmdVectorRef& pw, qmdVectorRef& pc );
    static ostream& alert( char* fn, char* msg );
    static void error( char* fn, char* msg );
    float calibData[8];
    float xinit[8];
    void compute2DQuickPoints( qmdVectorRef&, qmdMatrixRef&,
			       qmdVectorRef&, qmdVectorRef& );
    void computeQuickPoint( reald*, reald*, reald*, reald* );
    void yakkaNormalizeBlobs( int which, qmdVector** blobs );
 public:
    static float calibTol;
    static float calibAbstol;
    static int kalmanMode;
    static reald mmtNoiseVar;
    static reald dynNoiseVar;
    static int vtiIter;
    enum { NOMODE=0, TRIANGLE, BLOB, YAKKA };
    enum { HEAD=0, LEFTHAND=1, RIGHTHAND=2 };
    ComputeGeometry();
    void initCalibration( reald* t, reald* q, reald b );
    void initNoMode(int);
    void initTriangle(int);
    void initBlobs(int);
    void initYakka(int);
    inline void compute( qmdVectorRef* y,
			 int* val,
			 qmdVectorRef* t, 
			 qmdUnitQuaternionRef* q,
			 qmdVectorRef* l,
			 qmdVectorRef* tw, 
			 qmdUnitQuaternionRef* qw,
			 qmdVectorRef* lw );
    static reald yakkaDL;
    static reald yakkaDLs;
    static reald yakkaUV;
    static reald lmAbsTol;
    static int lmIterLimit;
    static reald yakkaT;
    static reald yakkaW;
    static reald yakkaL;
    static reald yakkaS;
    static int wakkaFooMode;
    static reald wakkaFoo[3];
    static reald wakkaHoo[3];
    void setMode( int, int );
    void setMode( int mode ) { for (int i=0; i<3; i++ ) setMode(i,mode); };

    void convertToWorld( int* val,
			 qmdVectorRef* t,
			 qmdUnitQuaternionRef* q,
			 qmdVectorRef* l,
			 qmdVectorRef* tw,
			 qmdUnitQuaternionRef* qw,
			 qmdVectorRef* lw );
    static reald yakkaDT;
    static reald yakkaDW;
    void setWorldTransform( qmdVectorRef& t, qmdMatrixRef& r, reald s );
    inline float* getCalibration() { return calibData; }
    void calibrate( int, float* );
    void calibrate( int, float* ,char *);
    int computeQuickPoints( qmdVectorRef* y, int* val, reald* );
};
inline
void
ComputeGeometry::compute( qmdVectorRef* y, 
			  int* val,
			  qmdVectorRef* t, 
			  qmdUnitQuaternionRef* q,
			  qmdVectorRef* l,
			  qmdVectorRef* tw, 
			  qmdUnitQuaternionRef* qw,
			  qmdVectorRef* lw )
{
  for( int i = 0; i < 3; i++ )
    {
      if( TRIANGLE == mode[i] )
	computeTriangle(i, y, val, t, q, l );
      else if( BLOB == mode[i] )
	computeBlob(i, y, val, t, q, l );
      else if( YAKKA == mode[i] )
	computeYakkaYakka(i, y, val, t, q, l );
      else
	computeNotrealdly(i, y, val, t, q, l );
      prevMode[i] = mode[i];
    }

  convertToWorld( val, t, q, l, tw, qw, lw );
}


#endif /* _ComputeGeometry_H_ */
