// Iv3DViewer.h++

#ifndef _Iv3DViewer_H_
#define _Iv3DViewer_H_

#include <stream.h>
#include <Inventor/Xt/viewers/SoXtExaminerViewer.h>
#include <Inventor/nodes/SoSelection.h>
#include <Inventor/nodes/SoPerspectiveCamera.h>
#include <Inventor/manips/SoHandleBoxManip.h>
#include <Inventor/manips/SoTrackballManip.h>
#include <Inventor/manips/SoTransformBoxManip.h>
#include <Inventor/nodes/SoEventCallback.h>
#include <Inventor/events/SoKeyboardEvent.h>
#include <Inventor/Xt/SoXt.h>
#include <X11/Xm/Xm.h>
#include <X11/Xm/Form.h>
#include <X11/Vendor.h>
#include <X11/IntrinsicP.h>
#include <Inventor/SoPickedPoint.h>
#include <Inventor/events/SoMouseButtonEvent.h>
#include <Inventor/SoOffscreenRenderer.h>
#include "IvCameraOverlay.h++"
class Iv3DViewer
{
 public:
    static void selectCB( void*, SoPath* );
    static void deselectCB( void*, SoPath* );
    static void keypressCB( void*, SoEventCallback* );
 protected:
    void initialize( Widget = NULL );
    SoXtExaminerViewer* examiner;
    SoSeparator* root;
    SoPerspectiveCamera* view3Dcamera;
    SoSelection* selectRoot;
    SoSeparator* sceneRoot;
    SoEventCallback* keyEventNode;
    int (*userKeyCB)( void*, SoEventCallback* );
    void* userKeyData;
    int altkey;
    int objectCount;
    int selectCount;
    SoSeparator* objRootLastSelected;
    int inited;
    IvCameraOverlay* overlay;
 public:
    Iv3DViewer( Widget = 0, int build = 1 );
    ~Iv3DViewer();
    int addObject( SoNode* );
    void removeObject( SoSeparator* );
    void removeObject( int );
    SoTransform* getObjectTransform( int );
    SoSeparator* getObject( int );
    inline void setKeyCB( int (*)(void*, SoEventCallback*), void* = 0 );
    inline void resize( int, int, int, int );
    inline SoSelection* getSelectRoot() { return selectRoot; }
    inline SoSeparator* getSceneRoot() { return sceneRoot; }
    inline SoPerspectiveCamera* getCamera() { return view3Dcamera; }
    inline int getObjectCount() { return objectCount; }
    inline SoXtExaminerViewer* getExaminer() { return examiner; }
    inline SoSeparator* getSelectedObject() { return objRootLastSelected; }
    inline void setBackgroundColor( float r, float g, float b );
    int printPS( FILE* fp );
    int writeRGB( FILE* fp, int w = 0, int h = 0 );
    ostream& printCameraTransform( ostream& );
    istream& scanCameraTransform( istream& );
    inline IvCameraOverlay* getOverlay() { return overlay; }
};

inline 
void
Iv3DViewer::setKeyCB( int (*fn)(void*, SoEventCallback* ),
				void* data )
{
    userKeyCB = fn;
    userKeyData = data;
}

inline
void
Iv3DViewer::resize( int x0, int y0, int w, int h )
{
    XtConfigureWidget( examiner->getWidget(), x0, y0, w, h, 0 );
    overlay -> setSize( h, w );
}

inline
void
Iv3DViewer::setBackgroundColor( float r, float g, float b )
{
    examiner -> setBackgroundColor( SbColor( r, g, b ) );
}

#endif /* _Iv3DViewer_H_ */
