// PersonViewer.c++

#include "PersonViewer.h"
#include "SoWriteAction.h"
#include "SoSearchAction.h"

float PersonViewer::triColor[3][3] = {
    { .8, .8, .0 }, { .2, .8, .2 }, { .8, .2, .2 }};
BlobAxes* PersonViewer::triangleAxes = 0;
int PersonViewer::stopInput = 0;
int PersonViewer::jogInput = 0;
int PersonViewer::quitFlag = 0;
int PersonViewer::collectFlag = 0;
int PersonViewer::calibrateFlag = 0;
int PersonViewer::mode = 0;
float PersonViewer::rayDepth = 3.;
int PersonViewer::qpMode = 0;
float PersonViewer::qpSize = .05;

PersonViewer::PersonViewer()
  : window( 0 ), view( 0 ), worldRoot( 0 ), worldXform( 0 )
{
}

int
PersonViewer::keyCB( void* /*vptr*/, SoEventCallback* ecb )
{
    // IvCamera* cam = (IvCamera*) vptr;
    const SoEvent* event = ecb -> getEvent();
    if( SO_KEY_PRESS_EVENT( event, SPACE ) )
       {
	   if( !stopInput )
	      stopInput = 1;
	   else
	      jogInput = 1;
       }
    if( SO_KEY_PRESS_EVENT( event, DELETE ) )
       {
	   if( !stopInput )
	      stopInput = 1;
	   else
	      jogInput = -1;
       }
    if( SO_KEY_PRESS_EVENT( event, C ) )
       {
	   collectFlag = 1;
       }
    if( SO_KEY_PRESS_EVENT( event, PAD_ENTER ) )
       {
	   calibrateFlag = 1;
       }
    if( SO_KEY_PRESS_EVENT( event, RIGHT_ARROW ) )
       {
	   stopInput = 0;
	   jogInput = 0;
       }
    if( SO_KEY_PRESS_EVENT( event, D ) )
       {
	   stopInput = 0;
	   jogInput = 0;
       }
    if( SO_KEY_PRESS_EVENT( event, N ) )
       {
	   cerr << "setting NOMODE mode" << endl;
	   mode = NOMODE;
       }
    if( SO_KEY_PRESS_EVENT( event, T ) )
       {
	   cerr << "setting TRIANGLE mode" << endl;
	   mode = TRIANGLE;
       }
    if( SO_KEY_PRESS_EVENT( event, B ) )
       {
	   cerr << "setting BLOB mode" << endl;
	   mode = BLOB;
       }
    if( SO_KEY_PRESS_EVENT( event, Y ) )
       {
	   cerr << "setting YAKKA mode" << endl;
	   mode = YAKKA;
       }
    if( SO_KEY_PRESS_EVENT( event, Q ) )
       {
	   quitFlag = 1;
       }
    return 1;  // prevents default Iv3DViewer behaviour
	       // (which leaves server thread running)
}


Widget
PersonViewer::init( char* name, int w, int h )
{
    if( window )
       return window;
    window = SoXt::init( name );
    view = new IvStereoCamera( window, w, h );
    view -> setKeyCB( keyCB, view );
    view -> getExaminer() -> setDecoration( 0 );
    view -> getOverlay() -> newTextField(); // frame number
    view -> getOverlay() -> newTextField(); // messages
    triangleAxes = new BlobAxes;
    triangleSphere = new SoSphere;
    for( int i = 0; i < 3; i++ )
       {
	   triangleRoot[i] = new SoSeparator;
	   triangleRoot[i] -> ref();
	   triangleRoot[i] -> addChild
	      ( triangleMaterial[i] = new SoMaterial );
	   triangleRoot[i] -> addChild( triangleSphere );
	   triangleRoot[i] -> addChild
	      ( triangleAxes -> getNode() );
	   if( i )
	      {
		  hands[i-1] = new Hand;
		  triangleRoot[i] -> addChild( hands[i-1]->getNode() );
	      }
	   triangleMaterial[i] -> diffuseColor.setValue
	      ( triColor[i][0], triColor[i][1], triColor[i][2] );
	   if( !qpMode )
	      {
		  triangleIndex[i] = view -> addObject( triangleRoot[i] );
		  triangleXform[i] = 
		     view -> getObjectTransform( triangleIndex[i] );
	      }
       }
    for( i = 0; i < 6; i++ )
       {
	   quickPoints[ i ] = new PurplePoint;
	   if( qpMode && i > 1 )
	      view -> addObject( quickPoints[ i ] -> getNode() );
       }
    SoXt::show( window );
    return window;
}

istream&
PersonViewer::scanViewTransform( istream& is )
{
    view -> scanCameraTransform( is );
    return is;
}

int
PersonViewer::scanViewTransform( char* filename )
{
    ifstream file( filename );
    scanViewTransform( file );
    file.close();
    return 0;
}

void
PersonViewer::setViewTransform( real* t, real* q, real ha )
{
  SoPerspectiveCamera* cam = view -> get3DCamera();
  //  SoPerspectiveCamera* cam = new SoPerspectiveCamera;
  cam -> position.setValue( t[0], t[1], t[2] );
  cam -> orientation.setValue( q[0], q[1], q[2], q[3] );
  cam -> heightAngle.setValue( ha );
}

void
PersonViewer::handle()
{
    XtInputMask m;
    while( m = XtAppPending( SoXt::getAppContext() ) )
       XtAppProcessEvent( SoXt::getAppContext(), m );
}

inline
void
drawCross( IvCamera& cam, float u, float v )
{
    const float color = 2; //white
    const float d = .03;
    cam.newLine( u - d, v, u + d, v, color );
    cam.newLine( u, v - d, u, v + d, color );
}

void
drawBlob( IvCamera* cam, qmVectorRef& y, int which, int color = 2 )
{
    static qmVector evalues( 2 );
    static qmMatrix evectors( 2, 2 );
    static qmMatrix C( 2, 2 );
    static real delta[2];
    C( 0, 0 ) = y( 2 );
    C( 0, 1 ) = y( 4 );
    C( 1, 0 ) = y( 4 );
    C( 1, 1 ) = y( 3 );
    qmEig::eigenSymmetric2( C, evalues, evectors );
    delta[0] = 2. * sqrt( evalues(0) );
    delta[1] = 2. * sqrt( evalues(1) );
    // color = ( p.valid ) ? 2 : 1;
    cam -> redrawLine( 2 * which,
		       y(0) - delta[0] * evectors(0, 0),
		       y(1) - delta[0] * evectors(0, 1),
		       y(0) + delta[0] * evectors(0, 0),
		       y(1) + delta[0] * evectors(0, 1), color );
    cam -> redrawLine( 2 * which + 1,
		       y(0) - delta[1] * evectors(1, 0),
		       y(1) - delta[1] * evectors(1, 1),
		       y(0) + delta[1] * evectors(1, 0),
		       y(1) + delta[1] * evectors(1, 1), color );
}

void
drawFeature( IvStereoCamera& sc, qmVectorRef& y, int inited, int which )
{
    if( !inited )
       {
	   drawCross( *sc.leftCamera, 0, 0 );
	   drawCross( *sc.rightCamera, 0, 0 );
       }
    drawBlob( sc.leftCamera, y, which );
    drawBlob( sc.rightCamera, y.subVector( 5, 5 ), which );
    sc.leftCamera -> drawRay( which, y(0), y(1), PersonViewer::rayDepth );
    sc.rightCamera -> drawRay( which, y(5), y(6), PersonViewer::rayDepth );
}

void
PersonViewer::set( qmVectorRef* y, 
		   int* valid,
		   qmVectorRef* t,
		   qmUnitQuaternionRef* q,
		   qmVectorRef* l )
{
    if( qpMode )
       return;
    const float root5 = sqrt( 5. );
    static int inited[3] = { 0, 0, 0 };
    for( int i = 0; i < 3; i++ )
       {
	   if( !valid[i] )
	      continue;
	   triangleXform[i] -> translation.setValue
	      ( t[i](0), t[i](1), t[i](2) );
	   triangleXform[i] -> rotation.setValue
	      ( q[i](1), q[i](2), q[i](3), q[i](0) );
	   triangleXform[i] -> scaleFactor.setValue
	      ( root5*l[i](0), root5*l[i](1), root5*l[i](2) );
	   drawFeature( *view, y[i], inited[i], i );
	   inited[i] = 1;
       }
}

void
PersonViewer::initCalibration( float tx, float ty, float tz,
			       float q0, float qx, float qy, float qz,
			       float  )
{
    view -> rightCamera -> setCoordinateTransform
       ( tx, ty, tz, q0, qx, qy, qz );
}

void
PersonViewer::setWorldTransform( real* t, real* r, real scale )
{
    float a = 1. / scale; // a is unit/inch
    worldS.setValue( a, a, a );
    worldT.setValue( t[0], t[1], t[2] );
    worldR.setValue( SbMatrix( r[0], r[1], r[2], 0.,
			       r[3], r[4], r[5], 0.,
			       r[6], r[7], r[8], 0.,
			       0., 0., 0., 1. ) );
    if( worldXform )
       {
	   worldXform -> scaleFactor.setValue( worldS );
	   worldXform -> rotation.setValue( SbRotation( worldR ) );
	   worldXform -> translation.setValue( worldT );
       }
}

void
PersonViewer::initWorld( char* fname, char* pname )
{
#if 1
    if( worldRoot )
       return;
    char fullpath[512];
    sprintf( fullpath, "%s/%s", pname, fname );
    SoInput input;
    if( FALSE == input.openFile( fullpath ) )
       {
	   alert( "initWorld", "can't read world.iv" );
	   error( "initWorld", "read failed" );
       }
    worldRoot = SoDB::readAll( &input );
    view -> addObject( worldRoot );
    SoSearchAction searchAction;
    searchAction.setName( "worldTransform" );
    searchAction.apply( worldRoot );
    worldXform = (SoTransform*) searchAction.getPath() -> getTail();
    if( !worldXform )
       error( "initWorld", "can't find worldTransform in scene graph" );
    worldXform -> scaleFactor.setValue( worldS );
    worldXform -> rotation.setValue( SbRotation( worldR ) );
    worldXform -> translation.setValue( worldT );
#else
    const float tableWidth_inch = 45.;
    const float tableHeight_inch = 1.;
    const float tableDepth_inch = 24.;

    if( worldRoot )
       return;
    worldRoot = new SoSeparator;
    worldXform = new SoTransform;

    tableRoot = new SoSeparator;
    tableMaterial = new SoMaterial;
    tableXform2 = new SoTransform;
    tableShape = new SoCube;

    axesRoot = new SoSeparator;
    axesMaterial = new SoMaterial;
    axesEndPoints = new SoCoordinate3;
    axesConnection = new SoIndexedLineSet;

    worldRoot -> ref();
    worldRoot -> addChild( worldXform );
    worldRoot -> addChild( tableRoot );
    worldRoot -> addChild( axesRoot );

    tableRoot -> addChild( tableMaterial );
    tableRoot -> addChild( tableXform2 );
    tableRoot -> addChild( tableShape );

    axesRoot -> addChild( axesMaterial );
    axesRoot -> addChild( axesEndPoints );
    axesRoot -> addChild( axesConnection );

    // Transform to world
    worldXform -> scaleFactor.setValue( worldS );
    worldXform -> rotation.setValue( SbRotation( worldR ) );
    worldXform -> translation.setValue( worldT );
    worldXform -> setName( "worldTransform" );

    // Material
    tableMaterial -> diffuseColor.setValue( .5, .5, .5 );
    tableMaterial -> ambientColor.setValue( .5, .5, .5 );

    // Set origin at edge rather than in middle
    tableXform2 -> translation.setValue
       ( 0, -tableHeight_inch, -tableDepth_inch/2. );
    tableShape -> width.setValue( tableWidth_inch );
    tableShape -> height.setValue( tableHeight_inch );
    tableShape -> depth.setValue( tableDepth_inch );

    // Axes
    static const float xyz[6][3] = { {0,0,0}, {18.,0,0},
					 {0,0,0}, {0,18,0},
					 {0,0,0}, {0,0,18} };
    static const int32_t conn[9] = { 0, 1, -1, 2, 3, -1, 4, 5, -1 };
    axesMaterial -> diffuseColor.setValue( 0,1,0 );
    axesMaterial -> ambientColor.setValue( 0,1,0 );
    axesEndPoints -> point.setValues( 0, 6, xyz );
    axesConnection -> coordIndex.setValues( 0, 9, conn );

    view -> addObject( worldRoot );
#endif
}

void
PersonViewer::dumpScene()
{
    SoWriteAction writer;
    writer.getOutput() -> openFile( "dump.iv" );
    writer.getOutput() -> setBinary( FALSE );
    writer.apply( worldRoot );
    writer.getOutput() -> closeFile();
}

ostream&
PersonViewer::alert( char* fn, char* msg )
{
    return( cerr << "PersonViewer: ERROR: " << fn << "(): " << msg << endl );
}

void
PersonViewer::error( char* fn, char* msg )
{
    alert( fn, msg );
    // exit( 1 );
    assert(0);
}

void
PersonViewer::setHands( int which[2], int minus[2] )
{
    hands[0] -> set( which[0], minus[0] );
    hands[1] -> set( which[1], minus[1] );
}

void
PersonViewer::setExistence( int exists )
{
#if 0
    static int color = 0;
    if( exists && color )
       {
	   view -> setBackgroundColor( 0, 0, 0 );
	   color = 0;
	   return;
       }
    if( !exists && !color )
       {
	   view -> setBackgroundColor( 1, 1, 1 );
	   color = 1;
	   return;
       }
#else
    static int existed = 0;
    if( exists && !existed )
       setMessage( "" );
    if( !exists && existed )
       setMessage( "NO PERSON IN SCENE" );
    existed = exists;
#endif
}

void
PersonViewer::setFrameNumber( int f )
{
    static char banner[ 256 ];
    sprintf( banner, "Frame %d", f );
    view -> getOverlay() -> setText( 0, banner );
}

void
PersonViewer::setMessage( char* s )
{
    view -> getOverlay() -> setText( 1, s );
}

void
PersonViewer::setQuickPoints( real* qp, real size )
{
    if( !qpMode )
       return;
    PurplePoint::setSize( size );
    for( int i = 0; i < 6; i++ )
       {
	   quickPoints[i] -> set( qp+i*3 );
	   if( i%2 )
	      quickPoints[i] -> setColor( 1, .15, .65 );
       }
}
