// SffinderClient.c++

#include "SffinderClient.h"
//#include <unistd.h>
#include <limits.h>
#include <iostream.h>
#include "iotools.h"
#include <rpc.h>
#include "sffinder.h"

SffinderClient::~SffinderClient()
{
    /*  The calls to the remote procedures are complete. */
    /*  Free the string and the binding handle           */
    RpcStringFree(&pszStringBinding);  // remote calls done; unbind
    RpcBindingFree(&sffinder_IfHandle);  // remote calls done; unbind
}

SffinderClient::SffinderClient( char* serverHostname )
  : lastFrame( -1 ), lastGot( GOTNONE )
{
  pszUuid = NULL;             //= NULL;
  strcpy((char *)pszProtocolSequence,"ncacn_np"); //= "ncacn_np";
  pszNetworkAddress = NULL;   //= NULL;
  strcpy((char*)pszEndpoint,"\\pipe\\sffinder");         //= "\\pipe\\hello";
  pszOptions = NULL;          //= NULL;
  pszStringBinding = NULL;    //= NULL;
  strcpy((char *)pszString,"hello, sffinder");           //= "hello, world";

    person.rightHand.valid = 0;
    person.leftHand.valid = 0;
    person.head.valid = 0;
    if( serverHostname )
       if( connect( serverHostname ) )
	        cerr << "connection failed" << endl;

}

int
SffinderClient::connect( char* serverHostname )
{
    RPC_STATUS status;

    pszNetworkAddress = (unsigned char *)serverHostname;

    /* Use a convenience function to concatenate the elements of */
    /* the string binding into the proper sequence.              */
    status = RpcStringBindingCompose(pszUuid,
                                     pszProtocolSequence,
                                     pszNetworkAddress,
                                     pszEndpoint,
                                     pszOptions,
                                     &pszStringBinding);
    printf("RpcStringBindingCompose returned 0x%x\n", status);
    printf("pszStringBinding = %s\n", pszStringBinding);
    if (status) {
        return 1;
    }

    /* Set the binding handle that will be used to bind to the server. */
    status = RpcBindingFromStringBinding(pszStringBinding,
                                         &sffinder_IfHandle);
    printf("RpcBindingFromStringBinding returned 0x%x\n", status);
    if (status) {
        return 1;
    }
    return 0;
}

int
SffinderClient::getPersonFromServer()
{
    personPlus = sffgetpersonplus();
    lastGot = GOTPLUS;
    return 0;
}

int
SffinderClient::getPart( int i, float* xyz, float* q, float* l )
{
    sffWorldPart* wp = (HEAD==i) ? &personPlus.head 
       : (LHAND==i) ? &personPlus.leftHand
       : (RHAND==i) ? &personPlus.rightHand : 0;
    if( !wp )
       return 1;
    if( !wp->valid )
       return 2;
    iotools_copyVector( 3, wp->world_ref.t, xyz );
    if( q )
       iotools_copyVector( 4, wp->world_ref.q, q );
    if( l )
       iotools_copyVector( 3, wp->world_ref.s, l );
    return 0;
}

sffPersonPlus*
SffinderClient::getPersonPlus()
{
    personPlus = sffgetpersonplus();
    lastFrame = personPlus.frameNumber;
    lastGot = GOTPLUS;
    return &personPlus;
}

sffPersonPlus*
SffinderClient::getNextPersonPlus()
{
    personPlus = sffgetnextpersonplus(lastFrame);;
    lastFrame = personPlus.frameNumber;
    lastGot = GOTPLUS;
    return &personPlus;
}

void
SffinderClient::nextUndump()
{
   sffundumpnext();
}

ostream&
SffinderClient::print( ostream& os )
{
    return os;
}

void
SffinderClient::sleep()
{
  sffsleep();
}

void
SffinderClient::wakeup()
{
  sffwakeup();
}

int
SffinderClient::getPersonExists()
{
  return sffgetpersonexists();
}

sffPerson*
SffinderClient::getNextPerson()
{
    person = sffgetnextperson(lastFrame);
    lastFrame = person.frameNumber;
    lastGot = GOTPERSON;
    return &person;
}

//////////////////////////////////////////////////////////////////////////////


void SffinderClient::quit()
{
  sffquit();
  RpcStringFree(&pszStringBinding);  // remote calls done; unbind
  RpcBindingFree(&sffinder_IfHandle);  // remote calls done; unbind
}
