// test.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h> 
#include <mil.h> 

void MilApplication(HWND UserWindowHandle)
{
  /* MIL variables */
  MIL_ID MilApplication,  /* MIL Application identifier.  */
         MilSystem,       /* MIL System identifier.       */
         MilDisplay,      /* MIL Display identifier.      */
         MilDigitizer,    /* MIL Digitizer identifier.    */
         MilImage;        /* MIL Image buffer identifier. */
         
  long BufSizeX;
  long BufSizeY;
  long BufSizeBand;

  /* Allocate a MIL application. */
  MappAlloc(M_DEFAULT, &MilApplication);

  /* Allocate a MIL system. */
  MsysAlloc(M_DEF_SYSTEM_TYPE, M_DEV0, M_DEFAULT, &MilSystem);

  /* Allocate a MIL display. */
  MdispAlloc(MilSystem, M_DEV0, M_DEF_DISPLAY_FORMAT, M_DEFAULT, &MilDisplay);

  /* Allocate a MIL digitizer if supported and sets the target image size. */
  if (MsysInquire(MilSystem, M_DIGITIZER_NUM, M_NULL) > 0)
    {
    MdigAlloc(MilSystem, M_DEV0, M_DEF_DIGITIZER_FORMAT, M_DEFAULT, &MilDigitizer);
    MdigInquire(MilDigitizer, M_SIZE_X,    &BufSizeX);
    MdigInquire(MilDigitizer, M_SIZE_Y,    &BufSizeY);
    MdigInquire(MilDigitizer, M_SIZE_BAND, &BufSizeBand);
    }
  else
    {
    MilDigitizer = M_NULL;
    BufSizeX     = BUFFERSIZEX;
    BufSizeY     = BUFFERSIZEY;
    BufSizeBand  = BUFFERSIZEBAND;
    }
  
     
  /* Do not allow example to run in dual screen mode */             
  if(MdispInquire(MilDisplay, M_DISPLAY_MODE, M_NULL) != M_WINDOWED)             
    {
     MessageBox(0,"""This example does no run in dual screen mode.",
                  "MIL application example",
                  MB_APPLMODAL | MB_ICONEXCLAMATION );
    goto end;
    }

  /* Allocate a MIL buffer. */
  MbufAllocColor(MilSystem, BufSizeBand, BufSizeX, BufSizeY, 8+M_UNSIGNED,
             (MilDigitizer? M_IMAGE+M_DISP+M_GRAB : M_IMAGE+M_DISP), &MilImage);

  /* Clear the buffer */
  MbufClear(MilImage,0);

  /* Select the MIL buffer to be displayed in the user specified window */
  MdispSelectWindow(MilDisplay, MilImage, UserWindowHandle);

  /* Print a string in the image buffer using MIL.
     Note: After a MIL command writing in a MIL buffer, the display 
     will automatically update the window given to MdispSelectWindow().
  */
  MgraText(M_DEFAULT, MilImage, (BufSizeX/8)*3, BufSizeY/2,   " ------------------ ");
  MgraText(M_DEFAULT, MilImage, (BufSizeX/8)*3, BufSizeY/2+25," Welcome to MIL !!! ");
  MgraText(M_DEFAULT, MilImage, (BufSizeX/8)*3, BufSizeY/2+50," ------------------ ");

  /* Windows code to open a message box to wait a key. */
  MessageBox(0,"""Welcome to MIL !!!"" was printed",
               "MIL application example",
               MB_APPLMODAL | MB_ICONEXCLAMATION );

  /* Grab in the user window if supported by the system. */
  if (MilDigitizer)
     {
     /* Grab continuously. */
     MdigGrabContinuous(MilDigitizer, MilImage);

     /* Windows code to open a message box to wait a key. */
     MessageBox(0,"Continuous grab in progress",
                  "MIL application example",
                  MB_APPLMODAL | MB_ICONEXCLAMATION );
                  
     /* Stop continuous grab. */
     MdigHalt(MilDigitizer);
     }

  /* Deselect the MIL buffer from the display. */
  MdispDeselect(MilDisplay, MilImage);

  /* Free allocated objects. */
  MbufFree(MilImage);

end:

  MdispFree(MilDisplay);
  if (MilDigitizer)
     MdigFree(MilDigitizer);
  MsysFree(MilSystem);
  MappFree(MilApplication);
}


void main(void)
{ 
  MIL_ID MilApplication,  /* Application identifier.  */
         MilSystem,       /* System identifier.       */
         MilDisplay,      /* Display identifier.      */
         MilDigitizer,    /* Digitizer identifier.    */ 
         MilImage;        /* Image buffer identifier. */

  /* Allocate defaults. */
  MappAllocDefault(M_SETUP, &MilApplication, &MilSystem,
                   &MilDisplay, &MilDigitizer, &MilImage);

  /* Grab continuously. */
  MdigGrabContinuous(MilDigitizer, MilImage);

  /* When a key is pressed, halt. */
  printf("Continuous grab in progress. Adjust your camera and\n");
  printf("press <Enter> to stop grabbing.\n");
  getchar();

  /* Stop continuous grab. */
  MdigHalt(MilDigitizer);

  /* Pause to show the result. */
  printf("\nDisplaying the last grabbed image.\n");
  printf("Press <Enter> to end.\n");
  getchar();

  /* Release defaults. */
  MappFreeDefault(MilApplication, MilSystem, MilDisplay,
                                 MilDigitizer, MilImage);
}


