// vlImage.c++

#include "vlImage.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//VLServer vlImage::server = 0;
//VLNode vlImage::drain = 0;
//VLNode vlImage::source = 0;
//VLPath vlImage::path = 0;
//VLBuffer vlImage::buffer = 0;
//VLInfoPtr vlImage::info = 0;
//VLControlValue vlImage::val;

vlImage::vlImage(){
  transferBegun = 0;
  captureWidth = 0;
  captureHeight = 0;
  captureX = 0;
  captureY = 0;
  captureSize = 0;
  captureType = vlImage::RGB;
  captureBytesPerPixel = vliRGBBytesPerPixel;
  verbose = 0;

  MilSystem = M_NULL;
  MilDisplay = M_NULL;
  MilDigitizer = M_NULL;
  MilImage = M_NULL;
  Mbufattribs = M_IMAGE+M_DISP+M_GRAB+M_PROC;
  Mbufformat = M_BGR32+M_PACKED;
}

vlImage::~vlImage(){
  close();
}

void
vlImage::error( char *message )
{
  char tag[128];
  sprintf( tag, "vlImage: %s", message );
  //vlPerror( tag );
  close();
  exit( 1 );
}

void
vlImage::warning( char *message )
{
  char tag[128];
  sprintf( tag, "vlImage: %s", message );
  //vlPerror( tag );
}
/*
int
findIndyAnalogVideoInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, "Analog Video Input" ) )
	  return devList.devices[0].nodes[i].number;
    return -1;
}

int
findO2SVideoInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, "SVideo Input" ) )
	  return devList.devices[0].nodes[i].number;
    return -1;
}

int
findO2CompositeInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, "Composite Video Input"))
	  return devList.devices[0].nodes[i].number;
    return -1;
}

int
findDigitalVideoInput( VLServer server )
{
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    char* name = "Bogus";
    if( !strcmp( devList.devices[0].name, "ev1" ) )
       // we are on Galileo
       {
	   name = "IndyCam";
       }
    else if( !strcmp( devList.devices[0].name, "vino" ) )
       // we are on Indy
       {
	   name = "Digital Video Input";
       }
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( !strcmp( devList.devices[0].nodes[i].name, name ) )
	  return devList.devices[0].nodes[i].number;
    return -1;
}
*/
#if 0
int
findAllInputs( VLServer server )
{
    int count = 0;
    VLDevList devList;
    vlGetDeviceList( server, &devList );
    for( int i = 0; i < devList.devices[0].numNodes; i++ )
       if( devList.devices[0].nodes[i].type == VL_SRC &&
	   devList.devices[0].nodes[i].kind == VL_VIDEO )
	  {
	      cerr << devList.devices[0].nodes[i].name << "= "
		   << devList.devices[0].nodes[i].number << endl;
	      count++;
	  }
    return count;
}
#endif

void
vlImage::allocateBuffers(){

  deallocateBuffers();

  int BufSizeBand;
  MdigInquire(MilDigitizer, M_SIZE_BAND, &BufSizeBand);

  MbufAllocColor(MilSystem, BufSizeBand, captureWidth, captureHeight, 8+M_UNSIGNED,
             Mbufattribs+Mbufformat, &MilImage);
  /* Clear the buffer */
  MbufClear(MilImage,0);
}

void
vlImage::deallocateBuffers(){
  if(MilImage!=M_NULL)
    MbufFree(MilImage);
  MilImage = M_NULL;
}

void
vlImage::init(int card, int zoom, int encoding, 
	       int cropx, int cropy, int cropw, int croph)
{
  if( transferBegun )
    return;

  /* Allocate a MIL video card. */
  /* Allocate a MIL digitizer if supported and sets the target image size. */
  switch(card){
  case 1: // M_SYSTEM_METEOR:
    MsysAlloc(M_SYSTEM_METEOR, M_DEV0, M_DEFAULT, &MilSystem);
    MdigAlloc(MilSystem, M_DEV0, "M_NTSC_YC", M_DEFAULT, &MilDigitizer);
    break;
  case 0: // M_SYSTEM_METEOR_II: 
  default:
    MsysAlloc(M_SYSTEM_METEOR_II, M_DEV0, M_DEFAULT, &MilSystem);
    MdigAlloc(MilSystem, M_DEV0, "M_NTSC", M_DEFAULT, &MilDigitizer);
    break;
  }
  //MdigControl(MilDigitizer,M_GRAB_SCALE,M_FILL_DISPLAY);
  //char buffer[1000];
  //MappGetError(M_CURRENT,buffer);
  // MessageBox(0,buffer,"Mil",MB_APPLMODAL|MB_ICONEXCLAMATION);
     

  /* Select the MIL buffer to be displayed in the user specified window */
  //MdispSelectWindow(MilDisplay, MilImage, UserWindowHandle);

  setEncoding( encoding );
  setZoom( zoom );
  if ((cropw != -1) && (croph != -1))
      setCrop( cropx, cropy, cropw, croph );

  allocateBuffers();

/*
  if( !( server = vlOpenVideo("") ) )
    error( "open video" );
  drain = vlGetNode( server, VL_DRN, VL_MEM, VL_ANY );
  int inumber;
  switch (aord) 
    {
    case O2SVIDEO: 
      inumber = findO2SVideoInput( server );
      break;
    case O2COMPOSITE:
      inumber = findO2CompositeInput( server );
      break;
    case INDYANALOG: 
      inumber = findIndyAnalogVideoInput( server );
      break;
    case DIGITAL: 
      inumber = findDigitalVideoInput( server );
      break;
    default:
      inumber = -1;
    }
  if( -1 == inumber )
    error( "failed to find analog or digital input" );
  // source = vlGetNode( server, VL_SRC, VL_VIDEO, VL_ANY );
  source = vlGetNode( server, VL_SRC, VL_VIDEO, inumber );
  path = vlCreatePath( server, VL_ANY, source, drain );
  if( ( vlSetupPaths( server, 
		      (VLPathList)&path, 1, VL_SHARE, VL_SHARE ) ) < 0 )
    error( "setup paths" );
  //setInput( input );
  setEncoding( encoding );
  setZoom( zoom );
  if ((cropw != -1) && (croph != -1))
      setCrop( cropx, cropy, cropw, croph );
  buffer = vlCreateBuffer( server, path, drain, nbuffers );
  if( !buffer )
    error( "create buffers" );
  vlRegisterBuffer( server, path, drain, buffer );
  if( vlBeginTransfer( server, path, 0, NULL ) )
    error( "begin transfer" );
    */
  transferBegun = 1;
}

void
vlImage::close()
{
    if( !transferBegun )
       return;
    
    /* Free allocated objects. */
    deallocateBuffers();

    //MdispFree(MilDisplay);
    MilDisplay = M_NULL;
    if (MilDigitizer != M_NULL)
       MdigFree(MilDigitizer);
    MilDigitizer = M_NULL;
    if (MilSystem != M_NULL)
      MsysFree(MilSystem);
    MilSystem = M_NULL;
    /*
    vlEndTransfer( server, path );
    vlDeregisterBuffer( server, path, drain, buffer );
    vlDestroyBuffer( server, buffer );
    vlDestroyPath( server, path );
    vlCloseVideo( server );
    */
    transferBegun = 0;
}

#if 0
int
vlImage::NBuffers( int n )
{
  if( QUERY != n && !transferBegun )
     nbuffers = n;
  return nbuffers;
}
#endif

int
vlImage::getEncoding()
{
  if(MilDigitizer==M_NULL)
    return -1;

  return Mbufattribs+Mbufformat;
  /*
    if( !server )
       return -1;
    vlGetControl( server, path, drain, VL_PACKING, &val );
    return val.intVal;
    */
}

void
vlImage::setEncoding( int mode )
{
  switch( mode )
     {
	case YUV: 
	 //val.intVal = VL_PACKING_YVYU_422_8; 
   Mbufattribs = M_IMAGE+M_GRAB;
   Mbufformat = M_YUV16+M_PACKED; // Force YUV16 packed (4:2:2) format.
	 captureType = YUV;
	 captureBytesPerPixel = vliYUVBytesPerPixel;
	 break;
	case RGB: default: 
	 //val.intVal = VL_PACKING_RGB_8; 
   Mbufattribs =  M_IMAGE+M_GRAB;
   Mbufformat = M_BGR32+M_PACKED; //	Force 32-bit pixels in BGRX 8888 format.
	 captureType = RGB;
	 captureBytesPerPixel = vliRGBBytesPerPixel;
	 break;
     }

  /*
  if( vlSetControl( server, path, drain, VL_PACKING, &val ) < 0 )
     error( "set packing control" );
     */


  if(MilDigitizer==M_NULL)
     return;

  // Only reallocate if it exists already
  if(MilImage!=M_NULL) {
    allocateBuffers();
  }
}

int
vlImage::getZoom( )
{
  double zoom;
  if(MilDigitizer == M_NULL)
     return -1;

  MdigInquire(MilDigitizer, M_GRAB_SCALE_X,    &zoom);

  return 1.0/zoom;
/*    if( !server )
       return -1;
    vlGetControl( server, path, drain, VL_ZOOM, &val );
    return val.intVal;
    */
}

void
vlImage::setZoom( int decimation )
{

  if(MilDigitizer == M_NULL)
     return;

  long w,h;
  MdigInquire(MilDigitizer, M_SIZE_X,    &w);
  MdigInquire(MilDigitizer, M_SIZE_Y,    &h);

  MdigControl(MilDigitizer,M_GRAB_SCALE,1.0/decimation);

  captureWidth = w/decimation;
  captureHeight = h/decimation;
  captureSize = captureHeight * captureWidth;

  // Only reallocate if it exists already
  if(MilImage!=M_NULL) {
    allocateBuffers();
  }
  
  /*
    if( !server )
       return;
    val.fractVal.numerator = 1;
    val.fractVal.denominator = decimation;
    if( vlSetControl( server, path, drain, VL_ZOOM, &val ) < 0 )
       error( "set zoom control" );
    vlGetControl( server, path, drain, VL_SIZE, &val );
    captureHeight = val.xyVal.y;
    captureWidth = val.xyVal.x;
    captureSize = captureHeight * captureWidth;
    */
}

void
vlImage::setCrop( int x, int y, int width, int height )
{
    if(MilDigitizer == M_NULL)
       return;

    MdigControl(MilDigitizer,M_SOURCE_OFFSET_X,x);
    MdigControl(MilDigitizer,M_SOURCE_OFFSET_Y,y);
    MdigControl(MilDigitizer,M_SOURCE_SIZE_X,width);
    MdigControl(MilDigitizer,M_SOURCE_SIZE_Y,height);

    captureY = y;
    captureX = x;
    captureHeight = height;
    captureWidth = width;

  /*
    if( !server )
       return;

    captureY = val.xyVal.y = y;
    captureX = val.xyVal.x = x;

    if( vlSetControl( server, path, drain, VL_OFFSET, &val ) < 0 )
       warning( "set OFFSET control" );

    vlGetControl( server, path, drain, VL_OFFSET, &val );
    captureY = val.xyVal.y;
    captureX = val.xyVal.x;
    cerr << "set OFFSET to " << captureX << " " << captureY << endl;

    val.xyVal.y = height;
    val.xyVal.x = width;

    if( vlSetControl( server, path, drain, VL_SIZE, &val ) < 0 )
      warning( "set SIZE control" );

    vlGetControl( server, path, drain, VL_SIZE, &val );
    captureHeight = val.xyVal.y;
    captureWidth = val.xyVal.x;
    cerr << "set SIZE to " << captureWidth << " " << captureHeight << endl;
    */
}

int
vlImage::getInput()
{
    if(MilDigitizer == M_NULL)
       return -1;

    return -1;
  /*
    if( !server )
       return -1;
    vlGetControl( server, path, source, VL_MUXSWITCH, &val );
    return val.intVal;
    */
}

void
vlImage::setInput( int in )
{
    if(MilDigitizer == M_NULL)
       return;

  /*
    if( !server )
       return;
    val.intVal = in;
    if ( vlSetControl( server, path, source, VL_MUXSWITCH, &val ) < 0 )
      {
	warning( "set source control" );
      }
      */
}

void
vlImage::setVerbose( int in )
{
    // Only set if 0 or 1; else just echo value
    if( 0 == in )
       verbose = 0;
    if( 1 == in )
       verbose = 1;
}

vliImageRef
vlImage::blindGrab()
{
  if( !transferBegun )
    error( "blindGrab: attempt to grab before initialized" );

  // Grab a frame
  MdigGrab(MilDigitizer,MilImage);

  // Get a ptr to the buffer
  static byte * data = NULL;
  //if(!data)
  //  data = new byte[captureWidth*captureHeight*captureBytesPerPixel];
//  MbufGetColor(MilImage,M_BGR24+M_PACKED,M_ALL_BAND,data);
  data = (byte *)MbufInquire(MilImage,M_HOST_ADDRESS,M_NULL);
  vliImageRef iref( captureWidth,
		  captureHeight,
		  captureBytesPerPixel,
		  data );
  return iref;

/*
  if( !transferBegun )
       error( "blindGrab: attempt to grab before initialized" );
    info = (VLInfoPtr) NULL;
    while( !( info = vlGetLatestValid( server, buffer ) ) )
      {
	if (vlBufferDone(buffer) < 0)
	  {
	    warning("didn't get a frame");
	  }
	else
	  {
	    if ( vlGetErrno() != VLSuccess )
	      warning("vlBufferDone");
	  }
	sginap( 1 );
      }
    byte* dataPtr = (byte *) vlGetActiveRegion( server, buffer, info );
    vliImageRef iref( captureWidth,
		      captureHeight,
		      captureBytesPerPixel,
		      dataPtr );
    return iref;
    */
}

void
vlImage::grab( vliRGBImageRef& rgb )
{
    if( RGB == captureType )
       rgb = blindGrab();
    else
       error( "grab: attempt to assign non-RGB image to vliRGBImageRef" );
}

void
vlImage::grab( vliYUVImageRef& yuv )
{
    if( YUV == captureType )
       yuv = blindGrab();
    else
       error( "grab: attempt to assign non-YUV image to vliYUVImageRef" );
}

void
vlImage::free()
{
}

void
vlImage::yuvToRgb( const vliYUVImageRef& yuv_image, vliRGBImageRef& rgb_image )
{
    // could check dimensions here
    vliYUVPixel yuv;
    int pixelCount = rgb_image.getPixelCount();
    for( int i = 0; i < pixelCount; i++ )
       {
	   vliRGBPixel& rgb = rgb_image.getPixel( i );
	   yuv = yuv_image.getPixel(i);
	   vliPixel::yuvToRgb
	      ( yuv.y, yuv.u, yuv.v, rgb.r, rgb.g, rgb.b );
       }
}

void
vlImage::yuvToRgb2( const vliYUVImageRef& yuv_image,
		    vliRGBImageRef& rgb_image )
{
    // could check dimensions here
    vliYUVPixel yuv;
    int pixelCount = rgb_image.getPixelCount();
    for( int i = 0; i < pixelCount; i++ )
       {
	   vliRGBPixel& rgb = rgb_image.getPixel( i );
	   yuv = yuv_image.getPixel(i);
	   vliPixel::yuvToRgb2
	      ( yuv.y, yuv.u, yuv.v, rgb.r, rgb.g, rgb.b );
       }
}

void
vlImage::rgbToYuv( const vliRGBImageRef& rgb_image, vliYUVImageRef& yuv_image )
{
    // could check dimensions here
    vliYUVPixel yuv;
    int pixelCount = yuv_image.getPixelCount();
    for( int i = 0; i < pixelCount; i++ )
       {
	   rgbToYuv( rgb_image.getPixel( i ), yuv );
	   yuv_image.setPixel( i, yuv );
       }
}

