// vlImage.h

#ifndef _vlImage_H_
#define _vlImage_H_
#include "vlImageGlobal.h"
#include "vliImageRef.h"
#include "vliPixel.h"
#include "vliRGBImage.h"
#include "vliRGBImageRef.h"
#include "vliRGBPixel.h"
#include "vliTypes.h"
#include "vliYUVImage.h"
#include "vliYUVImageRef.h"
#include "vliYUVPixel.h"

#include "mil.h"
#include "matrox.h"

class EXPORT_CLASS vlImage
{
public:
  vlImage();
  ~vlImage();
  MIL_ID MilSystem;
  MIL_ID MilDisplay;      /* MIL Display identifier.      */
  MIL_ID MilDigitizer;    /* MIL Digitizer identifier.    */
  MIL_ID MilImage;        /* MIL Image buffer identifier. */
  long Mbufattribs;
  long Mbufformat;
  
  void error(char* message);
  void warning(char* message);
  //  static VLServer server;	// Set by Init
  //  static VLNode drain;	// Set by Init
  //  static VLNode source;	// Set by Init
  //  static VLPath path;		// Set by Init
  //  static VLBuffer buffer;	// Set by Init
  //  static VLInfoPtr info;
  //  static VLControlValue val;
    int transferBegun;	// Set by Init and Close
    int captureWidth;	// Set by Zoom
    int captureHeight;	// Set by Zoom
    int captureSize;	// Set by Zoom (height * width)
    int captureX;	// Set by Crop
    int captureY;	// Set by Crop
    int captureType;
    int captureBytesPerPixel;
    int verbose;

 public:
    enum { QUERY = -1, RGB = 0, YUV = 1 };
    enum { O2SVIDEO = 0, O2COMPOSITE = 1, INDYANALOG = 2, DIGITAL = 3 };
    enum { SVIDEO_1 = 0, SVIDEO_2 = 1, SVIDEO_3 = 2,
	       COMPOSITE_1 = 3, COMPOSITE_2 = 4, COMPOSITE_3 = 5,
	       YRB_1 = 6, YRB_2 = 7 };

    void init( int card = 0,
		      int zoom = 1,
		      int encoding = RGB,
		      int cropx = 0,
		      int cropy = 0,
		      int cropw = -1, 
		      int croph = -1);
    void close();
    vliImageRef blindGrab();
    void grab( vliRGBImageRef& );
    void grab( vliYUVImageRef& );
    void free();
    void setEncoding( int );
    int getEncoding( );
    void setZoom( int );
    int getZoom( );
    void setCrop( int x, int y, int width, int height );
    void setInput( int );
    int getInput( );
    void setVerbose( int );
    static inline void yuvToRgb( const vliYUVPixel&, vliRGBPixel& );
    static void yuvToRgb( const vliYUVImageRef&, vliRGBImageRef& );
    static void yuvToRgb2( const vliYUVImageRef&, vliRGBImageRef& );
    static inline void rgbToYuv( const vliRGBPixel&, vliYUVPixel& );
    static void rgbToYuv( const vliRGBImageRef&, vliYUVImageRef& );

 private:
   void allocateBuffers();
   void deallocateBuffers();
};

inline
void
vlImage::yuvToRgb( const vliYUVPixel& yuv, vliRGBPixel& rgb )
{
    vliPixel::yuvToRgb( yuv.y, yuv.u, yuv.v, rgb.r, rgb.g, rgb.b );
}

inline
void
vlImage::rgbToYuv( const vliRGBPixel& rgb, vliYUVPixel& yuv )
{
    vliPixel::rgbToYuv( rgb.getR(), rgb.getG(), rgb.getB(),
			yuv.y, yuv.u, yuv.v );
}

#endif /* _vlImage_H_ */
