// vliRGBImageRef.c++

#include "vliRGBImageRef.h"
#include "vliPixel.h"
#include <math.h>

istream&
vliRGBImageRef::readRawRGB( istream& is )
{
    byte rgb[3];
    vliRGBPixel* pixp = getPixelPtr(0);
  
    for( int i = 0; i < pixelCount; i++ )
       {
	   is.read( rgb, 3 );
	   if( is.eof() )
	      return is;
	   pixp -> red() = rgb[0];
	   pixp -> green() = rgb[1];
	   pixp -> blue() = rgb[2];
	   pixp++;
       }
    return is;
}

ostream&
vliRGBImageRef::writeRawRGB( ostream& os )
{
    byte rgb[3];
    byte* p = data;
    for( int i = 0; i < pixelCount; i++ )
       {
	   rgb[0] = p[3];
	   rgb[1] = p[2];
	   rgb[2] = p[1];
	   p += 4;
	   os.write( rgb, 3 );
       }
    os.flush();
    return os;
}

void
vliRGBImageRef::drawBlob( double um, double vm,
			  double c11, double c22, double c12,
			  vliRGBPixel& color, double d )
{
    double determ = c11*c22 - c12*c12;
    if( determ <= 0 )
       {
	   cerr << "vliRGBImageRef::drawBlob: bad blob: determ <= 0" << endl;
	   return;
       }
    int u_plus;
    int u_minus;
    int v_plus;
    int v_minus;
    double discrim;
    double s;
    double t;
    for( double du = 0.; ( discrim = c22*d - du*du*determ ) >= 0.; du += 1. )
       {
	   s = -c12*du/c22;
	   t = sqrt( discrim ) / c22;
	   u_plus = um + int( du );
	   u_minus = um - int( du );
	   v_minus = vm + int( s - t );
	   v_plus = vm + int( s + t );
	   for( int v = v_minus; v <= v_plus; v++ )
	      setPixelSafe( u_plus, v, color );
	   if( !du )
	      continue;
	   v_minus = vm + int( -s - t );
	   v_plus = vm + int( -s + t );
	   for( v = v_minus; v <= v_plus; v++ )
	      setPixelSafe( u_minus, v, color );
       }
}

void
vliRGBImageRef::drawHLine( float u0, float u1, float v, 
			   const vliRGBPixel& color )
{
    int umin = vliPixel::min( u0, u1 );
    int umax = vliPixel::max( u0, u1 );
    for( int u = umin; u <= umax; u++ )
       setPixelSafe( u, v, color );
}

void
vliRGBImageRef::drawVLine( float v0, float v1, float u, 
			   const vliRGBPixel& color )
{
    int vmin = vliPixel::min( v0, v1 );
    int vmax = vliPixel::max( v0, v1 );
    for( int v = vmin; v <= vmax; v++ )
       setPixelSafe( u, v, color );
}

void
vliRGBImageRef::drawCross( float u, float v, 
			   const vliRGBPixel& color, int d )
{
    drawHLine( u-d, u+d, v, color );
    drawVLine( v-d, v+d, u, color );
}

void 
vliRGBImageRef::drawLine( float u0, float u1,  
			  float v0, float v1, 
			  const vliRGBPixel& color)
{
  float length = fabsf(u1-u0);
  if (fabsf(v1-v0) > length )
    length = fabsf(v1-v0);
  float u = u0 + 0.5;
  float v = v0 + 0.5;
  float du = (u1-u0)/length;
  float dv = (v1-v0)/length;
  for(int i=1; i<length; i++)
    {
      setPixelSafe( u, v, color );
      u += du;
      v += dv;
    }
}
