// vliRGBImageRef.h++

#ifndef _vliRGBImageRef_H_
#define _vliRGBImageRef_H_
#include "vlImageGlobal.h"
#include <iostream.h>
#include "vliTypes.h"
#include "vliImageRef.h"
#include "vliRGBPixel.h"

#define vliRGBBytesPerPixel (4)

class EXPORT_CLASS vliRGBImageRef : public vliImageRef
{
 public:
    vliRGBImageRef()
      : vliImageRef( 0, 0, 0, 0 ) {}
    vliRGBImageRef( int w, int h, byte* p )
      : vliImageRef( w, h, vliRGBBytesPerPixel, p ) {}
    vliRGBImageRef( const vliImageRef& src )
      : vliImageRef( src ) {}
    inline int ix( int x, int y ) const { return( y*width+x ); }
    inline byte getRed( int i ) const { return data[i*4 + 2]; }
    inline byte getGreen( int i ) const { return data[i*4 + 1]; }
    inline byte getBlue( int i ) const { return data[i*4 + 0]; }
    inline byte& r( int i ) { return data[i*4 + 2]; }
    inline byte& g( int i ) { return data[i*4 + 1]; }
    inline byte& b( int i ) { return data[i*4 + 0]; }
    inline vliRGBPixel& getPixel( int i ) const;
    inline vliRGBPixel* getPixelPtr( int i );
    inline vliRGBPixel* getPixelPtr( int x, int y );
    inline void setPixel( int i, const vliRGBPixel& );
    inline void setPixel( int x, int y, const vliRGBPixel& );
    inline void setPixelSafe( int x, int y, const vliRGBPixel& );
    virtual istream& readRawRGB( istream& );
    virtual ostream& writeRawRGB( ostream& );
    int getType() { return RGB; }
    void drawBlob( double u0, double v0,
		   double suu, double svv, double suv, // inv cov
		   vliRGBPixel&, double d = 4. );
    void drawHLine( float u0, float u1, float v, const vliRGBPixel& );
    void drawVLine( float v0, float v1, float u, const vliRGBPixel& );
    void drawCross( float u, float v, const vliRGBPixel&, int d=5 );
    void drawLine( float u0, float u1,  
		   float v0, float v1, 
		   const vliRGBPixel&);
};

inline
vliRGBPixel&
vliRGBImageRef::getPixel( int i ) const
{
    return( ((vliRGBPixel*)data)[i] );
}

inline
vliRGBPixel*
vliRGBImageRef::getPixelPtr( int i )
{
    return( ((vliRGBPixel*)data) + i );
}

inline
vliRGBPixel*
vliRGBImageRef::getPixelPtr( int x, int y )
{
    return( ((vliRGBPixel*)data) + y*width + x );
}

inline
void
vliRGBImageRef::setPixel( int i, const vliRGBPixel& pixel )
{
    getPixel( i ) = pixel;
}

inline
void
vliRGBImageRef::setPixel( int u, int v, const vliRGBPixel& pixel )
{
    setPixel( v*width + u, pixel );
}

inline
void
vliRGBImageRef::setPixelSafe( int u, int v, const vliRGBPixel& pixel )
{
    if( u < 0 || v < 0 || u >= width || v*width >= pixelCount )
       return;
    setPixel( v*width + u, pixel );
}

#endif /* _vliRGBImageRef_H_ */
