// vliYUVImageRef.h++

// #define vliYUVImageRefDEBUG 1

#ifndef _vliYUVImageRef_H_
#define _vliYUVImageRef_H_
#include "vlImageGlobal.h"
#include "vliTypes.h"
#include "vliImageRef.h"
#include "vliYUVPixel.h"
#include "vliPixel.h"

#define vliYUVBytesPerPixel (2)

class EXPORT_CLASS vliYUVImageRef : public vliImageRef
{
public:
    vliYUVImageRef()
      : vliImageRef( 0, 0, 0, 0 ) {}
    vliYUVImageRef( int w, int h, byte* p )
      : vliImageRef( w, h, vliYUVBytesPerPixel, p ) {}
    vliYUVImageRef( const vliImageRef& src )
    : vliImageRef( src ) {}
    inline int ix( int x, int y ) const { return( y*width+x ); }
    inline byte getY( int i ) const { return data[i*2 + 1]; }
    inline byte getU( int i ) const { return data[(i-i%2)*2]; }
    inline byte getV( int i ) const { return data[(i-i%2+1)*2]; } 

    inline byte& y( int i ) { return data[i*2 + 1]; }
    inline byte& u( int i ) { return data[(i-i%2)*2]; }
    inline byte& v( int i ) { return data[(i-i%2+1)*2]; } 
    
    inline vliYUVPixel getPixel( int i ) const;
    // vliYUVPixel getPixel( int i );
    inline vliYUVPixel getPixel( int x, int y ) { return getPixel(y*width+x); }
    inline void setPixel( int i, const vliYUVPixel& );
    inline void setPixel( int x, int y, const vliYUVPixel& );
    inline void setPixelSafe( int x, int y, const vliYUVPixel& );
    virtual istream& readRawRGB( istream& );
    virtual ostream& writeRawRGB( ostream& );
    virtual int getType() { return YUV; }
    virtual void black();
    void drawBlob( double u0, double v0,
		   double suu, double svv, double suv, // inv cov
		   vliYUVPixel&, double d = 2. );
    void drawHLine( float u0, float u1, float v, vliYUVPixel& );
    void drawVLine( float v0, float v1, float u, vliYUVPixel& );
    void drawLine( float u0, float u1,  
		   float v0, float v1, 
		   const vliYUVPixel&);
    void overlay( vliYUVImageRef& );
    void map( void (*fn)( vliYUVPixel& ) );
    void _toabgr( byte* ) const; // usually for sgi
    void _tobgra( byte* ) const; // usually for pc
    void _tobgr( byte* ) const; // usually for pc
    static float toas;
    static float toao;
};

inline
vliYUVPixel
vliYUVImageRef::getPixel( int i ) const
{
#ifdef vliYUVImageRefDEBUG
    if( i < 0 || i >= pixelCount )
       cerr << "vliYUVImageRef::getPixel: DEBUG: boundary error" << endl;
#endif
    vliYUVPixel pixel;
    pixel.y = getY(i);
    pixel.u = getU(i);
    pixel.v = getV(i);
    return pixel;
}

inline
void
vliYUVImageRef::setPixel( int i, const vliYUVPixel& pixel )
{
    y(i) = pixel.y;
    u(i) = pixel.u;
    v(i) = pixel.v;
}

inline
void
vliYUVImageRef::setPixel( int u, int v, const vliYUVPixel& pixel )
{
    setPixel( v*width + u, pixel );
}

inline
void
vliYUVImageRef::setPixelSafe( int u, int v, const vliYUVPixel& pixel )
{
    if( u < 0 || v < 0 || u >= width || v*width >= pixelCount )
       return;
    setPixel( v*width + u, pixel );
}

#endif /* _vliYUVImageRef_H_ */
