// worldBaby.c++

//#include "Parse/Parse.h++"

static int verbose = 0;

//#define QmDouble
#include "qmdStatistics.h"
#include "PersonPlus.h"

static
inline
void
readAndCompute( char* filename,
		qmdVectorRef& mean,
		qmdVectorRef& direction,
		reald* length = 0 )
{
    // read in points

    PersonPlus personPlus;
    sffPersonPlus person;
    qmdSampleAcc xline( 3 );
    qmdGaussian xlineStats( 3 );
    xline.setZero();
//     int forceFlag = 1;
//     personPlus.undump_force( filename, &forceFlag );
//     int ringCount = personPlus.getRingLength();
    personPlus.readRing( filename );
    int ringCount = personPlus.getRingLength();
    reald* data = new reald[ 3 * ringCount ];
    int pointCount = 0;
    person = personPlus.get();
    int firstFrame = person.frameNumber;
    do
       {
	   if( ! person.rightHand.valid )
	      {
		  cerr << "  INVALID POINT" << endl;
		  continue;
	      }
	   qmdVectorRef( 3, data+3*pointCount ).copy
	      ( qmdVectorRef( 3, person.rightHand.camera_ref.t ) );
	   // xline.addSample( qmdVectorRef( 3, person.rightHand.camera_ref.t ) );
	   xline.addSample( qmdVectorRef( 3, data+3*pointCount ) );
	   pointCount++;
	   // forceFlag = 1, personPlus.undump_force( filename, &forceFlag );
	   personPlus.undump_force();
	   person = personPlus.get();
       }
    while( person.frameNumber != firstFrame );
    xline.getGaussian( xlineStats );
    mean.copy( xlineStats.getMean() );
    qmdMatrix cov = xlineStats.getCovariance();
    qmdEigenStructure eig( cov );
    qmdMatrix evectors = eig.getEigVectors();
    qmdVector evalues = eig.getEigValues();
    if( verbose )
       {
	   cerr << "\teval= " << evalues << endl;
	   evectors.print( cerr << "\tevec= ", "\n\t" ) << endl;
       }
    int which;
    evalues.getMax( &which );
    direction.copy( evectors.getColumn(which) );
    
    if( length )
       {
	   qmdVector dist( pointCount );
	   qmdVector y_ym( 3 );
	   for( int i = 0; i < pointCount; i++ )
	      dist(i) = y_ym.setSubtract
		 ( qmdVectorRef(3,data+i*3), mean ).
		 getInnerProduct( direction );
	   *length = dist.getMax() - dist.getMin();
       }
}

int
main( int argc, char *argv[] )
{
    char xfilename[100] = "x.dump";
    char yfilename[100] = "y.dump";
    char sfilename[100] = "x.dump";
    reald inches = 18.5;
/*
    ParseRuleFlag( 'v', verbose, 1, "Verbose mode" );
    ParseRuleArg( xfilename, "xfile", "xfile" );
    ParseRuleArg( yfilename, "yfile", "yfile" );
    ParseRuleArg( sfilename, "sfile", "sfile" );
    ParseRuleArgFloat( inches, "inches", "inches" );
    
    if( Parse::ParseCmdLine( argc, argv ) )
       Parse::Usage( argv[0] ), exit(1);
*/
    // compute xm, xhat
    //  xm is just the mean
    //  xhat is the eigenvector of the covariance corresp to largest evalue
    qmdVector xmean(3);
    qmdVector xdir(3);
    readAndCompute( xfilename, xmean, xdir );
    if( verbose )
       {
	   cerr << "xmean= " << xmean << endl;
	   cerr << "xdir= " << xdir << endl << endl;
       }

    qmdVector ymean(3);
    qmdVector ydir(3);
    readAndCompute( yfilename, ymean, ydir );
    if( verbose )
       {
	   cerr << "ymean= " << ymean << endl;
	   cerr << "ydir= " << ydir << endl;
       }

    qmdVector smean(3);
    qmdVector sdir(3);
    reald lengthInUnits;
    readAndCompute( sfilename, smean, sdir, &lengthInUnits );
    if( verbose )
       {
	   cerr << "smean= " << smean << endl;
	   cerr << "sdir= " << sdir << endl;
       }

    xdir.setNormalize();  // even though they should both be normal already
    if( xdir(0) > 0 )
       xdir.setNegateEq();
    ydir.setNormalize();
    if( ydir(1) > 0 )
       ydir.setNegateEq();
    
    qmdVector yp( 3 );
    reald a = xdir.getInnerProduct( ydir );
    cerr << "\ta= " << a << endl;
    yp.copy( xdir );
    yp *= -a;
    yp += ydir;
    yp.setNormalize();

    qmdVector zdir( 3 );
    zdir.setCross( xdir, yp );

    // compute origin
    //  A = [ xhat  -yhat ]
    //  b = (ym - xm)
    //  s = (A`A)^-1 A`b
    //  origin = xm + s(0)*xhat
    qmdMatrix A( 3, 2 );
    qmdVector tmp( 3 );

    A.setColumn( 0, xdir );
    A.setColumn( 1, tmp.setNegate( yp ) );
    tmp.setSubtract( ymean, xmean );
    qmdMatrix AtA( 2, 2 );
    AtA.setTMultiply( A, A );
    AtA.setInvertEq();
    qmdVector tmp2( 2 );
    tmp2.setTMultiply( A, tmp );
    qmdVector s( 2 );
    s.setMultiply( AtA, tmp2 );
    qmdVector origin( 3 );
    origin.copy( xdir );
    origin *= s( 0 );
    origin += xmean;
    cout << origin << endl;

    // compute scale

    // output T = origin
    // output R = [xhat`
    //             yhat`
    //             zhat`]
    // output scale = 
    qmdMatrix R( 3, 3 );
    R.setRow( 0, xdir );
    R.setRow( 1, yp );
    R.setRow( 2, zdir );
    cout << R;

    reald scale = inches / lengthInUnits;
    cout << scale << endl;

    return 0;
}
